Package: osrm.backend
Title: Bindings for 'Open Source Routing Machine'
Version: 0.3.1
Authors@R: 
    person("Egor", "Kotov", , "kotov.egor@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-6690-5345"))
Description: Install and control 'Open Source Routing Machine' ('OSRM')
    backend executables to prepare routing data and run/stop a local
    'OSRM' server. For computations with the running server use the 'osrm'
    R package (<https://cran.r-project.org/package=osrm>).
License: MIT + file LICENSE
URL: https://github.com/e-kotov/osrm.backend,
        https://www.ekotov.pro/osrm.backend/
BugReports: https://github.com/e-kotov/osrm.backend/issues
Imports: digest, httr2, jsonlite, lifecycle, processx, ps, viridisLite
Suggests: DT, knitr, mapgl, osrm, rmarkdown, sf, shiny, testthat (>=
        3.0.0), tibble, withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Language: en
RoxygenNote: 7.3.3
SystemRequirements: OSRM backend binaries (>= v5.27.0)
        <https://github.com/Project-OSRM/osrm-backend>; package
        downloads binaries automatically if not found
NeedsCompilation: no
Packaged: 2026-04-26 06:45:32 UTC; ek
Author: Egor Kotov [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-6690-5345>)
Maintainer: Egor Kotov <kotov.egor@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-26 07:00:02 UTC
Built: R 4.6.0; ; 2026-05-01 21:37:11 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,osrm_server_list)
export(osrm_check_available_versions)
export(osrm_check_latest_version)
export(osrm_cleanup)
export(osrm_clear_path)
export(osrm_contract)
export(osrm_customize)
export(osrm_extract)
export(osrm_find_profile)
export(osrm_get_server_profile)
export(osrm_gui)
export(osrm_install)
export(osrm_partition)
export(osrm_prepare_graph)
export(osrm_servers)
export(osrm_start)
export(osrm_start_server)
export(osrm_stop)
export(osrm_stop_all)
export(osrm_uninstall)
export(osrm_which)
importFrom(lifecycle,deprecated)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# osrm.backend

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![CRAN
status](https://www.r-pkg.org/badges/version/osrm.backend)](https://CRAN.R-project.org/package=osrm.backend)
<a href="https://CRAN.R-project.org/package=osrm.backend"
target="_blank"><img
src="https://cranlogs.r-pkg.org/badges/grand-total/osrm.backend?color=blue"
alt="CRAN/METACRAN Total downloads" /></a>
<a href="https://CRAN.R-project.org/package=osrm.backend"
target="_blank"><img
src="https://cranlogs.r-pkg.org/badges/osrm.backend?color=blue"
alt="CRAN/METACRAN Downloads per month" /></a> [![osrm.backend status
badge](https://e-kotov.r-universe.dev/osrm.backend/badges/version)](https://e-kotov.r-universe.dev/osrm.backend)
[![R-CMD-check](https://github.com/e-kotov/osrm.backend/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/e-kotov/osrm.backend/actions/workflows/R-CMD-check.yaml)
[![pkgcheck](https://github.com/e-kotov/osrm.backend/workflows/pkgcheck/badge.svg)](https://github.com/e-kotov/osrm.backend/actions?query=workflow%3Apkgcheck)
[![Codecov test
coverage](https://codecov.io/gh/e-kotov/osrm.backend/graph/badge.svg)](https://app.codecov.io/gh/e-kotov/osrm.backend)

[![DOI](https://zenodo.org/badge/DOI/10.32614/CRAN.package.osrm.backend.svg)](https://doi.org/10.32614/CRAN.package.osrm.backend)
<!-- badges: end -->

![](https://www.ekotov.pro/osrm.backend/reference/figures/card.png)

The goal of `osrm.backend` is to be a companion to `{osrm}` R package
<https://github.com/riatelab/osrm>:

- provide single line quick start with `osmr_start()` function that does
  everything automatically to run local OSRM backend server, no other
  setup is required;

- easily install `osrm.backend` on major operating systems (Linux,
  Windows, MacOS);

- provide wrapper functions to prepare data for `osrm` routing;

- provide wrapper to start/stop local OSRM backend server.

## Installation

Install the latest stable release of `osrm.backend` from CRAN with:

``` r
install.packages("osrm.backend")
```

You can install the development version of `osrm.backend` from R
Universe with:

``` r
install.packages('osrm.backend',
 repos = c('https://e-kotov.r-universe.dev', 'https://cloud.r-project.org')
)
```

or from [GitHub repo](https://github.com/e-kotov/osrm.backend) with:

``` r
# install.packages("pak")
pak::pak("e-kotov/osrm.backend")
```

You can prepare the OpenStreetMap data for routing with one function
`osrm_start(<path_to_osm_file_or_folder_with_osm_file>)`, set the
`osrm.server` option to `http://localhost:5001/` and use all the
`{osrm}` functions as usual. You do not need to have `osrm-backend`
installed or have `Docker` to run `osrm-backend` from a container,
everything is handled automatically an all major operating systems. Just
follow the [Get
started](https://www.ekotov.pro/osrm.backend/articles/osrm-backend.html)
guide. For advanced control over each step of the process, see the full
[function
reference](https://www.ekotov.pro/osrm.backend/reference/index.html).

## OSRM Graphical User Interface demo

![osrm GUI demo
1](https://github.com/user-attachments/assets/d7f4c671-10f1-4373-9786-1356a8942a94)

![osrm GUI demo
2](https://github.com/user-attachments/assets/f0bd80e8-632e-47f1-9800-2131d5002204)

![osrm GUI demo
3](https://github.com/user-attachments/assets/7b524d02-4e28-4f22-8c1b-221c2e9c7204)

# Clean up the temporary binary folder after tests complete
test_osrm_path <- file.path(tempdir(), "osrm_bin_test")
if (dir.exists(test_osrm_path)) {
  unlink(test_osrm_path, recursive = TRUE, force = TRUE)
}

#' Prepare OSRM Graph for Routing (Extract + Partition/Contract)
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' High-level wrapper that first runs `osrm-extract` on an OSM file
#' to produce the base `.osrm` graph, then prepares it for routing via
#' either the MLD pipeline (`osrm-partition` + `osrm-customize`) or the CH pipeline (`osrm-contract`).
#'
#' @param input_osm A string. Path to the input OSM file (`.osm`, `.osm.bz2`, or `.osm.pbf`) or a directory containing exactly one OSM file with a supported extension.
#' @param profile A string. Path to the OSRM Lua profile (e.g. returned by `osrm_find_profile("car.lua")`).
#' @param threads An integer. Number of threads for extract and partition/contract; default `8`.
#' @param overwrite A logical. If `FALSE`, stops if any existing `.osrm*` files matching
#'   the base name are found alongside `input_osm`. Set to `TRUE` to overwrite them.
#' @param verbosity A string. Log verbosity for extract/partition/contract
#'   (one of `"NONE","ERROR","WARNING","INFO","DEBUG"`); default `"INFO"`.
#' @param data_version A string or `NULL`. Passed to `osrm-extract` via `-d`; default `NULL`.
#' @param small_component_size An integer. For extract & partition; default `1000`.
#' @param with_osm_metadata A logical. Adds `--with-osm-metadata` during extract; default `FALSE`.
#' @param parse_conditional_restrictions A logical. Adds `--parse-conditional-restrictions`; default `FALSE`.
#' @param location_dependent_data A string or `NULL`. Path to GeoJSON for extract; default `NULL`.
#' @param disable_location_cache A logical. Adds `--disable-location-cache`; default `FALSE`.
#' @param dump_nbg_graph A logical. Adds `--dump-nbg-graph`; default `FALSE`.
#' @param algorithm A string. One of `"MLD"` (default) or `"CH"` (case-insensitive).
#' @param balance A numeric. Balance for `osrm-partition`; default `1.2`.
#' @param boundary A numeric. Boundary percentage for `osrm-partition`; default `0.25`.
#' @param optimizing_cuts An integer. Optimizing cuts for `osrm-partition`; default `10`.
#' @param max_cell_sizes A numeric vector. Max cell sizes for `osrm-partition`; default `c(128,4096,65536,2097152)`.
#' @param quiet A logical. Master switch that suppresses package messages and
#'   process output when `TRUE`; default `FALSE`.
#' @param verbose A logical. When `TRUE` and `quiet = FALSE`, streams stdout and
#'   stderr from the underlying `processx::run` calls.
#' @param spinner A logical. When `TRUE` and `quiet = FALSE`, shows a spinner
#'   instead of live logs; default `TRUE`.
#' @param echo_cmd A logical. When `TRUE` and `quiet = FALSE`, prints each
#'   command before running; default `FALSE`.
#'
#' @return An object of class \code{osrm_job} with the following elements:
#' \describe{
#'   \item{osrm_job_artifact}{The path to the final routing-ready graph file (`.osrm.hsgr` for CH or `.osrm.mldgr` for MLD).}
#'   \item{osrm_working_dir}{The directory containing all OSRM files.}
#'   \item{logs}{A list of \code{processx::run} results for each stage:
#'     \code{extract}, \code{partition}/\code{contract}, and \code{customize} (if MLD).}
#' }
#'
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   # Prepare a routing-ready graph with the default MLD pipeline
#'   pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
#'   osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
#'   dir.create(osrm_dir, recursive = TRUE)
#'   tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
#'   file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
#'
#'   graph <- osrm_prepare_graph(
#'     input_osm = tmp_pbf,
#'     overwrite = TRUE,
#'     threads = 1L,
#'     algorithm = "mld"
#'   )
#'   graph$osrm_job_artifact
#'
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'   unlink(osrm_dir, recursive = TRUE)
#' }
#' }
#'
#' @export
osrm_prepare_graph <- function(
  input_osm,
  profile = osrm_find_profile("car.lua"),
  threads = 8L,
  overwrite = FALSE,
  verbosity = c("INFO", "NONE", "ERROR", "WARNING", "DEBUG"),
  data_version = NULL,
  small_component_size = 1000L,
  with_osm_metadata = FALSE,
  parse_conditional_restrictions = FALSE,
  location_dependent_data = NULL,
  disable_location_cache = FALSE,
  dump_nbg_graph = FALSE,
  algorithm = c("mld", "ch"),
  balance = 1.2,
  boundary = 0.25,
  optimizing_cuts = 10L,
  max_cell_sizes = c(128, 4096, 65536, 2097152),
  quiet = FALSE,
  verbose = FALSE,
  spinner = TRUE,
  echo_cmd = FALSE
) {
  if (!requireNamespace("processx", quietly = TRUE)) {
    stop("'processx' package is required for osrm_prepare_graph", call. = FALSE)
  }

  algorithm <- tryCatch(.normalize_algorithm(algorithm), error = function(e) {
    # If algorithm is a vector (default args), normalize first element
    if (length(algorithm) > 1) .normalize_algorithm(algorithm[1]) else stop(e)
  })

  # Check for algorithm conflicts before starting pipeline
  # Extract path from input
  input_osm_resolved <- resolve_osrm_path(
    input_osm,
    pattern = "\\.(osm|osm\\.bz2|osm\\.pbf)$",
    file_description = "OSM files (.osm, .osm.bz2, or .osm.pbf)"
  )

  # Determine base name and directory
  if (grepl("\\.osm\\.pbf$", input_osm_resolved, ignore.case = TRUE)) {
    base <- sub("\\.osm\\.pbf$", "", input_osm_resolved, ignore.case = TRUE)
  } else if (grepl("\\.osm\\.bz2$", input_osm_resolved, ignore.case = TRUE)) {
    base <- sub("\\.osm\\.bz2$", "", input_osm_resolved, ignore.case = TRUE)
  } else if (grepl("\\.osm$", input_osm_resolved, ignore.case = TRUE)) {
    base <- sub("\\.osm$", "", input_osm_resolved, ignore.case = TRUE)
  } else {
    stop(
      "'input_osm' must have extension .osm, .osm.bz2, or .osm.pbf",
      call. = FALSE
    )
  }

  dir_path <- dirname(base)
  base_name <- basename(base)

  # Check for existing algorithm files
  detection <- detect_osrm_algorithm(dir_path, base_name)

  if (detection$state %in% c("ch", "mld", "mixed") && !overwrite) {
    # Conflict exists and overwrite is FALSE
    check_algorithm_conflict(dir_path, base_name, algorithm, "prepare_graph")
  } else if (detection$state %in% c("ch", "mld", "mixed") && overwrite) {
    # Warn user about overwriting files from a different algorithm
    current_algo <- if (detection$state == "ch") {
      "CH"
    } else if (detection$state == "mld") {
      "MLD"
    } else {
      "mixed"
    }
    
    if (!quiet && current_algo != algorithm && current_algo != "mixed") {
      message(
        "Note: Directory contains existing ",
        current_algo,
        " algorithm files.\n",
        "overwrite=TRUE will remove them and create new ",
        algorithm,
        " files."
      )
    }
  }

  # 1) Extract
  extract_res <- osrm_extract(
    input_osm = input_osm,
    profile = profile,
    overwrite = overwrite,
    verbosity = verbosity,
    data_version = data_version,
    threads = threads,
    small_component_size = small_component_size,
    with_osm_metadata = with_osm_metadata,
    parse_conditional_restrictions = parse_conditional_restrictions,
    location_dependent_data = location_dependent_data,
    disable_location_cache = disable_location_cache,
    dump_nbg_graph = dump_nbg_graph,
    quiet = quiet,
    verbose = verbose,
    spinner = spinner,
    echo_cmd = echo_cmd
  )

  # 2) Partition+Customize or Contract
  # Algorithm already normalized
  if (algorithm == "MLD") {
    # MLD pipeline: extract -> partition -> customize
    # Use explicit intermediate assignment for readability and to avoid
    # the native pipe operator so older R versions are supported.
    partition_res <- osrm_partition(
      extract_res,
      threads = threads,
      verbosity = verbosity,
      balance = balance,
      boundary = boundary,
      optimizing_cuts = optimizing_cuts,
      small_component_size = small_component_size,
      max_cell_sizes = max_cell_sizes,
      quiet = quiet,
      verbose = verbose,
      spinner = spinner,
      echo_cmd = echo_cmd
    )

    osrm_graph <- osrm_customize(
      partition_res,
      threads = threads,
      verbosity = verbosity,
      segment_speed_file = NULL,
      turn_penalty_file = NULL,
      edge_weight_updates_over_factor = 0,
      parse_conditionals_from_now = 0,
      time_zone_file = NULL,
      quiet = quiet,
      verbose = verbose,
      spinner = spinner,
      echo_cmd = echo_cmd
    )
  } else {
    # CH pipeline: extract -> contract
    # Each function accumulates logs from the previous step
    osrm_graph <- osrm_contract(
      input_osrm = extract_res,
      threads = threads,
      verbosity = verbosity,
      segment_speed_file = NULL,
      turn_penalty_file = NULL,
      edge_weight_updates_over_factor = 0,
      parse_conditionals_from_now = 0,
      time_zone_file = NULL,
      quiet = quiet,
      verbose = verbose,
      spinner = spinner,
      echo_cmd = echo_cmd
    )
  }

  # The final osrm_graph already contains accumulated logs from all stages
  osrm_graph
}

# Analysis of CH and MLD pipeline file conflicts

# Helper function to get all osrm-related files with their modification times
get_osrm_files <- function(dir) {
  all_files <- list.files(dir, full.names = TRUE, recursive = FALSE)
  osrm_files <- all_files[grepl("\\.osrm\\.", basename(all_files))]

  if (length(osrm_files) == 0) {
    return(data.frame(
      file = character(0),
      basename = character(0),
      size = numeric(0),
      mtime = character(0),
      stringsAsFactors = FALSE
    ))
  }

  data.frame(
    file = osrm_files,
    basename = basename(osrm_files),
    size = file.size(osrm_files),
    mtime = format(file.mtime(osrm_files), "%Y-%m-%d %H:%M:%OS3"),
    stringsAsFactors = FALSE
  )
}

# Helper to compare file lists and identify changes
compare_file_snapshots <- function(before, after, label = "") {
  cat("\n", strrep("=", 70), "\n", sep = "")
  cat(label, "\n")
  cat(strrep("=", 70), "\n")

  if (nrow(before) == 0 && nrow(after) == 0) {
    cat("No OSRM files in either snapshot.\n")
    return(invisible(NULL))
  }

  if (nrow(before) == 0) {
    cat("\nNew files created:\n")
    for (i in seq_len(nrow(after))) {
      cat(sprintf("  + %s (size: %d bytes)\n",
                  after$basename[i], after$size[i]))
    }
    return(invisible(NULL))
  }

  # Find new files
  new_files <- after[!after$basename %in% before$basename, ]
  if (nrow(new_files) > 0) {
    cat("\nNew files created:\n")
    for (i in seq_len(nrow(new_files))) {
      cat(sprintf("  + %s (size: %d bytes)\n",
                  new_files$basename[i], new_files$size[i]))
    }
  }

  # Find modified files (same name but different mtime or size)
  common_files <- intersect(before$basename, after$basename)
  if (length(common_files) > 0) {
    modified <- character()
    for (fname in common_files) {
      before_file <- before[before$basename == fname, ]
      after_file <- after[after$basename == fname, ]

      if (before_file$mtime != after_file$mtime ||
          before_file$size != after_file$size) {
        modified <- c(modified, fname)
        cat(sprintf("\nModified file: %s\n", fname))
        cat(sprintf("  Before: size=%d, mtime=%s\n",
                    before_file$size, before_file$mtime))
        cat(sprintf("  After:  size=%d, mtime=%s\n",
                    after_file$size, after_file$mtime))
      }
    }

    if (length(modified) == 0 && nrow(new_files) == 0) {
      cat("\nNo files modified or created.\n")
    }
  }

  # Find removed files
  removed_files <- before[!before$basename %in% after$basename, ]
  if (nrow(removed_files) > 0) {
    cat("\nFiles removed:\n")
    for (i in seq_len(nrow(removed_files))) {
      cat(sprintf("  - %s\n", removed_files$basename[i]))
    }
  }

  invisible(NULL)
}

test_that("Analyze CH and MLD pipeline file conflicts", {
  skip_on_cran()
  skip_if(Sys.which("osrm-routed") == "", "osrm-routed not available")
  skip_if(
    !getOption("osrm.backend.skip_osrm_tests", TRUE) == FALSE,
    "OSRM tests skipped (binary not available)"
  )

  pbf_file <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  skip_if(!file.exists(pbf_file), "Test data not available")

  # Create base temp directory for all tests
  base_tmp <- tempfile()
  dir.create(base_tmp)
  on.exit(unlink(base_tmp, recursive = TRUE), add = TRUE)

  cat("\n\n")
  cat(strrep("#", 70), "\n")
  cat("# OSRM Pipeline File Conflict Analysis\n")
  cat(strrep("#", 70), "\n")

  # Test 1: CH pipeline only
  cat("\n\nTEST 1: CH Pipeline Only\n")
  cat(strrep("-", 70), "\n")

  ch_dir <- file.path(base_tmp, "ch_only")
  dir.create(ch_dir)
  file.copy(pbf_file, file.path(ch_dir, "cur.osm.pbf"))

  before_ch <- get_osrm_files(ch_dir)
  cat("\nRunning osrm_prepare_graph with algorithm='ch'...\n")

  ch_result <- try(
    osrm_prepare_graph(
      input_osm = file.path(ch_dir, "cur.osm.pbf"),
      algorithm = "ch",
      quiet = TRUE,
      verbose = FALSE
    ),
    silent = TRUE
  )

  skip_if(inherits(ch_result, "try-error"), "CH preparation failed")

  after_ch <- get_osrm_files(ch_dir)
  compare_file_snapshots(before_ch, after_ch, "CH Pipeline - Files Created")

  # Test 2: MLD pipeline only
  cat("\n\nTEST 2: MLD Pipeline Only\n")
  cat(strrep("-", 70), "\n")

  mld_dir <- file.path(base_tmp, "mld_only")
  dir.create(mld_dir)
  file.copy(pbf_file, file.path(mld_dir, "cur.osm.pbf"))

  before_mld <- get_osrm_files(mld_dir)
  cat("\nRunning osrm_prepare_graph with algorithm='mld'...\n")

  mld_result <- try(
    osrm_prepare_graph(
      input_osm = file.path(mld_dir, "cur.osm.pbf"),
      algorithm = "mld",
      quiet = TRUE,
      verbose = FALSE
    ),
    silent = TRUE
  )

  skip_if(inherits(mld_result, "try-error"), "MLD preparation failed")

  after_mld <- get_osrm_files(mld_dir)
  compare_file_snapshots(before_mld, after_mld, "MLD Pipeline - Files Created")

  # Test 3: CH then MLD in same directory
  cat("\n\nTEST 3: CH Pipeline, Then MLD Pipeline (Same Directory)\n")
  cat(strrep("-", 70), "\n")

  ch_then_mld_dir <- file.path(base_tmp, "ch_then_mld")
  dir.create(ch_then_mld_dir)
  file.copy(pbf_file, file.path(ch_then_mld_dir, "cur.osm.pbf"))

  # First run CH
  cat("\nStep 1: Running CH pipeline...\n")
  ch_first <- try(
    osrm_prepare_graph(
      input_osm = file.path(ch_then_mld_dir, "cur.osm.pbf"),
      algorithm = "ch",
      quiet = TRUE,
      verbose = FALSE
    ),
    silent = TRUE
  )
  skip_if(inherits(ch_first, "try-error"), "CH preparation failed")

  Sys.sleep(0.1)  # Small delay to ensure different timestamps

  after_ch_first <- get_osrm_files(ch_then_mld_dir)
  cat("\nFiles after CH pipeline:\n")
  print(after_ch_first$basename)

  # Then run MLD
  cat("\nStep 2: Running MLD pipeline on same directory...\n")
  mld_second <- try(
    osrm_prepare_graph(
      input_osm = file.path(ch_then_mld_dir, "cur.osm.pbf"),
      algorithm = "mld",
      quiet = TRUE,
      verbose = FALSE,
      overwrite = TRUE
    ),
    silent = TRUE
  )
  skip_if(inherits(mld_second, "try-error"), "MLD preparation failed")

  after_mld_second <- get_osrm_files(ch_then_mld_dir)
  compare_file_snapshots(
    after_ch_first,
    after_mld_second,
    "CH → MLD: Changes When MLD Runs After CH"
  )

  # Test 4: MLD then CH in same directory
  cat("\n\nTEST 4: MLD Pipeline, Then CH Pipeline (Same Directory)\n")
  cat(strrep("-", 70), "\n")

  mld_then_ch_dir <- file.path(base_tmp, "mld_then_ch")
  dir.create(mld_then_ch_dir)
  file.copy(pbf_file, file.path(mld_then_ch_dir, "cur.osm.pbf"))

  # First run MLD
  cat("\nStep 1: Running MLD pipeline...\n")
  mld_first <- try(
    osrm_prepare_graph(
      input_osm = file.path(mld_then_ch_dir, "cur.osm.pbf"),
      algorithm = "mld",
      quiet = TRUE,
      verbose = FALSE
    ),
    silent = TRUE
  )
  skip_if(inherits(mld_first, "try-error"), "MLD preparation failed")

  Sys.sleep(0.1)  # Small delay to ensure different timestamps

  after_mld_first <- get_osrm_files(mld_then_ch_dir)
  cat("\nFiles after MLD pipeline:\n")
  print(after_mld_first$basename)

  # Then run CH
  cat("\nStep 2: Running CH pipeline on same directory...\n")
  ch_second <- try(
    osrm_prepare_graph(
      input_osm = file.path(mld_then_ch_dir, "cur.osm.pbf"),
      algorithm = "ch",
      quiet = TRUE,
      verbose = FALSE,
      overwrite = TRUE
    ),
    silent = TRUE
  )
  skip_if(inherits(ch_second, "try-error"), "CH preparation failed")

  after_ch_second <- get_osrm_files(mld_then_ch_dir)
  compare_file_snapshots(
    after_mld_first,
    after_ch_second,
    "MLD → CH: Changes When CH Runs After MLD"
  )

  # Summary comparison
  cat("\n\n")
  cat(strrep("#", 70), "\n")
  cat("# SUMMARY: File Types by Pipeline\n")
  cat(strrep("#", 70), "\n")

  ch_only_files <- after_ch$basename
  mld_only_files <- after_mld$basename

  cat("\nCH-specific files (not in MLD):\n")
  ch_specific <- setdiff(ch_only_files, mld_only_files)
  if (length(ch_specific) > 0) {
    for (f in ch_specific) cat("  -", f, "\n")
  } else {
    cat("  (none)\n")
  }

  cat("\nMLD-specific files (not in CH):\n")
  mld_specific <- setdiff(mld_only_files, ch_only_files)
  if (length(mld_specific) > 0) {
    for (f in mld_specific) cat("  -", f, "\n")
  } else {
    cat("  (none)\n")
  }

  cat("\nShared files (in both pipelines):\n")
  shared <- intersect(ch_only_files, mld_only_files)
  if (length(shared) > 0) {
    for (f in shared) cat("  -", f, "\n")
  } else {
    cat("  (none)\n")
  }

  cat("\n")
  cat(strrep("#", 70), "\n")
  cat("# Analysis Complete\n")
  cat(strrep("#", 70), "\n\n")

  # Always succeed - this test is for analysis only
  expect_true(TRUE)
})

test_that("Algorithm conflict detection prevents mixing CH and MLD", {
  skip_on_cran()
  skip_if(Sys.which("osrm-routed") == "", "osrm-routed not available")
  skip_if(
    !getOption("osrm.backend.skip_osrm_tests", TRUE) == FALSE,
    "OSRM tests skipped (binary not available)"
  )

  pbf_file <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  skip_if(!file.exists(pbf_file), "Test data not available")

  # Test 1: Trying to run CH contract after MLD partition should error
  mld_then_ch_dir <- tempfile()
  dir.create(mld_then_ch_dir)
  on.exit(unlink(mld_then_ch_dir, recursive = TRUE), add = TRUE)

  file.copy(pbf_file, file.path(mld_then_ch_dir, "cur.osm.pbf"))

  # Run MLD pipeline stages
  extract_result <- osrm_extract(
    input_osm = file.path(mld_then_ch_dir, "cur.osm.pbf"),
    quiet = TRUE,
    overwrite = TRUE
  )

  partition_result <- osrm_partition(
    input_osrm = extract_result,
    quiet = TRUE
  )

  # Now try to run CH contract - should error
  expect_error(
    osrm_contract(
      input_osrm = extract_result,
      quiet = TRUE
    ),
    "Cannot run CH pipeline.*directory contains MLD algorithm files"
  )

  # Test 2: Trying to run MLD partition after CH contract should error
  ch_then_mld_dir <- tempfile()
  dir.create(ch_then_mld_dir)
  on.exit(unlink(ch_then_mld_dir, recursive = TRUE), add = TRUE)

  file.copy(pbf_file, file.path(ch_then_mld_dir, "cur.osm.pbf"))

  # Run CH pipeline stages
  extract_result2 <- osrm_extract(
    input_osm = file.path(ch_then_mld_dir, "cur.osm.pbf"),
    quiet = TRUE,
    overwrite = TRUE
  )

  contract_result <- osrm_contract(
    input_osrm = extract_result2,
    quiet = TRUE
  )

  # Now try to run MLD partition - should error
  expect_error(
    osrm_partition(
      input_osrm = extract_result2,
      quiet = TRUE
    ),
    "Cannot run MLD pipeline.*directory contains CH algorithm files"
  )

  # Test 3: osrm_prepare_graph should error when trying to switch without overwrite
  mixed_dir <- tempfile()
  dir.create(mixed_dir)
  on.exit(unlink(mixed_dir, recursive = TRUE), add = TRUE)

  file.copy(pbf_file, file.path(mixed_dir, "cur.osm.pbf"))

  # First prepare with CH
  osrm_prepare_graph(
    input_osm = file.path(mixed_dir, "cur.osm.pbf"),
    algorithm = "ch",
    quiet = TRUE,
    overwrite = TRUE
  )

  # Try to prepare with MLD without overwrite - should error
  expect_error(
    osrm_prepare_graph(
      input_osm = file.path(mixed_dir, "cur.osm.pbf"),
      algorithm = "mld",
      quiet = TRUE,
      overwrite = FALSE
    ),
    "Found existing OSRM files|Cannot run MLD pipeline"
  )
})

test_that("osrm_cleanup removes OSRM files correctly", {
  # Create a temporary directory with some OSRM files
  test_dir <- tempfile()
  dir.create(test_dir)
  on.exit(unlink(test_dir, recursive = TRUE), add = TRUE)

  # Create mock OSRM files
  file.create(file.path(test_dir, "data.osm.pbf"))
  file.create(file.path(test_dir, "data.osrm.timestamp"))
  file.create(file.path(test_dir, "data.osrm.hsgr"))
  file.create(file.path(test_dir, "data.osrm.geometry"))

  # Test dry run
  removed_files <- osrm_cleanup(test_dir, dry_run = TRUE, quiet = TRUE)
  expect_true(length(removed_files) >= 3) # Should find at least 3 .osrm.* files
  expect_true(all(file.exists(removed_files))) # Files should still exist after dry run

  # Test actual cleanup (keeping OSM file)
  removed_files <- osrm_cleanup(test_dir, keep_osm = TRUE, quiet = TRUE)
  expect_false(file.exists(file.path(test_dir, "data.osrm.timestamp")))
  expect_false(file.exists(file.path(test_dir, "data.osrm.hsgr")))
  expect_true(file.exists(file.path(test_dir, "data.osm.pbf"))) # OSM file should remain

  # Test cleanup including OSM file
  file.create(file.path(test_dir, "data.osrm.timestamp")) # Recreate one file
  removed_files <- osrm_cleanup(
    file.path(test_dir, "data.osm.pbf"),
    keep_osm = FALSE,
    quiet = TRUE
  )
  expect_false(file.exists(file.path(test_dir, "data.osm.pbf")))
})

# Tests for osrm_start.R

test_that("osrm_start errors when path doesn't exist", {
  with_mocked_bindings(
    {
      expect_error(
        osrm_start(path = "/nonexistent/path/xyz123", quiet = TRUE),
        "Input path does not exist"
      )
    },
    Sys.which = function(x) "/usr/bin/osrm-routed",
    .package = "base"
  )
})

test_that("osrm_start errors with invalid file type", {
  tmp_dir <- tempdir()
  invalid_path <- file.path(tmp_dir, "data.txt")
  file.create(invalid_path)
  on.exit(unlink(invalid_path), add = TRUE)

  with_mocked_bindings(
    {
      expect_error(
        osrm_start(path = invalid_path, quiet = TRUE),
        "Invalid input path"
      )
    },
    Sys.which = function(x) "/usr/bin/osrm-routed",
    .package = "base"
  )
})

test_that("osrm_start errors when directory has no graphs or OSM files", {
  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  # Create a dummy file that's not .osm.pbf or graph
  file.create(file.path(tmp_dir, "dummy.txt"))

  with_mocked_bindings(
    {
      expect_error(
        osrm_start(path = tmp_dir, quiet = TRUE),
        "Directory contains no prepared OSRM graphs"
      )
    },
    Sys.which = function(x) "/usr/bin/osrm-routed",
    .package = "base"
  )
})

test_that("osrm_start works with direct path to prepared graph file", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(Sys.which("osrm-routed") == "", "osrm-routed not available")

  # Use the test graph from setup
  pbf_file <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  skip_if(!file.exists(pbf_file), "Test data not available")

  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  # Copy test file
  file.copy(pbf_file, file.path(tmp_dir, "cur.osm.pbf"))

  # Prepare the graph
  skip_if(
    !getOption("osrm.backend.skip_osrm_tests", TRUE) == FALSE,
    "OSRM tests skipped (binary not available)"
  )

  graph_result <- try(
    osrm_prepare_graph(
      input_osm = file.path(tmp_dir, "cur.osm.pbf"),
      quiet = TRUE
    ),
    silent = TRUE
  )

  skip_if(inherits(graph_result, "try-error"), "Graph preparation failed")

  # Now test osrm_start with the prepared graph
  server <- try(
    osrm_start(
      path = graph_result$osrm_job_artifact,
      quiet = TRUE
    ),
    silent = TRUE
  )

  if (!inherits(server, "try-error") && server$is_alive()) {
    on.exit(osrm_stop(server, quiet = TRUE), add = TRUE)
    expect_s3_class(server, "process")
  } else {
    skip("Server failed to start")
  }
})

test_that("osrm_start handles directory with graph files", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(Sys.which("osrm-routed") == "", "osrm-routed not available")

  # Use the test graph from setup
  pbf_file <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  skip_if(!file.exists(pbf_file), "Test data not available")

  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  file.copy(pbf_file, file.path(tmp_dir, "cur.osm.pbf"))

  skip_if(
    !getOption("osrm.backend.skip_osrm_tests", TRUE) == FALSE,
    "OSRM tests skipped (binary not available)"
  )

  # Prepare the graph to create graph files in the directory
  graph_result <- try(
    osrm_prepare_graph(
      input_osm = file.path(tmp_dir, "cur.osm.pbf"),
      quiet = TRUE
    ),
    silent = TRUE
  )

  skip_if(inherits(graph_result, "try-error"), "Graph preparation failed")

  # Now test osrm_start with the directory
  server <- try(
    osrm_start(
      path = tmp_dir,
      quiet = TRUE
    ),
    silent = TRUE
  )

  if (!inherits(server, "try-error") && server$is_alive()) {
    on.exit(osrm_stop(server, quiet = TRUE), add = TRUE)
    expect_s3_class(server, "process")
  } else {
    skip("Server failed to start")
  }
})

test_that("osrm_start algorithm parameter validation", {
  # Test that it accepts valid algorithms
  expect_silent({
    match.arg("mld", c("mld", "ch"))
    match.arg("ch", c("mld", "ch"))
  })

  # Test that it rejects invalid algorithms
  expect_error(
    match.arg("invalid", c("mld", "ch")),
    "should be one of"
  )
})

test_that("osrm_start respects algorithm parameter when selecting existing graphs", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(Sys.which("osrm-routed") == "", "osrm-routed not available")

  pbf_file <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  skip_if(!file.exists(pbf_file), "Test data not available")

  # Create separate directories for preparation to avoid conflicts
  tmp_dir_mld <- tempfile("osrm_mld_pre")
  tmp_dir_ch <- tempfile("osrm_ch_pre")
  tmp_dir_final <- tempfile("osrm_final")
  dir.create(tmp_dir_mld)
  dir.create(tmp_dir_ch)
  dir.create(tmp_dir_final)
  on.exit(unlink(c(tmp_dir_mld, tmp_dir_ch, tmp_dir_final), recursive = TRUE), add = TRUE)

  file.copy(pbf_file, file.path(tmp_dir_mld, "cur.osm.pbf"))
  file.copy(pbf_file, file.path(tmp_dir_ch, "cur.osm.pbf"))

  skip_if(
    !getOption("osrm.backend.skip_osrm_tests", TRUE) == FALSE,
    "OSRM tests skipped (binary not available)"
  )

  # Prepare MLD graph in its own directory
  mld_result <- try(
    osrm_prepare_graph(
      input_osm = file.path(tmp_dir_mld, "cur.osm.pbf"),
      algorithm = "mld",
      quiet = TRUE
    ),
    silent = TRUE
  )
  skip_if(inherits(mld_result, "try-error"), "MLD graph preparation failed")

  # Prepare CH graph in its own directory
  ch_result <- try(
    osrm_prepare_graph(
      input_osm = file.path(tmp_dir_ch, "cur.osm.pbf"),
      algorithm = "ch",
      quiet = TRUE
    ),
    silent = TRUE
  )
  skip_if(inherits(ch_result, "try-error"), "CH graph preparation failed")

  # Merge files into final directory for testing selection logic
  all_mld <- list.files(tmp_dir_mld, full.names = TRUE)
  file.copy(all_mld, tmp_dir_final, recursive = TRUE)
  all_ch <- list.files(tmp_dir_ch, full.names = TRUE)
  file.copy(all_ch, tmp_dir_final, recursive = TRUE)

  # Both graph types should now exist in the final directory
  expect_true(file.exists(file.path(tmp_dir_final, "cur.osrm.mldgr")))
  expect_true(file.exists(file.path(tmp_dir_final, "cur.osrm.hsgr")))

  # Test 1: Start with algorithm = "ch" should use CH graph
  server_ch <- try(
    osrm_start(
      path = tmp_dir_final,
      algorithm = "ch",
      quiet = TRUE
    ),
    silent = TRUE
  )

  if (!inherits(server_ch, "try-error") && server_ch$is_alive()) {
    # Check that the server command line contains the CH algorithm flag
    cmdline <- server_ch$get_cmdline()
    # OSRM-routed uses '-a CH' or '-a MLD'
    expect_true(any(cmdline == "CH"))
    expect_false(any(cmdline == "MLD"))
    osrm_stop(server_ch, quiet = TRUE)
  } else {
    skip("CH server failed to start")
  }

  # Test 2: Start with algorithm = "mld" should use MLD graph
  server_mld <- try(
    osrm_start(
      path = tmp_dir_final,
      algorithm = "mld",
      quiet = TRUE
    ),
    silent = TRUE
  )

  if (!inherits(server_mld, "try-error") && server_mld$is_alive()) {
    # Check that the server command line contains the MLD algorithm flag
    cmdline <- server_mld$get_cmdline()
    expect_true(any(cmdline == "MLD"))
    expect_false(any(cmdline == "CH"))
    osrm_stop(server_mld, quiet = TRUE)
  } else {
    skip("MLD server failed to start")
  }
})

test_that("osrm_prepare_graph runs MLD pipeline correctly", {
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  input_osm <- file.path(tmp_dir, "test.osm.pbf")
  file.create(input_osm)
  on.exit(unlink(input_osm), add = TRUE)

  # Mock osrm_extract, osrm_partition, osrm_customize
  mock_extract <- function(input_osm, ...) {
    osrm.backend:::as_osrm_job(
      osrm_job_artifact = sub("\\.osm\\.pbf$", ".osrm.timestamp", input_osm),
      osrm_working_dir = dirname(input_osm),
      logs = list(extract = list(status = 0))
    )
  }
  mock_partition <- function(input_osrm, ...) {
    # Extract path and previous logs if input is an osrm_job
    if (inherits(input_osrm, "osrm_job")) {
      prev_logs <- input_osrm$logs
      input_path <- input_osrm$osrm_job_artifact
    } else {
      prev_logs <- list()
      input_path <- input_osrm
    }
    osrm.backend:::as_osrm_job(
      osrm_job_artifact = paste0(sub("\\.timestamp$", "", input_path), ".osrm.partition"),
      osrm_working_dir = dirname(input_path),
      logs = c(prev_logs, list(partition = list(status = 0)))
    )
  }
  mock_customize <- function(input_osrm, ...) {
    # Extract path and previous logs if input is an osrm_job
    if (inherits(input_osrm, "osrm_job")) {
      prev_logs <- input_osrm$logs
      input_path <- input_osrm$osrm_job_artifact
    } else {
      prev_logs <- list()
      input_path <- input_osrm
    }
    osrm.backend:::as_osrm_job(
      osrm_job_artifact = sub("\\.partition$", ".mldgr", input_path),
      osrm_working_dir = dirname(input_path),
      logs = c(prev_logs, list(customize = list(status = 0)))
    )
  }

  result <- with_mocked_bindings(
    osrm_prepare_graph(
      input_osm = input_osm,
      algorithm = "mld",
      quiet = TRUE
    ),
    osrm_extract = mock_extract,
    osrm_partition = mock_partition,
    osrm_customize = mock_customize
  )

  expect_true(grepl("\\.mldgr$", result$osrm_job_artifact))
  expect_named(result$logs, c("extract", "partition", "customize"))
})

test_that("osrm_prepare_graph runs CH pipeline correctly", {
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  input_osm <- file.path(tmp_dir, "test.osm.pbf")
  file.create(input_osm)
  on.exit(unlink(input_osm), add = TRUE)

  mock_extract <- function(input_osm, ...) {
    osrm.backend:::as_osrm_job(
      osrm_job_artifact = sub("\\.osm\\.pbf$", ".osrm.timestamp", input_osm),
      osrm_working_dir = dirname(input_osm),
      logs = list(extract = list(status = 0))
    )
  }
  mock_contract <- function(input_osrm, ...) {
    # Extract path and previous logs if input is an osrm_job
    if (inherits(input_osrm, "osrm_job")) {
      prev_logs <- input_osrm$logs
      input_path <- input_osrm$osrm_job_artifact
    } else {
      prev_logs <- list()
      input_path <- input_osrm
    }
    osrm.backend:::as_osrm_job(
      osrm_job_artifact = sub("\\.timestamp$", ".hsgr", input_path),
      osrm_working_dir = dirname(input_path),
      logs = c(prev_logs, list(contract = list(status = 0)))
    )
  }

  result <- with_mocked_bindings(
    osrm_prepare_graph(
      input_osm = input_osm,
      algorithm = "ch",
      quiet = TRUE
    ),
    osrm_extract = mock_extract,
    osrm_contract = mock_contract
  )

  expect_true(grepl("\\.hsgr$", result$osrm_job_artifact))
  expect_named(result$logs, c("extract", "contract"))
})

# R/osrm_stop_server.R

#' List OSRM servers
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Lists `osrm-routed` processes. By default, it returns a snapshot of servers
#' started by the current R session (registered via [osrm_start_server()] or [osrm_start()]).
#' You can optionally list all `osrm-routed` processes running on the system,
#' including those started by other sessions or manually.
#'
#' You can stop a server by passing its `id`, `port`, or `pid` to [osrm_stop()].
#'
#' @param include_all Logical; if `TRUE`, scans the system process table for
#'   all `osrm-routed` processes, including those not started by this package
#'   in the current session. Default is `FALSE`.
#' @param output Character string specifying the return format. Either `"data.frame"`
#'   (the default) which returns a tabular summary with a custom print method,
#'   or `"list"` which returns a detailed list of server metadata objects.
#'
#' @return If `output = "data.frame"`, returns a `data.frame` (class `osrm_server_list`)
#'   of OSRM job processes with columns:
#'   `id`, `pid`, `port`, `algorithm`, `started_at`, `alive`, `has_handle`, `log`, `input_osm`,
#'   `center_lon`, `center_lat`. External servers will have `id` prefixed with `sys-` and
#'   `log` set to `<external>`. If `output = "list"`, returns a named list of server metadata.
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   # copy example OSM PBF into a temporary workspace to avoid polluting pkg data
#'   pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
#'   osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
#'   dir.create(osrm_dir, recursive = TRUE)
#'   tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
#'   file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
#'   graph <- osrm_prepare_graph(tmp_pbf, overwrite = TRUE, threads = 1L)
#'
#'   srv <- osrm_start_server(graph$osrm_job_artifact, port = 6000, threads = 1L)
#'   osrm_servers()
#'   osrm_stop(srv)
#'
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'   unlink(osrm_dir, recursive = TRUE)
#' }
#' }
#' @export
osrm_servers <- function(include_all = FALSE, output = c("data.frame", "list")) {
  output <- match.arg(output)
  
  # 1. Background Cleanup: Always scan other registries to GC dead files
  # This ensures orphaned registry files from crashed sessions are pruned.
  orphans <- tryCatch(.osrm_registry_scan_others(), error = function(e) list())

  # 2. Internal Registry (Current Session)
  reg <- .osrm_state$registry
  
  # Helper to build an empty/structured data.frame
  build_df <- function(ids, pids, ports, algos, starts, alives, handles, logs, inputs, center_lons, center_lats) {
    df <- data.frame(
      id = as.character(ids),
      pid = as.integer(pids),
      port = as.integer(ports),
      algorithm = as.character(algos),
      started_at = as.POSIXct(starts),
      alive = as.logical(alives),
      has_handle = as.logical(handles),
      log = as.character(logs),
      input_osm = as.character(inputs),
      center_lon = as.numeric(center_lons),
      center_lat = as.numeric(center_lats),
      stringsAsFactors = FALSE
    )
    class(df) <- c("osrm_server_list", "data.frame")
    df
  }

  out_reg <- if (length(reg)) {
    alive_vec <- logical(length(reg))
    handle_vec <- logical(length(reg))
    for (i in seq_along(reg)) {
      e <- reg[[i]]
      if (!is.null(e$proc) && inherits(e$proc, "process")) {
        handle_vec[i] <- TRUE
        alive_vec[i] <- tryCatch(e$proc$is_alive(), error = function(...) FALSE)
      } else {
        handle_vec[i] <- FALSE
        alive_vec[i] <- .osrm_pid_is_running(e$pid)
      }
    }
    build_df(
      ids = vapply(reg, `[[`, "", "id"),
      pids = vapply(reg, function(e) as.integer(e$pid %||% NA_integer_), 0L),
      ports = vapply(reg, function(e) as.integer(e$port %||% NA_integer_), 0L),
      algos = vapply(reg, function(e) as.character(e$algorithm %||% ""), ""),
      starts = vapply(reg, function(e) as.character(e$started_at %||% NA_character_), ""),
      alives = alive_vec,
      handles = handle_vec,
      logs = vapply(reg, function(e) as.character(e$log %||% ""), ""),
      inputs = vapply(reg, function(e) as.character(e$input_osm %||% ""), ""),
      center_lons = vapply(reg, function(e) as.numeric(e$center_lon %||% NA_real_), 0.0),
      center_lats = vapply(reg, function(e) as.numeric(e$center_lat %||% NA_real_), 0.0)
    )
  } else {
    build_df(character(), integer(), integer(), character(), character(), logical(), logical(), character(), character(), numeric(), numeric())
  }

  # Merge: Current + Orphans + System
  res <- out_reg
  if (isTRUE(include_all)) {
    # 3. Process Orphans/Foreign (already scanned in Step 1)
    out_orph <- if (length(orphans)) {
      build_df(
        ids = vapply(orphans, `[[`, "", "id"),
        pids = vapply(orphans, `[[`, 0L, "pid"),
        ports = vapply(orphans, `[[`, 0L, "port"),
        algos = vapply(orphans, function(e) as.character(e$algorithm %||% ""), ""),
        starts = vapply(orphans, function(e) as.character(e$started_at %||% NA_character_), ""),
        alives = rep(TRUE, length(orphans)),
        handles = rep(FALSE, length(orphans)),
        logs = vapply(orphans, function(e) as.character(e$log %||% ""), ""),
        inputs = vapply(orphans, function(e) as.character(e$input_osm %||% ""), ""),
        center_lons = vapply(orphans, function(e) as.numeric(e$center_lon %||% NA_real_), 0.0),
        center_lats = vapply(orphans, function(e) as.numeric(e$center_lat %||% NA_real_), 0.0)
      )
    } else {
      build_df(character(), integer(), integer(), character(), character(), logical(), logical(), character(), character(), numeric(), numeric())
    }

    # 3. System Process Discovery via ps
    sys_procs <- tryCatch(ps::ps(), error = function(e) NULL)

    sys_rows <- if (!is.null(sys_procs)) {
      # Filter for osrm-routed
      is_osrm <- grepl("osrm-routed", sys_procs$name, ignore.case = TRUE)
      sys_osrm <- sys_procs[is_osrm, ]

      if (nrow(sys_osrm) > 0) {
        ext_ids <- character()
        ext_pids <- integer()
        ext_ports <- integer()
        ext_algos <- character()
        ext_starts <- as.POSIXct(character())
        ext_alives <- logical()
        ext_handles <- logical()
        ext_logs <- character()
        ext_inputs <- character()
        ext_center_lons <- numeric()
        ext_center_lats <- numeric()

        # PIDs we already know about (Current + Orphans)
        known_pids <- c(out_reg$pid, out_orph$pid)

        for (i in seq_len(nrow(sys_osrm))) {
          p <- sys_osrm[i, ]
          this_pid <- p$pid

          if (this_pid %in% known_pids) next

          # Parse command line
          cmd <- tryCatch(ps::ps_cmdline(p), error = function(e) character())

          # Defaults
          port <- 5000L
          algo <- "CH"
          input <- NA_character_

          # Simple parsing logic
          if (length(cmd) > 1) {
            # Port
            idx_p <- which(cmd %in% c("-p", "--port"))
            if (length(idx_p) && (idx_p[1] + 1 <= length(cmd))) {
              val <- suppressWarnings(as.integer(cmd[idx_p[1] + 1]))
              if (!is.na(val)) port <- val
            }

            # Algorithm
            idx_a <- which(cmd %in% c("-a", "--algorithm"))
            if (length(idx_a) && (idx_a[1] + 1 <= length(cmd))) {
              algo <- cmd[idx_a[1] + 1]
            }

            # Input: heuristic - last arg that doesn't start with "-"
            args <- cmd[-1]
            non_flag <- args[!startsWith(args, "-")]
            if (length(non_flag) > 0) {
              input <- utils::tail(non_flag, 1)
            }
          }

          ext_ids <- c(ext_ids, paste0("sys-", this_pid))
          ext_pids <- c(ext_pids, this_pid)
          ext_ports <- c(ext_ports, port)
          ext_algos <- c(ext_algos, algo)
          ext_starts <- c(ext_starts, NA)
          ext_alives <- c(ext_alives, TRUE)
          ext_handles <- c(ext_handles, FALSE)
          ext_logs <- c(ext_logs, "<external>")
          ext_inputs <- c(ext_inputs, if (is.na(input)) "" else input)
          ext_center_lons <- c(ext_center_lons, NA_real_)
          ext_center_lats <- c(ext_center_lats, NA_real_)
        }

        if (length(ext_pids) > 0) {
          build_df(ext_ids, ext_pids, ext_ports, ext_algos, ext_starts, ext_alives, ext_handles, ext_logs, ext_inputs, ext_center_lons, ext_center_lats)
        } else {
          NULL
        }
      } else {
        NULL
      }
    } else {
      NULL
    }

    if (nrow(out_orph) > 0) res <- rbind(res, out_orph)
    if (!is.null(sys_rows) && nrow(sys_rows) > 0) res <- rbind(res, sys_rows)
  }
  
  if (output == "list") {
    if (nrow(res) == 0) return(list())
    
    # Strip the custom class so split/as.list behave like a plain data.frame
    df_raw <- res
    class(df_raw) <- "data.frame"
    
    # We must explicitly convert the data.frame to a list of lists.
    # split() on plain data.frame creates a list of 1-row dataframes.
    row_dfs <- split(df_raw, seq_len(nrow(df_raw)))
    out_list <- lapply(row_dfs, function(df) {
      lst <- as.list(df)
      # Flatten 1-element vectors to scalars to match 'metadata list' expectation
      lapply(lst, function(x) x[[1]])
    })
    
    names(out_list) <- res$id
    return(out_list)
  }
  res
}

#' @export
print.osrm_server_list <- function(x, ...) {
  n <- nrow(x)
  if (n == 0) {
    cat("\u2139 No OSRM servers registered.\n")
    return(invisible(x))
  }
  
  cat("\u2139 This is a standard data.frame. Access full columns with `$` (e.g., `x$log`).\n\n")
  
  for (i in seq_len(n)) {
    status <- if (x$alive[i]) "RUNNING" else "DEAD (Orphaned)"
    has_handle <- if (x$has_handle[i]) "Yes" else "No"
    
    center_str <- if (is.na(x$center_lon[i]) || is.na(x$center_lat[i])) {
      "NA, NA"
    } else {
      sprintf("%.4f, %.4f", x$center_lon[i], x$center_lat[i])
    }
    
    started <- if (is.na(x$started_at[i])) "Unknown" else format(x$started_at[i], "%Y-%m-%d %H:%M:%S")
    
    cat(sprintf("[%d] OSRM Server (pid: %d, port: %d)\n", i, x$pid[i], x$port[i]))
    cat(sprintf("    Status    : %s\n", status))
    cat(sprintf("    Algorithm : %s\n", x$algorithm[i]))
    cat(sprintf("    Started   : %s\n", started))
    cat(sprintf("    Input OSM : %s\n", x$input_osm[i]))
    cat(sprintf("    Log File  : %s\n", x$log[i]))
    cat(sprintf("    Center    : %s\n", center_str))
    cat(sprintf("    Handle    : %s\n", has_handle))
    
    if (i < n) cat("\n")
  }
  
  invisible(x)
}

#' Stop an OSRM Server
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Terminates an `osrm-routed` process launched by `osrm_start()` or
#' `osrm_start_server()`. Can also stop external servers by PID or ID.
#'
#' @details
#' This function provides a flexible way to stop a running OSRM process. If no
#' arguments are specified, it defaults to stopping the most recently started
#' server that is still alive in the current session.
#'
#' You can also stop a specific server by providing:
#' \itemize{
#'   \item The OSRM job process (a `processx::process` object) returned by
#'     `osrm_start()` or `osrm_start_server()`.
#'   \item The server's `id`, `port`, or `pid` (use `osrm_servers()` to find these).
#' }
#' 
#' **Advanced Use:**
#' You can stop an external `osrm-routed` process (one not started by the current
#' R session) by passing its PID, or by finding it via `osrm_servers(include_all = TRUE)`
#' and passing its `id` or `port`. This requires permission to signal the process.
#'
#' @param server Optional OSRM job process (an `osrm_server` object
#'   inheriting from `processx::process`) returned by `osrm_start_server()`.
#' @param id Optional character id from `osrm_servers()`.
#' @param port Optional integer TCP port.
#' @param pid Optional integer process id.
#' @param wait Integer milliseconds to wait for clean shutdown (default `1000`).
#' @param quiet Logical; suppress messages (default `FALSE`).
#'
#' @return A list with fields `id`, `pid`, `port`, `stopped` (logical).
#' @export
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   # copy example OSM PBF into a temporary workspace to avoid polluting pkg data
#'   pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
#'   osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
#'   dir.create(osrm_dir, recursive = TRUE)
#'   tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
#'   file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
#'   graph <- osrm_prepare_graph(tmp_pbf, overwrite = TRUE, threads = 1L)
#'
#'   srv <- osrm_start_server(graph$osrm_job_artifact, port = 6000, threads = 1L)
#'
#'   # Stop by passing the process object
#'   osrm_stop(srv)
#'
#'   # Or stop by port after the process is registered
#'   osrm_stop(port = 6000)
#'
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'   unlink(osrm_dir, recursive = TRUE)
#' }
#' 
#' \dontrun{
#'   # Advanced: Stop an external server by PID
#'   # 1. Find the PID of an external server
#'   srvs <- osrm_servers(include_all = TRUE)
#'   # 2. Stop it by PID
#'   if (nrow(srvs) > 0) {
#'     osrm_stop(pid = srvs$pid[1])
#'   }
#'   
#'   # Or stop by its external ID (e.g., "sys-12345")
#'   osrm_stop(id = "sys-12345")
#' }
#' }
#' @seealso [osrm_start()], [osrm_servers()], [osrm_stop_all()]
#' 
osrm_stop <- function(
  server = NULL,
  id = NULL,
  port = NULL,
  pid = NULL,
  wait = 1000L,
  quiet = FALSE
) {
  quiet <- isTRUE(quiet)
  # Case 1: user supplied a process object directly
  if (!is.null(server)) {
    if (!inherits(server, "process")) {
      stop("'server' must be a processx::process object", call. = FALSE)
    }
    targ_pid <- suppressWarnings(try(server$get_pid(), silent = TRUE))
    if (!inherits(targ_pid, "try-error")) {
      # Try to find and deregister by pid from the internal registry
      reg <- .osrm_state$registry
      hit <- names(reg)[vapply(
        reg,
        function(e) identical(e$pid, targ_pid),
        logical(1)
      )]
      if (server$is_alive()) {
        try(server$kill(), silent = TRUE)
        try(server$wait(as.integer(wait)), silent = TRUE)
      }
      if (length(hit)) {
        try(.osrm_deregister(hit[[1]]), silent = TRUE)
      }
      if (!quiet) {
        message("Stopped OSRM server (pid ", targ_pid, ").")
      }
      return(list(
        id = if (length(hit)) hit[[1]] else NA_character_,
        pid = as.integer(targ_pid),
        port = NA_integer_,
        stopped = TRUE
      ))
    } else {
      # Could not read pid; best effort kill
      try(server$kill(), silent = TRUE)
      try(server$wait(as.integer(wait)), silent = TRUE)
      if (!quiet) {
        message("Stopped OSRM server (unknown pid).")
      }
      return(list(
        id = NA_character_,
        pid = NA_integer_,
        port = NA_integer_,
        stopped = TRUE
      ))
    }
  }

  # Case 2: Direct PID (External/Manual stop)
  if (!is.null(pid)) {
    pid_int <- as.integer(pid)
    
    # Check if it's in our local registry first (to clean up properly)
    reg <- .osrm_state$registry
    hit <- names(reg)[vapply(
      reg,
      function(e) identical(e$pid, pid_int),
      logical(1)
    )]
    
    if (length(hit)) {
      # Redirect to local stop logic (Case 3 logic will handle it if we just fall through, 
      # but let's be explicit)
      return(osrm_stop(id = hit[1], wait = wait, quiet = quiet))
    }
    
    # External stop
    if (.osrm_pid_is_running(pid_int)) {
      .osrm_kill_pid(pid_int)
      # Wait a moment
      Sys.sleep(wait / 1000)
      still_alive <- .osrm_pid_is_running(pid_int)
      if (!quiet) {
        if (!still_alive) {
          message("Stopped external OSRM server (pid ", pid_int, ").")
        } else {
          message("Attempted to stop external OSRM server (pid ", pid_int, "), but it may still be running.")
        }
      }
      return(list(
        id = NA_character_,
        pid = pid_int,
        port = NA_integer_,
        stopped = !still_alive
      ))
    } else {
      if (!quiet) message("Process with PID ", pid_int, " not found.")
      return(list(
        id = NA_character_,
        pid = pid_int,
        port = NA_integer_,
        stopped = TRUE
      ))
    }
  }

  # Case 3: Selection via Registry (Local OR External)
  reg <- .osrm_state$registry
  
  # Search Local Registry
  idx <- NULL
  if (length(reg)) {
    pick_index <- function() {
      if (!is.null(id)) {
        which(names(reg) == id)[1]
      } else if (!is.null(port)) {
        hits <- which(vapply(
          reg,
          function(e) identical(e$port, as.integer(port)),
          logical(1)
        ))
        if (length(hits)) utils::tail(hits, 1) else integer()
      } else {
        # Default: most recently started alive server
        alive <- vapply(
          reg,
          function(e) {
            if (!is.null(e$proc) && inherits(e$proc, "process")) {
              tryCatch(e$proc$is_alive(), error = function(...) FALSE)
            } else {
              .osrm_pid_is_running(e$pid)
            }
          },
          logical(1)
        )
        utils::tail(which(alive), 1)
      }
    }
    idx <- pick_index()
  }

  # Found in Local Registry?
  if (length(idx) && !is.na(idx) && idx >= 1) {
    entry <- reg[[idx]]
    stopped <- FALSE

    # Prefer to use the live process handle if it exists in the session
    if (!is.null(entry$proc) && inherits(entry$proc, "process")) {
      if (tryCatch(entry$proc$is_alive(), error = function(...) FALSE)) {
        try(entry$proc$kill(), silent = TRUE)
        try(entry$proc$wait(as.integer(wait)), silent = TRUE)
        stopped <- TRUE
      } else {
        stopped <- TRUE
      }
    } else if (!is.null(entry$pid)) {
      if (.osrm_pid_is_running(entry$pid)) {
        .osrm_kill_pid(entry$pid)
        stopped <- TRUE
      } else {
        stopped <- TRUE
      }
    }

    # Deregister regardless to keep the registry clean
    try(.osrm_deregister(entry$id), silent = TRUE)

    if (!quiet) {
      msg <- sprintf(
        "Stopped OSRM server id=%s (pid %s, port %s).",
        entry$id,
        as.character(entry$pid %||% "NA"),
        as.character(entry$port %||% "NA")
      )
      message(msg)
    }

    return(list(
      id = entry$id,
      pid = as.integer(entry$pid %||% NA_integer_),
      port = as.integer(entry$port %||% NA_integer_),
      stopped = isTRUE(stopped)
    ))
  }

  # Case 4: Not found locally. Try External Discovery.
  # Only try if specific criteria were given (don't default to killing random external servers)
  if (!is.null(id) || !is.null(port)) {
    # Special case: id="sys-PID" -> Kill PID
    if (!is.null(id) && grepl("^sys-(\\d+)$", id)) {
      pid_from_id <- as.integer(sub("^sys-(\\d+)$", "\\1", id))
      return(osrm_stop(pid = pid_from_id, wait = wait, quiet = quiet))
    }

    # Search all
    all_srvs <- osrm_servers(include_all = TRUE)
    
    target_pid <- NULL
    target_id <- NA_character_
    
    if (!is.null(id)) {
      match <- all_srvs[all_srvs$id == id, ]
      if (nrow(match) > 0) {
        target_pid <- match$pid[1]
        target_id <- match$id[1]
      }
    } else if (!is.null(port)) {
      # Find by port (most recent first if multiple? usually only one per port)
      match <- all_srvs[all_srvs$port == as.integer(port), ]
      if (nrow(match) > 0) {
        target_pid <- utils::tail(match$pid, 1)
        target_id <- utils::tail(match$id, 1)
      }
    }
    
    if (!is.null(target_pid)) {
      return(osrm_stop(pid = target_pid, wait = wait, quiet = quiet))
    }
  }

  # Case 5: Failure
  if (!is.null(id) || !is.null(port)) {
    stop(
      "Could not identify a server to stop with the specified criteria (scanned local and external).",
      call. = FALSE
    )
  }
  
  if (!quiet) {
    message("No running OSRM servers to stop in the current session.")
  }
  
  list(stopped = FALSE)
}

#' Stop all running OSRM servers started via this package
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' @param include_all A logical value. If `TRUE`, stop all OSRM servers found
#'   running on the system, including those not started by the current R session.
#'   Requires `ps` and `jsonlite` packages. Defaults to `FALSE`.
#' @return The number of servers attempted.
#' @export
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
#'   osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
#'   dir.create(osrm_dir, recursive = TRUE)
#'   tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
#'   file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
#'   graph <- osrm_prepare_graph(tmp_pbf, overwrite = TRUE, threads = 1L)
#'
#'   srv <- osrm_start_server(graph$osrm_job_artifact, port = 6000, threads = 1L)
#'   stopped <- osrm_stop_all()
#'   stopped
#'
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'   unlink(osrm_dir, recursive = TRUE)
#' }
#' }
#' 
osrm_stop_all <- function(include_all = FALSE) {
  # Get all servers (registry + optional external)
  srvs <- osrm_servers(include_all = include_all)
  
  if (nrow(srvs) == 0) {
    return(0L)
  }

  n <- 0L
  
  # 1. Stop registry servers first (by ID)
  reg_srvs <- srvs[srvs$id != "" & !startsWith(srvs$id, "sys-"), ]
  for (id in reg_srvs$id) {
    n <- n + 1L
    try(osrm_stop(id = id, quiet = TRUE), silent = TRUE)
  }
  
  # 2. Stop external servers (by PID)
  ext_srvs <- srvs[startsWith(srvs$id, "sys-"), ]
  for (pid in ext_srvs$pid) {
    n <- n + 1L
    try(osrm_stop(pid = pid, quiet = TRUE), silent = TRUE)
  }
  
  n
}

# ROBUST SKIP: Check if OSRM is actually available before running these tests.
osrm_exec_opt <- getOption("osrm.routed.exec", "")
has_osrm_option <- nzchar(osrm_exec_opt) && nzchar(Sys.which(osrm_exec_opt))
has_osrm_path <- nzchar(Sys.which("osrm-routed"))

if (!has_osrm_option && !has_osrm_path) {
  testthat::skip("OSRM binary not found (Skipping Server Additions Tests)")
}

test_that("osrm_start_server triggers hint on version incompatibility", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  # Create a mock log file with the incompatibility error content
  mock_log_content <- c(
    "[info] Starting OSRM server...",
    "[error] File is incompatible with this version of OSRM: /path/to/test.osrm.cells prepared with OSRM 6.0.0 but this is v26.4.1 (at include/storage/tar.hpp:203)",
    "[info] Maximum header size: 25024"
  )
  mock_log_file <- file.path(tmp_dir, "mock_incompatible.log")
  writeLines(mock_log_content, mock_log_file)

  MockProcess <- list(
    new = function(command, args, ..., stdout, stderr) {
      structure(
        list(
          is_alive = function() FALSE, # Simulate immediate failure
          get_pid = function() 12345,
          kill = function() TRUE,
          wait = function(...) TRUE,
          get_exit_status = function() 3
        ),
        class = c("process", "list")
      )
    }
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      old_opt <- options(osrm.server.log_file = mock_log_file)
      on.exit(options(old_opt), add = TRUE)

      expect_error(
        osrm_start_server(osrm_path = osrm_path, quiet = TRUE),
        "Hint: The OSRM graph files are incompatible"
      )
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_servers output='list' returns a named list of metadata", {
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  mock_proc <- list(
    is_alive = function() TRUE,
    get_pid = function() 1001L
  )
  class(mock_proc) <- c("process", "list")

  mock_reg <- list(
    "osrm-5001-1001" = list(
      id = "osrm-5001-1001",
      pid = 1001L,
      port = 5001L,
      algorithm = "mld",
      started_at = "2025-01-01T12:00:00.000Z",
      input_osm = "data1.osm.pbf",
      center_lon = 10.5,
      center_lat = 50.2,
      proc = mock_proc
    )
  )

  # Store original registry and session id to restore later
  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  on.exit(state_env$registry <- orig_registry, add = TRUE)
  
  # Inject mock registry
  state_env$registry <- mock_reg

  with_mocked_bindings(
    {
      result_list <- osrm_servers(output = "list")

      expect_type(result_list, "list")
      expect_equal(length(result_list), 1)
      expect_named(result_list, "osrm-5001-1001")
      expect_equal(result_list[[1]]$pid, 1001L)
      expect_equal(result_list[[1]]$port, 5001L)
      expect_equal(result_list[[1]]$center_lon, 10.5)
      expect_equal(result_list[[1]]$center_lat, 50.2)
    },
    .osrm_pid_is_running = function(...) TRUE
  )
})

test_that("print.osrm_server_list produces expected console output", {
  # Create a dummy data frame with the right class and structure
  dummy_df <- data.frame(
    id = "osrm-5001-1001",
    pid = 1001L,
    port = 5001L,
    algorithm = "MLD",
    started_at = as.POSIXct("2025-01-01 12:00:00", tz = "UTC"),
    alive = TRUE,
    has_handle = TRUE,
    log = "/tmp/test.log",
    input_osm = "/tmp/data.osm.pbf",
    center_lon = 10.1234,
    center_lat = 50.5678,
    stringsAsFactors = FALSE
  )
  class(dummy_df) <- c("osrm_server_list", "data.frame")
  
  # Capture the print output
  out <- utils::capture.output(print(dummy_df))
  
  # Verify key components of Option 2 formatting are present
  out_str <- paste(out, collapse = "\n")
  expect_match(out_str, "OSRM Server \\(pid: 1001, port: 5001\\)")
  expect_match(out_str, "Status    : RUNNING")
  expect_match(out_str, "Algorithm : MLD")
  expect_match(out_str, "Center    : 10.1234, 50.5678")
  expect_match(out_str, "Input OSM : /tmp/data.osm.pbf")
  expect_match(out_str, "Log File  : /tmp/test.log")
})

test_that("osrm_server class and metadata are correctly assigned", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  # Create a fake executable path to bypass osrm_start_server's Sys.which check.
  # We use a real file to avoid complex Sys.which mocking across platforms.
  fake_osrm_bin <- file.path(tmp_dir, if (.Platform$OS.type == "windows") "osrm-routed.exe" else "osrm-routed")
  file.create(fake_osrm_bin)
  if (.Platform$OS.type != "windows") Sys.chmod(fake_osrm_bin, mode = "0755")
  
  # Ensure the function uses our fake binary
  old_exec <- options(osrm.routed.exec = fake_osrm_bin)
  on.exit({
    options(old_exec)
    unlink(fake_osrm_bin)
  }, add = TRUE)

  # Mock processx::process
  MockProcess <- list(
    new = function(...) {
      structure(
        list(
          is_alive = function() TRUE,
          get_pid = function() 12345,
          kill = function() TRUE,
          wait = function(...) TRUE
        ),
        class = c("process", "list")
      )
    }
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_reg <- state_env$registry
  on.exit(state_env$registry <- orig_reg, add = TRUE)
  state_env$registry <- list()

  with_mocked_bindings(
    {
      server <- osrm_start_server(
        osrm_path = osrm_path,
        algorithm = "MLD",
        port = 5002L,
        quiet = TRUE
      )
    },
    process = MockProcess,
    .package = "processx"
  )

  expect_s3_class(server, "osrm_server")
  expect_s3_class(server, "process")
  
  meta <- attr(server, "osrm_metadata")
  expect_equal(meta$port, 5002L)
  expect_equal(meta$algorithm, "MLD")
  expect_equal(meta$path, osrm_path)
  
  # Check print method
  expect_output(print(server), "OSRM Server \\(Job Process\\)")
  expect_output(print(server), "Status:\\s+Running")
  expect_output(print(server), "Port:\\s+5002")
  
  # Verify inheritance (process methods still work)
  expect_true(server$is_alive())
  expect_equal(server$get_pid(), 12345)
})

test_that("osrm_server print method handles stopped state", {
  server <- structure(
    list(
      is_alive = function() FALSE,
      get_pid = function() 12345
    ),
    class = c("osrm_server", "process", "list"),
    osrm_metadata = list(
      port = 5001L,
      profile = "car",
      algorithm = "MLD",
      path = "test.osrm.mldgr"
    )
  )
  
  expect_output(print(server), "Status:\\s+Stopped")
  # PID should NOT be printed if stopped (based on our print logic)
  out <- capture.output(print(server))
  expect_false(any(grepl("PID", out)))
})

test_that("osrm_stop works with osrm_server object", {
  skip_if_not_installed("processx")
  
  # Create a mock that tracks if kill was called
  killed <- FALSE
  server <- structure(
    list(
      is_alive = function() TRUE,
      get_pid = function() 12345,
      kill = function() { killed <<- TRUE; TRUE },
      wait = function(...) TRUE
    ),
    class = c("osrm_server", "process", "list")
  )
  
  # Mock the registry state so osrm_stop can find it if needed
  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_reg <- state_env$registry
  on.exit(state_env$registry <- orig_reg, add = TRUE)
  
  state_env$registry <- list(
    "test" = list(id = "test", pid = 12345, port = 5001, proc = server)
  )
  
  osrm_stop(server = server, quiet = TRUE)
  
  expect_true(killed)
})

# ROBUST SKIP: Check if OSRM is actually available before running these tests.
# This prevents crashes locally AND ensures CRAN compliance.
# We check both the Option (set by setup) and the PATH.
osrm_exec_opt <- getOption("osrm.routed.exec", "")
has_osrm_option <- nzchar(osrm_exec_opt) && nzchar(Sys.which(osrm_exec_opt))
has_osrm_path <- nzchar(Sys.which("osrm-routed"))

if (!has_osrm_option && !has_osrm_path) {
  testthat::skip("OSRM binary not found (Skipping Server Tests)")
}

# Helper for logging tests - creates MockProcess stub
create_mock_process <- function(captured_env) {
  list(
    new = function(command, args, ..., stdout, stderr) {
      captured_env$captured <- list(
        command = command,
        args = args,
        stdout = stdout,
        stderr = stderr
      )
      structure(
        list(
          is_alive = function() TRUE,
          get_pid = function() 12345,
          kill = function() TRUE,
          wait = function(...) TRUE
        ),
        class = c("process", "list")
      )
    }
  )
}

test_that("osrm_start_server launches osrm-routed with correct arguments", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  captured <- list()

  # Mock processx::process
  # Mock processx::process without R6 dependency
  MockProcess <- list(
    new = function(command, args, ...) {
      captured <<- list(command = tools::file_path_sans_ext(basename(command)), args = args)
      structure(
        list(
          is_alive = function() TRUE,
          get_pid = function() 12345,
          kill = function() TRUE,
          wait = function(...) TRUE
        ),
        class = c("process", "list")
      )
    }
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      server <- osrm_start_server(
        osrm_path = osrm_path,
        algorithm = "MLD",
        port = 5002L,
        threads = 4L,
        max_table_size = 500L,
        quiet = TRUE
      )
    },
    process = MockProcess,
    .package = "processx"
  )

  expect_equal(tools::file_path_sans_ext(basename(captured$command)), "osrm-routed")
  expect_true("-a" %in% captured$args && "MLD" %in% captured$args)
  expect_true("-p" %in% captured$args && "5002" %in% captured$args)
  expect_true("-t" %in% captured$args && "4" %in% captured$args)
  expect_true("--max-table-size" %in% captured$args && "500" %in% captured$args)
})

test_that("osrm_start_server validates input file extension", {
  tmp_file <- "test.txt"
  file.create(tmp_file)
  on.exit(unlink(tmp_file))

  expect_error(
    osrm_start_server(tmp_file),
    "must end in .osrm.mldgr or .osrm.hsgr"
  )
})

test_that("osrm_stop handles stopping by object, id, port, and pid", {
  skip_if_not_installed("processx")
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  # Mock registry state
  mock_reg <- list(
    "server1" = list(id = "server1", pid = 1001L, port = 5001L, proc = NULL),
    "server2" = list(id = "server2", pid = 1002L, port = 5002L, proc = NULL)
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  on.exit(state_env$registry <- orig_registry, add = TRUE)
  state_env$registry <- mock_reg

  with_mocked_bindings(
    {
      # Stop by ID
      res1 <- osrm_stop(id = "server1", quiet = TRUE)
      expect_equal(res1$id, "server1")

      # Stop by Port
      res2 <- osrm_stop(port = 5002, quiet = TRUE)
      expect_equal(res2$port, 5002)

      # Stop non-existent
      expect_error(
        osrm_stop(id = "server3", quiet = TRUE),
        "Could not identify a server"
      )
    },
    .osrm_deregister = function(...) TRUE,
    .osrm_pid_is_running = function(...) TRUE,
    .osrm_kill_pid = function(...) TRUE
  )
})

# Additional tests for osrm_servers() ----
test_that("osrm_servers returns empty data frame when no servers", {
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  on.exit(state_env$registry <- orig_registry, add = TRUE)
  state_env$registry <- list()

  result <- osrm_servers()

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
  expect_named(
    result,
    c(
      "id",
      "pid",
      "port",
      "algorithm",
      "started_at",
      "alive",
      "has_handle",
      "log",
      "input_osm",
      "center_lon",
      "center_lat"
      ))
})

test_that("osrm_servers returns server information correctly", {
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  mock_proc <- list(
    is_alive = function() TRUE,
    get_pid = function() 1001L
  )
  class(mock_proc) <- c("process", "list")

  mock_reg <- list(
    "server1" = list(
      id = "server1",
      pid = 1001L,
      port = 5001L,
      algorithm = "mld",
      started_at = "2025-01-01T12:00:00.000Z",
      input_osm = "data1.osm.pbf",
      proc = mock_proc
    ),
    "server2" = list(
      id = "server2",
      pid = 1002L,
      port = 5002L,
      algorithm = "ch",
      started_at = "2025-01-01T12:00:01.000Z",
      input_osm = "data2.osm.pbf",
      proc = NULL
    )
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  on.exit(state_env$registry <- orig_registry, add = TRUE)
  state_env$registry <- mock_reg

  with_mocked_bindings(
    {
      result <- osrm_servers()

      expect_equal(nrow(result), 2)
      expect_equal(result$id, c("server1", "server2"))
      expect_equal(result$pid, c(1001L, 1002L))
      expect_equal(result$port, c(5001L, 5002L))
      expect_equal(result$algorithm, c("mld", "ch"))
      expect_equal(result$input_osm, c("data1.osm.pbf", "data2.osm.pbf"))
      expect_true(result$has_handle[1])
      expect_false(result$has_handle[2])
    },
    .osrm_pid_is_running = function(...) TRUE
  )
})

# Tests for osrm_stop_all() ----
test_that("osrm_stop_all stops all servers", {
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  mock_reg <- list(
    "server1" = list(id = "server1", pid = 1001L, port = 5001L, proc = NULL),
    "server2" = list(id = "server2", pid = 1002L, port = 5002L, proc = NULL)
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  on.exit(state_env$registry <- orig_registry, add = TRUE)
  state_env$registry <- mock_reg

  stop_calls <- character()

  with_mocked_bindings(
    {
      result <- osrm_stop_all()

      expect_equal(result, 2L)
      expect_length(stop_calls, 2)
    },
    osrm_stop = function(id, quiet) {
      stop_calls <<- c(stop_calls, id)
      list(id = id, stopped = TRUE)
    }
  )
})

test_that("osrm_stop_all returns 0 when no servers", {
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  on.exit(state_env$registry <- orig_registry, add = TRUE)
  state_env$registry <- list()

  result <- osrm_stop_all()
  expect_equal(result, 0L)
})

# Tests for server registry internal functions ----
test_that("registry saves and loads correctly", {
  skip_if_not_installed("jsonlite")
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempfile()
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$session_id <- "session-test-12345"
  state_env$registry <- list(
    "test-server" = list(
      id = "test-server",
      pid = 1234L,
      port = 5001L,
      algorithm = "mld",
      started_at = "2025-01-01T12:00:00.000Z",
      prefix = "/tmp/test.osrm"
    )
  )

  with_mocked_bindings(
    {
      # Test save
      .osrm_registry_save()

      # Check file exists
      registry_path <- file.path(tmp_dir, "session-test-12345.json")
      expect_true(file.exists(registry_path))

      # Test load
      .osrm_state$registry <- list() # Clear registry
      .osrm_registry_load()

      # Check loaded correctly (excluding proc field which is NULL after load)
      expect_true("test-server" %in% names(.osrm_state$registry))
      expect_equal(.osrm_state$registry$`test-server`$pid, 1234L)
    },
    .osrm_registry_dir = function() tmp_dir,
    .osrm_cleanup_orphans = function() invisible(NULL)
  )
})

# Tests for osrm_start_server error validation ----
test_that("osrm_start_server validates input file extension", {
  tmp_file <- "test.txt"
  file.create(tmp_file)
  on.exit(unlink(tmp_file))

  expect_error(
    osrm_start_server(tmp_file),
    "must end in .osrm.mldgr or .osrm.hsgr"
  )
})

test_that("osrm_start_server accepts dataset_name parameter", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  captured <- list()

  MockProcess <- list(
    new = function(command, args, ...) {
      captured <<- list(command = tools::file_path_sans_ext(basename(command)), args = args)
      structure(
        list(
          is_alive = function() TRUE,
          get_pid = function() 12345,
          kill = function() TRUE,
          wait = function(...) TRUE
        ),
        class = c("process", "list")
      )
    }
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      server <- osrm_start_server(
        osrm_path = osrm_path,
        dataset_name = "my_dataset",
        quiet = TRUE
      )

      expect_true("--dataset-name" %in% captured$args)
      expect_true("my_dataset" %in% captured$args)
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_start_server handles max size parameters", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  captured <- list()

  MockProcess <- list(
    new = function(command, args, ...) {
      captured <<- list(command = tools::file_path_sans_ext(basename(command)), args = args)
      structure(
        list(
          is_alive = function() TRUE,
          get_pid = function() 12345,
          kill = function() TRUE,
          wait = function(...) TRUE
        ),
        class = c("process", "list")
      )
    }
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      server <- osrm_start_server(
        osrm_path = osrm_path,
        max_table_size = 200L,
        max_trip_size = 50L,
        quiet = TRUE
      )

      expect_true("--max-table-size" %in% captured$args)
      expect_true("200" %in% captured$args)
      expect_true("--max-trip-size" %in% captured$args)
      expect_true("50" %in% captured$args)
    },
    process = MockProcess,
    .package = "processx"
  )
})

# Tests for logging configuration ----
test_that("osrm_start_server uses temp file by default for logging", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  captured_env <- new.env()
  captured_env$captured <- list()
  MockProcess <- create_mock_process(captured_env)

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      server <- osrm_start_server(
        osrm_path = osrm_path,
        quiet = TRUE
      )
      on.exit(try(server$kill(), silent = TRUE), add = TRUE)

      # Should use a temp file with .log extension
      expect_type(captured_env$captured$stdout, "character")
      expect_type(captured_env$captured$stderr, "character")
      expect_false(identical(captured_env$captured$stdout, ""))
      expect_match(captured_env$captured$stdout, "\\.log$")
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_start_server routes to console when verbose = TRUE", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  captured_env <- new.env()
  captured_env$captured <- list()
  MockProcess <- create_mock_process(captured_env)

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      server <- osrm_start_server(
        osrm_path = osrm_path,
        verbose = TRUE,
        quiet = TRUE
      )
      on.exit(try(server$kill(), silent = TRUE), add = TRUE)

      # Should route to console (empty string)
      expect_equal(captured_env$captured$stdout, "")
      expect_equal(captured_env$captured$stderr, "")
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_start_server uses osrm.server.log_file option when set (character)", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  custom_log <- file.path(tmp_dir, "custom_osrm.log")

  captured_env <- new.env()
  captured_env$captured <- list()
  MockProcess <- create_mock_process(captured_env)

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Set custom log file option
      old_opt <- options(osrm.server.log_file = custom_log)
      on.exit(options(old_opt), add = TRUE)

      server <- osrm_start_server(
        osrm_path = osrm_path,
        quiet = TRUE
      )
      on.exit(try(server$kill(), silent = TRUE), add = TRUE)

      # Should use the custom log path
      expect_equal(
        normalizePath(captured_env$captured$stdout, mustWork = FALSE, winslash = "/"),
        normalizePath(custom_log, mustWork = FALSE, winslash = "/")
      )
      expect_equal(
        normalizePath(captured_env$captured$stderr, mustWork = FALSE, winslash = "/"),
        normalizePath(custom_log, mustWork = FALSE, winslash = "/")
      )
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_start_server falls back to temp file when list option is used", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  captured_env <- new.env()
  captured_env$captured <- list()
  MockProcess <- create_mock_process(captured_env)

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Set list option (now deprecated) - should fall back to temp file
      old_opt <- options(
        osrm.server.log_file = list(
          stdout = "/tmp/out.log",
          stderr = "/tmp/err.log"
        )
      )
      on.exit(options(old_opt), add = TRUE)

      server <- osrm_start_server(
        osrm_path = osrm_path,
        quiet = TRUE
      )
      on.exit(try(server$kill(), silent = TRUE), add = TRUE)

      # Should fall back to temp file (not use the list paths)
      expect_type(captured_env$captured$stdout, "character")
      expect_type(captured_env$captured$stderr, "character")
      expect_false(identical(captured_env$captured$stdout, "/tmp/out.log"))
      expect_false(identical(captured_env$captured$stderr, "/tmp/err.log"))
      expect_match(captured_env$captured$stdout, "\\.log$")
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_start_server reads log file on startup failure", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  # Create a mock log file with error content
  mock_log_content <- c(
    "[info] Starting OSRM server...",
    "[error] Port 5001 is already in use",
    "[error] Failed to bind to socket"
  )
  mock_log_file <- file.path(tmp_dir, "mock_error.log")
  writeLines(mock_log_content, mock_log_file)

  call_count <- 0
  captured_stdout <- NULL

  MockProcess <- list(
    new = function(command, args, ..., stdout, stderr) {
      call_count <<- call_count + 1
      captured_stdout <<- stdout
      structure(
        list(
          is_alive = function() FALSE, # Simulate immediate failure
          get_pid = function() 12345,
          kill = function() TRUE,
          wait = function(...) TRUE,
          get_exit_status = function() 1
        ),
        class = c("process", "list")
      )
    }
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Use the mock log file directly via the option
      old_opt <- options(osrm.server.log_file = mock_log_file)
      on.exit(options(old_opt), add = TRUE)

      expect_error(
        osrm_start_server(osrm_path = osrm_path, quiet = TRUE),
        "Port 5001 is already in use"
      )
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("verbose = TRUE takes precedence over osrm.server.log_file option", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  custom_log <- file.path(tmp_dir, "custom_osrm.log")

  captured_env <- new.env()
  captured_env$captured <- list()
  MockProcess <- create_mock_process(captured_env)

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Set custom log file option but also verbose = TRUE
      old_opt <- options(osrm.server.log_file = custom_log)
      on.exit(options(old_opt), add = TRUE)

      server <- osrm_start_server(
        osrm_path = osrm_path,
        verbose = TRUE,
        quiet = TRUE
      )
      on.exit(try(server$kill(), silent = TRUE), add = TRUE)

      # verbose = TRUE should take precedence (console output)
      expect_equal(captured_env$captured$stdout, "")
      expect_equal(captured_env$captured$stderr, "")
    },
    process = MockProcess,
    .package = "processx"
  )
})

# Tests for invalid osrm.server.log_file values ----
test_that("osrm_start_server falls back to temp file for empty string log option", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  captured_env <- new.env()
  captured_env$captured <- list()
  MockProcess <- create_mock_process(captured_env)

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Set empty string option - should fall back to temp file
      old_opt <- options(osrm.server.log_file = "")
      on.exit(options(old_opt), add = TRUE)

      server <- osrm_start_server(
        osrm_path = osrm_path,
        quiet = TRUE
      )
      on.exit(try(server$kill(), silent = TRUE), add = TRUE)

      # Should fall back to temp file
      expect_type(captured_env$captured$stdout, "character")
      expect_type(captured_env$captured$stderr, "character")
      expect_false(identical(captured_env$captured$stdout, ""))
      expect_match(captured_env$captured$stdout, "\\.log$")
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_start_server falls back to temp file for NA log option", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  captured_env <- new.env()
  captured_env$captured <- list()
  MockProcess <- create_mock_process(captured_env)

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Set NA option - should fall back to temp file
      old_opt <- options(osrm.server.log_file = NA_character_)
      on.exit(options(old_opt), add = TRUE)

      server <- osrm_start_server(
        osrm_path = osrm_path,
        quiet = TRUE
      )
      on.exit(try(server$kill(), silent = TRUE), add = TRUE)

      # Should fall back to temp file
      expect_type(captured_env$captured$stdout, "character")
      expect_type(captured_env$captured$stderr, "character")
      expect_match(captured_env$captured$stdout, "\\.log$")
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_start_server falls back to temp file for multiple paths log option", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  captured_env <- new.env()
  captured_env$captured <- list()
  MockProcess <- create_mock_process(captured_env)

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Set multiple paths - should fall back to temp file
      old_opt <- options(
        osrm.server.log_file = c("/tmp/log1.log", "/tmp/log2.log")
      )
      on.exit(options(old_opt), add = TRUE)

      server <- osrm_start_server(
        osrm_path = osrm_path,
        quiet = TRUE
      )
      on.exit(try(server$kill(), silent = TRUE), add = TRUE)

      # Should fall back to temp file (not use the list paths)
      expect_type(captured_env$captured$stdout, "character")
      expect_type(captured_env$captured$stderr, "character")
      expect_false(identical(captured_env$captured$stdout, "/tmp/log1.log"))
      expect_false(identical(captured_env$captured$stderr, "/tmp/log2.log"))
      expect_match(captured_env$captured$stdout, "\\.log$")
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_start_server falls back to temp file for numeric log option", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  captured_env <- new.env()
  captured_env$captured <- list()
  MockProcess <- create_mock_process(captured_env)

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Set numeric option - should fall back to temp file
      old_opt <- options(osrm.server.log_file = 12345)
      on.exit(options(old_opt), add = TRUE)

      server <- osrm_start_server(
        osrm_path = osrm_path,
        quiet = TRUE
      )
      on.exit(try(server$kill(), silent = TRUE), add = TRUE)

      # Should fall back to temp file
      expect_type(captured_env$captured$stdout, "character")
      expect_type(captured_env$captured$stderr, "character")
      expect_match(captured_env$captured$stdout, "\\.log$")
    },
    process = MockProcess,
    .package = "processx"
  )
})

# Tests for error handling robustness in read_last_n_lines ----
test_that("osrm_start_server handles log file read errors gracefully during failure reporting", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  # Create a log file that will be deleted during error reporting
  mock_log_file <- file.path(tmp_dir, "disappearing.log")

  log_deleted <- FALSE

  MockProcess <- list(
    new = function(command, args, ..., stdout, stderr) {
      # Create initial log content
      writeLines(c("[error] Test error"), mock_log_file)

      structure(
        list(
          is_alive = function() {
            # Delete the log file when is_alive is checked
            # This simulates the file being deleted between creation and read
            if (!log_deleted && file.exists(mock_log_file)) {
              unlink(mock_log_file)
              log_deleted <<- TRUE
            }
            FALSE # Simulate immediate failure
          },
          get_pid = function() 12345,
          kill = function() TRUE,
          wait = function(...) TRUE,
          get_exit_status = function() 1
        ),
        class = c("process", "list")
      )
    }
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Use the mock log file
      old_opt <- options(osrm.server.log_file = mock_log_file)
      on.exit(options(old_opt), add = TRUE)

      # Should still get an error, but not crash due to log reading failure
      expect_error(
        osrm_start_server(osrm_path = osrm_path, quiet = TRUE),
        "osrm-routed failed to start"
      )

      # The error should be caught gracefully, not propagate file operation errors
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_start_server handles permission errors when reading log file", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_on_os("windows") # Permission changes work differently on Windows
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  # Create a log file with no read permissions
  mock_log_file <- file.path(tmp_dir, "no_read_perm.log")

  MockProcess <- list(
    new = function(command, args, ..., stdout, stderr) {
      # Create log file and make it unreadable
      writeLines(c("[error] Permission test"), mock_log_file)
      Sys.chmod(mock_log_file, mode = "000")

      structure(
        list(
          is_alive = function() FALSE, # Simulate immediate failure
          get_pid = function() 12345,
          kill = function() TRUE,
          wait = function(...) TRUE,
          get_exit_status = function() 1
        ),
        class = c("process", "list")
      )
    }
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Use the mock log file
      old_opt <- options(osrm.server.log_file = mock_log_file)
      on.exit(
        {
          options(old_opt)
          # Restore permissions so cleanup can work
          if (file.exists(mock_log_file)) {
            try(Sys.chmod(mock_log_file, mode = "644"), silent = TRUE)
            unlink(mock_log_file)
          }
        },
        add = TRUE
      )

      # Should still get an error message, even though log file is unreadable
      expect_error(
        osrm_start_server(osrm_path = osrm_path, quiet = TRUE),
        "osrm-routed failed to start"
      )
    },
    process = MockProcess,
    .package = "processx"
  )
})

test_that("osrm_start_server handles empty log file during error reporting", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(
    packageVersion("testthat") < "3.2.0",
    "Requires testthat >= 3.2.0 for object mocking"
  )

  tmp_dir <- tempdir()
  osrm_path <- file.path(tmp_dir, "test.osrm.mldgr")
  file.create(osrm_path)
  on.exit(unlink(osrm_path), add = TRUE)

  # Create an empty log file
  mock_log_file <- file.path(tmp_dir, "empty.log")

  MockProcess <- list(
    new = function(command, args, ..., stdout, stderr) {
      # Create empty log file
      file.create(mock_log_file)

      structure(
        list(
          is_alive = function() FALSE, # Simulate immediate failure
          get_pid = function() 12345,
          kill = function() TRUE,
          wait = function(...) TRUE,
          get_exit_status = function() 1
        ),
        class = c("process", "list")
      )
    }
  )

  state_env <- asNamespace("osrm.backend")$.osrm_state
  orig_registry <- state_env$registry
  orig_session_id <- state_env$session_id
  on.exit({
    state_env$registry <- orig_registry
    state_env$session_id <- orig_session_id
  }, add = TRUE)

  state_env$registry <- list()
  state_env$session_id <- "test-session"

  with_mocked_bindings(
    {
      # Use the mock log file
      old_opt <- options(osrm.server.log_file = mock_log_file)
      on.exit(options(old_opt), add = TRUE)

      # Should gracefully handle empty log file
      expect_error(
        osrm_start_server(osrm_path = osrm_path, quiet = TRUE),
        "osrm-routed failed to start"
      )
    },
    process = MockProcess,
    .package = "processx"
  )
})

#' Locate the OSRM Installation Used by `osrm.backend`
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Resolves the `osrm-routed` executable available on the current `PATH`
#' (or the override provided via `options(osrm.routed.exec)`). Runs
#' `osrm-routed --version` to verify availability, then prints the directory
#' containing the executable together with the backend version reported by
#' `osrm-routed` so you know what will be used in the current session.
#' @param quiet Logical; if `FALSE` (default), prints information about the located installation. If `TRUE`, suppresses printed output and only returns the information as a list.
#' @return A list with components `executable` (full path to
#'   `osrm-routed`), `directory` (its parent folder), `osrm_version` (character
#'   vector of non-empty lines emitted by `osrm-routed --version`), and the raw
#'   `processx::run` result in `logs`.
#' @export
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   # check which OSRM installation will be used
#'   osrm_which()
#'
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#' }
#' }
osrm_which <- function(quiet = FALSE) {
  osrm_exec <- resolve_osrm_bin("osrm-routed")
  
  # Resolve to full path if it's just a name on PATH
  resolved <- Sys.which(osrm_exec)
  if (!nzchar(resolved)) {
    # If Sys.which failed, check if osrm_exec itself is a valid path
    if (file.exists(osrm_exec)) {
      resolved <- normalizePath(osrm_exec, mustWork = TRUE)
    } else {
      # Still nothing found
      resolved <- ""
    }
  }

  if (!nzchar(resolved)) {
    stop(
      "Cannot find '",
      osrm_exec,
      "' on PATH. Install OSRM manually or with osrm_install() or set options(osrm.routed.exec = '/path/to/osrm-routed') if you already have OSRM installed.",
      call. = FALSE
    )
  }

  result <- processx::run(
    osrm_exec,
    args = "--version",
    echo = FALSE,
    spinner = FALSE,
    echo_cmd = FALSE,
    error_on_status = FALSE
  )

  if (!identical(result$status, 0L)) {
    stop(
      "Running '",
      osrm_exec,
      " --version' failed with status ",
      result$status,
      ".\n",
      result$stderr,
      call. = FALSE
    )
  }

  output <- c(result$stdout, result$stderr)
  output <- output[nzchar(output)]
  version_lines <- character()
  if (length(output)) {
    version_lines <- trimws(unlist(strsplit(
      paste(output, collapse = "\n"),
      "\r?\n"
    )))
    version_lines <- version_lines[nzchar(version_lines)]
  }

  install_dir <- dirname(resolved)

  if (!quiet) {
    message("OSRM installation directory: ", install_dir)
    if (length(version_lines)) {
      message("osrm-backend version: ", version_lines[1])
      if (length(version_lines) > 1) {
        for (extra in version_lines[-1]) {
          message("  ", extra)
        }
      }
    } else {
      message("osrm-backend version: <no output>")
    }
    message(
      "This is the installation that will be used by osrm_start() and other functions."
    )
  }

  return(invisible(list(
    executable = resolved,
    directory = install_dir,
    osrm_version = version_lines,
    logs = result
  )))
}

# Tests for osrm_which.R

test_that("osrm_which returns correct structure", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(Sys.which("osrm-routed") == "", "osrm-routed not available")

  result <- osrm_which(quiet = TRUE)

  expect_type(result, "list")
  expect_named(result, c("executable", "directory", "osrm_version", "logs"))
  expect_true(file.exists(result$executable))
  expect_true(dir.exists(result$directory))
})

test_that("osrm_which quiet parameter suppresses output", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(Sys.which("osrm-routed") == "", "osrm-routed not available")

  expect_silent(osrm_which(quiet = TRUE))
})

test_that("osrm_which non-quiet shows messages", {
  skip_if_not_installed("processx")
  skip_on_cran()
  skip_if(Sys.which("osrm-routed") == "", "osrm-routed not available")

  expect_message(
    osrm_which(quiet = FALSE),
    "OSRM installation directory"
  )
})

test_that("osrm_which errors when osrm-routed not found", {
  # This test works by temporarily setting an option to a non-existent path
  orig_opt <- getOption("osrm.routed.exec")
  on.exit(options(osrm.routed.exec = orig_opt))

  options(osrm.routed.exec = "definitely-not-a-real-osrm-executable-xyz123")

  expect_error(
    osrm_which(),
    "Cannot find '.*definitely-not-a-real-osrm-executable-xyz123'"
  )
})

test_that("osrm_which respects osrm.routed.exec option", {
  skip_if_not_installed("processx")

  orig_opt <- getOption("osrm.routed.exec")
  on.exit(options(osrm.routed.exec = orig_opt))

  # Set to a non-existent executable - should error
  options(osrm.routed.exec = "/custom/path/osrm-routed-fake")

  expect_error(
    osrm_which(),
    "Cannot find '.*osrm-routed-fake'"
  )
})

# # This file is part of the standard setup for testthat.
# # It is recommended that you do not modify it.
# #
# # Where should you do additional test configuration?
# # Learn more about the roles of various files in:
# # * https://r-pkgs.org/testing-design.html#sec-tests-files-overview
# # * https://testthat.r-lib.org/articles/special-files.html

library(testthat)
library(osrm.backend)

test_check("osrm.backend")

#' Clean Up OSRM Files in a Directory
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Remove OSRM-generated files from a directory. This is useful when switching
#' between algorithms (CH and MLD) or when you want to start fresh.
#'
#' @details
#' OSRM creates many `.osrm.*` files during the extract, contract, partition,
#' and customize stages. This function helps clean up these files.
#'
#' **Important:** The CH and MLD algorithms cannot safely coexist in the same
#' directory because the MLD partition stage modifies some extract-stage files.
#' Use this function to clean up before switching algorithms.
#'
#' @param path A string. Path to an OSRM file or directory containing OSRM files.
#'   If a file path is provided (e.g., `data.osm.pbf` or `data.osrm.hsgr`),
#'   the base name will be extracted and all related `.osrm.*` files will be removed.
#' @param keep_osm Logical. If `TRUE` (default), keeps the original `.osm.pbf`
#'   (or `.osm`, `.osm.bz2`) file. If `FALSE`, removes it as well.
#' @param dry_run Logical. If `TRUE`, shows what would be deleted without
#'   actually deleting. Default is `FALSE`.
#' @param quiet Logical. If `TRUE`, suppresses messages. Default is `FALSE`.
#'
#' @return Invisibly returns a character vector of removed file paths.
#'
#' @export
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   # Stage a temporary workspace with placeholder OSRM files
#'   pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
#'   osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
#'   dir.create(osrm_dir, recursive = TRUE)
#'   tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
#'   file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
#'   file.create(
#'     file.path(osrm_dir, "cur.osrm.timestamp"),
#'     file.path(osrm_dir, "cur.osrm.hsgr"),
#'     file.path(osrm_dir, "cur.osrm.mldgr"),
#'     file.path(osrm_dir, "cur.osrm.partition")
#'   )
#'
#'   # Preview what would be deleted
#'   osrm_cleanup(osrm_dir, dry_run = TRUE)
#'
#'   # Clean up OSRM artifacts (keep the OSM file)
#'   osrm_cleanup(osrm_dir)
#'
#'   # Remove everything including the OSM source
#'   osrm_cleanup(osrm_dir, keep_osm = FALSE)
#'
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'   unlink(osrm_dir, recursive = TRUE)
#' }
#' }
osrm_cleanup <- function(
  path,
  keep_osm = TRUE,
  dry_run = FALSE,
  quiet = FALSE
) {
  if (!is.character(path) || length(path) != 1) {
    stop("'path' must be a single string", call. = FALSE)
  }

  if (!file.exists(path) && !dir.exists(path)) {
    # If the exact path doesn't exist, we check if the parent directory exists.
    # This allows us to use osrm_cleanup with a base_name prefix.
    parent_dir <- dirname(path)
    if (!dir.exists(parent_dir)) {
      stop("Directory does not exist: ", parent_dir, call. = FALSE)
    }
  }

  # Determine directory and base name
  if (dir.exists(path)) {
    dir_path <- path
    # Find all unique base names in the directory
    all_files <- list.files(dir_path, full.names = FALSE)
    osrm_files <- all_files[grepl("\\.osrm\\.", all_files)]

    if (length(osrm_files) == 0) {
      if (!quiet) {
        message("No OSRM files found in directory: ", dir_path)
      }
      return(invisible(character(0)))
    }

    # Extract unique base names
    base_names <- unique(sub("\\.osrm\\..*$", "", osrm_files))

    if (length(base_names) > 1) {
      stop(
        "Multiple OSRM file sets found in directory:\n  ",
        paste(base_names, collapse = ", "),
        "\n\nPlease specify a single file path to clean up a specific set.",
        call. = FALSE
      )
    }

    base_name <- base_names[1]
  } else {
    # Path is a file
    dir_path <- dirname(path)
    file_name <- basename(path)

    # Extract base name from file
    if (grepl("\\.osm(\\.pbf|\\.bz2)?$", file_name, ignore.case = TRUE)) {
      base_name <- sub(
        "\\.osm(\\.pbf|\\.bz2)?$",
        "",
        file_name,
        ignore.case = TRUE
      )
    } else if (grepl("\\.osrm\\.", file_name)) {
      base_name <- sub("\\.osrm\\..*$", "", file_name)
    } else {
      # If no extension is recognized, assume the file_name itself is the base_name
      # This is useful when called with a prefix from other functions.
      base_name <- file_name
    }
  }

  # Get all files matching the base name
  all_files <- list.files(dir_path, full.names = TRUE)

  # Find all OSRM files
  osrm_pattern <- paste0("^", gsub("\\.", "\\\\.", base_name), "\\.osrm\\.")
  osrm_files_to_remove <- all_files[grepl(osrm_pattern, basename(all_files))]

  # Optionally include OSM file
  if (!keep_osm) {
    osm_pattern <- paste0(
      "^",
      gsub("\\.", "\\\\.", base_name),
      "\\.osm(\\.pbf|\\.bz2)?$"
    )
    osm_files <- all_files[grepl(
      osm_pattern,
      basename(all_files),
      ignore.case = TRUE
    )]
    osrm_files_to_remove <- c(osrm_files_to_remove, osm_files)
  }

  if (length(osrm_files_to_remove) == 0) {
    if (!quiet) {
      message("No OSRM files found to remove for base name: ", base_name)
    }
    return(invisible(character(0)))
  }

  # Sort for consistent display
  osrm_files_to_remove <- sort(osrm_files_to_remove)

  # Display what will be removed
  if (!quiet || dry_run) {
    if (dry_run) {
      message("DRY RUN: Would remove the following files:")
    } else {
      message("Removing the following OSRM files:")
    }
    for (f in osrm_files_to_remove) {
      message("  - ", basename(f))
    }
    message("\nDirectory: ", dir_path)
    message("Total files: ", length(osrm_files_to_remove))
  }

  # Actually remove files (unless dry run)
  if (!dry_run) {
    removed <- file.remove(osrm_files_to_remove)
    failed <- osrm_files_to_remove[!removed]

    if (length(failed) > 0) {
      warning(
        "Failed to remove ",
        length(failed),
        " file(s):\n  ",
        paste(basename(failed), collapse = ", "),
        call. = FALSE
      )
    }

    if (!quiet) {
      message("\nSuccessfully removed ", sum(removed), " file(s).")
    }

    return(invisible(osrm_files_to_remove[removed]))
  } else {
    return(invisible(osrm_files_to_remove))
  }
}

#' Uninstall OSRM Backend Binaries
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Removes the OSRM backend binaries and optionally clears the `PATH`
#' configuration from the project's `.Rprofile`.
#'
#' @param dest_dir A string specifying the directory from which to remove OSRM binaries.
#'   If `NULL` (the default), the function looks for an installation in the
#'   per-version subdirectories inside `tools::R_user_dir("osrm.backend", which = "cache")`
#'   and removes it. When multiple versions are installed, interactive sessions
#'   that are not `quiet` will be prompted (with a numbered menu and `0` to cancel)
#'   to choose a directory; otherwise, `dest_dir` must be supplied. Ignored if `all = TRUE`.
#' @param clear_path A logical value. If `TRUE` (default), also removes the
#'   `PATH` configuration from the project's `.Rprofile` by calling `osrm_clear_path()`.
#' @param quiet A logical value. If `TRUE`, suppresses informational messages
#'   and confirmation prompts. Defaults to `FALSE`.
#' @param all A logical value. If `TRUE`, removes all OSRM installations found
#'   in the default cache directory. Will prompt for confirmation unless `force = TRUE`.
#'   Defaults to `FALSE`. When `TRUE`, the `dest_dir` parameter is ignored.
#' @param force A logical value. If `TRUE`, skips all confirmation prompts,
#'   enabling non-interactive usage. Defaults to `FALSE`.
#' @return `TRUE` if one or more directories were successfully removed,
#'   and `FALSE` otherwise.
#' @export
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   # Install OSRM temporarily
#'   install_dir <- osrm_install(path_action = "session", quiet = TRUE)
#'
#'   # Uninstall that specific version and clear PATH changes
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'
#'   # If multiple installs exist, remove them all
#'   osrm_uninstall(all = TRUE, force = TRUE, quiet = TRUE)
#' }
#' }
osrm_uninstall <- function(dest_dir = NULL, clear_path = TRUE, quiet = FALSE, all = FALSE, force = FALSE) {
  quiet <- isTRUE(quiet)
  all <- isTRUE(all)
  force <- isTRUE(force)
  default_root <- osrm_default_install_root()
  default_root_display <- osrm_try_normalize_path(default_root)
  auto_selected <- FALSE

  # --- Handle 'all' parameter: remove all installations ---
  if (all) {
    if (!is.null(dest_dir) && !quiet) {
      warning(
        "The 'dest_dir' parameter is ignored when 'all = TRUE'.",
        call. = FALSE
      )
    }

    version_installs <- osrm_detect_installations(default_root)

    if (length(version_installs) == 0) {
      if (!quiet) {
        message("OSRM installation not found under: ", default_root_display)
      }
      if (clear_path) {
        osrm_clear_path(quiet = quiet)
      }
      return(FALSE)
    }

    # Ask for confirmation to remove all
    if (!quiet) {
      message(sprintf("Found %d OSRM installation%s:",
                      length(version_installs),
                      if (length(version_installs) == 1) "" else "s"))
      for (p in version_installs) {
        message("  - ", p)
      }
    }

    proceed_all <- FALSE
    if (!force) {
      if (interactive() && !quiet) {
        ans <- utils::askYesNo(
          sprintf("Do you want to permanently remove all %d installation%s?",
                  length(version_installs),
                  if (length(version_installs) == 1) "" else "s"),
          default = FALSE
        )
        if (isTRUE(ans)) {
          proceed_all <- TRUE
        }
      } else {
        proceed_all <- TRUE
      }
    } else {
      proceed_all <- TRUE
    }

    if (!proceed_all) {
      if (!quiet) {
        message("Uninstall aborted by user.")
      }
      if (clear_path) {
        osrm_clear_path(quiet = quiet)
      }
      return(FALSE)
    }

    # Remove all installations
    any_removed <- FALSE
    for (install_dir in version_installs) {
      # Use OS-native path for file system ops, forward-slash for display
      install_dir_fs <- normalizePath(install_dir, mustWork = FALSE)
      install_dir_display <- normalizePath(install_dir, winslash = "/", mustWork = FALSE)

      if (!dir.exists(install_dir_fs)) {
        if (!quiet) {
          message("Skipping (not found): ", install_dir_display)
        }
        next
      }

      if (!quiet) {
        message("Removing: ", install_dir_display)
      }

      tryCatch(
        {
          unlink(install_dir_fs, recursive = TRUE, force = TRUE)
          if (!quiet) {
            message("  Successfully removed.")
          }
          any_removed <- TRUE
        },
        error = function(e) {
          if (!quiet) {
            warning(
              "Failed to remove ", install_dir_display, ": ",
              e$message,
              call. = FALSE
            )
          }
        }
      )
    }

    if (clear_path) {
      osrm_clear_path(quiet = quiet)
    }

    if (any_removed && !quiet) {
      message("Successfully uninstalled OSRM backend binaries.")
    }

    return(any_removed)
  }

  # --- Original single-installation logic ---
  if (is.null(dest_dir)) {
    version_installs <- osrm_detect_installations(default_root)

    if (length(version_installs) == 0) {
      if (!quiet) {
        message("OSRM installation not found under: ", default_root_display)
      }
      if (clear_path) {
        osrm_clear_path(quiet = quiet)
      }
      return(FALSE)
    }

    if (length(version_installs) > 1) {
      if (interactive() && !quiet) {
        selection <- osrm_prompt_install_selection(version_installs)
        if (is.null(selection)) {
          if (!quiet) {
            message("Uninstall aborted; no selection made.")
          }
          if (clear_path) {
            osrm_clear_path(quiet = quiet)
          }
          return(FALSE)
        }
        dest_dir <- selection
        auto_selected <- TRUE
      } else {
        if (!quiet) {
          message("Multiple OSRM installations detected:")
          for (p in version_installs) {
            message("  - ", p)
          }
          message("Please supply dest_dir to choose which installation to remove.")
        }
        if (clear_path) {
          osrm_clear_path(quiet = quiet)
        }
        return(FALSE)
      }
    }

    if (is.null(dest_dir)) {
      dest_dir <- version_installs[[1]]
      auto_selected <- TRUE
    }
  }

  dest_dir_fs <- normalizePath(dest_dir, mustWork = FALSE)
  dest_dir_display <- normalizePath(dest_dir, winslash = "/", mustWork = FALSE)

  # --- 1. Check if the directory exists ---
  dir_existed <- dir.exists(dest_dir_fs)
  if (!dir_existed) {
    if (!quiet) {
      message("OSRM installation not found at: ", dest_dir_display)
      message("Nothing to uninstall.")
    }
    if (clear_path) {
      osrm_clear_path(quiet = quiet)
    }
    return(FALSE)
  }

  # --- 2. Ask for confirmation for file deletion ---
  if (auto_selected) {
    if (!quiet) {
      message("Uninstalling OSRM from detected installation: ", dest_dir_display)
    }
  }

  proceed <- FALSE
  if (!force) {
    if (!quiet) {
      message("This will permanently remove the OSRM installation from:")
      message(dest_dir_display)
    }
    if (interactive() && !quiet) {
      ans <- utils::askYesNo(
        "Do you want to proceed with deleting files?",
        default = FALSE
      )
      if (isTRUE(ans)) {
        proceed <- TRUE
      }
    } else {
      proceed <- TRUE
    }
  } else {
    proceed <- TRUE
  }

  # --- 3. Perform deletion if confirmed ---
  uninstalled_ok <- FALSE
  if (proceed) {
    if (!quiet) {
      message("Removing directory...")
    }
    tryCatch(
      {
        unlink(dest_dir_fs, recursive = TRUE, force = TRUE)
        if (!quiet) {
          message("Successfully uninstalled OSRM backend binaries.")
        }
        uninstalled_ok <- TRUE
      },
      error = function(e) {
        if (!quiet) {
          warning(
            "An error occurred while trying to remove the directory: ",
            e$message,
            call. = FALSE
          )
        }
      }
    )
  } else {
    if (!quiet) {
      message("Uninstall of binaries aborted by user.")
    }
  }

  # --- 4. Clear path if requested ---
  if (clear_path) {
    osrm_clear_path(quiet = quiet)
  }

  return(uninstalled_ok)
}

#' @noRd
osrm_prompt_install_selection <- function(paths) {
  if (!length(paths)) {
    return(NULL)
  }

  repeat {
    message("Multiple OSRM installations detected:")
    for (i in seq_along(paths)) {
      message(sprintf("  [%d] %s", i, paths[[i]]))
    }
    message("  [0] Cancel uninstall")
    ans <- readline("Enter the number of the installation to remove: ")
    if (!nzchar(ans)) {
      next
    }
    choice <- suppressWarnings(as.integer(ans))
    if (is.na(choice)) {
      message("Please enter a number between 0 and ", length(paths), ".")
      next
    }
    if (choice == 0) {
      return(NULL)
    }
    if (choice < 0 || choice > length(paths)) {
      message("Please enter a number between 0 and ", length(paths), ".")
      next
    }
    return(paths[[choice]])
  }
}

# Tests for osrm_uninstall.R

test_that("osrm_uninstall removes installation when dest_dir is provided", {
  # Create temporary directory structure
  tmp_root <- tempfile()
  dir.create(tmp_root, recursive = TRUE)
  on.exit(unlink(tmp_root, recursive = TRUE), add = TRUE)

  dest_dir <- file.path(tmp_root, "osrm-install")
  dir.create(dest_dir)
  file.create(file.path(dest_dir, "osrm-routed"))

  # Mock osrm_clear_path to avoid .Rprofile manipulation
  # Note: interactive() set via option to avoid base function mocking
  orig_inter <- getOption("testthat.force_interactive")
  options(testthat.force_interactive = FALSE)
  on.exit(options(testthat.force_interactive = orig_inter), add = TRUE)

  # Set non-interactive mode
  options(rlang_interactive = FALSE)

  result <- osrm_uninstall(
    dest_dir = dest_dir,
    clear_path = FALSE,
    quiet = TRUE
  )

  expect_true(result)
  expect_false(dir.exists(dest_dir))
})

test_that("osrm_uninstall auto-detects single installation", {
  tmp_root <- tempfile()
  dir.create(tmp_root, recursive = TRUE)
  on.exit(unlink(tmp_root, recursive = TRUE), add = TRUE)

  install_dir <- file.path(tmp_root, "v5.27.1")
  dir.create(install_dir)
  file.create(file.path(install_dir, "osrm-routed"))

  with_mocked_bindings(
    {
      options(rlang_interactive = FALSE)

      result <- osrm_uninstall(
        dest_dir = NULL,
        clear_path = FALSE,
        quiet = TRUE
      )

      expect_true(result)
      expect_false(dir.exists(install_dir))
    },
    osrm_default_install_root = function() tmp_root,
    osrm_clear_path = function(...) TRUE
  )
})

test_that("osrm_uninstall returns FALSE when no installation found", {
  tmp_root <- tempfile()
  dir.create(tmp_root, recursive = TRUE)
  on.exit(unlink(tmp_root, recursive = TRUE), add = TRUE)

  with_mocked_bindings(
    {
      result <- osrm_uninstall(
        dest_dir = NULL,
        clear_path = FALSE,
        quiet = TRUE
      )

      expect_false(result)
    },
    osrm_default_install_root = function() tmp_root,
    osrm_clear_path = function(...) TRUE
  )
})

test_that("osrm_uninstall returns FALSE when directory doesn't exist", {
  fake_dir <- "/this/path/does/not/exist"

  with_mocked_bindings(
    {
      result <- osrm_uninstall(
        dest_dir = fake_dir,
        clear_path = FALSE,
        quiet = TRUE
      )

      expect_false(result)
    },
    osrm_clear_path = function(...) TRUE
  )
})

test_that("osrm_uninstall calls osrm_clear_path when clear_path=TRUE", {
  tmp_root <- tempfile()
  dir.create(tmp_root, recursive = TRUE)
  on.exit(unlink(tmp_root, recursive = TRUE), add = TRUE)

  clear_path_called <- FALSE

  with_mocked_bindings(
    {
      osrm_uninstall(
        dest_dir = NULL,
        clear_path = TRUE,
        quiet = TRUE
      )

      expect_true(clear_path_called)
    },
    osrm_default_install_root = function() tmp_root,
    osrm_clear_path = function(...) {
      clear_path_called <<- TRUE
      TRUE
    }
  )
})

test_that("osrm_uninstall doesn't call osrm_clear_path when clear_path=FALSE", {
  tmp_root <- tempfile()
  dir.create(tmp_root, recursive = TRUE)
  on.exit(unlink(tmp_root, recursive = TRUE), add = TRUE)

  install_dir <- file.path(tmp_root, "v5.27.1")
  dir.create(install_dir)
  file.create(file.path(install_dir, "osrm-routed"))

  clear_path_called <- FALSE

  with_mocked_bindings(
    {
      options(rlang_interactive = FALSE)

      osrm_uninstall(
        dest_dir = install_dir,
        clear_path = FALSE,
        quiet = TRUE
      )

      expect_false(clear_path_called)
    },
    osrm_clear_path = function(...) {
      clear_path_called <<- TRUE
      TRUE
    }
  )
})

test_that("osrm_uninstall handles multiple installations in non-interactive mode", {
  tmp_root <- tempfile()
  dir.create(tmp_root, recursive = TRUE)
  on.exit(unlink(tmp_root, recursive = TRUE), add = TRUE)

  # Create two version directories
  v1_dir <- file.path(tmp_root, "v5.27.1")
  v2_dir <- file.path(tmp_root, "v6.0.0")
  dir.create(v1_dir)
  dir.create(v2_dir)
  file.create(file.path(v1_dir, "osrm-routed"))
  file.create(file.path(v2_dir, "osrm-routed"))

  with_mocked_bindings(
    {
      options(rlang_interactive = FALSE)

      # Should return FALSE because it can't decide which to uninstall
      result <- osrm_uninstall(
        dest_dir = NULL,
        clear_path = FALSE,
        quiet = TRUE
      )

      expect_false(result)
      expect_true(dir.exists(v1_dir))
      expect_true(dir.exists(v2_dir))
    },
    osrm_default_install_root = function() tmp_root,
    osrm_clear_path = function(...) TRUE
  )
})

test_that("osrm_prompt_install_selection returns NULL for empty paths", {
  result <- osrm_prompt_install_selection(character(0))
  expect_null(result)
})

#' Install OSRM Backend Binaries
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Downloads and installs pre-compiled binaries for the OSRM backend from the
#' official GitHub releases. The function automatically detects the user's
#' operating system and architecture to download the appropriate files. Only the
#' latest v5 release (`v5.27.1`), `v6.0.0`, `v26.4.0` and `v26.4.1` were manually
#' tested and are known to work well; other releases available on GitHub can be
#' installed but are not guranteed to function correctly.
#'
#' @details
#' The function performs the following steps:
#' 1.  Queries the GitHub API to find the specified release of
#'     `Project-OSRM/osrm-backend`.
#' 2.  Identifies the correct binary (`.tar.gz` archive) for the user's OS
#'     (Linux, macOS, or Windows) and architecture (x64, arm64).
#' 3.  Downloads the archive to a temporary location.
#' 4.  Extracts the archive and locates the OSRM executables (e.g.,
#'     `osrm-routed`, `osrm-extract`).
#' 5.  Copies these executables to a local directory (defaults to
#'     `file.path(tools::R_user_dir("osrm.backend", which = "cache"), <version>)`).
#' 6.  Downloads the matching Lua profiles from the release tarball and installs
#'     them alongside the binaries.
#' 7.  Optionally modifies the `PATH` environment variable for the current
#'     session or project.
#'
#' macOS users should note that upstream OSRM v6.x (and newer) binaries are
#' built for macOS 15.0 (Sequoia, Darwin 24.0.0) or newer. `osrm_install()`
#' automatically blocks v6+ installs on older macOS releases and, when
#' `version = "latest"`, selects the most recent v5 build instead while warning
#' about the requirement. Warnings include both the marketing version and Darwin
#' kernel so you'll see messages like `macOS 13 Ventura [Darwin 22.6.0]`.
#'
#' When installing OSRM v6.x or newer for Windows, the upstream release omits
#' the Intel Threading Building Blocks (TBB) runtime and a compatible `bz2` DLL.
#' To keep the executables runnable out of the box, `osrm_install()` fetches TBB
#' from \href{https://github.com/uxlfoundation/oneTBB/releases/tag/v2022.3.0}{oneTBB
#' v2022.3.0} and the BZip2 runtime from
#' \href{https://github.com/philr/bzip2-windows/releases/tag/v1.0.8.0}{bzip2-windows
#' v1.0.8.0}, verifying their SHA-256 checksums before extraction. Without these
#' extra libraries, the OSRM v6+ binaries shipped for Windows cannot start.
#'
#' On macOS, OSRM v6.x and newer binaries also miss the bundled TBB runtime.
#' The installer reuses the libraries from release `v5.27.1` to keep the
#' binaries functional and patches their `libbz2` linkage using
#' `install_name_tool` so that they load the system-provided BZip2 runtime.
#'
#' Power users (including package authors running cross-platform tests) can
#' override the auto-detected platform by setting the R options
#' `osrm.backend.override_os` and `osrm.backend.override_arch` (e.g.,
#' `options(osrm.backend.override_os = "linux", osrm.backend.override_arch = "arm64")`)
#' before calling `osrm_install()`. Overrides allow requesting binaries for any
#' OS and CPU combination that exists on the GitHub releases.
#'
#' @param version A string specifying the OSRM version tag to install.
#'   Defaults to `"latest"`. Use `"latest"` to automatically find the most
#'   recent stable version (internally calls [osrm_check_latest_version()]). Versions
#'   other than `v5.27.1`, `v6.0.0`, `v26.4.0` and `v26.4.1` will trigger a
#'   warning but are still attempted if binaries are available.
#' @param dest_dir A string specifying the directory where OSRM binaries should be
#'   installed. If `NULL` (the default), a user-friendly, persistent location is
#'   chosen via `tools::R_user_dir("osrm.backend", which = "cache")`, and the
#'   binaries are installed into a subdirectory named after the OSRM version
#'   (e.g. `.../cache/v26.4.1`).
#' @param force A logical value. If `TRUE`, reinstall OSRM even if it's already
#'   found in `dest_dir`. If `FALSE` (default), the function will stop if an
#'   existing installation is detected.
#' @param path_action A string specifying how to handle the system `PATH`. One of:
#'   \itemize{
#'     \item `"session"` (default): Adds the OSRM bin directory to the `PATH`
#'       for the current R session only.
#'     \item `"project"`: Modifies the `.Rprofile` in the current project to set
#'       the `PATH` for all future sessions in that project.
#'     \item `"none"`: Does not modify the `PATH`.
#'   }
#' @param quiet A logical value. If `TRUE`, suppresses installer messages and
#'   warnings. Defaults to `FALSE`.
#' @return The path to the installation directory.
#' @export
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   old <- setwd(tempdir())
#'   on.exit(setwd(old), add = TRUE)
#'
#'   # Install the default stable version and set PATH for this session
#'   install_dir <- osrm_install(path_action = "session", quiet = TRUE)
#'
#'   # Install for a project non-interactively (e.g., in a script)
#'   osrm_install(path_action = "project", quiet = TRUE, force = TRUE)
#'
#'   # Clean up the project's .Rprofile and uninstall binaries
#'   osrm_clear_path(quiet = TRUE)
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#' }
#' }
osrm_install <- function(
  version = "latest",
  dest_dir = NULL,
  force = FALSE,
  path_action = c("session", "project", "none"),
  quiet = FALSE
) {
  quiet <- isTRUE(quiet)
  emit_message <- function(...) {
    if (!quiet) {
      message(...)
    }
  }
  emit_warning <- function(...) {
    if (!quiet) {
      warning(...)
    }
  }
  path_action <- match.arg(path_action)
  sys_info <- Sys.info()
  platform <- get_platform_info(sys_info = sys_info)
  mac_release_info <- get_macos_release_info(sys_info)

  # --- 1. Determine installation root (final directory depends on version) ---
  using_default_dest <- is.null(dest_dir)
  install_root <- if (using_default_dest) {
    osrm_default_install_root()
  } else {
    dest_dir
  }

  # --- 2. Determine version and get release info ---
  tested_versions <- c("v5.27.1", "v6.0.0", "v26.4.0", "v26.4.1")
  requested_version <- version
  mac_release_display <- mac_release_info$display_name
  if (is.null(mac_release_display) || !nzchar(mac_release_display)) {
    raw_release <- mac_release_info$release
    if (!is.na(raw_release) && nzchar(raw_release)) {
      mac_release_display <- paste0("Darwin ", raw_release)
    } else {
      mac_release_display <- "unknown macOS release"
    }
  }
  mac_required_display <- "macOS 15.0 (Sequoia) [Darwin 24]"
  is_macos <- identical(platform$os, "darwin")
  mac_too_old_for_v6 <- is_macos &&
    isFALSE(mac_release_info$meets_v6_requirement)

  if (identical(version, "latest")) {
    emit_message("Finding latest stable version with available binaries...")
    if (mac_too_old_for_v6) {
      version <- find_latest_pre_v6_release(platform)
      emit_message("Latest compatible version is '", version, "'")
      emit_warning(
        sprintf(
          "%s detected. OSRM v6.x requires %s or newer. Upgrade macOS to install v6.",
          mac_release_display,
          mac_required_display
        ),
        call. = FALSE
      )
    } else {
      version <- osrm_check_latest_version()
      emit_message("Latest stable version is '", version, "'")
    }
  }

  if (mac_too_old_for_v6 && version_at_least(version, "v6.0.0")) {
    stop(
      sprintf(
        "%s detected. OSRM v6.x requires %s or newer. Please install a v5.x release instead.",
        mac_release_display,
        mac_required_display
      ),
      call. = FALSE
    )
  }
  if (!version %in% tested_versions) {
    warning_message <- sprintf(
      paste(
        "Version '%s' has not been validated by osrm.backend;",
        "only %s are tested."
      ),
      version,
      paste(tested_versions, collapse = ", ")
    )
    if (identical(requested_version, "latest")) {
      warning_message <- sprintf(
        paste(
          "Latest available release '%s' has not been validated by",
          "osrm.backend; only %s are tested."
        ),
        version,
        paste(tested_versions, collapse = ", ")
      )
    }
    emit_warning(warning_message, call. = FALSE)
  }

  # Validate requested tag exists for this platform before hitting the API.
  if (!identical(requested_version, "latest")) {
    available_versions <- tryCatch(
      osrm_check_available_versions(prereleases = TRUE),
      error = identity
    )

    if (inherits(available_versions, "error")) {
      emit_warning(
        sprintf(
          paste(
            "Unable to verify requested version '%s' against",
            "available releases: %s"
          ),
          version,
          available_versions$message
        ),
        call. = FALSE
      )
    } else if (!version %in% available_versions) {
      stop(
        sprintf(
          paste(
            "Version '%s' is not available for this platform.",
            "Run osrm_check_available_versions(prereleases = TRUE) to list",
            "supported tags."
          ),
          version
        ),
        call. = FALSE
      )
    }
  }

  # --- 3. Resolve final destination directory for this version ---
  if (using_default_dest) {
    dest_dir <- file.path(install_root, version)
  } else {
    dest_dir <- install_root
  }
  if (!dir.exists(dest_dir)) {
    dir.create(dest_dir, recursive = TRUE, showWarnings = FALSE)
  }
  dest_dir <- normalizePath(dest_dir, mustWork = FALSE)

  bin_candidates <- file.path(dest_dir, c("osrm-routed", "osrm-routed.exe"))
  has_existing_install <- any(file.exists(bin_candidates))
  profiles_dir <- file.path(dest_dir, "profiles")
  profiles_missing <- !dir.exists(profiles_dir)

  if (has_existing_install) {
    if (!force && !profiles_missing) {
      emit_message("OSRM backend already found in: ", dest_dir)
      emit_message("Use force = TRUE to reinstall.")
      # Handle path and return
      handle_path_setting(path_action, dest_dir, quiet)
      return(dest_dir)
    }

    if (profiles_missing) {
      emit_message(
        "Existing OSRM installation in ",
        dest_dir,
        " is missing Lua profiles. Reinstalling to restore them."
      )
    } else if (force) {
      emit_message("force = TRUE; reinstalling OSRM backend in ", dest_dir)
    }
  }

  # --- 4. Fetch release metadata ---
  release_info <- get_release_by_tag(version)

  # --- 5. Detect platform ---
  emit_message(sprintf("Detected platform: %s-%s", platform$os, platform$arch))
  emit_message(sprintf(
    "Found release: %s (%s)",
    release_info$tag_name,
    release_info$name
  ))

  # --- 6. Find matching asset download URL ---
  asset_url <- find_asset_url(release_info, platform)
  emit_message("Found matching binary: ", basename(asset_url))

  # --- 7. Download and extract ---
  emit_message("Downloading from ", asset_url)
  tmp_file <- tempfile(fileext = ".tar.gz")
  on.exit(unlink(tmp_file), add = TRUE)

  tryCatch(
    {
      req <- httr2::req_retry(httr2::request(asset_url), max_tries = 3)
      httr2::req_perform(req, path = tmp_file)
    },
    error = function(e) {
      stop("Failed to download file: ", e$message, call. = FALSE)
    }
  )

  tmp_extract_dir <- tempfile()
  dir.create(tmp_extract_dir)
  on.exit(unlink(tmp_extract_dir, recursive = TRUE), add = TRUE)
  emit_message("Extracting binaries...")
  tryCatch(
    utils::untar(tmp_file, exdir = tmp_extract_dir),
    error = function(e) {
      stop("Failed to extract archive: ", e$message, call. = FALSE)
    }
  )

  # --- 8. Locate and install binaries ---
  bin_regex <- "osrm-(routed|extract|contract|partition|customize|datastore)"

  all_files <- list.files(tmp_extract_dir, recursive = TRUE, full.names = TRUE)
  found_bins <- all_files[
    grepl(
      paste0("^", bin_regex, "(\\.exe)?$"),
      basename(all_files),
      ignore.case = TRUE
    )
  ]

  # Fallback for archives that place binaries alongside additional artefacts (e.g. node bindings)
  if (length(found_bins) == 0) {
    found_bins <- all_files[
      grepl(
        paste0(bin_regex, "(\\.exe)?$"),
        basename(all_files),
        ignore.case = TRUE
      )
    ]
  }

  if (length(found_bins) == 0) {
    stop(
      paste(
        "Could not find OSRM binaries in the downloaded archive.",
        "The archive structure may have changed."
      ),
      call. = FALSE
    )
  }

  bin_source_dir <- dirname(found_bins[1])
  files_to_copy <- list.files(bin_source_dir, full.names = TRUE)

  emit_message("Installing binaries to ", dest_dir)
  file.copy(from = files_to_copy, to = dest_dir, overwrite = TRUE)

  runtime_version <- release_info$tag_name
  if (is.null(runtime_version) || !nzchar(runtime_version)) {
    runtime_version <- version
  }
  maybe_install_windows_v6_runtime(
    runtime_version,
    platform,
    dest_dir,
    quiet = quiet
  )
  maybe_install_linux_v6_runtime(
    runtime_version,
    platform,
    dest_dir,
    quiet = quiet
  )
  maybe_install_macos_v6_runtime(
    runtime_version,
    platform,
    dest_dir,
    quiet = quiet
  )

  # --- 9. Download and install Lua profiles ---
  install_profiles_for_release(release_info, dest_dir, quiet = quiet)

  # --- 10. Set permissions and update PATH ---
  installed_bins <- file.path(dest_dir, basename(files_to_copy))
  if (.Platform$OS.type != "windows") {
    emit_message("Setting executable permissions...")
    Sys.chmod(
      installed_bins[
        grepl(
          paste0("^", bin_regex, "$"),
          basename(installed_bins),
          ignore.case = TRUE
        )
      ],
      mode = "0755"
    )
  }

  handle_path_setting(path_action, dest_dir, quiet)

  # --- 11. Final verification and user message ---
  osrm_path_check <- Sys.which("osrm-routed")
  if (path_action != "none" && !nzchar(osrm_path_check)) {
    emit_warning(
      paste(
        "Installation completed, but 'osrm-routed' was not found on the",
        "PATH immediately. You may need to restart your R session."
      ),
      call. = FALSE
    )
  } else if (path_action != "none") {
    emit_message("Installation successful!")
  } else {
    emit_message("Installation successful! Binaries are in ", dest_dir)
  }

  return(dest_dir)
}

#' Check for the Latest Stable OSRM Version
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Queries the GitHub API to find the most recent stable (non-pre-release)
#' version tag for the OSRM backend that has binaries available for the current platform.
#'
#' @return A string containing the latest version tag (e.g., `"v26.4.1"`).
#' @export
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   # Get the latest stable version number of OSRM backend
#'   osrm_check_latest_version()
#' }
#' }
osrm_check_latest_version <- function() {
  releases <- get_all_releases()
  platform <- get_platform_info()
  for (rel in releases) {
    if (!isTRUE(rel$prerelease) && release_has_binaries(rel, platform)) {
      return(rel$tag_name)
    }
  }
  stop(
    "Could not find any stable releases with binaries for your platform.",
    call. = FALSE
  )
}

#' @noRd
find_latest_pre_v6_release <- function(platform) {
  releases <- get_all_releases()
  for (rel in releases) {
    if (isTRUE(rel$prerelease)) {
      next
    }
    if (!release_has_binaries(rel, platform)) {
      next
    }
    if (version_at_least(rel$tag_name, "v6.0.0")) {
      next
    }
    return(rel$tag_name)
  }

  stop(
    paste(
      "No pre-v6 releases with compatible binaries were found for this",
      "platform. Upgrade macOS to install OSRM v6.x or later."
    ),
    call. = FALSE
  )
}

#' Check for Available OSRM Versions
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Queries the GitHub API to get a list of all available version tags for the
#' OSRM backend that have binaries for the current platform.
#'
#' @param prereleases A logical value. If `TRUE`, include pre-release versions
#'   in the returned list. Defaults to `FALSE`.
#' @return A character vector of available version tags.
#' @export
#' 
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   # Get all stable versions with binaries for this platform
#'   osrm_check_available_versions()
#' }
#' }
osrm_check_available_versions <- function(prereleases = FALSE) {
  releases <- get_all_releases()
  platform <- get_platform_info()

  # Filter for releases that have binaries for the current platform
  releases_with_binaries <- Filter(
    function(rel) release_has_binaries(rel, platform),
    releases
  )

  if (!prereleases) {
    releases_with_binaries <- Filter(
      function(rel) !isTRUE(rel$prerelease),
      releases_with_binaries
    )
  }

  tags <- vapply(
    releases_with_binaries,
    function(rel) rel$tag_name,
    FUN.VALUE = character(1)
  )
  return(tags)
}

#' Clear OSRM Path from Project's .Rprofile
#'
#' Scans the `.Rprofile` file in the current project's root directory and
#' removes any lines that were added by `osrm_install()` to modify the `PATH`.
#'
#' @param quiet A logical value. If `TRUE`, suppresses messages. Defaults to `FALSE`.
#' @return `TRUE` if the file was modified, `FALSE` otherwise.
#' @export
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   # Clean up a temporary project's .Rprofile
#'   old <- setwd(tempdir())
#'   on.exit(setwd(old), add = TRUE)
#'   writeLines(
#'     c(
#'       "#added-by-r-pkg-osrm.backend",
#'       'Sys.setenv(PATH = paste("dummy", Sys.getenv("PATH"), sep = .Platform$path.sep))'
#'     ),
#'     ".Rprofile"
#'   )
#'   osrm_clear_path(quiet = TRUE)
#'   unlink(".Rprofile")
#' }
#' }
osrm_clear_path <- function(quiet = FALSE) {
  r_profile_path <- file.path(getwd(), ".Rprofile")

  if (!file.exists(r_profile_path)) {
    if (!quiet) {
      message("No .Rprofile file found in the current project directory.")
    }
    return(FALSE)
  }

  lines <- readLines(r_profile_path, warn = FALSE)
  comment_tag <- "#added-by-r-pkg-osrm.backend"

  lines_to_keep <- lines[!grepl(comment_tag, lines, fixed = TRUE)]

  if (length(lines_to_keep) == length(lines)) {
    if (!quiet) {
      message(
        ".Rprofile does not contain any paths set by osrm.backend. No changes made."
      )
    }
    return(FALSE)
  }

  if (!quiet) {
    message("Removing osrm.backend PATH configuration from: ", r_profile_path)
  }
  writeLines(lines_to_keep, r_profile_path)
  if (!quiet) {
    message(
      "Successfully removed configuration. Please restart R for changes to take effect."
    )
  }

  return(TRUE)
}


# --- Helper functions (not exported) ---

#' @noRd
osrm_default_install_root <- function() {
  tools::R_user_dir("osrm.backend", which = "cache")
}

#' @noRd
osrm_try_normalize_path <- function(path) {
  if (!file.exists(path)) {
    return(path)
  }
  tryCatch(
    normalizePath(path),
    error = function(...) path
  )
}

#' @noRd
osrm_is_install_dir <- function(path) {
  if (is.null(path) || !length(path)) {
    return(FALSE)
  }
  candidate <- osrm_try_normalize_path(path)
  if (!dir.exists(candidate)) {
    return(FALSE)
  }
  bin_candidates <- file.path(candidate, c("osrm-routed", "osrm-routed.exe"))
  any(file.exists(bin_candidates))
}

#' @noRd
osrm_detect_installations <- function(root = osrm_default_install_root()) {
  root_norm <- osrm_try_normalize_path(root)
  installs <- character(0)

  if (dir.exists(root_norm)) {
    entries <- dir(
      root_norm,
      full.names = TRUE,
      recursive = FALSE,
      include.dirs = TRUE,
      no.. = TRUE
    )
    if (length(entries)) {
      info <- file.info(entries)
      dirs <- entries[!is.na(info$isdir) & info$isdir]
      if (length(dirs)) {
        dir_norms <- vapply(dirs, osrm_try_normalize_path, character(1))
        valid <- vapply(dir_norms, osrm_is_install_dir, logical(1))
        installs <- unique(dir_norms[valid])
      }
    }
  }

  installs
}

#' @noRd
get_platform_info <- function(sys_info = Sys.info()) {
  if (is.null(sys_info)) {
    sys_info <- list()
  }

  sys_info <- as.list(sys_info)
  sysname <- sys_info[["sysname"]]
  if (!is.null(sysname)) {
    sysname <- as.character(sysname)[1]
  }

  machine <- sys_info[["machine"]]
  if (!is.null(machine)) {
    machine <- as.character(machine)[1]
  }

  if (is.null(sysname) || !nzchar(sysname)) {
    os <- NULL
  } else {
    sysname_key <- tolower(sysname)
    os <- switch(
      sysname_key,
      linux = "linux",
      darwin = "darwin",
      windows = "win32"
    )
  }

  normalized_machine <- NULL
  if (!is.null(machine) && isTRUE(nzchar(machine))) {
    normalized_machine <- tolower(machine)
    normalized_machine <- gsub("[^a-z0-9]+", "_", normalized_machine)
    normalized_machine <- gsub("^_+|_+$", "", normalized_machine)
  }

  arch_map <- c(
    x86_64 = "x64",
    amd64 = "x64",
    x64 = "x64",
    x86 = "x86",
    i386 = "x86",
    i686 = "x86",
    aarch64 = "arm64",
    arm64 = "arm64"
  )
  arch <- if (!is.null(normalized_machine)) arch_map[[normalized_machine]]

  override_os <- getOption("osrm.backend.override_os")
  if (!is.null(override_os)) {
    if (length(override_os) != 1 || !is.character(override_os)) {
      override_os <- as.character(override_os)[1]
    }
    override_os <- trimws(override_os)
    if (!nzchar(override_os)) {
      stop(
        "Option 'osrm.backend.override_os' must be a non-empty string when set.",
        call. = FALSE
      )
    }
    os <- override_os
  }

  override_arch <- getOption("osrm.backend.override_arch")
  if (!is.null(override_arch)) {
    if (length(override_arch) != 1 || !is.character(override_arch)) {
      override_arch <- as.character(override_arch)[1]
    }
    override_arch <- trimws(override_arch)
    if (!nzchar(override_arch)) {
      stop(
        "Option 'osrm.backend.override_arch' must be a non-empty string when set.",
        call. = FALSE
      )
    }
    arch <- override_arch
  }

  if (is.null(os) || is.null(arch) || !nzchar(os) || !nzchar(arch)) {
    stop(
      sprintf(
        "Your platform is not supported by pre-compiled OSRM binaries. OS: %s, Arch: %s.",
        if (!is.null(override_os)) {
          override_os
        } else if (!is.null(sysname)) {
          sysname
        } else {
          "unknown"
        },
        if (!is.null(override_arch)) {
          override_arch
        } else if (!is.null(machine)) {
          machine
        } else {
          "unknown"
        }
      ),
      call. = FALSE
    )
  }

  list(os = os, arch = arch)
}

#' @noRd
get_macos_release_info <- function(sys_info) {
  if (is.null(sys_info)) {
    sys_info <- list()
  }
  sys_info <- as.list(sys_info)

  sysname <- sys_info[["sysname"]]
  if (!is.null(sysname) && length(sysname)) {
    sysname <- as.character(sysname)[1]
  }
  if (is.null(sysname) || tolower(sysname) != "darwin") {
    return(list(
      release = NA_character_,
      darwin_major = NA_integer_,
      marketing_version = NA_character_,
      marketing_codename = NA_character_,
      display_name = NULL,
      meets_v6_requirement = TRUE
    ))
  }

  release <- NA_character_
  raw_release <- sys_info[["release"]]
  if (!is.null(raw_release) && length(raw_release)) {
    release_candidate <- as.character(raw_release)[1]
    release_candidate <- trimws(release_candidate)
    if (nzchar(release_candidate)) {
      release <- release_candidate
    }
  }

  darwin_major <- NA_integer_
  meets_requirement <- NA
  if (!is.na(release)) {
    numeric_release <- tryCatch(
      suppressWarnings(as.numeric_version(release)),
      error = function(e) NULL
    )
    if (!is.null(numeric_release) && length(numeric_release)) {
      components <- unclass(numeric_release)[[1]]
      if (length(components)) {
        major_component <- components[1]
        if (!is.na(major_component)) {
          darwin_major <- major_component
          meets_requirement <- major_component >= 24
        }
      }
    }
  }

  darwin_map <- list(
    `14` = list(version = "10.10", codename = "Yosemite"),
    `15` = list(version = "10.11", codename = "El Capitan"),
    `16` = list(version = "10.12", codename = "Sierra"),
    `17` = list(version = "10.13", codename = "High Sierra"),
    `18` = list(version = "10.14", codename = "Mojave"),
    `19` = list(version = "10.15", codename = "Catalina"),
    `20` = list(version = "11", codename = "Big Sur"),
    `21` = list(version = "12", codename = "Monterey"),
    `22` = list(version = "13", codename = "Ventura"),
    `23` = list(version = "14", codename = "Sonoma"),
    `24` = list(version = "15", codename = "Sequoia")
  )

  marketing <- NULL
  if (!is.na(darwin_major)) {
    marketing <- darwin_map[[as.character(darwin_major)]]
  }

  marketing_label <- NULL
  if (!is.null(marketing)) {
    marketing_label <- paste0("macOS ", marketing$version)
    if (!is.null(marketing$codename) && nzchar(marketing$codename)) {
      marketing_label <- paste0(marketing_label, " ", marketing$codename)
    }
  }

  darwin_label <- NULL
  if (!is.na(release)) {
    darwin_label <- paste0("Darwin ", release)
  }

  display_name <- NULL
  if (!is.null(marketing_label) && !is.null(darwin_label)) {
    display_name <- paste0(marketing_label, " [", darwin_label, "]")
  } else if (!is.null(marketing_label)) {
    display_name <- marketing_label
  } else if (!is.null(darwin_label)) {
    display_name <- darwin_label
  }

  list(
    release = release,
    darwin_major = darwin_major,
    marketing_version = if (!is.null(marketing)) {
      marketing$version
    } else {
      NA_character_
    },
    marketing_codename = if (!is.null(marketing)) {
      marketing$codename
    } else {
      NA_character_
    },
    display_name = display_name,
    meets_v6_requirement = if (!is.na(darwin_major)) {
      darwin_major >= 24
    } else {
      meets_requirement
    }
  )
}

#' @noRd
github_api_request_with_retries <- function(url, error_message, verb = "GET") {
  # Base request with error handling and retry policy
  req <- httr2::request(url)
  req <- httr2::req_error(req, body = function(resp) {
    httr2::resp_body_json(resp)$message
  })
  # Retry on GitHub's rate-limiting status (403), as well as the
  # standard 429 (Too Many Requests) and 503 (Service Unavailable).
  req <- httr2::req_retry(
    req,
    max_tries = 4,
    is_transient = ~ httr2::resp_status(.x) %in% c(403, 429, 503)
  )

  # Check for a GitHub Personal Access Token to increase rate limits.
  # This is especially important for CI/CD environments.
  token <- Sys.getenv("GITHUB_PAT")
  if (nzchar(token)) {
    # Add authentication header if token is found
    req <- httr2::req_auth_bearer_token(req, token)
  }

  resp <- tryCatch(
    httr2::req_perform(req),
    error = function(e) {
      stop(error_message, e$message, call. = FALSE)
    }
  )
  return(resp)
}

#' @noRd
get_all_releases <- function() {
  repo <- "Project-OSRM/osrm-backend"
  url <- paste0("https://api.github.com/repos/", repo, "/releases")

  resp <- github_api_request_with_retries(
    url,
    error_message = "GitHub API request failed: "
  )

  httr2::resp_body_json(resp)
}

#' @noRd
get_release_by_tag <- function(version) {
  repo <- "Project-OSRM/osrm-backend"
  url <- paste0(
    "https://api.github.com/repos/",
    repo,
    "/releases/tags/",
    version
  )

  resp <- github_api_request_with_retries(
    url,
    error_message = "GitHub API request failed: "
  )

  httr2::resp_body_json(resp)
}

#' @noRd
release_has_binaries <- function(release, platform) {
  pattern <- sprintf(".*%s-%s-Release\\.tar\\.gz$", platform$os, platform$arch)
  for (asset in release$assets) {
    if (grepl(pattern, asset$name, ignore.case = TRUE)) {
      return(TRUE)
    }
  }
  return(FALSE)
}

#' @noRd
find_asset_url <- function(release_info, platform) {
  pattern <- sprintf(".*%s-%s-Release\\.tar\\.gz$", platform$os, platform$arch)

  matching_assets <- Filter(
    function(asset) grepl(pattern, asset$name, ignore.case = TRUE),
    release_info$assets
  )

  if (!length(matching_assets)) {
    stop(
      "Could not find a compatible binary for your platform in release '",
      release_info$tag_name,
      "'.\nLooked for asset name matching pattern: ",
      pattern,
      call. = FALSE
    )
  }

  node_versions <- vapply(
    matching_assets,
    function(asset) {
      matches <- regexec("node-v([0-9]+)", asset$name, ignore.case = TRUE)
      captured <- regmatches(asset$name, matches)[[1]]
      if (length(captured) >= 2) {
        return(suppressWarnings(as.integer(captured[2])))
      }
      NA_integer_
    },
    integer(1)
  )

  if (all(is.na(node_versions))) {
    selected_asset <- matching_assets[[1]]
  } else {
    max_node <- max(node_versions, na.rm = TRUE)
    idx <- which(node_versions == max_node)[1]
    selected_asset <- matching_assets[[idx]]
  }

  selected_asset$browser_download_url
}

#' @noRd
install_profiles_for_release <- function(
  release_info,
  dest_dir,
  quiet = FALSE
) {
  tarball_url <- release_info$tarball_url

  if (is.null(tarball_url) || !nzchar(tarball_url)) {
    if (!quiet) {
      warning(
        "Release metadata did not include a tarball URL. Skipping installation of profiles.",
        call. = FALSE
      )
    }
    return(invisible(NULL))
  }

  if (!quiet) {
    message("Downloading profiles from release tarball...")
  }
  tmp_tarball <- tempfile(fileext = ".tar.gz")
  on.exit(unlink(tmp_tarball), add = TRUE)

  tryCatch(
    {
      req <- httr2::req_retry(httr2::request(tarball_url), max_tries = 3)
      httr2::req_perform(req, path = tmp_tarball)
    },
    error = function(e) {
      stop(
        "Failed to download source tarball for profiles: ",
        e$message,
        call. = FALSE
      )
    }
  )

  tmp_profiles_extract <- tempfile()
  dir.create(tmp_profiles_extract)
  on.exit(unlink(tmp_profiles_extract, recursive = TRUE), add = TRUE)

  allow_tar_warnings_env <- tolower(Sys.getenv(
    "OSRM_BACKEND_ALLOW_TAR_WARNINGS",
    unset = ""
  ))
  suppress_tar_warnings <- !allow_tar_warnings_env %in% c("1", "true", "yes")
  run_untar <- function(arg_list) {
    if (suppress_tar_warnings) {
      suppressWarnings(do.call(utils::untar, arg_list))
    } else {
      do.call(utils::untar, arg_list)
    }
  }

  list_args <- list(tarfile = tmp_tarball, list = TRUE)
  list_attempts <- list(list_args)
  if (.Platform$OS.type == "windows") {
    list_attempts[[1]]$tar <- "internal"
    list_attempts[[2]] <- list_args
  }

  tar_members <- NULL
  last_list_error <- NULL
  for (args in list_attempts) {
    tar_members <- tryCatch(
      run_untar(args),
      error = function(e) {
        last_list_error <<- e
        NULL
      }
    )
    if (!is.null(tar_members)) {
      break
    }
  }

  if (is.null(tar_members)) {
    stop(
      "Failed to inspect source tarball for profiles: ",
      last_list_error$message,
      call. = FALSE
    )
  }

  # Extract only the top-level 'profiles' subtree shipped with the release.
  top_level_dirs <- unique(sub("/.*$", "", tar_members))
  top_level_dirs <- top_level_dirs[nzchar(top_level_dirs)]
  top_level_dir <- top_level_dirs[[1]]

  prefix <- paste0(top_level_dir, "/")
  relative_members <- tar_members
  matches_prefix <- startsWith(relative_members, prefix)
  relative_members[matches_prefix] <- substr(
    relative_members[matches_prefix],
    nchar(prefix) + 1,
    nchar(relative_members[matches_prefix])
  )

  profile_mask <- startsWith(relative_members, "profiles/") |
    relative_members %in% c("profiles", "profiles/")

  profile_members <- tar_members[profile_mask]

  if (length(profile_members) == 0) {
    stop(
      "The release tarball did not contain a 'profiles' directory.",
      call. = FALSE
    )
  }

  profile_dirs <- unique(
    sub("^(.*?/profiles)(?:/.*)?$", "\\1", profile_members, perl = TRUE)
  )
  profile_dirs <- profile_dirs[
    nzchar(profile_dirs) &
      !grepl("[[:cntrl:]]", profile_dirs, perl = TRUE)
  ]
  extract_members <- profile_members
  extract_members <- extract_members[
    nzchar(extract_members) &
      !grepl("[[:cntrl:]]", extract_members, perl = TRUE)
  ]

  if (!quiet) {
    message("Extracting profiles...")
  }
  untar_args <- list(
    tarfile = tmp_tarball,
    files = extract_members,
    exdir = tmp_profiles_extract
  )
  untar_attempts <- list(untar_args)
  if (.Platform$OS.type == "windows") {
    untar_attempts[[1]]$tar <- "internal"
    untar_attempts[[2]] <- untar_args
  }

  last_extract_error <- NULL
  extracted <- FALSE
  for (args in untar_attempts) {
    tryCatch(
      {
        run_untar(args)
        extracted <- TRUE
      },
      error = function(e) {
        last_extract_error <<- e
      }
    )
    if (isTRUE(extracted)) {
      break
    }
  }

  if (!isTRUE(extracted)) {
    stop(
      "Failed to extract profiles archive: ",
      last_extract_error$message,
      call. = FALSE
    )
  }

  profile_dirs_found <- list.dirs(
    tmp_profiles_extract,
    recursive = TRUE,
    full.names = TRUE
  )
  profile_dirs_found <- profile_dirs_found[
    basename(profile_dirs_found) == "profiles"
  ]
  if (length(profile_dirs_found) == 0) {
    stop(
      "The release tarball did not contain a 'profiles' directory.",
      call. = FALSE
    )
  }
  profile_source <- profile_dirs_found[[which.min(nchar(profile_dirs_found))]]

  dest_profiles_dir <- file.path(dest_dir, "profiles")
  if (dir.exists(dest_profiles_dir)) {
    unlink(dest_profiles_dir, recursive = TRUE)
  }

  copied <- file.copy(
    from = profile_source,
    to = dest_dir,
    recursive = TRUE,
    overwrite = TRUE
  )

  if (!all(copied)) {
    stop(
      "Failed to copy the profiles directory to the destination.",
      call. = FALSE
    )
  }

  if (!quiet) {
    message("Installed profiles to ", dest_profiles_dir)
  }
  invisible(dest_profiles_dir)
}

#' @noRd
maybe_install_windows_v6_runtime <- function(
  version,
  platform,
  dest_dir,
  quiet = FALSE
) {
  if (!identical(platform$os, "win32")) {
    return(invisible(NULL))
  }

  if (is.null(version) || !nzchar(version)) {
    return(invisible(NULL))
  }

  if (!version_at_least(version, "v6.0.0")) {
    return(invisible(NULL))
  }

  install_windows_v6_runtime(dest_dir, quiet = quiet)
}

#' @noRd
maybe_install_linux_v6_runtime <- function(
  version,
  platform,
  dest_dir,
  quiet = FALSE
) {
  if (!identical(platform$os, "linux")) {
    return(invisible(NULL))
  }

  if (!version_at_least(version, "v6.0.0")) {
    return(invisible(NULL))
  }

  ensure_linux_tbb_runtime(dest_dir, platform, quiet = quiet)
}

#' @noRd
install_windows_v6_runtime <- function(dest_dir, quiet = FALSE) {
  if (!quiet) {
    message("Fetching additional Windows runtime dependencies (TBB, BZip2)...")
  }

  tbb_url <- paste0(
    "https://github.com/uxlfoundation/oneTBB/releases/download/v2022.3.0/",
    "oneapi-tbb-2022.3.0-win.zip"
  )
  download_zip_asset(
    url = tbb_url,
    member_path = "oneapi-tbb-2022.3.0/redist/intel64/vc14/tbb12.dll",
    dest_path = file.path(dest_dir, "tbb12.dll"),
    sha256 = "e1b2373f25558bf47d16b4c89cf0a31e6689aaf7221400d209e8527afc7c9eee"
  )
  if (!quiet) {
    message("  - Installed tbb12.dll")
  }

  bzip_url <- paste0(
    "https://github.com/philr/bzip2-windows/releases/download/v1.0.8.0/",
    "bzip2-dll-1.0.8.0-win-x64.zip"
  )
  download_zip_asset(
    url = bzip_url,
    member_path = "libbz2.dll",
    dest_path = file.path(dest_dir, "bz2.dll"),
    sha256 = "50340fece047960f49cf869034c778ff9f6af27dde2f1ea9773cd89ddb326254"
  )
  if (!quiet) {
    message("  - Installed bz2.dll")
  }

  invisible(dest_dir)
}

#' @noRd
maybe_install_macos_v6_runtime <- function(
  version,
  platform,
  dest_dir,
  quiet = FALSE
) {
  if (!identical(platform$os, "darwin")) {
    return(invisible(NULL))
  }

  if (!version_at_least(version, "v6.0.0")) {
    return(invisible(NULL))
  }

  ensure_macos_tbb_runtime(dest_dir, platform, quiet = quiet)
  patch_macos_bzip2_rpath(dest_dir, quiet = quiet)
}

#' @noRd
ensure_linux_tbb_runtime <- function(
  dest_dir,
  platform,
  reference_version = "v5.27.1",
  quiet = FALSE
) {
  required_libs <- c(
    "libtbbmalloc.so.2.3",
    "libtbbmalloc.so.2",
    "libtbbmalloc.so",
    "libtbbmalloc_proxy.so.2.3",
    "libtbbmalloc_proxy.so.2",
    "libtbbmalloc_proxy.so",
    "libtbb.so.12.3",
    "libtbb.so.12",
    "libtbb.so"
  )

  missing_libs <- required_libs[
    !file.exists(file.path(dest_dir, required_libs))
  ]
  if (length(missing_libs)) {
    if (!quiet) {
      message(
        "Fetching Linux TBB runtime components from OSRM release ",
        reference_version,
        "..."
      )
    }
  } else {
    if (!quiet) {
      message(
        "Refreshing Linux TBB runtime components from OSRM release ",
        reference_version,
        "..."
      )
    }
  }

  release_info <- get_release_by_tag(reference_version)
  asset_url <- find_asset_url(release_info, platform)

  tmp_tar <- tempfile(fileext = ".tar.gz")
  on.exit(unlink(tmp_tar), add = TRUE)

  tryCatch(
    {
      req <- httr2::req_retry(httr2::request(asset_url), max_tries = 3)
      httr2::req_perform(req, path = tmp_tar)
    },
    error = function(e) {
      stop(
        "Failed to download Linux TBB runtime from OSRM release ",
        reference_version,
        ": ",
        e$message,
        call. = FALSE
      )
    }
  )

  tmp_extract_dir <- tempfile()
  dir.create(tmp_extract_dir)
  on.exit(unlink(tmp_extract_dir, recursive = TRUE), add = TRUE)

  tryCatch(
    utils::untar(tmp_tar, exdir = tmp_extract_dir),
    error = function(e) {
      stop(
        "Failed to extract OSRM release ",
        reference_version,
        " while retrieving Linux TBB runtime: ",
        e$message,
        call. = FALSE
      )
    }
  )

  extracted_files <- list.files(
    tmp_extract_dir,
    recursive = TRUE,
    full.names = TRUE
  )

  locate_lib <- function(lib) {
    matches <- extracted_files[
      tolower(basename(extracted_files)) == tolower(lib)
    ]
    if (!length(matches)) {
      return(NA_character_)
    }
    matches[[1]]
  }

  lib_sources <- vapply(required_libs, locate_lib, FUN.VALUE = character(1))

  if (anyNA(lib_sources)) {
    missing <- required_libs[is.na(lib_sources)]
    stop(
      "Could not locate required Linux TBB libraries in OSRM release ",
      reference_version,
      ": ",
      paste(missing, collapse = ", "),
      call. = FALSE
    )
  }

  dest_paths <- file.path(dest_dir, required_libs)
  copied <- mapply(
    function(src, dest) {
      if (is.na(src)) {
        return(TRUE)
      }
      file.copy(src, dest, overwrite = TRUE)
    },
    lib_sources,
    dest_paths,
    USE.NAMES = FALSE
  )

  if (!all(copied)) {
    stop(
      "Failed to copy one or more Linux TBB runtime libraries into ",
      dest_dir,
      call. = FALSE
    )
  }

  Sys.chmod(dest_paths, mode = "0644", use_umask = FALSE)

  invisible(dest_dir)
}

#' @noRd
ensure_macos_tbb_runtime <- function(
  dest_dir,
  platform,
  reference_version = "v5.27.1",
  quiet = FALSE
) {
  required_libs <- c(
    "libtbbmalloc.dylib",
    "libtbbmalloc.2.3.dylib",
    "libtbbmalloc_proxy.2.3.dylib",
    "libtbb.dylib",
    "libtbb.12.3.dylib",
    "libtbbmalloc.2.dylib",
    "libtbbmalloc_proxy.dylib",
    "libtbbmalloc_proxy.2.dylib",
    "libtbb.12.dylib"
  )

  missing_libs <- required_libs[
    !file.exists(file.path(dest_dir, required_libs))
  ]
  if (length(missing_libs)) {
    if (!quiet) {
      message(
        "Fetching macOS TBB runtime components from OSRM release ",
        reference_version,
        "..."
      )
    }
  } else {
    if (!quiet) {
      message(
        "Refreshing macOS TBB runtime components from OSRM release ",
        reference_version,
        "..."
      )
    }
  }

  release_info <- get_release_by_tag(reference_version)
  asset_url <- find_asset_url(release_info, platform)

  tmp_tar <- tempfile(fileext = ".tar.gz")
  on.exit(unlink(tmp_tar), add = TRUE)

  tryCatch(
    {
      req <- httr2::req_retry(httr2::request(asset_url), max_tries = 3)
      httr2::req_perform(req, path = tmp_tar)
    },
    error = function(e) {
      stop(
        "Failed to download macOS TBB runtime from OSRM release ",
        reference_version,
        ": ",
        e$message,
        call. = FALSE
      )
    }
  )

  tmp_extract_dir <- tempfile()
  dir.create(tmp_extract_dir)
  on.exit(unlink(tmp_extract_dir, recursive = TRUE), add = TRUE)

  tryCatch(
    utils::untar(tmp_tar, exdir = tmp_extract_dir),
    error = function(e) {
      stop(
        "Failed to extract OSRM release ",
        reference_version,
        " while retrieving macOS TBB runtime: ",
        e$message,
        call. = FALSE
      )
    }
  )

  extracted_files <- list.files(
    tmp_extract_dir,
    recursive = TRUE,
    full.names = TRUE
  )

  locate_lib <- function(lib) {
    matches <- extracted_files[
      tolower(basename(extracted_files)) == tolower(lib)
    ]
    if (!length(matches)) {
      return(NA_character_)
    }
    matches[[1]]
  }

  lib_sources <- vapply(required_libs, locate_lib, FUN.VALUE = character(1))

  if (anyNA(lib_sources)) {
    missing <- required_libs[is.na(lib_sources)]
    stop(
      "Could not locate required TBB libraries in OSRM release ",
      reference_version,
      ": ",
      paste(missing, collapse = ", "),
      call. = FALSE
    )
  }

  dest_paths <- file.path(dest_dir, required_libs)
  copied <- mapply(
    function(src, dest) {
      if (is.na(src)) {
        return(TRUE)
      }
      file.copy(src, dest, overwrite = TRUE)
    },
    lib_sources,
    dest_paths,
    USE.NAMES = FALSE
  )

  if (!all(copied)) {
    stop(
      "Failed to copy one or more macOS TBB runtime libraries into ",
      dest_dir,
      call. = FALSE
    )
  }

  if (!quiet) {
    message("Installed macOS TBB runtime libraries.")
  }
  invisible(dest_dir)
}

#' @noRd
patch_macos_bzip2_rpath <- function(dest_dir, quiet = FALSE) {
  tool_path <- Sys.which("install_name_tool")
  if (!nzchar(tool_path)) {
    stop(
      "Required tool 'install_name_tool' not found in PATH; cannot rewrite macOS BZip2 linkage.",
      call. = FALSE
    )
  }

  binaries <- file.path(
    dest_dir,
    c(
      "osrm-routed",
      "osrm-partition",
      "osrm-extract",
      "osrm-datastore",
      "osrm-customize",
      "osrm-contract",
      "osrm-components"
    )
  )
  binaries <- binaries[file.exists(binaries)]

  if (!length(binaries)) {
    if (!quiet) {
      warning(
        "No OSRM executables found to patch in ",
        dest_dir,
        call. = FALSE
      )
    }
    return(invisible(dest_dir))
  }

  if (!quiet) {
    message("Updating macOS BZip2 linkage...")
  }
  for (bin in binaries) {
    status <- system2(
      tool_path,
      args = c(
        "-change",
        "@rpath/libbz2.1.dylib",
        "/usr/lib/libbz2.1.0.dylib",
        bin
      )
    )

    if (!identical(status, 0L)) {
      stop(
        sprintf(
          "install_name_tool failed for '%s' with exit status %s.",
          basename(bin),
          status
        ),
        call. = FALSE
      )
    }
  }

  invisible(dest_dir)
}

#' @noRd
version_at_least <- function(version, minimum) {
  normalize_version <- function(x) {
    if (is.null(x) || !length(x)) {
      return(NA_character_)
    }
    x <- as.character(x)[1]
    x <- trimws(x)
    if (!nzchar(x)) {
      return(NA_character_)
    }
    x <- tolower(x)
    sub("^v", "", x)
  }

  v_norm <- normalize_version(version)
  min_norm <- normalize_version(minimum)

  if (anyNA(c(v_norm, min_norm))) {
    return(FALSE)
  }

  cmp <- tryCatch(
    suppressWarnings(utils::compareVersion(v_norm, min_norm)),
    error = function(e) NA_integer_
  )

  if (is.na(cmp)) {
    return(FALSE)
  }

  cmp >= 0
}

#' @noRd
# Download a zip archive, verify its checksum, and extract a single file.
download_zip_asset <- function(url, member_path, dest_path, sha256 = NULL) {
  member_path <- gsub("^/+", "", member_path)
  normalized_member <- gsub("\\\\", "/", member_path)

  tmp_zip <- tempfile(fileext = ".zip")
  on.exit(unlink(tmp_zip), add = TRUE)

  req <- httr2::req_retry(httr2::request(url), max_tries = 3)
  tryCatch(
    httr2::req_perform(req, path = tmp_zip),
    error = function(e) {
      stop(
        sprintf("Failed to download '%s': %s", url, e$message),
        call. = FALSE
      )
    }
  )

  if (!is.null(sha256) && nzchar(sha256)) {
    if (!requireNamespace("digest", quietly = TRUE)) {
      stop(
        "Package 'digest' is required to verify download checksums.",
        call. = FALSE
      )
    }
    expected <- tolower(gsub("[^0-9a-f]", "", sha256))
    actual <- digest::digest(file = tmp_zip, algo = "sha256", serialize = FALSE)
    if (!identical(actual, expected)) {
      stop(
        sprintf(
          "SHA-256 checksum mismatch for '%s'. Expected %s but got %s.",
          basename(url),
          expected,
          actual
        ),
        call. = FALSE
      )
    }
  }

  contents <- utils::unzip(tmp_zip, list = TRUE)
  if (is.null(contents) || nrow(contents) == 0) {
    stop(
      sprintf("Archive '%s' did not contain any files.", basename(url)),
      call. = FALSE
    )
  }

  matches <- contents$Name[tolower(contents$Name) == tolower(normalized_member)]
  if (length(matches) == 0) {
    stop(
      sprintf(
        "Archive '%s' did not contain expected file '%s'.",
        basename(url),
        member_path
      ),
      call. = FALSE
    )
  }

  selected_member <- matches[[1]]

  tmp_extract <- tempfile()
  dir.create(tmp_extract)
  on.exit(unlink(tmp_extract, recursive = TRUE), add = TRUE)

  tryCatch(
    utils::unzip(tmp_zip, files = selected_member, exdir = tmp_extract),
    error = function(e) {
      stop(
        sprintf(
          "Failed to extract '%s' from archive '%s': %s",
          selected_member,
          basename(url),
          e$message
        ),
        call. = FALSE
      )
    }
  )

  parts <- strsplit(selected_member, "/", fixed = TRUE)[[1]]
  src_path <- do.call(file.path, c(list(tmp_extract), parts))

  if (!file.exists(src_path)) {
    stop(
      sprintf("Extraction did not produce expected file '%s'.", src_path),
      call. = FALSE
    )
  }

  dest_parent <- dirname(dest_path)
  if (!dir.exists(dest_parent)) {
    dir.create(dest_parent, recursive = TRUE, showWarnings = FALSE)
  }

  if (!file.copy(src_path, dest_path, overwrite = TRUE)) {
    stop(
      sprintf("Failed to copy '%s' to '%s'.", src_path, dest_path),
      call. = FALSE
    )
  }

  invisible(dest_path)
}

#' @noRd
handle_path_setting <- function(action, dir, quiet = FALSE) {
  if (action == "session") {
    set_path_session(dir, quiet)
  } else if (action == "project") {
    set_path_project(dir, quiet)
  }
  # If "none", do nothing
}

#' @noRd
set_path_session <- function(dir, quiet = FALSE) {
  path_sep <- .Platform$path.sep
  current_path <- Sys.getenv("PATH")

  path_elements <- strsplit(current_path, path_sep)[[1]]

  # On Windows, use forward slashes for consistent comparison
  # This handles issues with mixed separators and 8.3 short names
  if (.Platform$OS.type == "windows") {
    normalized_dir <- normalizePath(dir, mustWork = FALSE, winslash = "/")
    normalized_paths <- normalizePath(
      path_elements,
      mustWork = FALSE,
      winslash = "/"
    )
    cache_root <- normalizePath(
      osrm_default_install_root(),
      mustWork = FALSE,
      winslash = "/"
    )
    # Use "/" for cache_root_prefix to match the normalized paths
    cache_root_prefix <- paste0(cache_root, "/")
  } else {
    normalized_dir <- normalizePath(dir, mustWork = FALSE)
    normalized_paths <- normalizePath(path_elements, mustWork = FALSE)
    cache_root <- normalizePath(osrm_default_install_root(), mustWork = FALSE)
    cache_root_prefix <- paste0(cache_root, .Platform$file.sep)
  }

  # Helper function for path comparison (case-insensitive on Windows)
  paths_equal <- function(path1, path2) {
    if (.Platform$OS.type == "windows") {
      tolower(path1) == tolower(path2)
    } else {
      path1 == path2
    }
  }

  # Remove any paths that point to other versions in our cache directory
  # but keep system installations (homebrew, manual installs, etc.)
  paths_to_keep <- vapply(
    seq_along(path_elements),
    function(i) {
      norm_path <- normalized_paths[i]
      # Keep this path if:
      # 1. It's the directory we're installing to, OR
      # 2. It's not under our cache root (i.e., it's a system/homebrew install)
      if (paths_equal(norm_path, normalized_dir)) {
        return(TRUE)
      }
      # On Windows, use case-insensitive comparison for path prefix check
      if (.Platform$OS.type == "windows") {
        !startsWith(tolower(norm_path), tolower(cache_root_prefix))
      } else {
        !startsWith(norm_path, cache_root_prefix)
      }
    },
    logical(1)
  )

  filtered_paths <- path_elements[paths_to_keep]

  # Always prepend the new installation directory to the front
  # (even if it was already in the PATH, we want it first)
  first_path_matches <- if (length(normalized_paths) > 0) {
    paths_equal(normalized_paths[1], normalized_dir)
  } else {
    FALSE
  }

  if (!first_path_matches) {
    # Remove the dir if it's already present (to avoid duplicates)
    kept_normalized <- normalized_paths[paths_to_keep]
    paths_not_matching <- !vapply(
      kept_normalized,
      paths_equal,
      logical(1),
      path2 = normalized_dir
    )
    filtered_paths <- filtered_paths[paths_not_matching]
    new_path <- paste(c(normalized_dir, filtered_paths), collapse = path_sep)
    Sys.setenv(PATH = new_path)
    if (!quiet) {
      message(sprintf("Added '%s' to PATH for this session.", normalized_dir))
    }
  } else if (!quiet) {
    message(sprintf("'%s' is already first in PATH.", normalized_dir))
  }
}

#' @noRd
set_path_project <- function(dir, quiet = FALSE) {
  r_profile_path <- file.path(getwd(), ".Rprofile")

  if (!quiet) {
    message("You chose to set the OSRM path for this project.")
    message("This will update the following file: ", r_profile_path)
  }

  # Use forward slashes for cross-platform compatibility in the .Rprofile
  path_for_r <- normalizePath(dir, mustWork = FALSE, winslash = "/")

  comment_tag <- "#added-by-r-pkg-osrm.backend"
  line_to_add <- sprintf(
    'Sys.setenv(PATH = paste("%s", Sys.getenv("PATH"), sep = "%s")) %s',
    path_for_r,
    .Platform$path.sep,
    comment_tag
  )

  if (!file.exists(r_profile_path)) {
    if (!quiet) {
      message("Creating new .Rprofile file.")
    }
    writeLines(line_to_add, r_profile_path)
  } else {
    lines <- readLines(r_profile_path, warn = FALSE)
    has_tag <- grepl(comment_tag, lines, fixed = TRUE)

    if (any(has_tag)) {
      existing_dirs <- vapply(
        lines[has_tag],
        function(line) {
          match <- regmatches(line, regexec('paste\\("([^"]+)"', line))
          if (length(match[[1]]) >= 2) {
            # Normalize with forward slashes for consistent comparison
            normalizePath(match[[1]][2], mustWork = FALSE, winslash = "/")
          } else {
            ""
          }
        },
        character(1)
      )

      # Case-insensitive comparison on Windows
      is_same_path <- if (.Platform$OS.type == "windows") {
        tolower(path_for_r) %in% tolower(existing_dirs)
      } else {
        path_for_r %in% existing_dirs
      }

      if (any(is_same_path)) {
        if (!quiet) {
          message(
            ".Rprofile already contains this OSRM path configuration. No changes made."
          )
        }
        return(invisible())
      }

      if (!quiet) {
        message("Replacing old OSRM path configuration with new one.")
      }
      lines <- lines[!has_tag]
    } else {
      if (!quiet) {
        message("Appending configuration to existing .Rprofile.")
      }
    }

    # Add a newline for separation if the file doesn't end with one
    if (length(lines) > 0 && nzchar(lines[length(lines)])) {
      lines <- c(lines, "")
    }
    lines <- c(lines, line_to_add)
    writeLines(lines, r_profile_path)
  }

  if (!quiet) {
    message(
      "\nSuccessfully modified .Rprofile. Please restart R for the changes to take effect."
    )
  }
}

test_that("Live installation and basic routing works for all supported OSRM versions", {
  skip_if_not(identical(Sys.getenv("RUN_OSRM_LIVE_TESTS"), "true"), "Live OSRM tests skipped")
  
  # 0. DYNAMIC VERSION DISCOVERY
  # Get all versions with binaries for this platform
  all_tags <- osrm_check_available_versions(prereleases = TRUE)
  
  if (length(all_tags) == 0) {
    skip("No OSRM releases with binaries found for this platform.")
  }
  
  # Filter logic:
  # 1. Latest v5.x (usually v5.27.1)
  # 2. All versions >= v6.0.0
  v5_tags <- all_tags[grepl("^v5\\.", all_tags)]
  latest_v5 <- if (length(v5_tags) > 0) v5_tags[1] else NULL
  
  v6_plus_tags <- all_tags[!grepl("^v5\\.", all_tags)]
  # Note: osrm_check_available_versions returns tags sorted newest first.
  
  versions_to_test <- unique(c(latest_v5, v6_plus_tags))
  
  # For Mac, we can only test v6+ if we are on Sequoia (Darwin 24)
  is_macos <- identical(tolower(Sys.info()[["sysname"]]), "darwin")
  darwin_version <- NA_integer_
  if (is_macos) {
    darwin_version <- as.integer(gsub("\\..*", "", Sys.info()[["release"]]))
  }
  
  # Example OSM PBF (minimal)
  pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  
  for (ver in versions_to_test) {
    # Skip v6+ on old macOS to avoid intended failures
    if (is_macos && darwin_version < 24 && ver != latest_v5) {
      message(sprintf("Skipping OSRM %s on macOS %d (requires Darwin 24+)", ver, darwin_version))
      next
    }
    
    message(sprintf("\n>>> LIVE TEST: OSRM %s <<<", ver))
    
    # 1. INSTALL
    # Use a clean subdirectory for every version
    test_dir <- file.path(tempdir(), paste0("osrm_live_", gsub("\\.", "_", ver)))
    if (dir.exists(test_dir)) unlink(test_dir, recursive = TRUE)
    dir.create(test_dir, recursive = TRUE)
    
    # Ensure cleanup happens even if the test fails
    withr::defer(unlink(test_dir, recursive = TRUE))
    
    # We use path_action="none" to avoid messing with .Rprofile in CI
    # This MUST succeed
    install_path <- osrm_install(version = ver, dest_dir = test_dir, path_action = "none", quiet = FALSE)
    
    # 2. VERIFY BINARIES
    routed_bin <- list.files(install_path, pattern = "^osrm-routed(\\.exe)?$", full.names = TRUE)
    expect_length(routed_bin, 1)
    
    # Check version output
    res_ver <- system2(routed_bin, "--version", stdout = TRUE, stderr = TRUE)
    # The --version output usually starts with the version number (e.g. "5.27.1")
    expect_match(res_ver, gsub("^v", "", ver), fixed = TRUE)
    
    # 3. PREPARE GRAPH
    # Copy PBF to unique name in version dir
    tmp_pbf <- file.path(test_dir, paste0("test_", ver, ".osm.pbf"))
    file.copy(pbf_path, tmp_pbf)
    
    # Ensure the binaries we just installed are the ones used
    old_opt <- options(osrm.routed.exec = routed_bin)
    withr::defer(options(old_opt))
    
    # This MUST succeed
    graph <- osrm_prepare_graph(tmp_pbf, threads = 1L, quiet = FALSE)
    
    # 4. START SERVER
    # Use a random high port to avoid conflicts
    port <- 10000 + sample(1:1000, 1)
    
    # This MUST succeed
    srv <- osrm_start_server(graph$osrm_job_artifact, port = port, threads = 1L)
    withr::defer(try(osrm_stop(srv, quiet = TRUE), silent = TRUE))
    
    # 5. LIVENESS CHECK
    expect_true(srv$is_alive())
    
    # 6. STOP SERVER
    expect_message(
      stopped <- osrm_stop(srv),
      "Stopped OSRM server"
    )
    expect_false(srv$is_alive())
  }
})

# Tests for osrm_install.R - comprehensive coverage

# Test platform detection ----
test_that("get_platform_info detects Windows x86-64 machines", {
  fake_sys_info <- c(
    sysname = "Windows",
    machine = "x86-64"
  )

  platform <- get_platform_info(sys_info = fake_sys_info)

  expect_identical(platform, list(os = "win32", arch = "x64"))
})

test_that("get_platform_info detects Linux x86_64", {
  fake_sys_info <- c(
    sysname = "Linux",
    machine = "x86_64"
  )

  platform <- get_platform_info(sys_info = fake_sys_info)
  expect_identical(platform, list(os = "linux", arch = "x64"))
})

test_that("get_platform_info detects macOS arm64", {
  fake_sys_info <- c(
    sysname = "Darwin",
    machine = "arm64"
  )

  platform <- get_platform_info(sys_info = fake_sys_info)
  expect_identical(platform, list(os = "darwin", arch = "arm64"))
})

test_that("get_platform_info detects macOS x86_64", {
  fake_sys_info <- c(
    sysname = "Darwin",
    machine = "x86_64"
  )

  platform <- get_platform_info(sys_info = fake_sys_info)
  expect_identical(platform, list(os = "darwin", arch = "x64"))
})

test_that("get_platform_info handles alternative architecture names", {
  # Test amd64 -> x64
  platform <- get_platform_info(
    sys_info = c(sysname = "Linux", machine = "amd64")
  )
  expect_identical(platform$arch, "x64")

  # Test i386 -> x86
  platform <- get_platform_info(
    sys_info = c(sysname = "Linux", machine = "i386")
  )
  expect_identical(platform$arch, "x86")

  # Test i686 -> x86
  platform <- get_platform_info(
    sys_info = c(sysname = "Linux", machine = "i686")
  )
  expect_identical(platform$arch, "x86")

  # Test aarch64 -> arm64
  platform <- get_platform_info(
    sys_info = c(sysname = "Linux", machine = "aarch64")
  )
  expect_identical(platform$arch, "arm64")
})

test_that("get_platform_info respects override options", {
  # Store original options
  orig_os <- getOption("osrm.backend.override_os")
  orig_arch <- getOption("osrm.backend.override_arch")
  on.exit({
    options(osrm.backend.override_os = orig_os)
    options(osrm.backend.override_arch = orig_arch)
  })

  # Set overrides
  options(osrm.backend.override_os = "linux")
  options(osrm.backend.override_arch = "arm64")

  platform <- get_platform_info(
    sys_info = c(sysname = "Windows", machine = "x86_64")
  )

  expect_identical(platform, list(os = "linux", arch = "arm64"))
})

test_that("get_platform_info errors on unsupported platforms", {
  expect_error(
    get_platform_info(
      sys_info = c(sysname = "FreeBSD", machine = "x86_64")
    ),
    "not supported"
  )
})

test_that("get_platform_info errors on empty override options", {
  orig_os <- getOption("osrm.backend.override_os")
  on.exit(options(osrm.backend.override_os = orig_os))

  options(osrm.backend.override_os = "")
  expect_error(
    get_platform_info(),
    "must be a non-empty string"
  )
})

# Test macOS version detection ----
test_that("get_macos_release_info detects macOS Sequoia", {
  fake_sys_info <- c(
    sysname = "Darwin",
    release = "24.0.0"
  )

  info <- get_macos_release_info(fake_sys_info)

  expect_identical(info$darwin_major, 24L)
  expect_identical(info$marketing_version, "15")
  expect_identical(info$marketing_codename, "Sequoia")
  expect_true(info$meets_v6_requirement)
})

test_that("get_macos_release_info detects macOS Ventura", {
  fake_sys_info <- c(
    sysname = "Darwin",
    release = "22.6.0"
  )

  info <- get_macos_release_info(fake_sys_info)

  expect_identical(info$darwin_major, 22L)
  expect_identical(info$marketing_version, "13")
  expect_identical(info$marketing_codename, "Ventura")
  expect_false(info$meets_v6_requirement)
})

test_that("get_macos_release_info returns NA for non-Darwin systems", {
  fake_sys_info <- c(
    sysname = "Linux",
    release = "5.15.0"
  )

  info <- get_macos_release_info(fake_sys_info)

  expect_true(is.na(info$release))
  expect_true(is.na(info$darwin_major))
  expect_true(info$meets_v6_requirement) # Should be TRUE for non-macOS
})

# Test asset URL finding ----
test_that("find_asset_url prefers latest Node ABI for matching platform", {
  release_info <- list(
    tag_name = "v5.27.1",
    assets = list(
      list(
        name = "node_osrm-v5.27.1-node-v93-darwin-arm64-Release.tar.gz",
        browser_download_url = "https://example.com/node-v93"
      ),
      list(
        name = "node_osrm-v5.27.1-node-v108-darwin-arm64-Release.tar.gz",
        browser_download_url = "https://example.com/node-v108"
      )
    )
  )

  platform <- list(os = "darwin", arch = "arm64")

  selected <- find_asset_url(release_info, platform)
  expect_identical(selected, "https://example.com/node-v108")
})

test_that("find_asset_url selects asset without node version", {
  release_info <- list(
    tag_name = "v6.0.0",
    assets = list(
      list(
        name = "osrm-v6.0.0-linux-x64-Release.tar.gz",
        browser_download_url = "https://example.com/linux-x64"
      )
    )
  )

  platform <- list(os = "linux", arch = "x64")

  selected <- find_asset_url(release_info, platform)
  expect_identical(selected, "https://example.com/linux-x64")
})

test_that("find_asset_url errors when no matching asset", {
  release_info <- list(
    tag_name = "v5.27.1",
    assets = list(
      list(
        name = "node_osrm-v5.27.1-linux-x64-Release.tar.gz",
        browser_download_url = "https://example.com/linux"
      )
    )
  )

  platform <- list(os = "darwin", arch = "arm64")

  expect_error(
    find_asset_url(release_info, platform),
    "Could not find a compatible binary"
  )
})

# Test version comparison ----
test_that("version_at_least correctly compares versions", {
  expect_true(version_at_least("v26.4.1", "v6.0.0"))
  expect_true(version_at_least("v6.0.0", "v5.27.1"))
  expect_true(version_at_least("v6.0.0", "v6.0.0"))
  expect_false(version_at_least("v5.27.1", "v6.0.0"))
  expect_true(version_at_least("v6.1.0", "v6.0.0"))
})

test_that("version_at_least handles version tags without 'v'", {
  expect_true(version_at_least("6.0.0", "5.27.1"))
  expect_true(version_at_least("v6.0.0", "5.27.1"))
  expect_true(version_at_least("6.0.0", "v5.27.1"))
})

test_that("version_at_least handles invalid versions", {
  expect_false(version_at_least(NULL, "v6.0.0"))
  expect_false(version_at_least("v6.0.0", NULL))
  expect_false(version_at_least("", "v6.0.0"))
  expect_false(version_at_least("invalid", "v6.0.0"))
})

# Test release filtering ----
test_that("release_has_binaries detects binaries correctly", {
  release <- list(
    assets = list(
      list(name = "node_osrm-v5.27.1-node-v108-darwin-arm64-Release.tar.gz"),
      list(name = "node_osrm-v5.27.1-node-v108-linux-x64-Release.tar.gz")
    )
  )

  platform_darwin <- list(os = "darwin", arch = "arm64")
  platform_linux <- list(os = "linux", arch = "x64")
  platform_win <- list(os = "win32", arch = "x64")

  expect_true(release_has_binaries(release, platform_darwin))
  expect_true(release_has_binaries(release, platform_linux))
  expect_false(release_has_binaries(release, platform_win))
})

# Test path helpers ----
test_that("osrm_default_install_root returns cache directory", {
  root <- osrm_default_install_root()
  expect_true(is.character(root))
  expect_true(nzchar(root))
  expect_true(grepl("osrm.backend", root, fixed = TRUE))
})

test_that("osrm_try_normalize_path handles valid paths", {
  path <- tempdir()
  normalized <- osrm_try_normalize_path(path)
  expect_true(is.character(normalized))
  expect_true(nzchar(normalized))
})

test_that("osrm_try_normalize_path handles errors gracefully", {
  path <- "/nonexistent/path/that/does/not/exist"
  normalized <- osrm_try_normalize_path(path)
  expect_identical(normalized, path)
})

test_that("osrm_is_install_dir detects valid installation", {
  # Create temporary directory with osrm-routed
  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  on.exit(unlink(tmp_dir, recursive = TRUE))

  file.create(file.path(tmp_dir, "osrm-routed"))

  expect_true(osrm_is_install_dir(tmp_dir))
})

test_that("osrm_is_install_dir returns FALSE for non-installation", {
  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  on.exit(unlink(tmp_dir, recursive = TRUE))

  expect_false(osrm_is_install_dir(tmp_dir))
  expect_false(osrm_is_install_dir(NULL))
  expect_false(osrm_is_install_dir(character(0)))
})

test_that("osrm_detect_installations finds valid installations", {
  # Create a mock install root
  tmp_root <- tempfile()
  dir.create(tmp_root)
  on.exit(unlink(tmp_root, recursive = TRUE))

  # Create two version directories
  v1_dir <- file.path(tmp_root, "v5.27.1")
  v2_dir <- file.path(tmp_root, "v6.0.0")
  dir.create(v1_dir)
  dir.create(v2_dir)

  file.create(file.path(v1_dir, "osrm-routed"))
  file.create(file.path(v2_dir, "osrm-routed"))

  installations <- osrm_detect_installations(tmp_root)

  expect_length(installations, 2)
  expect_true(all(grepl(
    "osrm-routed",
    list.files(installations, recursive = FALSE)
  )))
})

# Test PATH setting (without actually modifying PATH) ----
test_that("set_path_session adds directory to PATH", {
  orig_path <- Sys.getenv("PATH")
  on.exit(Sys.setenv(PATH = orig_path))

  tmp_dir <- tempdir()
  set_path_session(tmp_dir, quiet = TRUE)

  new_path <- Sys.getenv("PATH")
  path_elements <- strsplit(new_path, .Platform$path.sep)[[1]]
  first_path <- path_elements[1]

  # Normalize both paths for a robust comparison
  expected_norm <- normalizePath(tmp_dir, mustWork = FALSE)
  first_path_norm <- normalizePath(first_path, mustWork = FALSE)

  # On Windows, paths are case-insensitive
  if (.Platform$OS.type == "windows") {
    expect_equal(tolower(first_path_norm), tolower(expected_norm))
  } else {
    expect_equal(first_path_norm, expected_norm)
  }
})

test_that("set_path_session doesn't duplicate paths", {
  orig_path <- Sys.getenv("PATH")
  on.exit(Sys.setenv(PATH = orig_path))

  tmp_dir <- tempdir()
  set_path_session(tmp_dir)
  path_after_first <- Sys.getenv("PATH")

  set_path_session(tmp_dir)
  path_after_second <- Sys.getenv("PATH")

  expect_identical(path_after_first, path_after_second)
})

# Test .Rprofile manipulation ----
test_that("osrm_clear_path removes added paths from .Rprofile", {
  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  orig_wd <- getwd()
  on.exit({
    setwd(orig_wd)
    unlink(tmp_dir, recursive = TRUE)
  })

  setwd(tmp_dir)

  # Create .Rprofile with our tag
  rprofile <- file.path(tmp_dir, ".Rprofile")
  writeLines(
    c(
      "# Some existing code",
      'Sys.setenv(PATH = paste("/some/path", Sys.getenv("PATH"), sep = ":")) #added-by-r-pkg-osrm.backend',
      "# More code"
    ),
    rprofile
  )

  result <- osrm_clear_path(quiet = TRUE)

  expect_true(result)
  lines <- readLines(rprofile)
  expect_false(any(grepl("#added-by-r-pkg-osrm.backend", lines, fixed = TRUE)))
  expect_true(any(grepl("Some existing code", lines, fixed = TRUE)))
})

test_that("osrm_clear_path handles missing .Rprofile", {
  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  orig_wd <- getwd()
  on.exit({
    setwd(orig_wd)
    unlink(tmp_dir, recursive = TRUE)
  })

  setwd(tmp_dir)

  result <- osrm_clear_path(quiet = TRUE)
  expect_false(result)
})

test_that("osrm_clear_path handles .Rprofile without our tags", {
  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  orig_wd <- getwd()
  on.exit({
    setwd(orig_wd)
    unlink(tmp_dir, recursive = TRUE)
  })

  setwd(tmp_dir)

  rprofile <- file.path(tmp_dir, ".Rprofile")
  writeLines("# Some code", rprofile)

  result <- osrm_clear_path(quiet = TRUE)
  expect_false(result)
})

# Test set_path_project ----
# Test set_path_project ----
test_that("set_path_project creates .Rprofile if missing", {
  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  orig_wd <- getwd()
  on.exit({
    setwd(orig_wd)
    unlink(tmp_dir, recursive = TRUE)
  })

  setwd(tmp_dir)

  set_path_project(tempdir(), quiet = TRUE)

  rprofile <- file.path(tmp_dir, ".Rprofile")
  expect_true(file.exists(rprofile))
  lines <- readLines(rprofile)
  expect_true(any(grepl("#added-by-r-pkg-osrm.backend", lines, fixed = TRUE)))
})

test_that("set_path_project doesn't duplicate entries", {
  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  orig_wd <- getwd()
  on.exit({
    setwd(orig_wd)
    unlink(tmp_dir, recursive = TRUE)
  })

  setwd(tmp_dir)

  set_path_project(tempdir(), quiet = TRUE)
  set_path_project(tempdir(), quiet = TRUE)

  rprofile <- file.path(tmp_dir, ".Rprofile")
  lines <- readLines(rprofile)
  count <- sum(grepl("#added-by-r-pkg-osrm.backend", lines, fixed = TRUE))
  expect_identical(count, 1L)
})

# Test handle_path_setting dispatcher ----
test_that("handle_path_setting calls correct function", {
  orig_path <- Sys.getenv("PATH")
  on.exit(Sys.setenv(PATH = orig_path))

  tmp_dir <- tempdir()

  # Test "session"
  handle_path_setting("session", tmp_dir, quiet = TRUE)
  new_path <- Sys.getenv("PATH")
  path_elements <- strsplit(new_path, .Platform$path.sep)[[1]]
  first_path <- path_elements[1]

  # Normalize for comparison
  expected_norm <- normalizePath(tmp_dir, mustWork = FALSE)
  first_path_norm <- normalizePath(first_path, mustWork = FALSE)

  if (.Platform$OS.type == "windows") {
    expect_equal(tolower(first_path_norm), tolower(expected_norm))
  } else {
    expect_equal(first_path_norm, expected_norm)
  }

  # Test "none" - should not error
  expect_silent(handle_path_setting(
    "none",
    tmp_dir,
    quiet = TRUE
  ))
})

# Test runtime installation helpers ----
test_that("maybe_install_windows_v6_runtime only runs on Windows", {
  platform_linux <- list(os = "linux", arch = "x64")
  expect_silent(
    maybe_install_windows_v6_runtime(
      "v6.0.0",
      platform_linux,
      tempdir()
    )
  )
})

test_that("maybe_install_linux_v6_runtime only runs on Linux for v6+", {
  platform_darwin <- list(os = "darwin", arch = "arm64")
  expect_silent(
    maybe_install_linux_v6_runtime(
      "v6.0.0",
      platform_darwin,
      tempdir()
    )
  )

  platform_linux <- list(os = "linux", arch = "x64")
  expect_silent(
    maybe_install_linux_v6_runtime(
      "v5.27.1",
      platform_linux,
      tempdir()
    )
  )
})

test_that("maybe_install_macos_v6_runtime only runs on macOS for v6+", {
  platform_linux <- list(os = "linux", arch = "x64")
  expect_silent(
    maybe_install_macos_v6_runtime(
      "v6.0.0",
      platform_linux,
      tempdir()
    )
  )

  platform_darwin <- list(os = "darwin", arch = "arm64")
  expect_silent(
    maybe_install_macos_v6_runtime(
      "v5.27.1",
      platform_darwin,
      tempdir()
    )
  )
})

# More edge cases for macOS detection
test_that("get_macos_release_info handles edge cases", {
  # Invalid release format
  info <- get_macos_release_info(
    c(sysname = "Darwin", release = "invalid")
  )
  expect_true(is.na(info$darwin_major))

  # Empty release
  info <- get_macos_release_info(
    c(sysname = "Darwin", release = "")
  )
  expect_true(is.na(info$darwin_major))

  # NULL sys_info
  info <- get_macos_release_info(NULL)
  expect_true(is.na(info$release))
})

# Test edge cases for platform detection
test_that("get_platform_info handles NULL and empty inputs", {
  expect_error(
    get_platform_info(sys_info = NULL),
    "not supported"
  )

  expect_error(
    get_platform_info(
      sys_info = c(sysname = "Linux", machine = "")
    ),
    "not supported"
  )
})

# Test get_platform_info with empty arch override
test_that("get_platform_info errors on empty arch override", {
  orig_arch <- getOption("osrm.backend.override_arch")
  on.exit(options(osrm.backend.override_arch = orig_arch))

  options(osrm.backend.override_arch = "")
  expect_error(
    get_platform_info(),
    "must be a non-empty string"
  )
})

# Test osrm_detect_installations with empty root
test_that("osrm_detect_installations handles non-existent root", {
  fake_root <- "/path/that/definitely/does/not/exist/anywhere"
  installations <- osrm_detect_installations(fake_root)
  expect_length(installations, 0)
})

# Test osrm_detect_installations with files (not directories)
test_that("osrm_detect_installations ignores files", {
  tmp_root <- tempfile()
  dir.create(tmp_root)
  on.exit(unlink(tmp_root, recursive = TRUE))

  # Create a file, not a directory
  file.create(file.path(tmp_root, "not-a-dir"))

  installations <- osrm_detect_installations(tmp_root)
  expect_length(installations, 0)
})

# Test osrm_is_install_dir with Windows executable
test_that("osrm_is_install_dir detects Windows installation", {
  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  on.exit(unlink(tmp_dir, recursive = TRUE))

  file.create(file.path(tmp_dir, "osrm-routed.exe"))

  expect_true(osrm_is_install_dir(tmp_dir))
})

# Test PATH management when switching between versions ----
test_that("set_path_session removes old cached versions and prioritizes new version", {
  orig_path <- Sys.getenv("PATH")
  on.exit(Sys.setenv(PATH = orig_path))

  # Use the actual cache root to match production behavior
  cache_root <- osrm_default_install_root()

  # Create two version directories in the cache
  v6_dir <- file.path(cache_root, "test_v6.0.0")
  v5_dir <- file.path(cache_root, "test_v5.27.1")
  dir.create(v6_dir, recursive = TRUE, showWarnings = FALSE)
  dir.create(v5_dir, recursive = TRUE, showWarnings = FALSE)
  on.exit({
    unlink(v6_dir, recursive = TRUE)
    unlink(v5_dir, recursive = TRUE)
  }, add = TRUE)

  # Create a system installation directory (outside cache)
  system_dir <- tempfile()
  dir.create(system_dir)
  on.exit(unlink(system_dir, recursive = TRUE), add = TRUE)

  # Add system installation to PATH first
  Sys.setenv(PATH = paste(system_dir, orig_path, sep = .Platform$path.sep))

  # Step 1: Add v6.0.0 to PATH
  set_path_session(v6_dir, quiet = TRUE)
  path_with_v6 <- Sys.getenv("PATH")
  path_elements <- strsplit(path_with_v6, .Platform$path.sep)[[1]]
  normalized_paths <- normalizePath(path_elements, mustWork = FALSE)
  v6_normalized <- normalizePath(v6_dir, mustWork = FALSE)

  # v6.0.0 should be first in PATH
  expect_identical(normalized_paths[1], v6_normalized)
  # System installation should still be in PATH
  expect_true(normalizePath(system_dir, mustWork = FALSE) %in% normalized_paths)

  # Step 2: Add v5.27.1 to PATH (simulating a version switch)
  set_path_session(v5_dir, quiet = TRUE)
  path_with_v5 <- Sys.getenv("PATH")
  path_elements <- strsplit(path_with_v5, .Platform$path.sep)[[1]]
  normalized_paths <- normalizePath(path_elements, mustWork = FALSE)
  v5_normalized <- normalizePath(v5_dir, mustWork = FALSE)

  # v5.27.1 should now be first in PATH
  expect_identical(normalized_paths[1], v5_normalized)
  # v6.0.0 should have been REMOVED (it's a cached version)
  expect_false(v6_normalized %in% normalized_paths)
  # System installation should still be in PATH
  expect_true(normalizePath(system_dir, mustWork = FALSE) %in% normalized_paths)
})

test_that("set_path_session preserves system installations when adding cached version", {
  orig_path <- Sys.getenv("PATH")
  on.exit(Sys.setenv(PATH = orig_path))

  # Use the actual cache root
  cache_root <- osrm_default_install_root()

  # Create cached version directory
  v6_dir <- file.path(cache_root, "test_v6.0.0_preserve")
  dir.create(v6_dir, recursive = TRUE, showWarnings = FALSE)
  on.exit(unlink(v6_dir, recursive = TRUE), add = TRUE)

  # Create multiple system installation directories
  homebrew_dir <- "/usr/local/bin"  # Common homebrew location
  custom_dir <- tempfile()
  dir.create(custom_dir)
  on.exit(unlink(custom_dir, recursive = TRUE), add = TRUE)

  # Set up PATH with system installations
  Sys.setenv(PATH = paste(
    c(homebrew_dir, custom_dir, orig_path),
    collapse = .Platform$path.sep
  ))

  # Add cached version to PATH
  set_path_session(v6_dir, quiet = TRUE)
  new_path <- Sys.getenv("PATH")
  path_elements <- strsplit(new_path, .Platform$path.sep)[[1]]
  normalized_paths <- normalizePath(path_elements, mustWork = FALSE)

  # Cached version should be first
  expect_identical(
    normalized_paths[1],
    normalizePath(v6_dir, mustWork = FALSE)
  )

  # System installations should still be present
  expect_true(homebrew_dir %in% path_elements)
  expect_true(normalizePath(custom_dir, mustWork = FALSE) %in% normalized_paths)
})

test_that("set_path_project replaces old cached version with new one", {
  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  orig_wd <- getwd()
  on.exit({
    setwd(orig_wd)
    unlink(tmp_dir, recursive = TRUE)
  })

  setwd(tmp_dir)

  # Create mock cache directories
  tmp_cache <- tempfile()
  dir.create(tmp_cache, recursive = TRUE)
  on.exit(unlink(tmp_cache, recursive = TRUE), add = TRUE)

  v6_dir <- file.path(tmp_cache, "v6.0.0")
  v5_dir <- file.path(tmp_cache, "v5.27.1")
  dir.create(v6_dir)
  dir.create(v5_dir)

  # Add v6.0.0 to .Rprofile
  set_path_project(v6_dir, quiet = TRUE)

  rprofile <- file.path(tmp_dir, ".Rprofile")
  lines <- readLines(rprofile)
  # Use forward slashes to match what set_path_project writes to .Rprofile
  expect_true(any(grepl(normalizePath(v6_dir, mustWork = FALSE, winslash = "/"), lines, fixed = TRUE)))

  # Now add v5.27.1 - it should replace v6.0.0
  set_path_project(v5_dir, quiet = TRUE)

  lines <- readLines(rprofile)
  # v5.27.1 should be in .Rprofile
  expect_true(any(grepl(normalizePath(v5_dir, mustWork = FALSE, winslash = "/"), lines, fixed = TRUE)))
  # v6.0.0 should NOT be in .Rprofile anymore
  expect_false(any(grepl(normalizePath(v6_dir, mustWork = FALSE, winslash = "/"), lines, fixed = TRUE)))
  # Should only have one osrm.backend line
  count <- sum(grepl("#added-by-r-pkg-osrm.backend", lines, fixed = TRUE))
  expect_identical(count, 1L)
})

# Integration tests with mocking would go here if we had a mocking framework
# For now, the actual osrm_install function is tested in the setup-osrm.R

# Internal helpers for osrm_gui
# These functions are not exported

#' Check for required GUI dependencies
#' @noRd
gui_check_dependencies <- function() {
  required_pkgs <- c("shiny", "mapgl", "osrm", "sf", "DT", "viridisLite")
  missing_pkgs <- required_pkgs[
    !vapply(required_pkgs, requireNamespace, logical(1), quietly = TRUE)
  ]

  if (length(missing_pkgs) > 0) {
    stop(
      "The following packages are required for the GUI but are not installed: ",
      paste(missing_pkgs, collapse = ", "),
      ".
",
      "Please install them using install.packages(c(",
      paste(sprintf("'%s'", missing_pkgs), collapse = ", "),
      "))",
      call. = FALSE
    )
  }
}

#' Manage OSRM Server Lifecycle for GUI
#' @noRd
gui_setup_server <- function(input_osrm, port) {
  if (!identical(port, "auto")) {
    port_val <- suppressWarnings(as.integer(port))
    if (is.na(port_val) || port_val <= 0) {
      stop("Invalid port specified. Must be 'auto' or a positive integer.", call. = FALSE)
    }
  }

  server_process <- NULL
  kill_on_exit <- FALSE
  host <- "http://127.0.0.1"
  active_port <- port

  if (inherits(input_osrm, "process")) {
    if (!input_osrm$is_alive()) {
      stop("The provided OSRM server process is not running.", call. = FALSE)
    }
    if (identical(port, "auto")) {
      stop(
        "When providing a process object, you must also specify the port explicitly.
",
        "Example: osrm_gui(my_process, port = 5001)",
        call. = FALSE
      )
    }
    active_port <- as.integer(port)
    message("Using provided OSRM server process on port ", active_port)
  } else if (is.character(input_osrm)) {
    message("Starting temporary OSRM server for ", basename(input_osrm), "...")
    server_port <- if (identical(port, "auto")) 5001L else as.integer(port)
    server_process <- osrm_start(
      path = input_osrm,
      port = server_port,
      quiet = FALSE
    )
    if (!server_process$is_alive()) {
      stop("Failed to start OSRM server. Check logs.", call. = FALSE)
    }
    active_port <- server_port
    kill_on_exit <- TRUE
  } else if (identical(port, "auto")) {
    servers <- osrm_servers()
    alive_servers <- servers[servers$alive, ]

    if (nrow(alive_servers) == 0) {
      stop(
        "No running OSRM servers detected. ",
        "Start a server with osrm_start() or specify an explicit port.",
        call. = FALSE
      )
    } else if (nrow(alive_servers) == 1) {
      active_port <- alive_servers$port[1]
      message("Connected to OSRM server on port ", active_port)
    } else {
      most_recent <- alive_servers[which.max(alive_servers$started_at), ]
      active_port <- most_recent$port

      warning(
        "Multiple OSRM servers running (ports: ",
        paste(alive_servers$port, collapse = ", "),
        "). Using most recent (port ",
        active_port,
        ", started at ",
        format(most_recent$started_at),
        "). ",
        "Specify 'port' explicitly to select a different server.",
        call. = FALSE
      )
    }
  } else {
    active_port <- as.integer(port)
    message(
      "Attempting to connect to existing OSRM server at ",
      host,
      ":",
      active_port
    )
  }

  cleanup_fn <- function() {
    if (kill_on_exit && !is.null(server_process) && server_process$is_alive()) {
      message("Stopping temporary OSRM server...")
      server_process$kill()
    }
  }

  list(
    active_port = active_port,
    host = host,
    server_process = server_process,
    cleanup_fn = cleanup_fn
  )
}

#' Resolve Map Center and Zoom
#' @noRd
gui_resolve_map_view <- function(center, zoom, input_osrm) {
  auto_center <- NULL
  auto_zoom <- NULL

  if (is.null(center)) {
    # 1. Try to extract from osrm_server object metadata directly
    if (inherits(input_osrm, "osrm_server")) {
      meta <- attr(input_osrm, "osrm_metadata")
      if (!is.null(meta$center) && length(meta$center) == 2 && !any(is.na(meta$center))) {
        auto_center <- meta$center
      }
    }
    
    # 2. Try to extract from registry
    if (is.null(auto_center)) {
      srv_info <- osrm_servers()
      alive_srv <- srv_info[srv_info$alive, ]
      
      # Match by port if input is a process or "auto"
      if (nrow(alive_srv) > 0) {
        selected <- NULL
        
        # If input_osrm is an object with a specific port, match it
        if (inherits(input_osrm, "osrm_server")) {
          target_port <- attr(input_osrm, "osrm_metadata")$port
          if (!is.null(target_port)) {
            selected <- alive_srv[alive_srv$port == target_port, ]
          }
        }
        
        # Otherwise take the most recent
        if (is.null(selected) || nrow(selected) == 0) {
          selected <- alive_srv[which.max(alive_srv$started_at), ]
        }
        
        # Check if the registry actually stored the center
        if (!is.null(selected$center_lon) && !is.null(selected$center_lat) &&
            !is.na(selected$center_lon[1]) && !is.na(selected$center_lat[1])) {
          auto_center <- c(selected$center_lon[1], selected$center_lat[1])
        }
      }
    }

    # 3. Fallback: PBF Parsing
    if (is.null(auto_center)) {
      pbf_path <- NULL
      if (
        is.character(input_osrm) &&
          grepl("\\.osm\\.pbf$", input_osrm, ignore.case = TRUE)
      ) {
        pbf_path <- input_osrm
      } else if (exists("selected") && !is.null(selected) && nrow(selected) > 0) {
        if (nzchar(selected$input_osm[1])) pbf_path <- selected$input_osm[1]
      }

      if (!is.null(pbf_path) && file.exists(pbf_path)) {
        pbf_info <- .get_pbf_extent(pbf_path)
        if (!is.null(pbf_info)) {
          auto_center <- pbf_info$center
        } else {
          message("Could not auto-detect map center from PBF file. Defaulting to world view.")
        }
      } else {
        message("Source .osm.pbf file not found. Could not auto-detect map center. Defaulting to world view.")
      }
    }
    
    if (!is.null(auto_center)) {
      auto_zoom <- if (is.null(zoom)) 9 else zoom
      message(
        "Auto-centered map on bounding box center: ",
        paste(round(auto_center, 4), collapse = ", ")
      )
    }
  }

  # Normalize
  final_center <- center %||% auto_center
  final_zoom <- zoom %||% auto_zoom

  if (!is.null(final_center)) {
    if (is.list(final_center)) {
      final_center <- c(
        final_center$lng %||% final_center$lon %||% final_center$x,
        final_center$lat %||% final_center$y
      )
    }
    final_center <- as.numeric(final_center)
    if (length(final_center) != 2 || any(is.na(final_center))) {
      stop(
        "'center' must be a numeric vector of length 2 (lng, lat) or a named list.",
        call. = FALSE
      )
    }
  }

  list(center = final_center, zoom = final_zoom)
}

#' Robustly Parse Isochrone Breaks from Text Input
#' @noRd
gui_parse_breaks <- function(input_str, default = c(5, 10, 15)) {
  if (is.null(input_str) || !nzchar(trimws(input_str))) {
    return(default)
  }
  
  # Split, trim, and convert
  parts <- unlist(strsplit(input_str, ","))
  nums <- suppressWarnings(as.numeric(trimws(parts)))
  
  # Clean up
  nums <- nums[!is.na(nums) & nums > 0]
  
  if (length(nums) == 0) {
    return(default)
  }
  
  sort(unique(nums))
}

#' UI Resources (CSS/JS)
#' @noRd
gui_ui_resources <- function() {
  css <- "
      html, body { height: 100%; margin: 0; overflow: hidden; }
      .container-fluid { height: 100%; display: flex; flex-direction: column; }
      #shiny-notification-panel { top: 70px; right: 10px; left: auto; bottom: auto; }
      
      .sidebar-layout { flex: 1; display: flex; overflow: hidden; min-height: 0; }
      
      .sidebar-panel { 
        padding: 15px; 
        background: #f8f9fa; 
        border-right: 1px solid #dee2e6; 
        overflow-y: auto; 
      }
      .sidebar-panel h4:first-child { margin-top: 0; }
      .sidebar-panel h4 { margin-top: 20px; }
      
      .main-panel { 
        height: 100%; 
        display: flex; 
        flex-direction: column; 
        padding: 15px; 
        overflow: hidden;
      }
      
      .map-wrapper { position: relative; flex: 3; min-height: 300px; }
      #map { height: 100% !important; }
      
      .table-wrapper { 
        flex: 2; 
        overflow-y: auto; 
        margin-top: 10px; 
        border-top: 1px solid #eee;
        padding-top: 10px;
      }
      
      .route-stats-overlay {
        position: absolute;
        top: 10px;
        left: 10px;
        z-index: 1000;
        background: rgba(245, 245, 245, 0.9);
        padding: 8px 12px;
        border-radius: 4px;
        box-shadow: 0 2px 4px rgba(0,0,0,0.15);
        pointer-events: none;
        display: flex;
        gap: 15px;
        font-size: 14px;
        border: 1px solid #ccc;
      }
      .stat-val { font-weight: bold; }

      .trip-marker-label {
        background-color: #984ea3;
        color: white;
        border: 2px solid white;
        border-radius: 50%;
        width: 24px;
        height: 24px;
        display: flex;
        justify-content: center;
        align-items: center;
        font-weight: bold;
        font-size: 12px;
        box-shadow: 0 2px 4px rgba(0,0,0,0.3);
        cursor: pointer;
      }

      .exec-time-overlay {
        position: absolute;
        top: 10px;
        right: 50px; /* Offset to avoid overlap with standard zoom/fullscreen controls */
        z-index: 1000;
        background: rgba(255, 255, 255, 0.8);
        padding: 2px 6px;
        border-radius: 4px;
        font-size: 11px;
        color: #777;
        pointer-events: none;
        border: 1px solid rgba(0,0,0,0.1);
      }
      
      .segments-header {
        display: flex;
        justify-content: space-between;
        align-items: center;
        margin-bottom: 5px;
      }
      
      /* Mobile/Responsive Styles */
      .hamburger-btn { display: none; margin-right: 10px; font-size: 20px; cursor: pointer; }
      
      /* Remove slider fill for discrete look */
      .irs-bar, .irs-bar-edge {
        background: none !important;
        border: none !important;
      }
      .irs-single {
        background: #555 !important;
      }
      
      @media (max-width: 768px) {
        .sidebar-panel {
          display: none;
          width: 250px !important;
          flex-shrink: 0;
        }
        .sidebar-panel.show-sidebar {
          display: block;
        }
        .main-panel {
          width: 100% !important;
          flex: 1;
          min-width: 0;
        }
        .hamburger-btn {
          display: inline-block;
        }
        /* Hide the overlay since we are pushing the map instead */
        .sidebar-overlay {
          display: none !important;
        }
        .route-stats-overlay {
          flex-direction: column;
          gap: 2px;
        }
      }
    "

  js <- "
    $(document).on('click', '#hamburger_btn', function() {
      $('.sidebar-panel').toggleClass('show-sidebar');
    });

    function initializeMapListeners(mapId) {
      const mapElement = document.getElementById(mapId);
      if (!mapElement) return;

      const observer = new MutationObserver((mutations, obs) => {
        const map = mapElement.map;
        if (map) {
          // Disable default context menu to allow right-click
          map.getCanvas().addEventListener('contextmenu', (e) => e.preventDefault());

          map.on('contextmenu', (e) => {
            Shiny.setInputValue('js_right_click', {
              lng: e.lngLat.lng,
              lat: e.lngLat.lat,
              nonce: Math.random()
            });
          });

          let startMarker = null;
          let endMarker = null;
          let isoStartMarker = null;
          const tripMarkers = {};

          Shiny.addCustomMessageHandler('updateMarker', function(message) {
            const lngLat = [message.lng, message.lat];
            const markerId = message.id;

            const createDragEndCallback = (id) => {
              return (marker) => {
                const coords = marker.getLngLat();
                Shiny.setInputValue('marker_dragged', {
                  id: id,
                  lng: coords.lng,
                  lat: coords.lat,
                  nonce: Math.random()
                });
              };
            };
            
            const createDragCallback = (id) => {
              return (marker) => {
                const coords = marker.getLngLat();
                Shiny.setInputValue('marker_moving', {
                  id: id,
                  lng: coords.lng,
                  lat: coords.lat,
                  nonce: Math.random()
                });
              };
            };

            if (markerId === 'start') {
              if (!startMarker) {
                startMarker = new maplibregl.Marker({ draggable: true, color: '#009E73' })
                  .setLngLat(lngLat)
                  .addTo(map);
                startMarker.on('dragend', () => createDragEndCallback('start')(startMarker));
                startMarker.on('drag', () => createDragCallback('start')(startMarker));
              } else {
                startMarker.setLngLat(lngLat);
              }
            } else if (markerId === 'end') {
              if (!endMarker) {
                endMarker = new maplibregl.Marker({ draggable: true, color: '#D55E00' })
                  .setLngLat(lngLat)
                  .addTo(map);
                endMarker.on('dragend', () => createDragEndCallback('end')(endMarker));
                endMarker.on('drag', () => createDragCallback('end')(endMarker));
              } else {
                endMarker.setLngLat(lngLat);
              }
            } else if (markerId === 'iso_start') {
              if (!isoStartMarker) {
                isoStartMarker = new maplibregl.Marker({ draggable: true, color: '#CC79A7' })
                  .setLngLat(lngLat)
                  .addTo(map);
                isoStartMarker.on('dragend', () => createDragEndCallback('iso_start')(isoStartMarker));
                isoStartMarker.on('drag', () => createDragCallback('iso_start')(isoStartMarker));
              } else {
                isoStartMarker.setLngLat(lngLat);
              }
            }
          });
          
          Shiny.addCustomMessageHandler('updateTripMarker', function(message) {
            const id = message.id;
            
            if (message.action === 'add') {
               const lngLat = [message.lng, message.lat];
               
               const el = document.createElement('div');
               el.className = 'trip-marker-label';
               el.innerText = '?'; // Default until optimized
               
               const marker = new maplibregl.Marker({ draggable: true, element: el })
                 .setLngLat(lngLat)
                 .addTo(map);
                 
               marker.on('dragend', () => {
                 const coords = marker.getLngLat();
                 Shiny.setInputValue('move_trip_point', {
                   id: id,
                   lng: coords.lng,
                   lat: coords.lat,
                   nonce: Math.random()
                 });
               });
               
               marker.on('drag', () => {
                 const coords = marker.getLngLat();
                 Shiny.setInputValue('marker_moving', {
                   id: id,
                   lng: coords.lng,
                   lat: coords.lat,
                   nonce: Math.random()
                 });
               });
               
               // Add click listener to element for removal
               el.addEventListener('click', (e) => {
                 e.stopPropagation(); // Prevent map click
                 Shiny.setInputValue('remove_trip_point', {id: id, nonce: Math.random()});
               });
               
               tripMarkers[id] = marker;
               
            } else if (message.action === 'remove') {
               if (tripMarkers[id]) {
                 tripMarkers[id].remove();
                 delete tripMarkers[id];
               }
            } else if (message.action === 'clear') {
               for (const id in tripMarkers) {
                 tripMarkers[id].remove();
               }
               // Clear object
               for (const key in tripMarkers) delete tripMarkers[key];
            }
          });

          Shiny.addCustomMessageHandler('updateTripLabels', function(message) {
            // message is an array of {id: ..., label: ...}
            message.forEach(item => {
              const marker = tripMarkers[item.id];
              if (marker) {
                const el = marker.getElement();
                if (el) el.innerText = item.label;
              }
            });
          });

          Shiny.addCustomMessageHandler('clearAllMarkers', function(message) {
              if(startMarker) {
                  startMarker.remove();
                  startMarker = null;
              }
              if(endMarker) {
                  endMarker.remove();
                  endMarker = null;
              }
              if(isoStartMarker) {
                  isoStartMarker.remove();
                  isoStartMarker = null;
              }
              // Also clear trip markers
              for (const id in tripMarkers) {
                 tripMarkers[id].remove();
              }
              for (const key in tripMarkers) delete tripMarkers[key];
          });

          obs.disconnect();
        }
      });

      observer.observe(mapElement, { childList: true, subtree: true });
    }

    $(document).on('shiny:connected', () => {
      initializeMapListeners('map');
      
      // Customize the Isochrone slider labels to show real values
      // We wait briefly to ensure the slider is initialized
      setTimeout(function() {
        [\"#iso_res\", \"#iso_live_res\"].forEach(function(id) {
          const $el = $(id);
          const slider = $el.data(\"ionRangeSlider\");
          if (slider) {
            const valMap = {
              1: '100', 2: '200', 3: '500', 4: '1k',
              5: '2k', 6: '5k', 7: '10k', 8: '20k', 9: '50k'
            };
            slider.update({
              prettify: function(n) {
                return valMap[Math.round(n)] || n;
              }
            });
          }
        });
      }, 500);
    });
  "
  list(css = css, js = js)
}

#' Construct Main UI
#' @noRd
gui_ui_layout <- function() {
  res <- gui_ui_resources()

  shiny::fluidPage(
    # Header
    shiny::div(
      style = "display: flex; justify-content: space-between; align-items: center; padding: 10px 0; flex-wrap: wrap; gap: 10px;",
      shiny::div(
        style = "display: flex; align-items: center;",
        shiny::HTML(
          '<div id="hamburger_btn" class="hamburger-btn">&#9776;</div>'
        ),
        shiny::h3(
          shiny::HTML("<b>osrm.backend</b> GUI"),
          style = "margin: 0;"
        )
      ),
      shiny::div(
        style = "display: flex; gap: 10px; align-items: center;",
        shiny::uiOutput("mode_button_ui", inline = TRUE),
        shiny::uiOutput("autozoom_button_ui", inline = TRUE),
        shiny::uiOutput("tracking_button_ui", inline = TRUE),
        shiny::actionButton(
          "quit_app",
          "Quit",
          style = "background-color: #d9534f; color: white; border-width: 0px;"
        )
      )
    ),
    shiny::tags$head(
      shiny::tags$style(shiny::HTML(res$css)),
      shiny::tags$script(shiny::HTML(res$js))
    ),

    shiny::div(
      class = "sidebar-overlay"
    ),

    shiny::div(
      class = "sidebar-layout",
      shiny::div(
        class = "sidebar-panel",
        style = "width: 25%;",
        shiny::h4("OSRM Controls"),
        
        # Hidden but functional for R/JS conditions
        shiny::div(
          style = "display: none;",
          shiny::selectInput(
            "mode",
            "Mode",
            choices = c("Route" = "route", "Isochrone" = "iso", "Trip" = "trip")
          )
        ),

        shiny::conditionalPanel(
          condition = "input.mode == 'iso'",
          shiny::textInput(
            "iso_breaks",
            "Time Breaks (min, comma sep):",
            value = "5, 10, 15"
          ),
          shiny::sliderInput(
            "iso_res",
            "Resolution (Detail):",
            min = 1,
            max = 9,
            value = 3,
            step = 1,
            ticks = TRUE
          ),
          shiny::sliderInput(
            "iso_live_res",
            "Live Resolution (Drag):",
            min = 1,
            max = 9,
            value = 2,
            step = 1,
            ticks = TRUE
          )
        ),

        shiny::hr(),
        shiny::h4("Locations"),
        shiny::helpText(
          shiny::conditionalPanel(
            "input.mode != 'trip'",
            "Left-click map: Start point",
            shiny::tags$br(),
            "Right-click map: End point (Route mode)",
            shiny::tags$br(),
            "Drag markers to adjust."
          ),
          shiny::conditionalPanel(
            "input.mode == 'trip'",
            "Left-click map: Add waypoints",
            shiny::tags$br(),
            "Click marker: Remove waypoint",
            shiny::tags$br(),
            "Drag markers to adjust."
          )
        ),
        shiny::actionButton(
          "reset",
          "Reset / Clear",
          style = "width: 100%; margin-bottom: 10px;"
        ),

        shiny::textInput(
          "start_coords_input",
          "Start (Lat, Lon)",
          placeholder = "-30.03, -51.22"
        ),
        shiny::textInput(
          "end_coords_input",
          "End (Lat, Lon)",
          placeholder = "-30.05, -51.18"
        )
      ),

      shiny::div(
        class = "main-panel",
        style = "width: 75%;",
        shiny::div(
          class = "map-wrapper",
          shiny::conditionalPanel(
            condition = "input.mode == 'route' || input.mode == 'trip' || input.mode == 'iso'",
            shiny::uiOutput("route_stats")
          ),
          shiny::uiOutput("map_edit_controls"),
          shiny::uiOutput("exec_time_overlay"),
          mapgl::maplibreOutput("map"),
          shiny::div(
            style = "position: absolute; bottom: 40px; right: 10px; z-index: 1000;",
            shiny::actionButton(
              "copy_code",
              "Copy R Code",
              icon = shiny::icon("code"),
              style = "background: white; border: 1px solid #ccc; font-size: 12px; padding: 4px 8px;"
            )
          )
        ),
        shiny::div(
          class = "table-wrapper",
          shiny::conditionalPanel(
            condition = "input.mode == 'route'",
            shiny::div(
              class = "segments-header",
              shiny::h4("Route Segments", style = "margin: 0;"),
              shiny::actionButton(
                "clear_selection",
                "Deselect All",
                class = "btn-xs btn-default",
                style = "font-size: 11px; padding: 2px 8px;"
              )
            ),
            DT::dataTableOutput("itinerary_table")
          ),
          shiny::conditionalPanel(
            condition = "input.mode == 'trip'",
            shiny::h4("Trip Sequence"),
            DT::dataTableOutput("trip_table")
          )
        )
      )
    )
  )
}

#' Fetch Detailed Route Steps
#' @noRd
api_fetch_route_detailed <- function(
  src,
  dst,
  overview = "false",
  debug = FALSE
) {
  server_url <- getOption("osrm.server")
  profile <- getOption("osrm.profile")

  # Use geometries=geojson to get coordinate arrays directly
  url <- sprintf(
    "%sroute/v1/%s/%f,%f;%f,%f?steps=true&overview=%s&geometries=geojson",
    server_url,
    profile,
    src$lng,
    src$lat,
    dst$lng,
    dst$lat,
    overview
  )

  if (debug) {
    message("DEBUG [Detailed Route]: Fetching URL: ", url)
  }

  tryCatch(
    {
      req <- httr2::request(url)
      resp <- httr2::req_perform(req)
      httr2::resp_body_json(resp)
    },
    error = function(e) NULL
  )
}

#' Fetch Trip via Direct HTTP Request
#' @description Direct HTTP request to OSRM trip endpoint, bypassing osrm::osrmTrip.
#' This is a workaround for issues where osrmTrip fails inside Shiny reactive contexts.
#' @param pts_df data.frame with lon and lat columns
#' @param debug Logical, print debug messages
#' @return List with trip (sf LINESTRING) and summary, or NULL on error
#' @noRd
api_fetch_trip <- function(pts_df, debug = FALSE) {
  server_url <- getOption("osrm.server")
  profile <- getOption("osrm.profile")

  # Build coordinate string: "lon1,lat1;lon2,lat2;..."
  coords_str <- paste(
    sprintf("%.6f,%.6f", pts_df$lon, pts_df$lat),
    collapse = ";"
  )

  url <- sprintf(
    "%strip/v1/%s/%s?steps=false&geometries=geojson&overview=full&generate_hints=false",
    server_url,
    profile,
    coords_str
  )

  if (debug) {
    message("DEBUG [Trip API]: Fetching URL: ", url)
  }

  tryCatch(
    {
      req <- httr2::request(url)
      resp <- httr2::req_perform(req)
      res <- httr2::resp_body_json(resp)

      if (is.null(res$trips) || length(res$trips) == 0) {
        if (debug) {
          message("DEBUG [Trip API]: No trips in response")
        }
        return(NULL)
      }

      # Parse the first trip
      trip_data <- res$trips[[1]]

      # Extract waypoint order (which input index maps to which trip position)
      # res$waypoints is in the order of input coordinates.
      # res$waypoints[[i]]$waypoint_index is the position in the optimized trip.
      waypoint_order <- vapply(res$waypoints, function(w) as.integer(w$waypoint_index) + 1L, integer(1))

      # Extract overall geometry (full route as one line)
      coords <- trip_data$geometry$coordinates
      if (length(coords) < 2) {
        if (debug) {
          message("DEBUG [Trip API]: Not enough coordinates for LINESTRING")
        }
        return(NULL)
      }

      coord_matrix <- do.call(
        rbind,
        lapply(coords, as.numeric)
      )

      # Validate matrix dimensions and contents to prevent segfaults in st_linestring
      if (
        is.null(coord_matrix) ||
          !is.matrix(coord_matrix) ||
          nrow(coord_matrix) < 2 ||
          ncol(coord_matrix) != 2 ||
          any(is.na(coord_matrix))
      ) {
        if (debug) {
          message("DEBUG [Trip API]: Invalid coordinate matrix")
        }
        return(NULL)
      }

      # Build SINGLE-ROW sf LINESTRING for mapping (most stable)
      trip_line <- sf::st_linestring(coord_matrix)
      trip_sfc <- sf::st_sfc(trip_line, crs = 4326)
      trip_sf_map <- sf::st_sf(geometry = trip_sfc)

      # 2. Extract Legs for Table Display (Pure data, no geometry)
      legs <- trip_data$legs
      n_legs <- length(legs)

      if (n_legs > 0) {
        legs_df <- data.frame(
          Start = seq_len(n_legs),
          End = c(seq(2, n_legs), 1), # Circular: last leg returns to start
          `Duration (min)` = round(
            unname(vapply(
              legs,
              function(l) as.numeric(l$duration) / 60,
              numeric(1)
            )),
            2
          ),
          `Distance (km)` = round(
            unname(vapply(
              legs,
              function(l) as.numeric(l$distance) / 1000,
              numeric(1)
            )),
            3
          ),
          stringsAsFactors = FALSE,
          check.names = FALSE
        )
      } else {
        # Fallback: single entry
        legs_df <- data.frame(
          Start = 1,
          End = 2,
          `Duration (min)` = round(as.numeric(trip_data$duration) / 60, 2),
          `Distance (km)` = round(as.numeric(trip_data$distance) / 1000, 3),
          stringsAsFactors = FALSE,
          check.names = FALSE
        )
      }

      summary <- list(
        duration = as.numeric(trip_data$duration) / 60,
        distance = as.numeric(trip_data$distance) / 1000
      )

      list(trip = trip_sf_map, legs = legs_df, summary = summary, waypoint_order = waypoint_order)
    },
    error = function(e) {
      if (debug) {
        message("DEBUG [Trip API]: Error: ", e$message)
      }
      NULL
    }
  )
}

test_that("gui_parse_breaks parses input correctly", {
  expect_equal(gui_parse_breaks("5, 10, 15"), c(5, 10, 15))
  expect_equal(gui_parse_breaks(""), c(5, 10, 15))
  expect_equal(gui_parse_breaks(NULL), c(5, 10, 15))
  expect_equal(gui_parse_breaks("5"), 5)
  expect_equal(gui_parse_breaks("10, 5, 15, 15"), c(5, 10, 15))
  expect_equal(gui_parse_breaks("5, abc, 10"), c(5, 10))
  expect_equal(gui_parse_breaks("abc"), c(5, 10, 15))
  expect_equal(gui_parse_breaks(" 12 , 24 "), c(12, 24))
})

test_that("gui_resolve_map_view resolves center and zoom", {
  # Direct center and zoom
  res <- gui_resolve_map_view(center = c(10, 20), zoom = 5, input_osrm = NULL)
  expect_equal(res$center, c(10, 20))
  expect_equal(res$zoom, 5)

  # Named list for center
  res2 <- gui_resolve_map_view(center = list(lng = 10, lat = 20), zoom = 5, input_osrm = NULL)
  expect_equal(res2$center, c(10, 20))
  
  res3 <- gui_resolve_map_view(center = list(lon = 10, lat = 20), zoom = 5, input_osrm = NULL)
  expect_equal(res3$center, c(10, 20))
  
  res4 <- gui_resolve_map_view(center = list(x = 10, y = 20), zoom = 5, input_osrm = NULL)
  expect_equal(res4$center, c(10, 20))

  # Invalid center
  expect_error(gui_resolve_map_view(center = c(10), zoom = 5, input_osrm = NULL), "numeric vector of length 2")
  expect_warning(
    expect_error(gui_resolve_map_view(center = c("a", "b"), zoom = 5, input_osrm = NULL), "numeric vector of length 2"),
    "NAs introduced by coercion"
  )
})

test_that("gui_setup_server error handling works", {
  expect_error(gui_setup_server(NULL, port = "invalid"), "Invalid port specified")
  expect_error(gui_setup_server(NULL, port = -5), "Invalid port specified")
  
  # Mock dead process
  dummy_proc <- structure(list(is_alive = function() FALSE), class = "process")
  expect_error(gui_setup_server(dummy_proc, port = 5000), "The provided OSRM server process is not running")

  # Mock alive process with auto port
  dummy_proc2 <- structure(list(is_alive = function() TRUE), class = "process")
  expect_error(gui_setup_server(dummy_proc2, port = "auto"), "specify the port explicitly")
})

test_that("gui_check_dependencies checks required packages", {
  skip_if_not_installed("osrm")
  skip_if_not_installed("DT")
  skip_if_not_installed("shiny")
  skip_if_not_installed("mapgl")
  skip_if_not_installed("sf")
  skip_if_not_installed("viridisLite")
  
  expect_silent(gui_check_dependencies())
})

test_that("osrm_gui prepares arguments and handles missing servers", {
  skip_if_not_installed("osrm")
  skip_if_not_installed("DT")
  skip_if_not_installed("shiny")
  skip_if_not_installed("mapgl")
  skip_if_not_installed("sf")
  skip_if_not_installed("viridisLite")
  
  # When no servers are running, port='auto' throws error
  local_mocked_bindings(
    osrm_servers = function() {
      data.frame(
        name = character(),
        port = integer(),
        pid = integer(),
        alive = logical(),
        input_osm = character(),
        started_at = .POSIXct(double(0)),
        stringsAsFactors = FALSE
      )
    }
  )
  
  expect_error(osrm_gui(), "No running OSRM servers detected")
})

skip_if_no_osrm <- function() {
  testthat::skip_on_cran()

  if (isTRUE(getOption("osrm.backend.skip_osrm_tests", FALSE))) {
    reason <- getOption("osrm.backend.install_error", "OSRM backend binaries are unavailable")
    testthat::skip(reason)
  }
}

test_that("osrm_extract runs osrm-extract with expected arguments", {
  skip_if_no_osrm()
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  input_osm <- file.path(
    tmp_dir,
    paste0("osrm-test-", Sys.getpid(), ".osm.pbf")
  )
  generated_prefix <- sub("\\.osm\\.pbf$", "", input_osm)
  on.exit(
    {
      artifacts <- list.files(
        tmp_dir,
        pattern = paste0("^", basename(generated_prefix), ""),
        full.names = TRUE
      )
      unlink(artifacts, recursive = TRUE)
    },
    add = TRUE
  )

  file.create(input_osm)
  normalized_input <- normalizePath(input_osm)
  expected_timestamp <- paste0(
    sub("\\.osm\\.pbf$", "", normalized_input, ignore.case = TRUE),
    ".osrm.timestamp"
  )

  captured <- list()
  mock_run <- function(command, args, echo, spinner, echo_cmd, ...) {
    captured <<- list(
      command = command,
      args = args,
      echo = echo,
      spinner = spinner,
      echo_cmd = echo_cmd
    )
    file.create(expected_timestamp)
    list(status = 0, stdout = "", stderr = "")
  }

  result <- with_mocked_bindings(
    osrm_extract(
      input_osm = input_osm,
      profile = "car.lua",
      threads = 2L,
      data_version = "2024-01",
      dump_nbg_graph = TRUE,
      quiet = FALSE,
      spinner = FALSE,
      echo_cmd = FALSE
    ),
    run = mock_run,
    .package = "processx"
  )

  expect_equal(tools::file_path_sans_ext(basename(captured$command)), "osrm-extract")
  expect_true(all(c("-p", "car.lua") %in% captured$args))
  expect_true("--dump-nbg-graph" %in% captured$args)
  expect_true("-d" %in% captured$args && "2024-01" %in% captured$args)
  expect_equal(result$osrm_job_artifact, expected_timestamp)
  expect_true(file.exists(expected_timestamp))
})

test_that("osrm_extract enforces overwrite flag for existing outputs", {
  skip_if_no_osrm()
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  input_osm <- file.path(
    tmp_dir,
    paste0("osrm-overwrite-", Sys.getpid(), ".osm.pbf")
  )
  generated_prefix <- sub("\\.osm\\.pbf$", "", input_osm)
  on.exit(
    {
      artifacts <- list.files(
        tmp_dir,
        pattern = paste0("^", basename(generated_prefix), ""),
        full.names = TRUE
      )
      unlink(artifacts, recursive = TRUE)
    },
    add = TRUE
  )

  file.create(input_osm)
  existing_artifact <- paste0(generated_prefix, ".osrm.hsgr")
  file.create(existing_artifact)

  expect_error(
    with_mocked_bindings(
      osrm_extract(
        input_osm = input_osm,
        profile = "car.lua"
      ),
      run = function(...) stop("processx::run should not be called"),
      .package = "processx"
    ),
    "Found existing OSRM files"
  )

  normalized_input <- normalizePath(input_osm)
  expected_timestamp <- paste0(
    sub("\\.osm\\.pbf$", "", normalized_input, ignore.case = TRUE),
    ".osrm.timestamp"
  )

  result <- with_mocked_bindings(
    osrm_extract(
      input_osm = input_osm,
      profile = "car.lua",
      overwrite = TRUE,
      quiet = FALSE,
      spinner = FALSE
    ),
    run = function(...) {
      file.create(expected_timestamp)
      list(status = 0, stdout = "", stderr = "")
    },
    .package = "processx"
  )

  expect_equal(result$osrm_job_artifact, expected_timestamp)
  expect_true(file.exists(expected_timestamp))
})

test_that("osrm_extract accepts a directory with one OSM file", {
  skip_if_no_osrm()
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-dir-test-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  input_osm <- file.path(tmp_dir, "test.osm.pbf")
  file.create(input_osm)

  normalized_input <- normalizePath(input_osm)
  expected_timestamp <- paste0(
    sub("\\.osm\\.pbf$", "", normalized_input, ignore.case = TRUE),
    ".osrm.timestamp"
  )

  mock_run <- function(command, args, echo, spinner, echo_cmd, ...) {
    file.create(expected_timestamp)
    list(status = 0, stdout = "", stderr = "")
  }

  result <- with_mocked_bindings(
    osrm_extract(
      input_osm = tmp_dir,
      profile = "car.lua",
      quiet = TRUE
    ),
    run = mock_run,
    .package = "processx"
  )

  expect_equal(result$osrm_job_artifact, expected_timestamp)
})

test_that("osrm_extract errors when directory has no OSM files", {
  skip_if_no_osrm()
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-empty-dir-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  expect_error(
    osrm_extract(
      input_osm = tmp_dir,
      profile = "car.lua"
    ),
    "No OSM files.*found in directory"
  )
})

test_that("osrm_extract errors when directory has multiple OSM files", {
  skip_if_no_osrm()
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-multi-dir-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  file.create(file.path(tmp_dir, "test1.osm.pbf"))
  file.create(file.path(tmp_dir, "test2.osm.pbf"))

  expect_error(
    osrm_extract(
      input_osm = tmp_dir,
      profile = "car.lua"
    ),
    "Multiple.*found"
  )
})

#' Contract OSRM Graph for Contraction Hierarchies (CH)
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Run the `osrm-contract` tool to contract an OSRM graph for the CH pipeline.
#' After running, a companion `<base>.osrm.hsgr` file must exist to confirm success.
#'
#' @param input_osrm A string. Path to a `.osrm.timestamp` file, the base path to the `.osrm` files (without extension),
#'   or a directory containing exactly one `.osrm.timestamp` file.
#' @param threads An integer. Number of threads to use; default `8`.
#' @param verbosity A string. Log verbosity level passed to `-l/--verbosity`
#'   (one of `"NONE","ERROR","WARNING","INFO","DEBUG"`); default `"INFO"`.
#' @param segment_speed_file A string or `NULL`. Path to nodeA,nodeB,speed CSV; default `NULL`.
#' @param turn_penalty_file A string or `NULL`. Path to from_,to_,via_nodes,penalties CSV; default `NULL`.
#' @param edge_weight_updates_over_factor A numeric. Threshold for logging large weight updates; default `0`.
#' @param parse_conditionals_from_now A numeric. UTC timestamp for conditional restrictions; default `0`.
#' @param time_zone_file A string or `NULL`. GeoJSON file for time zone boundaries; default `NULL`.
#' @inheritParams osrm_prepare_graph
#'
#' @return An object of class \code{osrm_job} with the following elements:
#' \describe{
#'   \item{osrm_job_artifact}{The path to the contracted `.osrm.hsgr` file.}
#'   \item{osrm_working_dir}{The directory containing all OSRM files.}
#'   \item{logs}{The \code{processx::run} result object.}
#' }
#'
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   # Prepare a small graph then contract it for the CH pipeline
#'   pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
#'   osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
#'   dir.create(osrm_dir, recursive = TRUE)
#'   tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
#'   file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
#'   profile <- osrm_find_profile("car.lua")
#'
#'   extract_job <- osrm_extract(
#'     input_osm = tmp_pbf,
#'     profile = profile,
#'     overwrite = TRUE,
#'     threads = 1L
#'   )
#'
#'   ch_graph <- osrm_contract(extract_job, threads = 1L, verbose = TRUE)
#'   ch_graph$osrm_job_artifact
#'
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'   unlink(osrm_dir, recursive = TRUE)
#' }
#' }
#'
#' @export
osrm_contract <- function(
  input_osrm,
  threads = 8L,
  verbosity = c("INFO", "NONE", "ERROR", "WARNING", "DEBUG"),
  segment_speed_file = NULL,
  turn_penalty_file = NULL,
  edge_weight_updates_over_factor = 0,
  parse_conditionals_from_now = 0,
  time_zone_file = NULL,
  quiet = FALSE,
  verbose = FALSE,
  spinner = TRUE,
  echo_cmd = FALSE
) {
  if (!requireNamespace("processx", quietly = TRUE)) {
    stop("'processx' package is required for osrm_contract", call. = FALSE)
  }

  # Extract previous logs if input is an osrm_job object
  input_logs <- if (inherits(input_osrm, "osrm_job")) {
    if (is.list(input_osrm$logs) && length(input_osrm$logs) > 0) {
      input_osrm$logs
    } else {
      list()
    }
  } else {
    list()
  }

  # Extract path from osrm_job object if needed
  input_osrm <- get_osrm_path_from_input(input_osrm)

  if (!is.character(input_osrm) || length(input_osrm) != 1) {
    stop(
      "'input_osrm' must be a single string (path to .osrm base)",
      call. = FALSE
    )
  }

  # Check if user is trying to use contract after partition
  if (grepl("\\.partition$", input_osrm, ignore.case = TRUE)) {
    stop(
      "`osrm_contract` cannot be used after `osrm_partition`.\n",
      "These are part of different pipelines:\n",
      "  - CH pipeline: extract -> contract\n",
      "  - MLD pipeline: extract -> partition -> customize\n",
      "After `osrm_partition`, use `osrm_customize` instead.",
      call. = FALSE
    )
  }

  # Resolve input path (file or directory)
  input_osrm <- resolve_osrm_path(
    input_osrm,
    pattern = "\\.osrm\\.timestamp$",
    file_description = ".osrm.timestamp files",
    error_context = "Please check that you have run `osrm_extract` first."
  )

  if (!grepl("\\.timestamp$", input_osrm, ignore.case = TRUE)) {
    stop(
      "'input_osrm' must have extension .timestamp",
      call. = FALSE
    )
  }

  if (!file.exists(input_osrm)) {
    stop(
      "File does not exist: ",
      input_osrm,
      "\nPlease check that you have run `osrm_extract` first.",
      call. = FALSE
    )
  }

  osrm_path <- gsub("\\.timestamp$", "", input_osrm)

  # Check for algorithm conflicts (MLD files in directory)
  base_name <- sub("\\.osrm$", "", basename(osrm_path))
  dir_path <- dirname(osrm_path)
  check_algorithm_conflict(dir_path, base_name, "ch", "contract")

  verbosity <- match.arg(verbosity)
  arguments <- c(
    osrm_path,
    "-l",
    verbosity,
    "-t",
    as.character(threads)
  )
  if (!is.null(segment_speed_file)) {
    arguments <- c(arguments, "--segment-speed-file", segment_speed_file)
  }
  if (!is.null(turn_penalty_file)) {
    arguments <- c(arguments, "--turn-penalty-file", turn_penalty_file)
  }
  if (!is.null(edge_weight_updates_over_factor)) {
    arguments <- c(
      arguments,
      "--edge-weight-updates-over-factor",
      as.character(edge_weight_updates_over_factor)
    )
  }
  if (!is.null(parse_conditionals_from_now)) {
    arguments <- c(
      arguments,
      "--parse-conditionals-from-now",
      as.character(parse_conditionals_from_now)
    )
  }
  if (!is.null(time_zone_file)) {
    arguments <- c(arguments, "--time-zone-file", time_zone_file)
  }

  # Determine final processx parameters
  show_echo <- !quiet && verbose
  show_spinner <- !quiet && spinner
  show_echo_cmd <- !quiet && echo_cmd

  logs <- processx::run(
    resolve_osrm_bin("osrm-contract"),
    args = arguments,
    echo = show_echo,
    spinner = show_spinner,
    echo_cmd = show_echo_cmd
  )

  hsgr_file <- paste0(gsub("\\.osrm$", "", osrm_path), ".osrm.hsgr")
  if (!file.exists(hsgr_file)) {
    stop(
      "Contracting did not produce CH hierarchy file: ",
      basename(hsgr_file),
      call. = FALSE
    )
  }

  # Accumulate logs from previous stages
  accumulated_logs <- c(input_logs, list(contract = logs))

  as_osrm_job(
    osrm_job_artifact = normalizePath(hsgr_file),
    osrm_working_dir = dirname(normalizePath(hsgr_file)),
    logs = accumulated_logs
  )
}

test_that("osrm_contract runs osrm-contract with expected arguments", {
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  input_osrm <- file.path(
    tmp_dir,
    paste0("osrm-contract-test-", Sys.getpid(), ".osrm.timestamp")
  )
  on.exit(unlink(input_osrm), add = TRUE)
  file.create(input_osrm)

  captured <- list()
  mock_run <- function(command, args, echo, spinner, echo_cmd, ...) {
    captured <<- list(
      command = command,
      args = args,
      echo = echo,
      spinner = spinner,
      echo_cmd = echo_cmd
    )
    # Create the expected output file to satisfy the check
    hsgr_file <- sub("\\.timestamp$", ".hsgr", input_osrm)
    file.create(hsgr_file)
    list(status = 0, stdout = "", stderr = "")
  }

  result <- with_mocked_bindings(
    osrm_contract(
      input_osrm = input_osrm,
      threads = 4L,
      verbosity = "DEBUG",
      segment_speed_file = "speeds.csv",
      turn_penalty_file = "penalties.csv",
      edge_weight_updates_over_factor = 1.5,
      parse_conditionals_from_now = 1234567890,
      time_zone_file = "zones.geojson",
      quiet = FALSE,
      verbose = TRUE,
      spinner = FALSE,
      echo_cmd = TRUE
    ),
    run = mock_run,
    .package = "processx"
  )

  expect_equal(tools::file_path_sans_ext(basename(captured$command)), "osrm-contract")
  expect_true("-t" %in% captured$args && "4" %in% captured$args)
  expect_true("-l" %in% captured$args && "DEBUG" %in% captured$args)
  expect_true(
    "--segment-speed-file" %in% captured$args && "speeds.csv" %in% captured$args
  )
  expect_true(
    "--turn-penalty-file" %in%
      captured$args &&
      "penalties.csv" %in% captured$args
  )
  expect_true(
    "--edge-weight-updates-over-factor" %in%
      captured$args &&
      "1.5" %in% captured$args
  )
  expect_true(
    "--parse-conditionals-from-now" %in%
      captured$args &&
      "1234567890" %in% captured$args
  )
  expect_true(
    "--time-zone-file" %in% captured$args && "zones.geojson" %in% captured$args
  )

  # Check processx options
  expect_true(captured$echo)
  expect_false(captured$spinner)
  expect_true(captured$echo_cmd)
})

test_that("osrm_contract handles missing input file", {
  expect_error(
    osrm_contract("nonexistent.osrm.timestamp"),
    "File does not exist"
  )
})

test_that("osrm_contract checks for output file", {
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  input_osrm <- file.path(
    tmp_dir,
    paste0("osrm-fail-", Sys.getpid(), ".osrm.timestamp")
  )
  on.exit(unlink(input_osrm), add = TRUE)
  file.create(input_osrm)

  mock_run <- function(...) {
    list(status = 0, stdout = "", stderr = "")
  }

  expect_error(
    with_mocked_bindings(
      osrm_contract(input_osrm),
      run = mock_run,
      .package = "processx"
    ),
    "Contracting did not produce CH hierarchy file"
  )
})

test_that("osrm_contract accepts a directory with one .osrm.timestamp file", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-contract-dir-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  input_osrm <- file.path(tmp_dir, "test.osrm.timestamp")
  file.create(input_osrm)

  normalized_input <- normalizePath(input_osrm)
  expected_hsgr <- sub("\\.timestamp$", ".hsgr", normalized_input)

  mock_run <- function(command, args, echo, spinner, echo_cmd, ...) {
    file.create(expected_hsgr)
    list(status = 0, stdout = "", stderr = "")
  }

  result <- with_mocked_bindings(
    osrm_contract(
      input_osrm = tmp_dir,
      quiet = TRUE
    ),
    run = mock_run,
    .package = "processx"
  )

  expect_equal(result$osrm_job_artifact, expected_hsgr)
})

test_that("osrm_contract errors when directory has no .osrm.timestamp files", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-contract-empty-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  expect_error(
    osrm_contract(input_osrm = tmp_dir),
    "No .osrm.timestamp files found"
  )
})

test_that("osrm_contract errors when directory has multiple .osrm.timestamp files", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-contract-multi-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  file.create(file.path(tmp_dir, "test1.osrm.timestamp"))
  file.create(file.path(tmp_dir, "test2.osrm.timestamp"))

  expect_error(
    osrm_contract(input_osrm = tmp_dir),
    "Multiple .osrm.timestamp files found"
  )
})

test_that("osrm_contract gives helpful error when used after osrm_partition", {
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  partition_file <- file.path(
    tmp_dir,
    paste0("osrm-partition-test-", Sys.getpid(), ".osrm.partition")
  )
  on.exit(unlink(partition_file), add = TRUE)
  file.create(partition_file)

  # Create a mock osrm_job object with a partition file (simulating osrm_partition output)
  mock_job <- structure(
    list(
      osrm_job_artifact = partition_file,
      osrm_working_dir = tmp_dir,
      logs = list()
    ),
    class = "osrm_job"
  )

  expect_error(
    osrm_contract(input_osrm = mock_job),
    "cannot be used after.*partition"
  )
  expect_error(
    osrm_contract(input_osrm = mock_job),
    "MLD pipeline"
  )
  expect_error(
    osrm_contract(input_osrm = mock_job),
    "osrm_customize"
  )
})

test_that("CH pipeline: extract -> contract works correctly", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-ch-pipeline-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  input_osm_path <- file.path(tmp_dir, "test.osm.pbf")
  file.create(input_osm_path)

  mock_run <- function(command, args, ...) {
    input_path <- args[1]
    cmd <- tools::file_path_sans_ext(basename(command))

    if (cmd == "osrm-extract") {
      base <- sub("\\.osm\\.pbf$", "", input_osm_path)
      timestamp_file <- paste0(base, ".osrm.timestamp")
      file.create(timestamp_file)
    } else if (cmd == "osrm-partition") {
      partition_file <- paste0(input_path, ".partition")
      file.create(partition_file)
    } else if (cmd == "osrm-contract") {
      # Input is test.osrm (without extension), create test.osrm.hsgr
      hsgr_file <- paste0(input_path, ".hsgr")
      file.create(hsgr_file)
    } else if (cmd == "osrm-customize") {
      mldgr_file <- paste0(input_path, ".mldgr")
      file.create(mldgr_file)
    }
    list(status = 0, stdout = "", stderr = "")
  }

  # Test the pipeline
  result <- with_mocked_bindings(
    {
      extract_result <- osrm_extract(
        input_osm_path,
        profile = "car.lua",
        quiet = TRUE
      )
      osrm_contract(extract_result, quiet = TRUE)
    },
    run = mock_run,
    .package = "processx"
  )

  # Verify the result
  expect_s3_class(result, "osrm_job")
  expect_true(grepl("\\.hsgr$", result$osrm_job_artifact))
  expect_true(file.exists(result$osrm_job_artifact))
  expect_named(result$logs, c("extract", "contract"))
})

test_that("MLD pipeline: extract -> partition -> customize works correctly", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-mld-pipeline-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  input_osm_path <- file.path(tmp_dir, "test.osm.pbf")
  file.create(input_osm_path)

  mock_run <- function(command, args, ...) {
    input_path <- args[1]
    cmd <- tools::file_path_sans_ext(basename(command))

    if (cmd == "osrm-extract") {
      base <- sub("\\.osm\\.pbf$", "", input_osm_path)
      timestamp_file <- paste0(base, ".osrm.timestamp")
      file.create(timestamp_file)
    } else if (cmd == "osrm-partition") {
      partition_file <- paste0(input_path, ".partition")
      file.create(partition_file)
    } else if (cmd == "osrm-contract") {
      # Input is test.osrm (without extension), create test.osrm.hsgr
      hsgr_file <- paste0(input_path, ".hsgr")
      file.create(hsgr_file)
    } else if (cmd == "osrm-customize") {
      mldgr_file <- paste0(input_path, ".mldgr")
      file.create(mldgr_file)
    }
    list(status = 0, stdout = "", stderr = "")
  }

  result <- with_mocked_bindings(
    {
      extract_result <- osrm_extract(
        input_osm_path,
        profile = "car.lua",
        quiet = TRUE
      )
      partition_result <- osrm_partition(extract_result, quiet = TRUE)
      osrm_customize(partition_result, quiet = TRUE)
    },
    run = mock_run,
    .package = "processx"
  )

  # Verify the result
  expect_s3_class(result, "osrm_job")
  expect_true(grepl("\\.mldgr$", result$osrm_job_artifact))
  expect_true(file.exists(result$osrm_job_artifact))
  expect_named(result$logs, c("extract", "partition", "customize"))
})

test_that("Mixed pipelines fail with helpful errors: extract -> partition -> contract", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-mixed-pipeline-1-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  input_osm_path <- file.path(tmp_dir, "test.osm.pbf")
  file.create(input_osm_path)

  mock_run <- function(command, args, ...) {
    input_path <- args[1]
    cmd <- tools::file_path_sans_ext(basename(command))

    if (cmd == "osrm-extract") {
      base <- sub("\\.osm\\.pbf$", "", input_osm_path)
      timestamp_file <- paste0(base, ".osrm.timestamp")
      file.create(timestamp_file)
    } else if (cmd == "osrm-partition") {
      partition_file <- paste0(input_path, ".partition")
      file.create(partition_file)
    } else if (cmd == "osrm-contract") {
      # Input is test.osrm (without extension), create test.osrm.hsgr
      hsgr_file <- paste0(input_path, ".hsgr")
      file.create(hsgr_file)
    } else if (cmd == "osrm-customize") {
      mldgr_file <- paste0(input_path, ".mldgr")
      file.create(mldgr_file)
    }
    list(status = 0, stdout = "", stderr = "")
  }

  expect_error(
    with_mocked_bindings(
      {
        extract_result <- osrm_extract(
          input_osm_path,
          profile = "car.lua",
          quiet = TRUE
        )
        partition_result <- osrm_partition(extract_result, quiet = TRUE)
        osrm_contract(partition_result, quiet = TRUE)
      },
      run = mock_run,
      .package = "processx"
    ),
    "cannot be used after.*partition"
  )
})

test_that("Mixed pipelines fail with helpful errors: extract -> contract -> partition", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-mixed-pipeline-2-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  input_osm_path <- file.path(tmp_dir, "test.osm.pbf")
  file.create(input_osm_path)

  # Create a mock osrm_job object with .hsgr file (simulating result after contract)
  hsgr_file <- file.path(tmp_dir, "test.osrm.hsgr")
  file.create(hsgr_file)

  contract_result <- structure(
    list(
      osrm_job_artifact = hsgr_file,
      osrm_working_dir = tmp_dir,
      logs = list(extract = list(), contract = list())
    ),
    class = "osrm_job"
  )

  # Trying to partition after contract should fail
  # (partition expects a .timestamp file, not .hsgr)
  # This will fail either with a "timestamp" error or a command execution error
  expect_error(
    osrm_partition(contract_result, quiet = TRUE)
  )
})

test_that("Mixed pipelines fail with helpful errors: extract -> customize (without partition)", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-mixed-pipeline-3-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  input_osm_path <- file.path(tmp_dir, "test.osm.pbf")
  file.create(input_osm_path)

  mock_run <- function(command, args, ...) {
    cmd <- tools::file_path_sans_ext(basename(command))
    if (cmd == "osrm-extract") {
      base <- sub("\\.osm\\.pbf$", "", input_osm_path)
      timestamp_file <- paste0(base, ".osrm.timestamp")
      file.create(timestamp_file)
    }
    list(status = 0, stdout = "", stderr = "")
  }

  expect_error(
    with_mocked_bindings(
      {
        extract_result <- osrm_extract(
          input_osm_path,
          profile = "car.lua",
          quiet = TRUE
        )
        osrm_customize(extract_result, quiet = TRUE)
      },
      run = mock_run,
      .package = "processx"
    ),
    "requires a partitioned graph"
  )
})

#' Locate an OSRM Lua profile (e.g. car.lua) in a host installation
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' By default OSRM ships profiles for "car", "bike" and "foot" in a
#' `profiles/` directory alongside the binaries.  This function will try to
#' locate `osrm-routed` on the `PATH`, resolve symlinks, and look first for a
#' `profiles/` directory next to the binary (as placed there by
#' `osrm_install()`). If that fails, it looks for sibling directories
#' `share/osrm/profiles` and `share/osrm-backend/profiles`. IF that fails, it
#' will try to fall back on `/usr/local/share/osrm/profiles`,
#' `/usr/local/share/osrm-backend/profiles`, `/usr/share/osrm/profiles`, and
#' `/usr/share/osrm-backend/profiles`.
#'
#' @param profile A single string, the name of the Lua profile file
#'   (e.g. \code{"car.lua"}). Defaults to \code{"car.lua"}.
#' @return The normalized filesystem path to the profile.
#'
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'   osrm_find_profile("car.lua")
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#' }
#' }
#'
#' @export
osrm_find_profile <- function(
  profile = "car.lua"
) {
  # locate and resolve osrm-routed
  osrm_bin <- resolve_osrm_bin("osrm-routed")
  if (!nzchar(osrm_bin) || (!file.exists(osrm_bin) && !nzchar(Sys.which(osrm_bin)))) {
    stop(
      "`osrm-routed` not found in PATH; please install OSRM backend",
      call. = FALSE
    )
  }
  real_bin <- try(normalizePath(osrm_bin), silent = TRUE)
  if (inherits(real_bin, "try-error")) {
    real_bin <- osrm_bin
  }
  bindir <- dirname(real_bin)

  # probe common relative locations
  candidates <- c(
    file.path(bindir, "profiles", profile),
    file.path(bindir, "..", "share", "osrm", "profiles", profile),
    file.path(bindir, "..", "share", "osrm-backend", "profiles", profile),
    file.path(bindir, "..", "lib", "osrm", "profiles", profile),
    file.path(bindir, "..", "lib", "osrm-backend", "profiles", profile)
  )

  if (.Platform$OS.type != "windows") {
    unix_candidates <- c(
      file.path("/", "usr", "local", "share", "osrm", "profiles", profile),
      file.path(
        "/",
        "usr",
        "local",
        "share",
        "osrm-backend",
        "profiles",
        profile
      ),
      file.path("/", "usr", "share", "osrm", "profiles", profile),
      file.path("/", "usr", "share", "osrm-backend", "profiles", profile)
    )
    candidates <- c(candidates, unix_candidates)
  }
  for (p in candidates) {
    if (file.exists(p)) {
      return(normalizePath(p))
    }
  }

  stop(
    sprintf(
      "Could not locate OSRM profile '%s'. Please locate it manually and provide the full path.",
      profile
    ),
    call. = FALSE
  )
}

# R/osrm_server_registry.R  -- registry with tools::R_user_dir() and jsonlite JSON persistence

# Package-private state --------------------------------------------------------

.osrm_state <- new.env(parent = emptyenv())
.osrm_state$registry <- list() # id -> list(id, pid, port, prefix, algorithm, started_at, proc?)
.osrm_state$session_id <- NULL # Unique ID for this R session

# Small helpers ---------------------------------------------------------------

`%||%` <- function(a, b) {
  if (is.null(a) || (is.character(a) && length(a) == 1 && !nzchar(a))) b else a
}

.osrm_registry_dir <- function() {
  # CI/override first
  opt <- getOption("osrm.server.state_dir")
  if (!is.null(opt) && nzchar(opt)) {
    return(opt)
  }

  env <- Sys.getenv("OSRM_BACKEND_STATE_DIR", unset = "")
  if (nzchar(env)) {
    return(env)
  }

  # Non-interactive: tempdir() is safest to avoid collisions on CI
  if (!interactive()) {
    return(tempdir())
  }

  # Interactive: stable per-user cache dir (base R)
  tools::R_user_dir("osrm.backend", which = "cache")
}

.osrm_registry_path <- function() {
  # If session ID isn't set (shouldn't happen if loaded), generate one
  if (is.null(.osrm_state$session_id)) {
    .osrm_state$session_id <- sprintf(
      "session-%s-%s-%s", 
      Sys.getpid(), 
      format(Sys.time(), "%Y%m%d%H%M%OS3"),
      paste0(sample(c(0:9, letters), 6, replace = TRUE), collapse = "")
    )
  }
  file.path(.osrm_registry_dir(), paste0(.osrm_state$session_id, ".json"))
}

# Persistence (atomic JSON via jsonlite) --------------------------------------

.osrm_registry_save <- function() {
  dir <- .osrm_registry_dir()
  if (!dir.exists(dir)) {
    dir.create(dir, recursive = TRUE, showWarnings = FALSE)
  }

  fn <- .osrm_registry_path()
  tmp <- paste0(fn, ".tmp")

  # Strip non-serializable fields (processx::process) before writing
  reg <- .osrm_state$registry
  reg_serializable <- lapply(reg, function(x) {
    x$proc <- NULL
    x
  })

  # Write JSON atomically
  jsonlite::write_json(reg_serializable, tmp, auto_unbox = TRUE, pretty = TRUE)
  ok <- suppressWarnings(file.rename(tmp, fn))
  if (!ok) {
    suppressWarnings(file.copy(tmp, fn, overwrite = TRUE))
    unlink(tmp, force = TRUE)
  }
}

.osrm_registry_load <- function() {
  fn <- .osrm_registry_path()
  if (!file.exists(fn)) {
    .osrm_state$registry <- list()
    return(invisible(NULL))
  }
  reg <- .osrm_read_registry_file(fn)
  .osrm_state$registry <- reg
  .osrm_cleanup_orphans() # Clean up dead processes in *this* session's registry
  invisible(NULL)
}

.osrm_read_registry_file <- function(fn) {
  reg <- try(jsonlite::read_json(fn, simplifyVector = TRUE), silent = TRUE)
  if (inherits(reg, "try-error") || !is.list(reg)) {
    return(list())
  }
  # Ensure list with names = ids when possible
  if (!is.null(names(reg)) && all(nzchar(names(reg)))) {
    return(reg)
  } else {
    idx <- vapply(
      reg,
      function(e) is.list(e) && length(e$id) == 1 && nzchar(e$id),
      logical(1)
    )
    if (any(idx)) {
      return(stats::setNames(
        reg[idx],
        vapply(reg[idx], `[[`, "", "id")
      ))
    }
  }
  list()
}

# Cross-session Management ----------------------------------------------------

# Scan all registry files in the cache dir to find servers from other sessions
.osrm_registry_scan_others <- function() {
  dir <- .osrm_registry_dir()
  if (!dir.exists(dir)) return(list())
  
  files <- list.files(dir, pattern = "^session-.*\\.json$", full.names = TRUE)
  current_fn <- .osrm_registry_path()
  
  # Exclude current session
  files <- setdiff(files, current_fn)
  
  all_external <- list()
  
  for (f in files) {
    reg <- .osrm_read_registry_file(f)
    if (length(reg)) {
      # Filter for alive processes only
      alive_reg <- list()
      has_alive <- FALSE
      for (entry in reg) {
        if (!is.null(entry$pid) && .osrm_pid_is_running(entry$pid)) {
          # Mark as external from registry
          entry$is_external_registry <- TRUE 
          alive_reg[[entry$id]] <- entry
          has_alive <- TRUE
        }
      }
      if (has_alive) {
        all_external <- c(all_external, alive_reg)
        # If the owner R session is dead, we can safely prune the file on disk
        if (has_alive && length(alive_reg) < length(reg)) {
          owner_pid <- as.integer(sub("^session-([0-9]+)-.*", "\\1", basename(f)))
          if (!is.na(owner_pid) && !.osrm_pid_is_running(owner_pid)) {
            # Strip the temporary flag before saving
            save_reg <- lapply(alive_reg, function(x) {
              x$is_external_registry <- NULL
              x
            })
            # Atomic write to prevent race conditions
            tmp_f <- paste0(f, ".tmp-", Sys.getpid())
            try({
              jsonlite::write_json(save_reg, tmp_f, auto_unbox = TRUE, pretty = TRUE)
              file.rename(tmp_f, f)
            }, silent = TRUE)
            if (file.exists(tmp_f)) unlink(tmp_f)
          }
        }
      } else {
        # File contains only dead processes -> garbage collect it
        try(unlink(f), silent = TRUE)
      }
    } else {
      # Empty file -> garbage collect
      try(unlink(f), silent = TRUE)
    }
  }
  
  all_external
}

# Process utilities (ps optional) ---------------------------------------------

.osrm_has_ps <- function() {
  requireNamespace("ps", quietly = TRUE)
}

.osrm_pid_is_running <- function(pid) {
  pid <- suppressWarnings(as.integer(pid))
  if (is.na(pid) || pid <= 0) {
    return(FALSE)
  }

  if (.osrm_has_ps()) {
    h <- try(ps::ps_handle(pid), silent = TRUE)
    if (inherits(h, "try-error")) {
      return(FALSE)
    }
    return(tryCatch(ps::ps_is_running(h), error = function(...) FALSE))
  }

  # Without ps: best-effort
  if (.Platform$OS.type == "unix" && file.exists("/proc")) {
    return(dir.exists(file.path("/proc", pid)))
  }
  # Conservative fallback (can't reliably check): assume TRUE
  TRUE
}

.osrm_kill_pid <- function(pid) {
  pid <- suppressWarnings(as.integer(pid))
  if (is.na(pid) || pid <= 0) {
    return(invisible(FALSE))
  }

  if (.osrm_has_ps()) {
    h <- try(ps::ps_handle(pid), silent = TRUE)
    if (!inherits(h, "try-error")) {
      if (tryCatch(ps::ps_is_running(h), error = function(...) FALSE)) {
        try(ps::ps_kill(h), silent = TRUE)
      }
    }
    return(invisible(TRUE))
  }

  if (.Platform$OS.type == "unix") {
    # SIGTERM-equivalent; OSRM should exit promptly
    try(tools::pskill(pid), silent = TRUE)
  } else {
    tk <- Sys.which("taskkill")
    if (nzchar(tk)) {
      system2(tk, c("/PID", pid, "/F"), stdout = FALSE, stderr = FALSE)
    }
  }
  invisible(TRUE)
}

.osrm_cleanup_orphans <- function() {
  reg <- .osrm_state$registry
  if (!length(reg)) {
    return(invisible(NULL))
  }

  dead_ids <- character(0)
  for (nm in names(reg)) {
    e <- reg[[nm]]
    alive <- FALSE
    if (!is.null(e$proc) && inherits(e$proc, "process")) {
      alive <- tryCatch(e$proc$is_alive(), error = function(...) FALSE)
    } else if (!is.null(e$pid)) {
      alive <- .osrm_pid_is_running(e$pid)
    }
    if (!alive) dead_ids <- c(dead_ids, nm)
  }

  if (length(dead_ids)) {
    for (id in dead_ids) {
      reg[[id]] <- NULL
    }
    .osrm_state$registry <- reg
    .osrm_registry_save()
  }
  invisible(NULL)
}

# Registration API used by osrm_start_server() --------------------------------

.osrm_register <- function(proc, port, prefix, algorithm, log = NULL, input_osm = NULL, profile = NULL, center = NULL) {
  ts <- format(Sys.time(), "%Y-%m-%dT%H:%M:%OS3Z", tz = "UTC", usetz = FALSE)
  pid <- tryCatch(
    if (!is.null(proc)) proc$get_pid() else NA_integer_,
    error = function(...) NA_integer_
  )
  prt <- suppressWarnings(as.integer(port))
  if (is.na(prt)) {
    prt <- NA_integer_
  }

  id <- sprintf(
    "osrm-%s-%s",
    if (!is.na(prt)) prt else "na",
    if (!is.na(pid)) pid else "na"
  )

  # Validate center is numeric length 2 or NA
  center_val <- if (!is.null(center) && length(center) == 2 && all(is.numeric(center))) {
    as.numeric(center)
  } else {
    c(NA_real_, NA_real_)
  }

  .osrm_state$registry[[id]] <- list(
    id = id,
    pid = pid,
    port = prt,
    prefix = as.character(prefix %||% ""),
    algorithm = as.character(algorithm %||% ""),
    started_at = ts,
    log = as.character(log %||% ""),
    input_osm = as.character(input_osm %||% ""),
    profile = as.character(profile %||% ""),
    center_lon = center_val[1],
    center_lat = center_val[2],
    proc = proc
  )

  .osrm_registry_save()
  id
}

.osrm_deregister <- function(id) {
  if (length(.osrm_state$registry)) {
    .osrm_state$registry[[id]] <- NULL
    .osrm_registry_save()
  }
  invisible(NULL)
}

# Internal stop-all for unload -------------------------------------------------

.osrm_stop_all_internal <- function() {
  reg <- .osrm_state$registry
  if (!length(reg)) {
    # Even if our registry is empty, we should check if we left a file on disk
    # and clean it up on unload if it's empty
    if (!is.null(.osrm_state$session_id)) {
      path <- .osrm_registry_path()
      if (file.exists(path)) try(unlink(path), silent = TRUE)
    }
    return(invisible(NULL))
  }

  for (e in reg) {
    if (!is.null(e$proc) && inherits(e$proc, "process")) {
      if (tryCatch(e$proc$is_alive(), error = function(...) FALSE)) {
        try(e$proc$kill(), silent = TRUE)
        try(e$proc$wait(500), silent = TRUE)
      }
    } else if (!is.null(e$pid)) {
      if (.osrm_pid_is_running(e$pid)) .osrm_kill_pid(e$pid)
    }
  }
  .osrm_state$registry <- list()
  
  # On unload, try to remove the registry file completely
  if (!is.null(.osrm_state$session_id)) {
    path <- .osrm_registry_path()
    if (file.exists(path)) try(unlink(path), silent = TRUE)
  }
  invisible(NULL)
}

# Lifecycle hooks -------------------------------------------------------------

.onLoad <- function(libname, pkgname) {
  # Initialize session ID
  .osrm_state$session_id <- sprintf(
    "session-%s-%s-%s", 
    Sys.getpid(), 
    format(Sys.time(), "%Y%m%d%H%M%OS3"),
    paste0(sample(c(0:9, letters), 6, replace = TRUE), collapse = "")
  )
  .osrm_registry_load()
}

.onUnload <- function(libpath) {
  stop_on_unload <- getOption(
    "osrm.server.stop_on_unload",
    default = !interactive()
  )
  if (isTRUE(stop_on_unload)) {
    try(.osrm_stop_all_internal(), silent = TRUE)
  } else {
    # If not stopping, verify if registry file is empty and clean up if so
    if (length(.osrm_state$registry) == 0 && !is.null(.osrm_state$session_id)) {
      path <- .osrm_registry_path()
      if (file.exists(path)) try(unlink(path), silent = TRUE)
    }
  }
}

#' Retrieve the OSRM Profile for a Running Server
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Attempts to determine the profile (e.g., "car", "bike", "foot") used by an OSRM server.
#' It follows a priority list:
#' 1. Checks the active server registry for the given port or ID.
#' 2. Checks for a `dataset.meta.json` file in the directory of the graph file.
#' 3. Checks the graph filename for hints (e.g. `berlin-car.osrm`).
#' 4. Falls back to `getOption("osrm.profile")`.
#'
#' @param input_osrm Optional. Can be an OSRM job process (an `osrm_server`
#'   object inheriting from `processx::process`), a path string, or NULL.
#' @param port Optional integer. The port of the server.
#' @return A character string representing the profile name (default "car").
#' @export
osrm_get_server_profile <- function(input_osrm = NULL, port = NULL) {
  # 1. Check Registry (if port is known or input_osrm is process)
  reg_profile <- NULL
  
  if (inherits(input_osrm, "process")) {
    # It's an OSRM job process
    pid <- tryCatch(input_osrm$get_pid(), error = function(e) NA)
    if (!is.na(pid)) {
      reg <- .osrm_state$registry
      match <- Filter(function(x) identical(x$pid, pid), reg)
      if (length(match) > 0 && nzchar(match[[1]]$profile)) {
        return(match[[1]]$profile)
      }
    }
  } else if (!is.null(port)) {
    # Check by port
    prt <- as.integer(port)
    reg <- .osrm_state$registry
    # Find active server on this port
    match <- Filter(function(x) identical(x$port, prt), reg)
    if (length(match) > 0) {
      # Use the most recently started one
      last <- match[[length(match)]]
      if (nzchar(last$profile)) return(last$profile)
      
      # If registry entry exists but has no profile, maybe we can look up its path?
      # (Implementation detail: 'prefix' in registry stores the path)
      if (nzchar(last$prefix)) input_osrm <- last$prefix
    }
  }

  # 2. Check Metadata File (if we have a path)
  path <- NULL
  if (is.character(input_osrm) && length(input_osrm) == 1) {
    path <- input_osrm
  }
  
  if (!is.null(path)) {
    # Check dataset.meta.json
    meta_path <- file.path(dirname(path), "dataset.meta.json")
    if (file.exists(meta_path)) {
      meta <- tryCatch(jsonlite::read_json(meta_path), error = function(e) NULL)
      if (!is.null(meta) && !is.null(meta$profile) && nzchar(meta$profile)) {
        return(meta$profile)
      }
    }
    
    # 3. Heuristic: Check filename (e.g. "berlin-car.osrm")
    base <- basename(path)
    if (grepl("car", base, ignore.case = TRUE)) return("car")
    if (grepl("bike|bicycle", base, ignore.case = TRUE)) return("bike")
    if (grepl("foot|walk", base, ignore.case = TRUE)) return("foot")
  }

  # 4. Fallback
  getOption("osrm.profile", default = "car")
}

#' Resolve OSRM File Path from Directory or File
#'
#' Internal helper that resolves a file path, accepting either a direct file
#' path or a directory containing exactly one file matching the specified pattern.
#'
#' @param input_path A string. Path to a file or directory.
#' @param pattern A string. Regex pattern to match files (e.g., "\\.osm\\.pbf$").
#' @param file_description A string. Human-readable description of expected files
#'   for error messages (e.g., "OSM files (.osm, .osm.bz2, or .osm.pbf)").
#' @param error_context A string or NULL. Additional context for error messages
#'   (e.g., "Please check that you have run `osrm_extract` first.").
#'
#' @return A string. The resolved normalized file path.
#' @keywords internal
#' @noRd
resolve_osrm_path <- function(input_path,
                               pattern,
                               file_description,
                               error_context = NULL) {
  # Check if path exists (let calling function handle missing files)
  if (!file.exists(input_path) && !dir.exists(input_path)) {
    # Return as-is if doesn't exist - calling function will handle the error
    return(input_path)
  }

  # Normalize path
  input_path <- normalizePath(input_path, mustWork = TRUE)

  # If input is a directory, search for matching files
  if (dir.exists(input_path)) {
    matching_files <- list.files(
      input_path,
      pattern = pattern,
      ignore.case = TRUE,
      full.names = TRUE
    )

    if (length(matching_files) == 0) {
      error_msg <- paste0(
        "No ", file_description, " found in directory: ", input_path
      )
      if (!is.null(error_context)) {
        error_msg <- paste0(error_msg, "\n", error_context)
      }
      stop(error_msg, call. = FALSE)
    } else if (length(matching_files) > 1) {
      stop(
        "Multiple ", file_description, " found in directory: ", input_path,
        "\n  Files: ", paste(basename(matching_files), collapse = ", "),
        "\n  Please specify a single file path instead of a directory.",
        call. = FALSE
      )
    }

    # Normalize the selected file path to ensure consistent separators (especially on Windows)
    input_path <- normalizePath(matching_files[1], mustWork = TRUE)
  }

  input_path
}

#' Print summary for an `osrm_server` object
#'
#' Displays status and metadata for an OSRM job process (server) returned
#' by [osrm_start()] or [osrm_start_server()].
#'
#' @param x An `osrm_server` object.
#' @param ... Passed to methods; currently ignored.
#'
#' @return Invisibly returns `x`.
#' @keywords internal
#' @noRd
print.osrm_server <- function(x, ...) {
  meta <- attr(x, "osrm_metadata")
  alive <- tryCatch(x$is_alive(), error = function(e) FALSE)
  status <- if (alive) "Running" else "Stopped"

  cat("------------------------------------------------------\n")
  cat("OSRM Server (Job Process)\n")
  cat("------------------------------------------------------\n")
  cat("Status:    ", status, "\n", sep = "")
  if (alive) {
    cat("PID:       ", x$get_pid(), "\n", sep = "")
  }
  cat("Port:      ", meta$port, "\n", sep = "")
  cat("Profile:   ", meta$profile, "\n", sep = "")
  cat("Algorithm: ", meta$algorithm, "\n", sep = "")
  cat("Graph:     ", basename(meta$path), "\n", sep = "")

  if (!is.null(meta$log) && !is.na(meta$log)) {
    cat("Logs:      ", meta$log, "\n", sep = "")
  }

  cat("------------------------------------------------------\n")
  invisible(x)
}

#' Print summary for an `osrm_job` object
#'
#' Displays pipeline state, outputs, and next steps for an `osrm_job` returned
#' by functions such as [osrm_extract()], [osrm_prepare_graph()], [osrm_partition()],
#' or [osrm_contract()].
#'
#' @param x An `osrm_job` object.
#' @param ... Passed to methods; currently ignored.
#'
#' @return Invisibly returns `x`.
#' @keywords internal
#' @noRd
print.osrm_job <- function(x, ...) {
  cat("------------------------------------------------------\n")
  cat("OSRM Job Completed\n")
  cat("------------------------------------------------------\n")
  cat("Working Directory: ", x$osrm_working_dir, "\n")
  cat("Job Artifact:      ", x$osrm_job_artifact, "\n")

  # Check if logs is a processx object or a list of them
  if (inherits(x$logs, "process_result")) {
    status <- x$logs$status
    cat("Status: ", ifelse(status == 0, "Success", paste("Error code", status)), "\n")
  } else if (is.list(x$logs)) {
    completed_stages <- names(x$logs)
    cat("Stages completed:  ", paste(completed_stages, collapse = ", "), "\n")

    # Analyze pipeline state
    pipeline_info <- .analyze_pipeline_state(completed_stages)

    # Show pipeline path and remaining stages
    if (!is.null(pipeline_info$pipeline)) {
      cat("Pipeline:          ", pipeline_info$pipeline, "\n")

      if (length(pipeline_info$remaining) > 0) {
        cat("Remaining stages:  ", paste(pipeline_info$remaining, collapse = ", "), "\n")
      }
    }

    # Show next step
    if (!is.null(pipeline_info$next_step)) {
      cat("\n")
      cat("Next step:         ", pipeline_info$next_step, "\n")
    } else if (length(pipeline_info$alternative_steps) > 0) {
      cat("\n")
      cat("Next step options:\n")
      for (step in pipeline_info$alternative_steps) {
        cat("  - ", step, "\n", sep = "")
      }
    }

    # Show routing readiness
    if (pipeline_info$ready_for_routing) {
      cat("\n")
      cat("Status:            Ready for routing!\n")
      cat("Usage:             Use osrm_routed() to start routing server,\n")
      cat("                   then use the 'osrm' R package for routing queries.\n")
    }
  }

  cat("\nLogs available in: <object>$logs\n")
  cat("------------------------------------------------------\n")
  invisible(x)
}

#' Analyze OSRM pipeline state
#' @noRd
.analyze_pipeline_state <- function(completed_stages) {
  result <- list(
    pipeline = NULL,
    remaining = character(0),
    next_step = NULL,
    alternative_steps = character(0),
    ready_for_routing = FALSE
  )

  has_extract <- "extract" %in% completed_stages
  has_partition <- "partition" %in% completed_stages
  has_contract <- "contract" %in% completed_stages
  has_customize <- "customize" %in% completed_stages

  # Determine pipeline and state
  if (has_extract && !has_partition && !has_contract) {
    # Just extract - two paths available
    result$alternative_steps <- c(
      "osrm_contract() for CH pipeline (extract -> contract)",
      "osrm_partition() for MLD pipeline (extract -> partition -> customize)"
    )
  } else if (has_extract && has_partition && !has_customize) {
    # MLD pipeline in progress
    result$pipeline <- "MLD (Multi-Level Dijkstra)"
    result$remaining <- "customize"
    result$next_step <- "osrm_customize()"
  } else if (has_extract && has_contract) {
    # CH pipeline complete
    result$pipeline <- "CH (Contraction Hierarchies)"
    result$ready_for_routing <- TRUE
  } else if (has_extract && has_partition && has_customize) {
    # MLD pipeline complete
    result$pipeline <- "MLD (Multi-Level Dijkstra)"
    result$ready_for_routing <- TRUE
  }

  result
}

#' Internal helper to extract path from string or osrm_job object
#' @noRd
get_osrm_path_from_input <- function(input) {
  if (inherits(input, "osrm_job")) {
    # Try artifact first if it exists
    if (!is.null(input$osrm_job_artifact) && file.exists(input$osrm_job_artifact)) {
      return(input$osrm_job_artifact)
    }
    # Fall back to working dir
    if (!is.null(input$osrm_working_dir)) {
      return(input$osrm_working_dir)
    }
    stop("osrm_job object has no valid paths", call. = FALSE)
  }
  return(input)
}

#' Internal helper to class return objects
#' @noRd
as_osrm_job <- function(osrm_job_artifact, osrm_working_dir, logs) {
  structure(
    list(
      osrm_job_artifact = osrm_job_artifact,
      osrm_working_dir = osrm_working_dir,
      logs = logs
    ),
    class = "osrm_job"
  )
}

#' Known OSRM file patterns by pipeline stage
#' @noRd
.osrm_file_patterns <- function() {
  list(
    # CH-specific files (created by osrm-contract)
    # Note: datasource_names is created by both CH and MLD, so not included here
    ch_specific = c("hsgr"),

    # MLD-specific files (created by osrm-partition and osrm-customize)
    # Note: datasource_names is created by both CH and MLD, so not included here
    mld_specific = c("cells", "partition", "cell_metrics", "mldgr"),

    # Files modified by osrm-partition (originally from extract)
    # These are shared with extract but have different content after partition
    mld_modified = c("cnbg_to_ebg", "ebg", "ebg_nodes", "enw", "fileIndex"),

    # Extract-only files (created by osrm-extract, not modified by later stages in CH)
    extract_base = c(
      "cnbg", "cnbg_to_ebg", "ebg", "ebg_nodes", "edges", "enw",
      "fileIndex", "geometry", "icd", "maneuver_overrides", "names",
      "nbg_nodes", "properties", "ramIndex", "restrictions", "timestamp",
      "tld", "tls", "turn_duration_penalties", "turn_penalties_index",
      "turn_weight_penalties"
    )
  )
}

#' Detect which OSRM algorithm pipeline has been used in a directory
#'
#' @param dir_path A string. Path to directory containing OSRM files
#' @param base_name A string. Base name of OSRM files (e.g., "data" for "data.osrm.*")
#'
#' @return A list with components:
#'   \describe{
#'     \item{state}{One of "empty", "extract_only", "ch", "mld", "mixed"}
#'     \item{has_ch_files}{Logical. Whether CH-specific files exist}
#'     \item{has_mld_files}{Logical. Whether MLD-specific files exist}
#'     \item{ch_files}{Character vector of found CH-specific files}
#'     \item{mld_files}{Character vector of found MLD-specific files}
#'   }
#'
#' @noRd
detect_osrm_algorithm <- function(dir_path, base_name) {
  patterns <- .osrm_file_patterns()

  # Check for CH-specific files
  ch_files <- character()
  for (ext in patterns$ch_specific) {
    file_path <- file.path(dir_path, paste0(base_name, ".osrm.", ext))
    if (file.exists(file_path)) {
      ch_files <- c(ch_files, basename(file_path))
    }
  }

  # Check for MLD-specific files
  mld_files <- character()
  for (ext in patterns$mld_specific) {
    file_path <- file.path(dir_path, paste0(base_name, ".osrm.", ext))
    if (file.exists(file_path)) {
      mld_files <- c(mld_files, basename(file_path))
    }
  }

  # Check for extract timestamp (indicates extract has been run)
  has_extract <- file.exists(file.path(dir_path, paste0(base_name, ".osrm.timestamp")))

  has_ch_files <- length(ch_files) > 0
  has_mld_files <- length(mld_files) > 0

  # Determine state
  state <- if (!has_extract && !has_ch_files && !has_mld_files) {
    "empty"
  } else if (has_extract && !has_ch_files && !has_mld_files) {
    "extract_only"
  } else if (has_ch_files && !has_mld_files) {
    "ch"
  } else if (has_mld_files && !has_ch_files) {
    "mld"
  } else if (has_ch_files && has_mld_files) {
    "mixed"
  } else {
    "unknown"
  }

  list(
    state = state,
    has_ch_files = has_ch_files,
    has_mld_files = has_mld_files,
    ch_files = ch_files,
    mld_files = mld_files
  )
}

#' Normalize algorithm argument to standard uppercase
#'
#' @param algorithm A string. "mld", "ch", "corech" (case-insensitive)
#' @return A string. "MLD", "CH", or "CoreCH"
#' @noRd
.normalize_algorithm <- function(algorithm) {
  if (is.null(algorithm)) return(NULL)
  
  # Normalize input
  algo <- toupper(algorithm)
  
  # Map valid values
  if (algo == "MLD") return("MLD")
  if (algo == "CH") return("CH")
  if (algo == "CORECH") return("CoreCH") # Special casing for OSRM standard
  
  # If validation is required here, we can stop() or let match.arg handle it later
  # For internal use, strict validation is better
  stop(
    "Invalid algorithm: '", algorithm, "'. Must be 'MLD', 'CH', or 'CoreCH'.", 
    call. = FALSE
  )
}

#' Check for algorithm conflicts and throw informative errors
#'
#' @param dir_path A string. Path to directory containing OSRM files
#' @param base_name A string. Base name of OSRM files
#' @param target_algorithm A string. The algorithm the user wants to use ("ch" or "mld")
#' @param stage A string. The pipeline stage being called (e.g., "contract", "partition")
#'
#' @noRd
check_algorithm_conflict <- function(dir_path, base_name, target_algorithm, stage) {
  detection <- detect_osrm_algorithm(dir_path, base_name)
  
  # Normalize target for comparison
  # We handle NULL/missing by skipping checks (should be caught elsewhere)
  if (missing(target_algorithm) || is.null(target_algorithm)) return(invisible(NULL))
  
  algo <- tryCatch(.normalize_algorithm(target_algorithm), error = function(e) "UNKNOWN")

  # If state is empty or extract_only, no conflict
  if (detection$state %in% c("empty", "extract_only")) {
    return(invisible(NULL))
  }

  # Check for mixed state (critical error)
  if (detection$state == "mixed") {
    stop(
      "Directory contains BOTH CH and MLD algorithm files, which is not supported.\n",
      "Found CH files: ", paste(detection$ch_files, collapse = ", "), "\n",
      "Found MLD files: ", paste(detection$mld_files, collapse = ", "), "\n\n",
      "This occurs when algorithms are mixed in the same directory.\n",
      "To fix this, use `osrm_cleanup()` to remove all OSRM files and start fresh.",
      call. = FALSE
    )
  }

  # Check for algorithm mismatch
  # detection$state returns lowercase "ch"/"mld". We compare against uppercase standard
  # so we normalize detection state for comparison
  detected_algo <- toupper(detection$state)
  
  # CoreCH acts like CH for conflict purposes
  is_target_ch <- algo %in% c("CH", "CoreCH")
  is_target_mld <- algo == "MLD"
  
  if (is_target_ch && detected_algo == "MLD") {
    stop(
      "Cannot run CH pipeline (", stage, " stage): directory contains MLD algorithm files.\n",
      "Found MLD files: ", paste(detection$mld_files, collapse = ", "), "\n\n",
      "The MLD pipeline modifies extract-stage files in a way that breaks CH compatibility.\n",
      "To switch from MLD to CH:\n",
      "  1. Use `osrm_cleanup()` to remove all OSRM files\n",
      "  2. Re-run the full CH pipeline: osrm_extract() -> osrm_contract()\n\n",
      "Alternatively, continue with the MLD pipeline using osrm_customize().",
      call. = FALSE
    )
  }

  if (is_target_mld && detected_algo == "CH") {
    stop(
      "Cannot run MLD pipeline (", stage, " stage): directory contains CH algorithm files.\n",
      "Found CH files: ", paste(detection$ch_files, collapse = ", "), "\n\n",
      "To switch from CH to MLD:\n",
      "  1. Use `osrm_cleanup()` to remove algorithm-specific files\n",
      "  2. Run the MLD pipeline: osrm_partition() -> osrm_customize()\n\n",
      "Note: Switching to MLD will modify some extract-stage files.\n",
      "Alternatively, continue with the CH pipeline using osrm_contract().",
      call. = FALSE
    )
  }

  invisible(NULL)
}

#' Resolve OSRM Binary Path
#'
#' Internal helper that resolves the path to an OSRM binary (e.g., `osrm-extract`).
#' It first checks if a full path to `osrm-routed` is provided in options. If so,
#' it looks for the target binary in the same directory. Otherwise, it falls
#' back to `Sys.which()`.
#'
#' @param bin_name A string. Name of the binary (e.g., "osrm-extract").
#' @return A string. The resolved full path to the binary, or `bin_name` if not found.
#' @keywords internal
#' @noRd
resolve_osrm_bin <- function(bin_name) {
  osrm_exec <- getOption("osrm.routed.exec")

  if (!is.null(osrm_exec) && nzchar(osrm_exec)) {
    # If it looks like a path (contains slashes), normalize it
    if (grepl("[/\\\\]", osrm_exec)) {
      osrm_exec <- normalizePath(osrm_exec, mustWork = FALSE, winslash = "/")

      if (bin_name == "osrm-routed") {
        return(osrm_exec)
      }

      bin_dir <- dirname(osrm_exec)
      ext <- if (.Platform$OS.type == "windows") ".exe" else ""
      target_bin <- file.path(bin_dir, paste0(bin_name, ext))
      if (file.exists(target_bin)) {
        return(normalizePath(target_bin, mustWork = TRUE, winslash = "/"))
      }
    } else {
      # It's just a name, return as is
      if (bin_name == "osrm-routed") {
        return(osrm_exec)
      }
    }
  }

  # Fallback to Sys.which
  resolved <- Sys.which(bin_name)
  if (nzchar(resolved)) {
    if (file.exists(resolved)) {
      return(normalizePath(resolved, mustWork = TRUE, winslash = "/"))
    }
    return(gsub("\\\\", "/", resolved))
  }

  # Return original name as a last resort (processx will throw if not on PATH)
  bin_name
}

#' Launch a GUI to View and Debug OSRM Routing
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Launches a lightweight Shiny application to interactively visualize routing
#' on a local OSRM server. This interface mimics the `r5rgui` experience,
#' supporting left-click for start, right-click for end, and draggable markers.
#'
#' @details
#' The function checks for optional dependencies `shiny`, `mapgl`, `osrm`, `sf`, and `DT`.
#' If missing, it prompts the user to install them.
#'
#' It attempts to detect an active OSRM server. If an OSRM job process (from
#' `osrm_start()`) is passed, it uses that configuration. If a path is passed,
#' it will start a temporary server for the session.
#'
#' @param input_osrm Optional. Can be:
#'   \itemize{
#'     \item An OSRM job process (an `osrm_server` object inheriting from
#'       `processx::process`) returned by `osrm_start()` or `osrm_start_server()`.
#'       When providing a process, you must also specify `port` explicitly.
#'     \item A path string to an `.osrm.hsgr` or `.osrm.mldgr` file.
#'     \item A path string to an `.osm.pbf` file (will be prepared and started).
#'     \item `NULL` (default): Auto-detects a running OSRM server using [osrm_servers()].
#'       Errors if no servers are running.
#'   }
#' @param port Integer or `"auto"`. The port the server is running on (or should run on).
#'   Defaults to `"auto"`, which attempts to auto-detect a running OSRM server using
#'   [osrm_servers()]. If multiple servers are running, the most recent one is selected
#'   with a warning. If no servers are running, an error is raised.
#' @param style Character. Map style for `mapgl`. Defaults to "https://basemaps.cartocdn.com/gl/voyager-gl-style/style.json".
#' @param center Numeric vector of length 2 (`c(lng, lat)`), or named list
#'   (`list(lng = ..., lat = ...)`), or `NULL` (default). Initial map center.
#'   If `NULL` and `input_osrm` is a `.osm.pbf` file, attempts to auto-center
#'   on the PBF extent. Priority is given to a fast pure R header parser and
#'   `osmium fileinfo` (fast); otherwise estimates the extent by sampling a
#'   small number of features via `sf::st_read()` (for example, reading with
#'   a `LIMIT 10` query).
#' @param zoom Numeric. Initial zoom level. If `NULL` (default) and center is
#'   auto-detected from PBF, defaults to 9. Otherwise uses map default.
#' @param autozoom Logical. Whether to enable auto-zoom by default. Defaults to `TRUE`.
#' @param update_while_drag Logical. Whether to enable live tracking mode by default (updates route while dragging). Defaults to `FALSE`.
#' @param debug Logical. Whether to enable debug mode (prints OSRM requests to console). Defaults to `FALSE`.
#' @return No return value; launches a Shiny Gadget.
#' @export
#' @examples
#' if (interactive()) {
#'   # 1. Auto-detect running server (errors if none running):
#'   osrm_gui()
#'
#'   # 2. Connect to specific port:
#'   # osrm_gui(port = 5001)
#'
#'   # 3. Start from a graph file (auto-center on PBF):
#'   # osrm_gui("berlin.osrm.mldgr")
#'
#'   # 4. Start from PBF with auto-center:
#'   # osrm_gui("berlin.osm.pbf")
#'
#'   # 5. Explicit center and zoom:
#'   # osrm_gui(port = 5001, center = c(13.4, 52.5), zoom = 12)
#'
#'   # 6. Use an existing process (must specify port):
#'   # srv <- osrm_start("graph.osrm.mldgr", port = 6000)
#'   # osrm_gui(srv, port = 6000)
#'
#'   # 7. Enable debug mode:
#'   # osrm_gui(debug = TRUE)
#' }
osrm_gui <- function(
  input_osrm = NULL,
  port = "auto",
  style = "https://basemaps.cartocdn.com/gl/voyager-gl-style/style.json",
  center = NULL,
  zoom = NULL,
  autozoom = TRUE,
  update_while_drag = FALSE,
  debug = FALSE
) {
  # 1. Check Dependencies
  gui_check_dependencies()

  # 2. Manage OSRM Server
  srv_context <- gui_setup_server(input_osrm, port)
  on.exit(srv_context$cleanup_fn(), add = TRUE)

  # 3. Configure Map View
  view <- gui_resolve_map_view(center, zoom, input_osrm)

  # 4. Configure 'osrm' package options for this session context
  # Detect profile from server registry or metadata
  detected_profile <- osrm_get_server_profile(input_osrm, srv_context$active_port)
  
  old_opts <- options(
    osrm.server = paste0(srv_context$host, ":", srv_context$active_port, "/"),
    osrm.profile = detected_profile
  )
  on.exit(options(old_opts), add = TRUE)

  # 5. Define UI
  ui <- gui_ui_layout()

  # 6. Define Server
  # We define this factory-style to pass in static arguments
  server <- function(input, output, session) {
    # Initialize Map
    output$map <- mapgl::renderMaplibre({
      map_args <- list(style = style)
      if (!is.null(view$center)) {
        map_args$center <- view$center
        map_args$zoom <- view$zoom %||% 9
      }
      m <- do.call(mapgl::maplibre, map_args)
      m <- mapgl::add_navigation_control(m)
      m <- mapgl::add_fullscreen_control(m)
      mapgl::add_scale_control(m)
    })

    # State
    locations <- shiny::reactiveValues(
      start = NULL,
      end = NULL,
      iso_start = NULL,
      trip = list()
    )
    init <- shiny::reactiveValues(
      route = FALSE,
      iso = FALSE,
      highlight = FALSE,
      trip = FALSE
    )
    autozoom_enabled <- shiny::reactiveVal(autozoom)
    live_update_enabled <- shiny::reactiveVal(update_while_drag)
    is_dragging <- shiny::reactiveVal(FALSE)

    # --- Debug Helper ---
    debug_msg <- function(...) {
      if (debug) {
        message("DEBUG [", format(Sys.time(), "%H:%M:%S"), "]: ", ...)
      }
    }

    if (debug) {
      debug_msg("Starting GUI in DEBUG mode.")
      debug_msg("OSRM Server: ", getOption("osrm.server"))
      debug_msg("OSRM Profile: ", getOption("osrm.profile"))
    }

    # Store latest summaries for different modes
    route_mode_summary <- shiny::reactiveVal(NULL)
    trip_mode_summary <- shiny::reactiveVal(NULL)
    # Store current route steps for highlighting
    current_steps <- shiny::reactiveVal(NULL)
    # Store intermediate coords for tracking
    tracking_coords <- shiny::reactiveValues(
      start = NULL,
      end = NULL,
      iso_start = NULL
    )
    # Store trip result for table
    trip_result <- shiny::reactiveVal(NULL)
    # Store OSRM execution time
    osrm_exec_time <- shiny::reactiveVal(NULL)

    # Time of last mode switch (for robust autozoom suppression)
    last_mode_switch <- shiny::reactiveVal(Sys.time())

    # Track which marker type is currently being dragged
    drag_type <- shiny::reactiveVal(NULL)

    # --- History Manager ---
    history <- shiny::reactiveValues(past = list(), future = list())

    # Snapshot current state
    get_state_snapshot <- function() {
      shiny::reactiveValuesToList(locations)
    }

    # Commit current state to history BEFORE making changes
    commit <- function() {
      history$past <- c(history$past, list(get_state_snapshot()))
      history$future <- list() # Clear future on new branch
    }

    # Restore state from snapshot
    restore <- function(snapshot) {
      # 1. Update Internal State
      locations$start <- snapshot$start
      locations$end <- snapshot$end
      locations$iso_start <- snapshot$iso_start
      locations$trip <- snapshot$trip

      # 2. Clear Tracking Overrides (Fix for Undo/Redo glitch with live updates)
      tracking_coords$start <- NULL
      tracking_coords$end <- NULL
      tracking_coords$iso_start <- NULL
      for (n in names(tracking_coords)) {
        if (startsWith(n, "trip_")) tracking_coords[[n]] <- NULL
      }

      # 3. Sync Map Visuals
      session$sendCustomMessage("clearAllMarkers", "clear")

      if (!is.null(snapshot$start)) {
        session$sendCustomMessage(
          'updateMarker',
          list(id = 'start', lng = snapshot$start$lng, lat = snapshot$start$lat)
        )
      }
      if (!is.null(snapshot$end)) {
        session$sendCustomMessage(
          'updateMarker',
          list(id = 'end', lng = snapshot$end$lng, lat = snapshot$end$lat)
        )
      }
      if (!is.null(snapshot$iso_start)) {
        session$sendCustomMessage(
          'updateMarker',
          list(
            id = 'iso_start',
            lng = snapshot$iso_start$lng,
            lat = snapshot$iso_start$lat
          )
        )
      }

      # Update inputs based on active mode for clarity (optional, but good for UX)
      if (input$mode == "route") {
        if (!is.null(snapshot$start)) {
          shiny::updateTextInput(
            session,
            "start_coords_input",
            value = paste(snapshot$start$lat, snapshot$start$lng, sep = ", ")
          )
        }
        if (!is.null(snapshot$end)) {
          shiny::updateTextInput(
            session,
            "end_coords_input",
            value = paste(snapshot$end$lat, snapshot$end$lng, sep = ", ")
          )
        }
      } else if (input$mode == "iso") {
        if (!is.null(snapshot$iso_start)) {
          shiny::updateTextInput(
            session,
            "start_coords_input",
            value = paste(
              snapshot$iso_start$lat,
              snapshot$iso_start$lng,
              sep = ", "
            )
          )
        }
      }

      if (!is.null(snapshot$trip)) {
        for (pt in snapshot$trip) {
          session$sendCustomMessage(
            'updateTripMarker',
            list(action = 'add', id = pt$id, lng = pt$lng, lat = pt$lat)
          )
        }
      }

      # Trigger route recalculation by invalidating/clearing result cache
      route_mode_summary(NULL)
      trip_mode_summary(NULL)
      iso_result(NULL)
      current_steps(NULL)
      trip_result(NULL)

      # Force map layer updates (Reset clears them, Observers re-add them)
      proxy <- mapgl::maplibre_proxy("map")

      # Ensure all initialized layers are visible
      if (init$route) {
        mapgl::set_layout_property(
          proxy,
          "route_layer",
          "visibility",
          "visible"
        )
      }
      if (init$iso) {
        mapgl::set_layout_property(proxy, "iso_layer", "visibility", "visible")
      }
      if (init$trip) {
        mapgl::set_layout_property(proxy, "trip_layer", "visibility", "visible")
      }
      if (init$highlight) {
        mapgl::set_layout_property(
          proxy,
          "highlight_layer",
          "visibility",
          "visible"
        )
      }

      has_content <- !is.null(snapshot$start) ||
        !is.null(snapshot$end) ||
        !is.null(snapshot$iso_start) ||
        length(snapshot$trip) > 0
      if (!has_content) {
        if (init$route) {
          mapgl::set_layout_property(proxy, "route_layer", "visibility", "none")
        }
        if (init$iso) {
          mapgl::set_layout_property(proxy, "iso_layer", "visibility", "none")
        }
        if (init$trip) {
          mapgl::set_layout_property(proxy, "trip_layer", "visibility", "none")
        }
        if (init$highlight) {
          mapgl::set_layout_property(
            proxy,
            "highlight_layer",
            "visibility",
            "none"
          )
        }
        mapgl::clear_legend(proxy)
      }
    }

    output$map_edit_controls <- shiny::renderUI({
      has_history <- length(history$past) > 0
      has_future <- length(history$future) > 0
      has_content <- !is.null(locations$start) ||
        !is.null(locations$end) ||
        !is.null(locations$iso_start) ||
        length(locations$trip) > 0

      if (!has_history && !has_future && !has_content) {
        return(NULL)
      }

      btns <- list()

      style_base <- "background: white; border: none; border-radius: 4px; box-shadow: 0 0 0 2px rgba(0,0,0,0.1); width: 30px; height: 30px; padding: 0; color: #555; margin-bottom: 5px;"

      if (has_history) {
        btns[[length(btns) + 1]] <- shiny::actionButton(
          "undo_btn",
          shiny::icon("rotate-left"),
          style = style_base,
          title = "Undo"
        )
      }

      if (has_future) {
        btns[[length(btns) + 1]] <- shiny::actionButton(
          "redo_btn",
          shiny::icon("rotate-right"),
          style = style_base,
          title = "Redo"
        )
      }

      if (has_content) {
        btns[[length(btns) + 1]] <- shiny::actionButton(
          "clear_map_icon",
          shiny::icon("trash"),
          style = style_base,
          title = "Clear Map"
        )
      }

      shiny::div(
        style = "position: absolute; top: 180px; right: 10px; z-index: 1000; display: flex; flex-direction: column;",
        btns
      )
    })

    shiny::observeEvent(input$undo_btn, {
      shiny::req(length(history$past) > 0)
      current <- get_state_snapshot()
      # Push current to future
      history$future <- c(list(current), history$future)
      # Pop from past
      prev <- history$past[[length(history$past)]]
      history$past <- history$past[-length(history$past)]
      restore(prev)
    })

    shiny::observeEvent(input$redo_btn, {
      shiny::req(length(history$future) > 0)
      current <- get_state_snapshot()
      # Push current to past
      history$past <- c(history$past, list(current))
      # Pop from future
      next_state <- history$future[[1]]
      history$future <- history$future[-1]
      restore(next_state)
    })

    # --- UI Helpers ---
    output$mode_button_ui <- shiny::renderUI({
      current_mode <- input$mode
      labels <- c("route" = "Route", "iso" = "Isochrone", "trip" = "Trip")
      colors <- c("route" = "#3b82f6", "iso" = "#CC79A7", "trip" = "#984ea3")
      
      label <- labels[current_mode]
      color <- colors[current_mode]
      
      if (is.na(label)) {
        label <- "Route"
        color <- "#3b82f6"
      }

      shiny::actionButton(
        "cycle_mode",
        paste("Mode:", label),
        style = sprintf("background-color: %s; color: white; border-width: 0px;", color)
      )
    })

    shiny::observeEvent(input$cycle_mode, {
      modes <- c("route", "iso", "trip")
      current_idx <- match(input$mode, modes)
      if (is.na(current_idx)) {
        current_idx <- 1
      }
      next_idx <- if (current_idx >= length(modes)) 1 else current_idx + 1
      shiny::updateSelectInput(session, "mode", selected = modes[next_idx])
    })

    # Mode Switch: Update Inputs (UX)
    shiny::observeEvent(input$mode, {
      if (input$mode == "route") {
        val_start <- if (!is.null(locations$start)) {
          paste(locations$start$lat, locations$start$lng, sep = ", ")
        } else {
          ""
        }
        val_end <- if (!is.null(locations$end)) {
          paste(locations$end$lat, locations$end$lng, sep = ", ")
        } else {
          ""
        }
        shiny::updateTextInput(session, "start_coords_input", value = val_start)
        shiny::updateTextInput(session, "end_coords_input", value = val_end)
      } else if (input$mode == "iso") {
        val_start <- if (!is.null(locations$iso_start)) {
          paste(locations$iso_start$lat, locations$iso_start$lng, sep = ", ")
        } else {
          ""
        }
        shiny::updateTextInput(session, "start_coords_input", value = val_start)
        shiny::updateTextInput(session, "end_coords_input", value = "")
      }
    })

    # Mode Switch: Record timestamp to suppress immediate auto-zooms
    shiny::observeEvent(
      input$mode,
      {
        last_mode_switch(Sys.time())
      },
      priority = 100
    )

    output$autozoom_button_ui <- shiny::renderUI({
      state <- autozoom_enabled()
      label <- if (state) "Autozoom: ON" else "Autozoom: OFF"
      color <- if (state) "#5cb85c" else "#777"
      shiny::actionButton(
        "toggle_autozoom",
        label,
        style = sprintf(
          "background-color: %s; color: white; border-width: 0px;",
          color
        )
      )
    })

    output$tracking_button_ui <- shiny::renderUI({
      state <- live_update_enabled()
      label <- if (state) "Update on Drag: ON" else "Update on Drag: OFF"
      color <- if (state) "#5cb85c" else "#777"
      shiny::actionButton(
        "toggle_update_on_drag",
        label,
        style = sprintf(
          "background-color: %s; color: white; border-width: 0px;",
          color
        )
      )
    })

    # Store latest isochrone result for area calculation
    iso_result <- shiny::reactiveVal(NULL)

    output$route_stats <- shiny::renderUI({
      if (input$mode == "iso") {
        iso <- iso_result()
        if (is.null(iso) || nrow(iso) == 0) return(NULL)
        
        # Calculate cumulative area for each interval
        iso_data <- tryCatch({
          iso_valid <- sf::st_make_valid(iso)
          # Ensure they are sorted by isomax
          iso_valid <- iso_valid[order(iso_valid$isomax), ]
          
          areas <- as.numeric(sf::st_area(iso_valid)) / 1e6
          cumulative_areas <- cumsum(areas)
          
          data.frame(
            isomax = iso_valid$isomax,
            cum_area = cumulative_areas,
            stringsAsFactors = FALSE
          )
        }, error = function(e) {
          if (debug) debug_msg("Area calculation failed: ", e$message)
          NULL
        })
        
        if (is.null(iso_data)) return(NULL)
        
        # Generate colors matching the map scale (inverted viridis)
        # We need to match the map's interpolation: 0 to max(isomax) -> viridis(direction = -1)
        max_iso <- max(iso_data$isomax)
        label_colors <- if (max_iso > 0) {
          # Interpolate between the viridis stops based on isomax position
          v_stops <- viridisLite::viridis(5, direction = -1)
          # Map each isomax to a color by interpolating the 5 stops
          v_func <- grDevices::colorRamp(v_stops)
          
          vapply(iso_data$isomax, function(m) {
            # Normalize m to 0-1 range
            rel_pos <- m / max_iso
            rgb_val <- v_func(rel_pos)
            grDevices::rgb(rgb_val[1], rgb_val[2], rgb_val[3], maxColorValue = 255)
          }, character(1))
        } else {
          viridisLite::viridis(nrow(iso_data), direction = -1)
        }
        
        # Create a list of stats for each threshold
        stats_items <- lapply(seq_len(nrow(iso_data)), function(i) {
          shiny::div(
            style = "white-space: nowrap;",
            shiny::tags$b(
              sprintf("Within %g min: ", iso_data$isomax[i]),
              style = sprintf("color: %s;", label_colors[i])
            ),
            shiny::span(
              paste(round(iso_data$cum_area[i], 2), "km^2"),
              class = "stat-val"
            )
          )
        })
        
        return(shiny::div(
          class = "route-stats-overlay",
          style = "flex-direction: column; align-items: flex-start; gap: 5px;",
          stats_items
        ))
      }
      
      stats <- if (input$mode == "route") route_mode_summary() else trip_mode_summary()
      if (is.null(stats)) {
        return(NULL)
      }
      
      mode_color <- if (input$mode == "route") "#3b82f6" else "#984ea3"
      
      shiny::div(
        class = "route-stats-overlay",
        shiny::div(
          shiny::tags$b("Duration: ", style = sprintf("color: %s;", mode_color)),
          shiny::span(
            paste(round(stats$duration, 1), "min"),
            class = "stat-val"
          )
        ),
        shiny::div(
          shiny::tags$b("Distance: ", style = sprintf("color: %s;", mode_color)),
          shiny::span(
            paste(round(stats$distance, 2), "km"),
            class = "stat-val"
          )
        )
      )
    })

    shiny::observeEvent(input$toggle_autozoom, {
      autozoom_enabled(!autozoom_enabled())
    })
    shiny::observeEvent(input$toggle_update_on_drag, {
      live_update_enabled(!live_update_enabled())
    })

    # --- Marker Helpers ---
    update_start <- function(lng, lat) {
      commit()
      is_dragging(FALSE)
      coords <- list(lat = round(lat, 5), lng = round(lng, 5))
      locations$start <- coords
      tracking_coords$start <- NULL
      session$sendCustomMessage(
        'updateMarker',
        list(id = 'start', lng = lng, lat = lat)
      )
      shiny::updateTextInput(
        session,
        "start_coords_input",
        value = paste(coords$lat, coords$lng, sep = ", ")
      )
      route_mode_summary(NULL)
      current_steps(NULL)
    }

    update_end <- function(lng, lat) {
      commit()
      is_dragging(FALSE)
      coords <- list(lat = round(lat, 5), lng = round(lng, 5))
      locations$end <- coords
      tracking_coords$end <- NULL
      session$sendCustomMessage(
        'updateMarker',
        list(id = 'end', lng = lng, lat = lat)
      )
      shiny::updateTextInput(
        session,
        "end_coords_input",
        value = paste(coords$lat, coords$lng, sep = ", ")
      )
      route_mode_summary(NULL)
      current_steps(NULL)
    }

    update_iso_start <- function(lng, lat) {
      commit()
      is_dragging(FALSE)
      coords <- list(lat = round(lat, 5), lng = round(lng, 5))
      locations$iso_start <- coords
      tracking_coords$iso_start <- NULL
      session$sendCustomMessage(
        'updateMarker',
        list(id = 'iso_start', lng = lng, lat = lat)
      )
      shiny::updateTextInput(
        session,
        "start_coords_input",
        value = paste(coords$lat, coords$lng, sep = ", ")
      )
    }

    add_trip_point <- function(lng, lat) {
      commit()
      id <- paste0("trip_", as.integer(Sys.time()), "_", sample(1000:9999, 1))
      locations$trip[[id]] <- list(id = id, lat = lat, lng = lng)
      session$sendCustomMessage(
        'updateTripMarker',
        list(action = 'add', id = id, lng = lng, lat = lat)
      )
    }

    remove_trip_point <- function(id) {
      commit()
      is_dragging(FALSE)
      locations$trip[[id]] <- NULL
      tracking_coords[[id]] <- NULL
      session$sendCustomMessage(
        'updateTripMarker',
        list(action = 'remove', id = id)
      )
    }

    move_trip_point <- function(id, lng, lat) {
      if (!is.null(locations$trip[[id]])) {
        commit()
        is_dragging(FALSE)
        locations$trip[[id]]$lat <- lat
        locations$trip[[id]]$lng <- lng
        tracking_coords[[id]] <- NULL
      }
    }

    # --- Interaction Handlers ---
    shiny::observeEvent(input$map_click, {
      shiny::req(input$map_click)
      if (input$mode == 'trip') {
        add_trip_point(input$map_click$lng, input$map_click$lat)
      } else if (input$mode == 'iso') {
        update_iso_start(input$map_click$lng, input$map_click$lat)
      } else {
        update_start(input$map_click$lng, input$map_click$lat)
      }
    })

    shiny::observeEvent(input$js_right_click, {
      shiny::req(input$js_right_click)
      if (input$mode == 'route') {
        update_end(input$js_right_click$lng, input$js_right_click$lat)
      }
    })

    # --- Live Events (Direct for responsiveness) ---

    # 1. Marker Dragged (Drop)
    shiny::observeEvent(input$marker_dragged, {
      drag <- input$marker_dragged
      if (is.null(drag)) {
        return()
      }

      if (drag$id == "start") {
        update_start(drag$lng, drag$lat)
      } else if (drag$id == "end") {
        update_end(drag$lng, drag$lat)
      } else if (drag$id == "iso_start") {
        update_iso_start(drag$lng, drag$lat)
      }
    })

    # 2. Marker Moving (Live Drag)
    shiny::observeEvent(input$marker_moving, {
      is_dragging(TRUE)
      moving <- input$marker_moving
      if (is.null(moving)) {
        return()
      }

      # Determine drag type to guide background updates
      if (moving$id == "start" || moving$id == "end") {
        drag_type("route")
      } else if (moving$id == "iso_start") {
        drag_type("iso")
      } else if (startsWith(moving$id, "trip_")) {
        drag_type("trip")
      }

      tracking_coords[[moving$id]] <- list(
        lat = round(moving$lat, 5),
        lng = round(moving$lng, 5)
      )
    })

    shiny::observeEvent(input$remove_trip_point, {
      remove_trip_point(input$remove_trip_point$id)
    })

    shiny::observeEvent(input$move_trip_point, {
      mv <- input$move_trip_point
      move_trip_point(mv$id, mv$lng, mv$lat)
    })

    # Reset Logic
    reset_all <- function() {
      commit()
      locations$start <- NULL
      locations$end <- NULL
      locations$iso_start <- NULL
      locations$trip <- list()
      for (n in names(tracking_coords)) {
        tracking_coords[[n]] <- NULL
      }
      route_mode_summary(NULL)
      trip_mode_summary(NULL)
      iso_result(NULL)
      current_steps(NULL)
      trip_result(NULL)
      osrm_exec_time(NULL)
      session$sendCustomMessage("clearAllMarkers", "clear")
      shiny::updateTextInput(session, "start_coords_input", value = "")
      shiny::updateTextInput(session, "end_coords_input", value = "")

      proxy <- mapgl::maplibre_proxy("map")
      if (init$route) {
        mapgl::set_layout_property(proxy, "route_layer", "visibility", "none")
      }
      if (init$iso) {
        mapgl::set_layout_property(proxy, "iso_layer", "visibility", "none")
      }
      if (init$trip) {
        mapgl::set_layout_property(proxy, "trip_layer", "visibility", "none")
      }
      if (init$highlight) {
        mapgl::set_layout_property(
          proxy,
          "highlight_layer",
          "visibility",
          "none"
        )
      }
      mapgl::clear_legend(proxy)
    }

    shiny::observeEvent(input$reset, {
      reset_all()
    })
    shiny::observeEvent(input$clear_map_icon, {
      reset_all()
    })

    # --- Calculation Logic ---

    # --- Specialized Reactive Coordinates (Decoupled & Throttled) ---
    # We split these to ensure dragging a Route marker doesn't trigger a Trip/Iso calc.

    active_route_coords <- shiny::throttle(
      shiny::reactive({
        res <- list(start = locations$start, end = locations$end)
        if (!is.null(tracking_coords$start)) {
          res$start <- tracking_coords$start
        }
        if (!is.null(tracking_coords$end)) {
          res$end <- tracking_coords$end
        }
        res
      }),
      250
    )

    active_trip_coords <- shiny::throttle(
      shiny::reactive({
        # Start with the stable base
        pts_list <- locations$trip
        if (is.null(pts_list) || length(pts_list) == 0) {
          return(list())
        }

        # Overlay with any active tracking data
        tracking_list <- shiny::reactiveValuesToList(tracking_coords)
        for (id in names(tracking_list)) {
          if (
            startsWith(id, "trip_") &&
              !is.null(pts_list[[id]]) &&
              !is.null(tracking_list[[id]]$lat) &&
              !is.null(tracking_list[[id]]$lng)
          ) {
            # Only overlay if the tracking coordinate is actually present/valid
            pts_list[[id]]$lat <- as.numeric(tracking_list[[id]]$lat)
            pts_list[[id]]$lng <- as.numeric(tracking_list[[id]]$lng)
          }
        }
        pts_list
      }),
      250
    )

    active_iso_coords <- shiny::throttle(
      shiny::reactive({
        res <- locations$iso_start
        if (!is.null(tracking_coords$iso_start)) {
          res <- tracking_coords$iso_start
        }
        res
      }),
      250
    )

    # Route Calculation: Live Tracking
    shiny::observe({
      shiny::req(live_update_enabled(), is_dragging(), drag_type() == "route")
      coords <- active_route_coords()
      shiny::req(coords$start, coords$end)
      tryCatch(
        {
          debug_msg(
            "Route (Live) request: ",
            coords$start$lng,
            ",",
            coords$start$lat,
            " -> ",
            coords$end$lng,
            ",",
            coords$end$lat
          )
          t0 <- Sys.time()
          route <- osrm::osrmRoute(
            src = c(coords$start$lng, coords$start$lat),
            dst = c(coords$end$lng, coords$end$lat),
            overview = "full"
          )
          osrm_exec_time(as.numeric(difftime(Sys.time(), t0, units = "secs")))
          if (!is.null(route)) {
            route_mode_summary(list(
              duration = route$duration[1],
              distance = route$distance[1]
            ))
            proxy <- mapgl::maplibre_proxy("map")
            if (!init$route) {
              mapgl::add_source(proxy, id = "route_source", data = route)
              mapgl::add_line_layer(
                proxy,
                id = "route_layer",
                source = "route_source",
                line_color = "#3b82f6",
                line_width = 5,
                line_opacity = 0.8
              )
              init$route <- TRUE
            } else {
              mapgl::set_source(proxy, layer_id = "route_layer", source = route)
              mapgl::set_layout_property(
                proxy,
                "route_layer",
                "visibility",
                "visible"
              )
            }
          }
        },
        error = function(e) NULL
      )
    })

    # Route Calculation: Stable Updates
    shiny::observeEvent(
      list(locations$start, locations$end),
      {
        shiny::req(locations$start, locations$end)
        calc_route <- function() {
          tryCatch(
            {
              debug_msg(
                "Route (Stable) request: ",
                locations$start$lng,
                ",",
                locations$start$lat,
                " -> ",
                locations$end$lng,
                ",",
                locations$end$lat
              )
              # Use consolidated API fetch to get both geometry and steps in one call
              t0 <- Sys.time()
              res <- api_fetch_route_detailed(
                locations$start,
                locations$end,
                overview = "full",
                debug = debug
              )
              osrm_exec_time(as.numeric(difftime(Sys.time(), t0, units = "secs")))

              if (!is.null(res) && length(res$routes) > 0) {
                route_data <- res$routes[[1]]

                # 1. Update Route Summary
                route_mode_summary(list(
                  duration = route_data$duration / 60, # OSRM returns seconds, convert to minutes
                  distance = route_data$distance / 1000 # OSRM returns meters, convert to km
                ))

                # 2. Update Current Steps (for Itinerary Table and Highlighting)
                if (!is.null(route_data$legs[[1]]$steps)) {
                  current_steps(route_data$legs[[1]]$steps)
                }

                # 3. Create sf geometry for mapping
                coords <- route_data$geometry$coordinates
                coord_matrix <- do.call(rbind, lapply(coords, as.numeric))
                route <- sf::st_sf(
                  geometry = sf::st_sfc(
                    sf::st_linestring(coord_matrix),
                    crs = 4326
                  )
                )
                proxy <- mapgl::maplibre_proxy("map")
                if (!init$route) {
                  mapgl::add_source(proxy, id = "route_source", data = route)
                  mapgl::add_line_layer(
                    proxy,
                    id = "route_layer",
                    source = "route_source",
                    line_color = "#3b82f6",
                    line_width = 5,
                    line_opacity = 0.8
                  )
                  init$route <- TRUE
                } else {
                  mapgl::set_source(
                    proxy,
                    layer_id = "route_layer",
                    source = route
                  )
                  mapgl::set_layout_property(
                    proxy,
                    "route_layer",
                    "visibility",
                    "visible"
                  )
                }
                # Check if enough time has passed since mode switch (isloated to prevent re-calc on switch)
                time_since_switch <- as.numeric(difftime(
                  Sys.time(),
                  shiny::isolate(last_mode_switch()),
                  units = "secs"
                ))
                should_autozoom <- time_since_switch > 1.0

                if (shiny::isolate(autozoom_enabled()) && should_autozoom) {
                  pts_sf <- sf::st_as_sf(
                    data.frame(
                      lon = c(locations$start$lng, locations$end$lng),
                      lat = c(locations$start$lat, locations$end$lat)
                    ),
                    coords = c("lon", "lat"),
                    crs = 4326
                  )
                  combined_sf <- rbind(
                    sf::st_sf(geometry = sf::st_geometry(route)),
                    sf::st_sf(geometry = sf::st_geometry(pts_sf))
                  )
                  map_width <- shiny::isolate(
                    session$clientData$output_map_width
                  ) %||%
                    1000
                  padding <- if (map_width < 768) 50 else 150
                  mapgl::fit_bounds(
                    proxy,
                    combined_sf,
                    animate = TRUE,
                    padding = padding
                  )
                }
              }
            },
            error = function(e) {
              shiny::showNotification(
                paste("Routing failed:", e$message),
                type = "error"
              )
            }
          )
        }
        shiny::withProgress(message = "Calculating Route...", calc_route())
      }
    )

    # Trip Calculation: Live Tracking
    shiny::observe({
      shiny::req(live_update_enabled(), is_dragging(), drag_type() == "trip")
      trip_pts <- active_trip_coords()

      if (debug) {
        debug_msg(
          "Trip (Live) observer firing with ",
          length(trip_pts),
          " points."
        )
      }
      if (length(trip_pts) < 2) {
        return()
      }

      # Handle extraction with explicit NA checks to avoid dying on req
      pt_ids <- names(trip_pts)
      lons <- numeric(length(pt_ids))
      lats <- numeric(length(pt_ids))

      for (i in seq_along(pt_ids)) {
        id <- pt_ids[i]
        lons[i] <- as.numeric(trip_pts[[id]]$lng %||% NA)
        lats[i] <- as.numeric(trip_pts[[id]]$lat %||% NA)
      }

      # If any are NA, we prefer to skip this frame rather than dying on req
      # and we definitely don't want to fetch
      if (any(is.na(lons)) || any(is.na(lats))) {
        if (debug) {
          debug_msg(
            "Trip (Live): Skipping frame due to incomplete coordinates."
          )
        }
        return()
      }

      shiny::req(length(lons) >= 2)

      # Ensure data frame is as clean as possible
      pts_df <- data.frame(lon = lons, lat = lats)
      rownames(pts_df) <- NULL

      if (debug) {
        debug_msg("Trip (Live) request points:")
        print(pts_df)
      }

      # 1. Calculate trip using direct HTTP API
      t0 <- Sys.time()
      trip_result_data <- api_fetch_trip(pts_df, debug = debug)
      osrm_exec_time(as.numeric(difftime(Sys.time(), t0, units = "secs")))

      if (!is.null(trip_result_data) && !is.null(trip_result_data$trip)) {
        trip_geom <- trip_result_data$trip

        summary <- trip_result_data$summary
        dur <- if (!is.null(summary$duration)) {
          as.numeric(summary$duration)
        } else {
          0
        }
        dis <- if (!is.null(summary$distance)) {
          as.numeric(summary$distance)
        } else {
          0
        }
        trip_mode_summary(list(duration = dur, distance = dis))

        # 2. Update Marker Labels
        if (!is.null(trip_result_data$waypoint_order)) {
          labels_payload <- lapply(seq_along(pt_ids), function(i) {
            list(id = pt_ids[i], label = as.character(trip_result_data$waypoint_order[i]))
          })
          session$sendCustomMessage('updateTripLabels', labels_payload)
        }

        # 3. Render on map (with fallback if set_source fails)
        tryCatch(
          {
            proxy <- mapgl::maplibre_proxy("map")
            if (!init$trip) {
              mapgl::add_source(proxy, id = "trip_source", data = trip_geom)
              mapgl::add_line_layer(
                proxy,
                id = "trip_layer",
                source = "trip_source",
                line_color = "#984ea3",
                line_width = 5,
                line_opacity = 0.8
              )
              init$trip <- TRUE
            } else {
              mapgl::set_source(
                proxy,
                layer_id = "trip_layer",
                source = trip_geom
              )
              mapgl::set_layout_property(
                proxy,
                "trip_layer",
                "visibility",
                "visible"
              )
            }
          },
          error = function(e) {
            if (debug) {
              debug_msg(
                "Trip (Live) map update error: ",
                e$message,
                " -- resetting layer"
              )
            }
            tryCatch(
              {
                proxy <- mapgl::maplibre_proxy("map")
                init$trip <- FALSE
                mapgl::add_source(proxy, id = "trip_source", data = trip_geom)
                mapgl::add_line_layer(
                  proxy,
                  id = "trip_layer",
                  source = "trip_source",
                  line_color = "#984ea3",
                  line_width = 5,
                  line_opacity = 0.8
                )
                init$trip <- TRUE
              },
              error = function(e2) NULL
            )
          }
        )
      }
    })

    # Trip Calculation: Stable Updates
    shiny::observeEvent(
      locations$trip,
      {
        shiny::req(!is.null(locations$trip))
        trip_pts <- locations$trip

        # Unified robust extraction mapping
        coords_list <- unname(lapply(trip_pts, function(p) {
          lng <- as.numeric(p$lng %||% NA)
          lat <- as.numeric(p$lat %||% NA)
          if (is.na(lng) || is.na(lat)) {
            return(NULL)
          }
          c(lng, lat)
        }))

        # Filter NULLs and validate count
        coords_list <- coords_list[!vapply(coords_list, is.null, logical(1))]
        shiny::req(length(coords_list) >= 2)

        coord_matrix <- do.call(rbind, coords_list)
        lons <- coord_matrix[, 1]
        lats <- coord_matrix[, 2]

        shiny::req(length(lons) >= 2, length(lats) == length(lons))

        # Ensure data frame is as clean as possible
        pts_df <- data.frame(lon = lons, lat = lats)
        rownames(pts_df) <- NULL

        calc_trip <- function() {
          local_lons <- lons
          local_lats <- lats

          if (debug) {
            debug_msg("Trip (Stable) request points:")
            print(data.frame(lon = local_lons, lat = local_lats))
            debug_msg(
              "  class(local_lons)=",
              class(local_lons),
              " length=",
              length(local_lons)
            )
            debug_msg(
              "  class(local_lats)=",
              class(local_lats),
              " length=",
              length(local_lats)
            )
            debug_msg("  osrm.server=", getOption("osrm.server"))
            debug_msg("  osrm.profile=", getOption("osrm.profile"))
            debug_msg("  init$trip=", init$trip)
          }

          # Create the data.frame outside the tryCatch for cleaner debugging
          trip_df <- data.frame(
            lon = as.numeric(local_lons),
            lat = as.numeric(local_lats)
          )
          rownames(trip_df) <- NULL

          if (debug) {
            debug_msg(
              "  trip_df created successfully, calling api_fetch_trip..."
            )
          }

          # 1. Calculate trip using direct HTTP API (bypasses osrm::osrmTrip issues)
          t0 <- Sys.time()
          trip_result_data <- api_fetch_trip(trip_df, debug = debug)
          osrm_exec_time(as.numeric(difftime(Sys.time(), t0, units = "secs")))

          if (is.null(trip_result_data) || is.null(trip_result_data$trip)) {
            shiny::showNotification(
              "Trip failed: No route returned (Is OSRM Trip service enabled? Are points within map coverage?)",
              type = "error"
            )
            return()
          }

          trip_geom <- trip_result_data$trip
          trip_result(trip_result_data)

          # Update Marker Labels
          if (!is.null(trip_result_data$waypoint_order)) {
            pt_ids <- names(trip_pts)
            labels_payload <- lapply(seq_along(pt_ids), function(i) {
              list(id = pt_ids[i], label = as.character(trip_result_data$waypoint_order[i]))
            })
            session$sendCustomMessage('updateTripLabels', labels_payload)
          }

          summary <- trip_result_data$summary
          dur <- if (!is.null(summary$duration)) {
            as.numeric(summary$duration)
          } else {
            0
          }
          dis <- if (!is.null(summary$distance)) {
            as.numeric(summary$distance)
          } else {
            0
          }
          trip_mode_summary(list(duration = dur, distance = dis))

          # 2. Render on map (with fallback if set_source fails)
          tryCatch(
            {
              proxy <- mapgl::maplibre_proxy("map")
              if (!init$trip) {
                mapgl::add_source(proxy, id = "trip_source", data = trip_geom)
                mapgl::add_line_layer(
                  proxy,
                  id = "trip_layer",
                  source = "trip_source",
                  line_color = "#984ea3",
                  line_width = 5,
                  line_opacity = 0.8
                )
                init$trip <- TRUE
              } else {
                mapgl::set_source(
                  proxy,
                  layer_id = "trip_layer",
                  source = trip_geom
                )
                mapgl::set_layout_property(
                  proxy,
                  "trip_layer",
                  "visibility",
                  "visible"
                )
              }
              # Check if enough time has passed since mode switch
              time_since_switch <- as.numeric(difftime(
                Sys.time(),
                shiny::isolate(last_mode_switch()),
                units = "secs"
              ))
              should_autozoom <- time_since_switch > 1.0

              if (shiny::isolate(autozoom_enabled()) && should_autozoom) {
                pts_sf <- sf::st_as_sf(
                  data.frame(
                    lon = as.numeric(local_lons),
                    lat = as.numeric(local_lats)
                  ),
                  coords = c("lon", "lat"),
                  crs = 4326
                )
                combined_sf <- rbind(
                  sf::st_sf(geometry = sf::st_geometry(trip_geom)),
                  sf::st_sf(geometry = sf::st_geometry(pts_sf))
                )
                map_width <- shiny::isolate(
                  session$clientData$output_map_width
                ) %||%
                  1000
                padding <- if (map_width < 768) 50 else 150
                mapgl::fit_bounds(
                  proxy,
                  combined_sf,
                  animate = TRUE,
                  padding = padding
                )
              }
            },
            error = function(e) {
              if (debug) {
                debug_msg(
                  "Trip map update error: ",
                  e$message,
                  " -- resetting layer"
                )
              }
              tryCatch(
                {
                  proxy <- mapgl::maplibre_proxy("map")
                  init$trip <- FALSE
                  mapgl::add_source(proxy, id = "trip_source", data = trip_geom)
                  mapgl::add_line_layer(
                    proxy,
                    id = "trip_layer",
                    source = "trip_source",
                    line_color = "#984ea3",
                    line_width = 5,
                    line_opacity = 0.8
                  )
                  init$trip <- TRUE
                },
                error = function(e2) {
                  if (debug) {
                    debug_msg("Trip layer reset also failed: ", e2$message)
                  }
                }
              )
            }
          )
        }
        shiny::withProgress(message = "Calculating Trip...", calc_trip())
      }
    )

    # Isochrone: Live Tracking (Custom Resolution)
    shiny::observe({
      shiny::req(live_update_enabled(), is_dragging(), drag_type() == "iso")
      iso_start <- active_iso_coords()
      shiny::req(iso_start)

      tryCatch(
        {
          breaks <- gui_parse_breaks(input$iso_breaks)

          # Live resolution
          debug_msg(
            "Isochrone (Live) request: ",
            iso_start$lng,
            ",",
            iso_start$lat,
            " breaks: ",
            paste(breaks, collapse = ",")
          )
          n_vals <- c(100, 200, 500, 1000, 2000, 5000, 10000, 20000, 50000)
          n_val <- tryCatch(
            n_vals[as.integer(input$iso_live_res)],
            error = function(e) 200
          )
          if (is.na(n_val)) {
            n_val <- 200
          }

          t0 <- Sys.time()
          iso <- osrm::osrmIsochrone(
            loc = c(iso_start$lng, iso_start$lat),
            breaks = breaks,
            n = n_val
          )
          osrm_exec_time(as.numeric(difftime(Sys.time(), t0, units = "secs")))

          if (!is.null(iso) && nrow(iso) > 0) {
            iso_result(iso)
            proxy <- mapgl::maplibre_proxy("map")
            if (!init$iso) {
              mapgl::add_source(proxy, id = "iso_source", data = iso)
              mapgl::add_fill_layer(
                proxy,
                id = "iso_layer",
                source = "iso_source",
                fill_color = mapgl::interpolate(
                  column = "isomax",
                  values = seq(0, max(breaks), length.out = 5),
                  stops = viridisLite::viridis(5, direction = -1)
                ),
                fill_opacity = 0.5,
                fill_outline_color = "white"
              )
              init$iso <- TRUE
            } else {
              mapgl::set_source(proxy, layer_id = "iso_layer", source = iso)
              mapgl::set_layout_property(
                proxy,
                "iso_layer",
                "visibility",
                "visible"
              )
            }
          }
        },
        error = function(e) NULL
      )
    })

    # Isochrone: Stable Updates
    shiny::observeEvent(
      list(locations$iso_start, input$iso_breaks, input$iso_res),
      {
        shiny::req(locations$iso_start)
        breaks <- gui_parse_breaks(input$iso_breaks)
        shiny::withProgress(message = 'Computing Isochrones...', {
          tryCatch(
            {
              # Valid 'n' values for osrmIsochrone
              n_vals <- c(100, 200, 500, 1000, 2000, 5000, 10000, 20000, 50000)

              # Use slider index (1-9) to pick value, default to 500 (index 3) if out of bounds
              n_val <- tryCatch(
                n_vals[as.integer(input$iso_res)],
                error = function(e) 500
              )
              if (is.na(n_val)) {
                n_val <- 500
              }

              debug_msg(
                "Isochrone (Stable) request: ",
                locations$iso_start$lng,
                ",",
                locations$iso_start$lat,
                " breaks: ",
                paste(breaks, collapse = ","),
                " n: ",
                n_val
              )
              t0 <- Sys.time()
              iso <- osrm::osrmIsochrone(
                loc = c(locations$iso_start$lng, locations$iso_start$lat),
                breaks = breaks,
                n = n_val
              )
              osrm_exec_time(as.numeric(difftime(Sys.time(), t0, units = "secs")))
              if (!is.null(iso) && nrow(iso) > 0) {
                iso_result(iso)
                proxy <- mapgl::maplibre_proxy("map")
                if (!init$iso) {
                  mapgl::add_source(proxy, id = "iso_source", data = iso)
                  mapgl::add_fill_layer(
                    proxy,
                    id = "iso_layer",
                    source = "iso_source",
                    fill_color = mapgl::interpolate(
                      column = "isomax",
                      values = seq(0, max(breaks), length.out = 5),
                      stops = viridisLite::viridis(5, direction = -1)
                    ),
                    fill_opacity = 0.5,
                    fill_outline_color = "white"
                  )
                  init$iso <- TRUE
                } else {
                  mapgl::set_source(proxy, layer_id = "iso_layer", source = iso)
                  mapgl::set_layout_property(
                    proxy,
                    "iso_layer",
                    "visibility",
                    "visible"
                  )
                  mapgl::set_paint_property(
                    proxy,
                    layer_id = "iso_layer",
                    name = "fill-color",
                    value = mapgl::interpolate(
                      column = "isomax",
                      values = seq(0, max(breaks), length.out = 5),
                      stops = viridisLite::viridis(5, direction = -1)
                    )
                  )
                }
                # Check if enough time has passed since mode switch
                time_since_switch <- as.numeric(difftime(
                  Sys.time(),
                  shiny::isolate(last_mode_switch()),
                  units = "secs"
                ))
                should_autozoom <- time_since_switch > 1.0

                if (shiny::isolate(autozoom_enabled()) && should_autozoom) {
                  map_width <- shiny::isolate(
                    session$clientData$output_map_width
                  ) %||%
                    1000
                  padding <- if (map_width < 768) 20 else 50
                  mapgl::fit_bounds(
                    proxy,
                    iso,
                    animate = TRUE,
                    padding = padding
                  )
                }
              }
            },
            error = function(e) {
              shiny::showNotification(
                paste("Isochrone failed:", e$message),
                type = "error"
              )
            }
          )
        })
      }
    )

    # --- Table Output ---
    output$itinerary_table <- DT::renderDataTable({
      shiny::req(input$mode == "route")
      steps <- current_steps()
      shiny::req(steps)

      # Use the cached steps from the stable Route observer instead of re-fetching
      # This is the key fix for redundant requests on mode switch.
      df <- do.call(
        rbind,
        lapply(steps, function(s) {
          instr <- s$maneuver$type
          if (!is.null(s$maneuver$modifier)) {
            instr <- paste(instr, s$maneuver$modifier)
          }
          data.frame(
            Instruction = instr,
            Road = if (is.null(s$name) || s$name == "") "-" else s$name,
            `Distance (km)` = round(s$distance / 1000, 3),
            `Duration (min)` = round(s$duration / 60, 2),
            stringsAsFactors = FALSE,
            check.names = FALSE
          )
        })
      )
      DT::datatable(
        df,
        selection = "multiple",
        options = list(pageLength = 10, scrollX = TRUE),
        rownames = FALSE
      )
    })

    shiny::observeEvent(input$clear_selection, {
      DT::selectRows(DT::dataTableProxy("itinerary_table"), NULL)
    })

    shiny::observeEvent(
      input$itinerary_table_rows_selected,
      {
        proxy <- mapgl::maplibre_proxy("map")
        idx <- input$itinerary_table_rows_selected
        steps <- current_steps()
        if (is.null(idx) || length(idx) == 0 || is.null(steps)) {
          if (init$highlight) {
            mapgl::set_layout_property(
              proxy,
              "highlight_layer",
              "visibility",
              "none"
            )
          }
          return()
        }
        segment_list <- lapply(idx, function(i) {
          step <- steps[[i]]
          coords <- matrix(
            unlist(step$geometry$coordinates),
            ncol = 2,
            byrow = TRUE
          )
          if (nrow(coords) < 2 || any(is.na(coords))) {
            return(NULL)
          }
          sf::st_linestring(coords)
        })
        # Filter out invalid segments
        segment_list <- segment_list[!vapply(segment_list, is.null, logical(1))]
        if (length(segment_list) == 0) {
          return()
        }

        segment_sfc <- sf::st_sfc(segment_list, crs = 4326)
        combined_geom <- sf::st_combine(segment_sfc)
        segment_sf <- sf::st_sf(geometry = combined_geom)
        if (!init$highlight) {
          mapgl::add_source(proxy, id = "highlight_source", data = segment_sf)
          mapgl::add_line_layer(
            proxy,
            id = "highlight_layer",
            source = "highlight_source",
            line_color = "#facc15",
            line_width = 8,
            line_opacity = 0.9
          )
          init$highlight <- TRUE
        } else {
          mapgl::set_source(
            proxy,
            layer_id = "highlight_layer",
            source = segment_sf
          )
          mapgl::set_layout_property(
            proxy,
            "highlight_layer",
            "visibility",
            "visible"
          )
        }
      },
      ignoreNULL = FALSE
    )

    output$trip_table <- DT::renderDataTable({
      shiny::req(input$mode == "trip")
      trip_data <- trip_result()
      shiny::req(trip_data, trip_data$legs)

      df <- trip_data$legs

      DT::datatable(
        df,
        selection = "none",
        options = list(pageLength = 10, scrollX = TRUE),
        rownames = FALSE
      )
    })

    shiny::observeEvent(input$copy_code, {
      mode <- input$mode
      
      # Determine if we have enough data to generate code
      has_data <- FALSE
      if (mode == "route") {
        has_data <- !is.null(locations$start) && !is.null(locations$end)
      } else if (mode == "trip") {
        has_data <- length(locations$trip) >= 2
      } else if (mode == "iso") {
        has_data <- !is.null(locations$iso_start)
      }

      if (!has_data) {
        shiny::showNotification(
          "Please add the required points to the map first.",
          type = "warning"
        )
        return()
      }

      # Gather options
      osrm_server <- getOption("osrm.server")
      osrm_profile <- getOption("osrm.profile")

      header_code <- sprintf(
        "library(osrm)\n\n# --- OSRM Configuration ---\noptions(osrm.server = \"%s\")\noptions(osrm.profile = \"%s\")\n\n",
        osrm_server, osrm_profile
      )

      main_code <- ""
      if (mode == "route") {
        main_code <- sprintf(
          "# --- Calculate Route ---\nroute <- osrm::osrmRoute(\n  src = c(%.6f, %.6f),\n  dst = c(%.6f, %.6f),\n  overview = \"full\"\n)\n\n# Visualize\nm <- mapgl::maplibre(style = mapgl::carto_style(\"voyager\"))\nmapgl::add_line_layer(m, id = \"route\", source = route, line_color = \"#3b82f6\", line_width = 5)",
          locations$start$lng, locations$start$lat,
          locations$end$lng, locations$end$lat
        )
      } else if (mode == "trip") {
        pts_list <- locations$trip
        # Use simple numeric extraction to avoid names/attributes in the paste
        lons <- vapply(pts_list, function(p) as.numeric(p$lng), numeric(1))
        lats <- vapply(pts_list, function(p) as.numeric(p$lat), numeric(1))
        
        main_code <- sprintf(
          "# --- Calculate Optimized Trip ---\npoints <- data.frame(\n  lon = c(%s),\n  lat = c(%s)\n)\n\ntrip <- osrm::osrmTrip(\n  loc = points,\n  overview = \"full\"\n)\n\n# Visualize\nm <- mapgl::maplibre(style = mapgl::carto_style(\"voyager\"))\nmapgl::add_line_layer(m, id = \"trip\", source = trip[[1]]$trip, line_color = \"#984ea3\", line_width = 5)",
          paste(sprintf("%.6f", lons), collapse = ", "),
          paste(sprintf("%.6f", lats), collapse = ", ")
        )
      } else if (mode == "iso") {
        breaks <- gui_parse_breaks(input$iso_breaks)
        n_vals <- c(100, 200, 500, 1000, 2000, 5000, 10000, 20000, 50000)
        n_val <- n_vals[as.integer(input$iso_res)]
        
        main_code <- sprintf(
          "# --- Calculate Isochrones ---\niso <- osrm::osrmIsochrone(\n  loc = c(%.6f, %.6f),\n  breaks = c(%s),\n  n = %d\n)\n\n# Visualize\nm <- mapgl::maplibre(style = mapgl::carto_style(\"voyager\"))\nmapgl::add_fill_layer(\n  m,\n  id = \"iso\",\n  source = iso,\n  fill_color = mapgl::interpolate(\n    column = \"isomax\",\n    values = seq(0, %g, length.out = 5),\n    stops = viridisLite::viridis(5, direction = -1)\n  ),\n  fill_opacity = 0.5\n)",
          locations$iso_start$lng, locations$iso_start$lat,
          paste(breaks, collapse = ", "),
          n_val,
          max(breaks)
        )
      }

      shiny::showModal(shiny::modalDialog(
        title = "R Code for Reproducible Analysis",
        shiny::tags$pre(
          id = "copy_code_block",
          style = "background-color: #f8f9fa; padding: 15px; border: 1px solid #dee2e6; border-radius: 4px; max-height: 400px; overflow-y: auto; white-space: pre-wrap; font-family: monospace; font-size: 13px;",
          paste0(header_code, main_code)
        ),
        shiny::helpText("Copy and run this code in your R console to reproduce the current analysis."),
        footer = shiny::modalButton("Dismiss"),
        easyClose = TRUE,
        size = "l"
      ))
    })

    output$exec_time_overlay <- shiny::renderUI({
      et <- osrm_exec_time()
      if (is.null(et)) {
        return(NULL)
      }
      shiny::div(
        class = "exec-time-overlay",
        shiny::div("Last request:"),
        shiny::div(paste0(round(et * 1000, 0), "ms"), style = "font-weight: bold; color: #333;")
      )
    })

    shiny::observeEvent(input$quit_app, {
      shiny::stopApp()
    })
  }

  # 7. Run Gadget
  shiny::runGadget(ui, server, viewer = shiny::browserViewer())
}

# Allow tar-related warnings during tests so we can detect regressions when
# Windows-specific extraction issues resurface.
Sys.setenv(OSRM_BACKEND_ALLOW_TAR_WARNINGS = "true")

test_osrm_path <- file.path(tempdir(), "osrm_bin_test")

# Clean up any previous runs to ensure a fresh state
if (dir.exists(test_osrm_path)) unlink(test_osrm_path, recursive = TRUE)
dir.create(test_osrm_path, showWarnings = FALSE, recursive = TRUE)

# POLITE INSTALLATION:
# Only install the binary if we are running LOCALLY or on CI (NOT_CRAN="true").
# We skip installation on CRAN to avoid timeouts and policy violations.
if (identical(Sys.getenv("NOT_CRAN"), "true")) {
  message("Running locally/CI: Installing OSRM binary to tempdir...")
  try(osrm_install(version = "v5.27.1", path = test_osrm_path, quiet = TRUE), silent = TRUE)
} else {
  message("Running on CRAN: Skipping OSRM binary installation.")
}

# CONFIGURATION (Only runs if binary exists):
bin_path <- list.files(test_osrm_path, pattern = "^osrm-routed(\\.exe)?$", full.names = TRUE, recursive = TRUE)

if (length(bin_path) > 0) {
  exe_path <- normalizePath(bin_path[1], mustWork = FALSE)

  # CRITICAL: Ensure the binary is executable (fixes Sys.which failure on Linux/Mac)
  if (.Platform$OS.type == "unix") {
    Sys.chmod(exe_path, mode = "0755")
  }

  # Set the option required by osrm_start_server
  options(osrm.routed.exec = exe_path)

  # Add to PATH so Sys.which finds it
  old_path <- Sys.getenv("PATH")
  Sys.setenv(PATH = paste(dirname(exe_path), old_path, sep = .Platform$path.sep))
}

# Helper functions to check for required binaries and profiles
has_osrm_binary <- function() {
  extract <- resolve_osrm_bin("osrm-extract")
  nzchar(extract) && (file.exists(extract) || nzchar(Sys.which(extract)))
}

has_osrm_profile <- function() {
  prof <- try(osrm_find_profile("car.lua"), silent = TRUE)
  if (inherits(prof, "try-error")) {
    return(FALSE)
  }
  options(osrm.backend.default_profile = prof)
  TRUE
}

# Set skip flag based on availability
if (!has_osrm_binary() || !has_osrm_profile()) {
  options(osrm.backend.skip_osrm_tests = TRUE)
} else {
  options(osrm.backend.skip_osrm_tests = FALSE)
}

#' Extract OSM into OSRM Graph Files
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Run the `osrm-extract` tool to preprocess an OSM file
#' (`.osm`, `.osm.bz2`, or `.osm.pbf`) into the base `.osrm` graph files
#' using a specified Lua profile.  After running, a companion
#' `<base>.osrm.timestamp` file must exist to confirm success.
#'
#' @param input_osm A string. Path to the input OSM file
#'   (`.osm`, `.osm.bz2`, or `.osm.pbf`) or a directory containing exactly one
#'   OSM file with a supported extension.
#' @param profile A string. Path to the OSRM Lua profile
#'   (e.g. returned by \code{osrm_find_profile("car.lua")}).
#' @param threads An integer. Number of threads for
#'   \code{-t/--threads}; default \code{8} (OSRM's default).
#' @param overwrite A logical. If \code{FALSE} (default), stops when any
#'   existing `.osrm*` files matching the base name are found alongside
#'   \code{input_osm}. Set to \code{TRUE} to proceed regardless.
#' @param verbosity A string. Log verbosity level passed to
#'   \code{-l/--verbosity} (one of \code{"NONE","ERROR","WARNING","INFO","DEBUG"}),
#'   default \code{"INFO"}.
#' @param data_version A string or \code{NULL}. Passed to
#'   \code{-d/--data_version}; default \code{NULL}, in which case the option
#'   is omitted.
#' @param small_component_size An integer. For
#'   \code{--small-component-size}; default \code{1000} (OSRM's default).
#' @param with_osm_metadata A logical. If \code{TRUE}, adds
#'   \code{--with-osm-metadata}; default \code{FALSE}.
#' @param parse_conditional_restrictions A logical. If \code{TRUE}, adds
#'   \code{--parse-conditional-restrictions}; default \code{FALSE}.
#' @param location_dependent_data A string or \code{NULL}. Path to GeoJSON,
#'   passed to \code{--location-dependent-data}; default \code{NULL}, in which
#'   case the option is omitted.
#' @param disable_location_cache A logical. If \code{TRUE}, adds
#'   \code{--disable-location-cache}; default \code{FALSE}.
#' @param dump_nbg_graph A logical. If \code{TRUE}, adds
#'   \code{--dump-nbg-graph}; default \code{FALSE}.
#' @inheritParams osrm_prepare_graph
#'
#' @return An object of class \code{osrm_job} with the following elements:
#' \describe{
#'   \item{osrm_job_artifact}{The path to the generated `.osrm.timestamp` file.}
#'   \item{osrm_working_dir}{The directory containing all OSRM files.}
#'   \item{logs}{The \code{processx::run} result object.}
#' }
#'
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   # Install OSRM (temporary, session PATH)
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   # copy example OSM PBF into a temporary workspace to avoid polluting pkg data
#'   pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
#'   osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
#'   dir.create(osrm_dir, recursive = TRUE)
#'   tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
#'   file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
#'
#'   # Find the path to the profile first
#'   car_profile <- osrm_find_profile("car.lua")
#'
#'   # extract OSRM graph files
#'   result <- osrm_extract(
#'     input_osm                  = tmp_pbf,
#'     profile                    = car_profile,
#'     overwrite                  = TRUE,
#'     threads                    = 1L
#'   )
#'   # path to generated .osrm files (specifically, the .osrm.timestamp file)
#'   result$osrm_job_artifact
#'
#'   # Clean up binaries and workspace
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'   unlink(osrm_dir, recursive = TRUE)
#' }
#' }
#' @export
osrm_extract <- function(
  input_osm,
  profile = osrm_find_profile("car.lua"),
  threads = 8L,
  overwrite = FALSE,
  verbosity = c("INFO", "NONE", "ERROR", "WARNING", "DEBUG"),
  data_version = NULL,
  small_component_size = 1000L,
  with_osm_metadata = FALSE,
  parse_conditional_restrictions = FALSE,
  location_dependent_data = NULL,
  disable_location_cache = FALSE,
  dump_nbg_graph = FALSE,
  quiet = FALSE,
  verbose = FALSE,
  spinner = TRUE,
  echo_cmd = FALSE
) {
  if (!requireNamespace("processx", quietly = TRUE)) {
    stop("'processx' package is required for osrm_extract", call. = FALSE)
  }

  # Extract previous logs if input is an osrm_job object
  input_logs <- if (inherits(input_osm, "osrm_job")) {
    if (is.list(input_osm$logs) && length(input_osm$logs) > 0) {
      input_osm$logs
    } else {
      list()
    }
  } else {
    list()
  }

  # Extract path from osrm_job object if needed
  input_osm <- get_osrm_path_from_input(input_osm)

  # Resolve input path (file or directory)
  input_osm <- resolve_osrm_path(
    input_osm,
    pattern = "\\.(osm|osm\\.bz2|osm\\.pbf)$",
    file_description = "OSM files (.osm, .osm.bz2, or .osm.pbf)"
  )

  # strip recognized extensions to derive base path
  if (grepl("\\.osm\\.pbf$", input_osm, ignore.case = TRUE)) {
    base <- sub("\\.osm\\.pbf$", "", input_osm, ignore.case = TRUE)
  } else if (grepl("\\.osm\\.bz2$", input_osm, ignore.case = TRUE)) {
    base <- sub("\\.osm\\.bz2$", "", input_osm, ignore.case = TRUE)
  } else if (grepl("\\.osm$", input_osm, ignore.case = TRUE)) {
    base <- sub("\\.osm$", "", input_osm, ignore.case = TRUE)
  } else {
    stop(
      "'input_osm' must have extension .osm, .osm.bz2, or .osm.pbf",
      call. = FALSE
    )
  }

  # check for existing .osrm* files
  existing <- list.files(
    dirname(base),
    pattern = paste0("^", basename(base), ".*\\.osrm"),
    ignore.case = TRUE
  )
  if (length(existing) > 0) {
    if (!overwrite) {
      stop(
        "Found existing OSRM files: ",
        paste(existing, collapse = ", "),
        ".\nSet overwrite = TRUE to proceed.",
        call. = FALSE
      )
    } else {
      if (!quiet) message("Existing OSRM files found. Cleaning up...")
      osrm_cleanup(base, quiet = quiet)
    }
  }

  # build command arguments
  verbosity <- match.arg(verbosity)
  arguments <- c(
    input_osm,
    "-p",
    profile,
    "-l",
    verbosity,
    "-t",
    as.character(threads),
    "--small-component-size",
    as.character(small_component_size)
  )
  if (!is.null(data_version)) {
    arguments <- c(arguments, "-d", data_version)
  }
  if (with_osm_metadata) {
    arguments <- c(arguments, "--with-osm-metadata")
  }
  if (parse_conditional_restrictions) {
    arguments <- c(arguments, "--parse-conditional-restrictions")
  }
  if (!is.null(location_dependent_data)) {
    arguments <- c(
      arguments,
      "--location-dependent-data",
      location_dependent_data
    )
  }
  if (disable_location_cache) {
    arguments <- c(arguments, "--disable-location-cache")
  }
  if (dump_nbg_graph) {
    arguments <- c(arguments, "--dump-nbg-graph")
  }

  # Determine final processx parameters
  show_echo <- !quiet && verbose
  show_spinner <- !quiet && spinner
  show_echo_cmd <- !quiet && echo_cmd

  # run extraction
  logs <- processx::run(
    resolve_osrm_bin("osrm-extract"),
    args = arguments,
    echo = show_echo,
    spinner = show_spinner,
    echo_cmd = show_echo_cmd
  )

  # verify timestamp file
  timestamp_file <- paste0(base, ".osrm.timestamp")
  if (!file.exists(timestamp_file)) {
    stop(
      "Extraction did not produce timestamp file: ",
      basename(timestamp_file),
      call. = FALSE
    )
  }

  # --- WRITE METADATA ---
  # Determine profile name: "path/to/car.lua" -> "car"
  profile_name <- tools::file_path_sans_ext(basename(profile))
  
  meta_file <- file.path(dirname(base), "dataset.meta.json")
  meta_data <- list(
    profile = profile_name,
    created_at = format(Sys.time(), "%Y-%m-%dT%H:%M:%SZ", tz = "UTC"),
    input_osm = basename(input_osm)
  )
  try(
    jsonlite::write_json(meta_data, meta_file, auto_unbox = TRUE, pretty = TRUE), 
    silent = TRUE
  )

  # Accumulate logs from previous stages
  accumulated_logs <- c(input_logs, list(extract = logs))

  as_osrm_job(
    osrm_job_artifact = timestamp_file,
    osrm_working_dir = dirname(timestamp_file),
    logs = accumulated_logs
  )
}

#' Read Bounding Box from OSM PBF Header
#'
#' Extracts the bbox from the HeaderBlock of an OSM PBF file using pure R.
#' This reads only the first header blob (OSMHeader), not the data blocks.
#'
#' @param pbf_file Path to an .osm.pbf file
#' @return Named list with `center` (c(lng, lat)) and `bbox` (c(lng_min, lat_min, lng_max, lat_max))
#'   or NULL if bbox not found or file invalid
#' @keywords internal
read_pbf_header_bbox <- function(pbf_file) {
  if (!file.exists(pbf_file)) {
    return(NULL)
  }

  con <- NULL
  tryCatch(
    {
      con <- file(pbf_file, "rb")

      # Read BlobHeader length (4 bytes, big endian)
      header_len <- readBin(con, "integer", n = 1, size = 4, endian = "big")
      if (header_len <= 0 || header_len > 65536) {
        return(NULL)
      } # Sanity check

      # Read BlobHeader
      header_bytes <- readBin(con, "raw", n = header_len)

      # Parse BlobHeader to get datasize
      pos <- 1
      datasize <- NULL

      while (pos <= length(header_bytes)) {
        tag <- as.integer(header_bytes[pos])
        field_num <- bitwShiftR(tag, 3)
        wire_type <- bitwAnd(tag, 7)
        pos <- pos + 1

        if (field_num == 1 && wire_type == 2) {
          # type string - skip
          vi <- .read_varint(header_bytes, pos)
          pos <- vi$next_pos + vi$value
        } else if (field_num == 2 && wire_type == 2) {
          # indexdata - skip
          vi <- .read_varint(header_bytes, pos)
          pos <- vi$next_pos + vi$value
        } else if (field_num == 3 && wire_type == 0) {
          # datasize
          datasize <- .read_varint(header_bytes, pos)$value
          break
        } else {
          break
        }
      }

      if (is.null(datasize) || datasize <= 0 || datasize > 33554432) {
        return(NULL)
      }

      # Read Blob
      blob_bytes <- readBin(con, "raw", n = datasize)
      close(con)
      con <- NULL

      # Parse Blob to get zlib_data
      pos <- 1
      zlib_data <- NULL

      while (pos <= length(blob_bytes)) {
        if (pos > length(blob_bytes)) {
          break
        }
        tag <- as.integer(blob_bytes[pos])
        field_num <- bitwShiftR(tag, 3)
        wire_type <- bitwAnd(tag, 7)
        pos <- pos + 1

        if (field_num == 1 && wire_type == 2) {
          # raw - skip
          vi <- .read_varint(blob_bytes, pos)
          pos <- vi$next_pos + vi$value
        } else if (field_num == 2 && wire_type == 0) {
          # raw_size - skip
          pos <- .read_varint(blob_bytes, pos)$next_pos
        } else if (field_num == 3 && wire_type == 2) {
          # zlib_data - this is what we want
          len <- .read_varint(blob_bytes, pos)
          pos <- len$next_pos
          zlib_data <- blob_bytes[pos:(pos + len$value - 1)]
          break
        } else {
          break
        }
      }

      if (is.null(zlib_data)) {
        return(NULL)
      }

      # Decompress zlib data
      decompressed <- tryCatch(
        {
          .zlib_decompress(zlib_data)
        },
        error = function(e) NULL
      )

      if (is.null(decompressed)) {
        return(NULL)
      }

      # Parse HeaderBlock to get bbox
      pos <- 1
      bbox <- NULL

      while (pos <= length(decompressed)) {
        if (pos > length(decompressed)) {
          break
        }
        tag <- as.integer(decompressed[pos])
        field_num <- bitwShiftR(tag, 3)
        wire_type <- bitwAnd(tag, 7)
        pos <- pos + 1

        if (field_num == 1 && wire_type == 2) {
          # bbox message
          vi <- .read_varint(decompressed, pos)
          len <- vi$value
          pos <- vi$next_pos
          bbox <- .parse_header_bbox(decompressed[pos:(pos + len - 1)])
          break
        } else if (wire_type == 2) {
          # Skip other string/bytes fields
          vi <- .read_varint(decompressed, pos)
          pos <- vi$next_pos + vi$value
        } else if (wire_type == 0) {
          # Skip varint fields
          pos <- .read_varint(decompressed, pos)$next_pos
        } else {
          break
        }
      }

      return(bbox)
    },
    error = function(e) {
      NULL
    },
    finally = {
      if (!is.null(con)) try(close(con), silent = TRUE)
    }
  )
}

#' @keywords internal
.read_varint <- function(bytes, start = 1) {
  result <- 0
  shift <- 0
  pos <- start

  while (pos <= length(bytes)) {
    byte <- as.integer(bytes[pos])
    result <- result + bitwAnd(byte, 127) * (2^shift)
    if (bitwAnd(byte, 128) == 0) {
      break
    }
    shift <- shift + 7
    pos <- pos + 1
  }

  list(value = result, next_pos = pos + 1)
}

#' @keywords internal
.parse_header_bbox <- function(bbox_bytes) {
  pos <- 1
  coords <- list(left = NULL, right = NULL, top = NULL, bottom = NULL)

  while (pos <= length(bbox_bytes)) {
    tag <- as.integer(bbox_bytes[pos])
    field_num <- bitwShiftR(tag, 3)
    wire_type <- bitwAnd(tag, 7)
    pos <- pos + 1

    if (wire_type == 0) {
      # varint
      vi <- .read_varint(bbox_bytes, pos)
      pos <- vi$next_pos
      # Zigzag decode using arithmetic to handle 64-bit nanodegree values
      decoded <- .zigzag_decode_64(as.numeric(vi$value))

      if (field_num == 1) {
        coords$left <- decoded
      } else if (field_num == 2) {
        coords$right <- decoded
      } else if (field_num == 3) {
        coords$top <- decoded
      } else if (field_num == 4) {
        coords$bottom <- decoded
      }
    } else {
      break
    }
  }

  if (!any(sapply(coords, is.null)) && !any(sapply(coords, is.na))) {
    NANO <- 1e-9
    lng_min <- as.numeric(coords$left) * NANO
    lng_max <- as.numeric(coords$right) * NANO
    lat_min <- as.numeric(coords$bottom) * NANO
    lat_max <- as.numeric(coords$top) * NANO

    list(
      center = c(mean(c(lng_min, lng_max)), mean(c(lat_min, lat_max))),
      bbox = c(lng_min, lat_min, lng_max, lat_max)
    )
  } else {
    NULL
  }
}

#' @keywords internal
.zigzag_decode_64 <- function(n) {
  if (n %% 2 == 0) {
    n / 2
  } else {
    -(n + 1) / 2
  }
}

#' @keywords internal
.zlib_decompress <- function(data) {
  if (length(data) == 0) {
    stop("Empty data to decompress")
  }

  # Method 1: Use zlib-flate if available
  zf <- Sys.which("zlib-flate")
  if (nzchar(zf)) {
    temp_in <- tempfile()
    temp_out <- tempfile()
    writeBin(data, temp_in)
    on.exit(unlink(c(temp_in, temp_out)), add = TRUE)

    # Use system2 for better cross-platform support
    # Redirect both stdout and stderr to avoid issues
    exit_code <- system2(
      zf,
      args = c("-uncompress"),
      stdin = temp_in,
      stdout = temp_out,
      stderr = FALSE
    )

    if (exit_code == 0 && file.exists(temp_out) && file.size(temp_out) > 0) {
      return(readBin(temp_out, "raw", n = file.size(temp_out)))
    }
  }

  # Method 2: Try Python zlib
  py <- Sys.which("python3")
  if (nzchar(py)) {
    temp_in <- tempfile()
    temp_out <- tempfile()
    py_script <- tempfile(fileext = ".py")
    on.exit(unlink(c(temp_in, temp_out, py_script)), add = TRUE)

    cat(
      "import sys, zlib; open(sys.argv[1],'wb').write(zlib.decompress(open(sys.argv[2],'rb').read()))",
      file = py_script
    )
    writeBin(data, temp_in)

    exit_code <- system2(
      py,
      args = c(shQuote(py_script), shQuote(temp_out), shQuote(temp_in))
    )
    if (exit_code == 0 && file.exists(temp_out) && file.size(temp_out) > 0) {
      return(readBin(temp_out, "raw", n = file.size(temp_out)))
    }
  }

  # Method 3: Try R's memDecompress by converting zlib to gzip format
  result <- tryCatch(
    {
      if (
        length(data) > 6 &&
          data[1] == as.raw(0x78) &&
          data[2] %in% c(as.raw(0x9c), as.raw(0x01), as.raw(0xda))
      ) {
        # Strip 2-byte zlib header and 4-byte ADLER32 checksum to get raw DEFLATE
        deflate_data <- data[3:(length(data) - 4)]
        # Wrap in minimal gzip container (header + raw deflate + dummy trailer)
        gzip_header <- as.raw(c(
          0x1f,
          0x8b,
          0x08,
          0x00,
          0x00,
          0x00,
          0x00,
          0x00,
          0x00,
          0x03
        ))
        gzip_trailer <- as.raw(rep(0x00, 8))
        memDecompress(c(gzip_header, deflate_data, gzip_trailer), type = "gzip")
      } else {
        memDecompress(data, type = "gzip")
      }
    },
    error = function(e) NULL
  )
  if (!is.null(result)) {
    return(result)
  }

  stop("Unable to decompress PBF header")
}

#' Get PBF Bounding Box with Fallbacks
#'
#' Tries multiple methods to extract the bounding box from an OSM PBF file:
#' 1. Pure R header parsing (fastest, no dependencies)
#' 2. osmium CLI (if available in PATH)
#' 3. sf sampling (if sf is installed, reads first 10 features)
#'
#' @param pbf_file Path to an .osm.pbf file
#' @return Named list with `center` (c(lng, lat)) and `bbox` (c(lng_min, lat_min, lng_max, lat_max))
#'   or NULL if bbox could not be determined
#' @keywords internal
.get_pbf_extent <- function(pbf_file) {
  if (!file.exists(pbf_file)) {
    return(NULL)
  }

  # Method 1: Pure R header parsing (fastest, no external deps)
  result <- tryCatch(read_pbf_header_bbox(pbf_file), error = function(e) NULL)
  if (!is.null(result)) {
    return(result)
  }

  # Method 2: osmium fileinfo (reads header only)
  osmium_path <- Sys.which("osmium")
  if (nzchar(osmium_path)) {
    tryCatch(
      {
        info <- system2(
          osmium_path,
          args = c("fileinfo", shQuote(pbf_file)),
          stdout = TRUE,
          stderr = FALSE
        )
        bbox_line <- grep("^\\s+\\([-\\d\\.]+,", info, value = TRUE)
        if (length(bbox_line) > 0) {
          bbox_str <- sub("^\\s+\\(([^)]+)\\).*", "\\1", bbox_line[1])
          coords <- as.numeric(strsplit(bbox_str, ",")[[1]])
          if (length(coords) == 4 && !any(is.na(coords))) {
            return(list(
              center = c(mean(coords[c(1, 3)]), mean(coords[c(2, 4)])),
              bbox = coords
            ))
          }
        }
      },
      error = function(e) NULL
    )
  }

  # Method 3: sf sampling (try points first, then lines for road-only PBFs)
  if (requireNamespace("sf", quietly = TRUE)) {
    old_idx <- Sys.getenv("OSM_USE_CUSTOM_INDEXING", unset = NA)
    Sys.setenv(OSM_USE_CUSTOM_INDEXING = "NO")
    on.exit(
      {
        if (is.na(old_idx)) {
          Sys.unsetenv("OSM_USE_CUSTOM_INDEXING")
        } else {
          Sys.setenv(OSM_USE_CUSTOM_INDEXING = old_idx)
        }
      },
      add = TRUE
    )
    for (layer in c("points", "lines")) {
      tryCatch(
        {
          feats <- sf::st_read(
            pbf_file,
            query = sprintf("SELECT * FROM %s LIMIT 10", layer),
            quiet = TRUE
          )
          if (nrow(feats) > 0) {
            bbox <- sf::st_bbox(feats)
            return(list(
              center = c(
                mean(c(bbox$xmin, bbox$xmax)),
                mean(c(bbox$ymin, bbox$ymax))
              ),
              bbox = as.numeric(bbox[c("xmin", "ymin", "xmax", "ymax")])
            ))
          }
        },
        error = function(e) NULL
      )
    }
  }

  NULL
}

test_that("read_pbf_header_bbox works with cur.osm.pbf", {
  pbf_file <- system.file("extdata", "cur.osm.pbf", package = "osrm.backend")
  skip_if(pbf_file == "", message = "cur.osm.pbf not found")

  res <- read_pbf_header_bbox(pbf_file)

  expect_type(res, "list")
  expect_named(res, c("center", "bbox"))
  expect_length(res$center, 2)
  expect_length(res$bbox, 4)
  expect_true(all(is.numeric(res$center)))
  expect_true(all(is.numeric(res$bbox)))

  # Semi-hardcoded expectations based on Curitiba (cur.osm.pbf location)
  # lng: -49.27, lat: -25.43 approx
  expect_equal(res$center[1], -49.2699, tolerance = 0.001)
  expect_equal(res$center[2], -25.4305, tolerance = 0.001)
  expect_equal(res$bbox[1], -49.2889, tolerance = 0.001) # lng_min
  expect_equal(res$bbox[2], -25.4413, tolerance = 0.001) # lat_min
  expect_equal(res$bbox[3], -49.2509, tolerance = 0.001) # lng_max
  expect_equal(res$bbox[4], -25.4197, tolerance = 0.001) # lat_max
})

test_that(".get_pbf_extent works with cur.osm.pbf", {
  pbf_file <- system.file("extdata", "cur.osm.pbf", package = "osrm.backend")
  skip_if(pbf_file == "", message = "cur.osm.pbf not found")

  res <- .get_pbf_extent(pbf_file)

  expect_type(res, "list")
  expect_named(res, c("center", "bbox"))
  expect_equal(res$center[1], -49.2699, tolerance = 0.001)
})

test_that("read_pbf_header_bbox handles missing files", {
  expect_null(read_pbf_header_bbox("nonexistent.pbf"))
})

test_that(".get_pbf_extent handles missing files", {
  expect_null(.get_pbf_extent("nonexistent.pbf"))
})

test_that("read_pbf_header_bbox handles invalid files", {
  tmp <- tempfile(fileext = ".pbf")
  writeBin(as.raw(c(0, 0, 0, 1, 0)), tmp) # Small invalid header
  on.exit(unlink(tmp))

  expect_null(read_pbf_header_bbox(tmp))
})

#' Start an OSRM Server with Automatic Setup
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' A high-level, "one-shot" function to start an OSRM server that automatically
#' handles OSRM installation and graph preparation. This is the recommended
#' function for most users to get a server running quickly with minimal steps.
#'
#' @details
#' This function is designed for convenience and automates the entire setup
#' process. By default, it is not verbose and only prints high-level status
#' messages.
#'
#' \enumerate{
#'   \item \strong{Check for OSRM Installation:} It first verifies if the `osrm-routed`
#'     executable is available in the system's `PATH`. If not, it automatically
#'     calls `osrm_install()` to download and install the latest stable version.
#'   \item \strong{Process Input Path and Prepare Graph:} The function intelligently
#'     handles the `path` argument to find or create the necessary graph files.
#'     If the graph files do not exist, it automatically runs `osrm_prepare_graph()`
#'     to build them, which may take some time.
#'   \item \strong{Start Server:} Once the graph files are located or prepared, it
#'     launches the `osrm-routed` server and prints a confirmation message
#'     with the server's PID and port.
#' }
#' For advanced users or for debugging, set `verbose = TRUE` to see the detailed
#' console output from the installation and graph preparation steps. For full
#' manual control, use the lower-level functions like `osrm_prepare_graph()` and
#' `osrm_start_server()` directly.
#' @inheritParams osrm_install
#' @param path A string. Path to the input data. Can be one of:
#'   \itemize{
#'     \item A path to an OSM file (e.g., `/path/to/data.osm.pbf`).
#'     \item A path to a directory containing OSRM graph files or an OSM file.
#'     \item A direct path to a final graph file (`.osrm.mldgr` or `.osrm.hsgr`).
#'   }
#' @param algorithm A string specifying the routing algorithm to use for graph
#'   preparation, either `"mld"` (Multi-Level Dijkstra, default) or `"ch"`
#'   (Contraction Hierarchies). This is only used when `osrm_prepare_graph` is
#'   automatically called.
#' @param force_rebuild A logical value. If `TRUE`, force the rebuilding of the
#'   OSRM graph even if it already exists.
#' @param show_progress A logical value. If `TRUE` (default), show a progress
#'   spinner during graph preparation.
#' @param verbose A logical. If `FALSE` (default), suppresses detailed console
#'   output from backend commands. If `TRUE`, shows all output, which is useful
#'   for debugging.
#' @param ... Additional arguments passed on to `osrm_prepare_graph()` (e.g.,
#'   `overwrite = TRUE`) and `osrm_start_server()` (e.g., `port = 5001`).
#'
#' @return An OSRM job process (an `osrm_server` object inheriting from
#'   `processx::process`) for the running server.
#' @export
#' @seealso [osrm_stop()], [osrm_start_server()] for manual server startup.
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   # copy example OSM PBF into a temporary workspace to avoid polluting pkg data
#'   pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
#'   osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
#'   dir.create(osrm_dir, recursive = TRUE)
#'   local_pbf <- file.path(osrm_dir, "cur.osm.pbf")
#'   file.copy(from = pbf_path, to = local_pbf, overwrite = TRUE)
#'
#'   # Start the server with one command.
#'   # It will quietly install OSRM and prepare the graph if needed.
#'   osrm_process <- osrm_start(local_pbf)
#'
#'   # Stop the server when done.
#'   osrm_stop()
#'
#'   # To see all backend logs during setup, use verbose = TRUE
#'   osrm_process_verbose <- osrm_start(local_pbf, verbose = TRUE)
#'   osrm_stop()
#'
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'   unlink(osrm_dir, recursive = TRUE)
#' }
#' }
osrm_start <- function(
  path,
  algorithm = c("mld", "ch"),
  force_rebuild = FALSE,
  show_progress = TRUE,
  quiet = FALSE,
  verbose = FALSE,
  ...
) {
  # Match algorithm argument
  algorithm <- match.arg(algorithm)
  quiet <- isTRUE(quiet)
  verbose <- isTRUE(verbose)
  force_rebuild <- isTRUE(force_rebuild)
  show_progress <- isTRUE(show_progress)

  # Capture additional arguments
  dot_args <- list(...)

  # Separate args for prepare_graph and start_server based on their formals
  # We use match.arg or similar logic to route them correctly
  prepare_formals <- names(formals(osrm_prepare_graph))
  server_formals <- names(formals(osrm_start_server))
  
  # Arguments that should go to prepare_graph
  # We perform a strict check to avoiding passing server args to build
  build_args <- dot_args[names(dot_args) %in% prepare_formals]
  
  # Arguments that should go to start_server
  server_args <- dot_args[names(dot_args) %in% server_formals]
  
  # Common arguments handling: threads, verbosity might be in both
  # If 'threads' is passed in ..., it goes to both via the above logic
  
  # 1. Resolve input path and prepare graph if necessary
  if (!file.exists(path)) {
    stop("Input path does not exist: ", path, call. = FALSE)
  }

  # 2. Check if OSRM is installed, if not, install it
  osrm_exec <- resolve_osrm_bin("osrm-routed")
  if (!nzchar(Sys.which(osrm_exec)) && !file.exists(osrm_exec)) {
    if (!quiet) {
      message("OSRM backend not found. Installing latest version...")
    }
    osrm_install(version = "latest", path_action = "session", quiet = quiet)
    if (!quiet) message("Installation complete.")
  }

  final_graph_path <- NULL
  original_osm_path <- NULL
  should_build <- FALSE
  
  # --- LOGIC TO DETERMINE IF WE NEED TO BUILD ---
  
  # Case A: Input is a directory
  if (dir.exists(path)) {
    # Check for existing graphs
    suffix <- if (algorithm == "mld") "\\.osrm\\.mldgr$" else "\\.osrm\\.hsgr$"
    existing_graphs <- list.files(path, pattern = suffix, full.names = TRUE)
    
    if (length(existing_graphs) > 0 && !force_rebuild) {
      final_graph_path <- existing_graphs[1]
      # Try to find original OSM for tracking if possible (heuristic)
      heuristic_osm <- sub("\\.osrm\\.(mldgr|hsgr)$", ".osm.pbf", final_graph_path)
      if (file.exists(heuristic_osm)) {
        original_osm_path <- heuristic_osm
      } else {
        # Fallback: take any .osm.pbf in the same directory
        osm_files <- list.files(path, pattern = "\\.osm(\\.pbf|\\.bz2)?$", full.names = TRUE)
        if (length(osm_files) > 0) original_osm_path <- osm_files[1]
      }
    } else {
      # Need to build. Find source PBF.
      osm_files <- list.files(path, pattern = "\\.osm(\\.pbf|\\.bz2)?$", full.names = TRUE)
      if (length(osm_files) == 0) {
        if (force_rebuild) stop("Force rebuild requested but no source OSM file found in directory.", call. = FALSE)
        stop("Directory contains no prepared OSRM graphs or source OSM files.", call. = FALSE)
      }
      original_osm_path <- osm_files[1]
      should_build <- TRUE
    }
  } 
  # Case B: Input is a source file (.osm.pbf, .osm)
  else if (grepl("\\.osm(\\.pbf|\\.bz2)?$", path, ignore.case = TRUE)) {
    original_osm_path <- path
    base_name <- sub("\\.osm(\\.pbf|\\.bz2)?$", "", path, ignore.case = TRUE)
    suffix <- if (algorithm == "mld") ".osrm.mldgr" else ".osrm.hsgr"
    expected_graph <- paste0(base_name, suffix)
    
    if (file.exists(expected_graph) && !force_rebuild) {
      final_graph_path <- expected_graph
    } else {
      should_build <- TRUE
    }
  }
  # Case C: Input is already a graph file
  else if (grepl("\\.osrm\\.(mldgr|hsgr)$", path, ignore.case = TRUE)) {
    if (force_rebuild) {
       stop("Cannot force rebuild when input 'path' is already a compiled graph file. Point to the source .osm.pbf instead.", call. = FALSE)
    }
    final_graph_path <- path
    # Try to find original OSM for tracking if possible (heuristic)
    heuristic_osm <- sub("\\.osrm\\.(mldgr|hsgr)$", ".osm.pbf", path)
    if (file.exists(heuristic_osm)) {
      original_osm_path <- heuristic_osm
    } else {
      # Fallback: take any .osm.pbf in the same directory
      osm_files <- list.files(dirname(path), pattern = "\\.osm(\\.pbf|\\.bz2)?$", full.names = TRUE)
      if (length(osm_files) > 0) original_osm_path <- osm_files[1]
    }
  } else {
    stop("Invalid input path. Must be a directory, OSM source file, or OSRM graph.", call. = FALSE)
  }

  # --- PERFORM BUILD IF NEEDED ---
  
  if (should_build) {
    if (!quiet) {
      msg <- if (force_rebuild) "Rebuilding OSRM graph..." else "OSRM graph not found. Preparing..."
      message(msg, " This may take a while.")
    }
    
    # Construct arguments for osrm_prepare_graph
    # We enforce known args + user provided build_args
    build_call_args <- utils::modifyList(
      build_args,
      list(
        input_osm = original_osm_path,
        algorithm = algorithm,
        quiet = quiet,
        verbose = verbose,
        overwrite = force_rebuild || isTRUE(build_args$overwrite) # Ensure overwrite if forced
      )
    )
    
    # Run build
    prepared_graph <- do.call(osrm_prepare_graph, build_call_args)
    final_graph_path <- prepared_graph$osrm_job_artifact
    
    if (!quiet) message("Graph preparation complete.")
  }

  if (is.null(final_graph_path)) {
    stop("Could not find or prepare a valid OSRM graph file.", call. = FALSE)
  }

  # 3. Start the server
  # Construct arguments for osrm_start_server
  server_call_args <- utils::modifyList(
    server_args,
    list(
      osrm_path = final_graph_path,
      quiet = quiet,
      verbose = verbose,
      input_osm = original_osm_path
    )
  )
  
  # Only pass algorithm if explicitly provided or inferred? 
  # osrm_start_server usually auto-detects from extension, but we can pass it to be safe if mld/ch match
  if (is.null(server_call_args$algorithm)) {
     server_call_args$algorithm <- if (grepl("mldgr$", final_graph_path)) "MLD" else "CH"
  }

  if (!quiet) {
    message(
      "Starting OSRM server with graph '",
      basename(final_graph_path),
      "'..."
    )
  }

  osrm_process <- do.call(osrm_start_server, server_call_args)

  # Extract info for the final confirmation message
  pid <- osrm_process$get_pid()
  # Port might be in server_args or default
  port <- server_call_args$port %||% 5001L

  if (!quiet) {
    message(sprintf(
      "OSRM server started successfully (pid %s, port %s).",
      pid,
      port
    ))
  }

  return(osrm_process)
}

#' Partition OSRM Graph for Multi-Level Dijkstra (MLD)
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Run the `osrm-partition` tool to partition an OSRM graph for the MLD pipeline.
#' After running, a companion `<base>.osrm.partition` file must exist to confirm success.
#'
#' @param input_osrm A string. Path to a `.osrm.timestamp` file, the base path to the `.osrm` files (without extension),
#'   or a directory containing exactly one `.osrm.timestamp` file.
#' @param threads An integer. Number of threads to use; default `8` (osrm-partition's default).
#' @param verbosity A string. Log verbosity level passed to `-l/--verbosity`
#'   (one of `"NONE","ERROR","WARNING","INFO","DEBUG"`); default `"INFO"`.
#' @param balance A numeric. Balance for left and right side in single bisection; default `1.2`.
#' @param boundary A numeric. Percentage of embedded nodes to contract as sources and sinks; default `0.25`.
#' @param optimizing_cuts An integer. Number of cuts to use for optimizing a single bisection; default `10`.
#' @param small_component_size An integer. Size threshold for small components; default `1000`.
#' @param max_cell_sizes A numeric vector. Maximum cell sizes starting from level 1; default `c(128,4096,65536,2097152)`.
#' @inheritParams osrm_prepare_graph
#'
#' @return An object of class \code{osrm_job} with the following elements:
#' \describe{
#'   \item{osrm_job_artifact}{The path to the partitioned `.osrm.partition` file.}
#'   \item{osrm_working_dir}{The directory containing all OSRM files.}
#'   \item{logs}{The \code{processx::run} result object.}
#' }
#'
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   # Prepare a small graph then partition it for the MLD pipeline
#'   pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
#'   osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
#'   dir.create(osrm_dir, recursive = TRUE)
#'   tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
#'   file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
#'   profile <- osrm_find_profile("car.lua")
#'
#'   extract_job <- osrm_extract(
#'     input_osm = tmp_pbf,
#'     profile = profile,
#'     overwrite = TRUE,
#'     threads = 1L
#'   )
#'
#'   partition_job <- osrm_partition(extract_job, threads = 1L, verbose = TRUE)
#'   partition_job$osrm_job_artifact
#'
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'   unlink(osrm_dir, recursive = TRUE)
#' }
#' }
#'
#' @export
osrm_partition <- function(
  input_osrm,
  threads = 8L,
  verbosity = c("INFO", "NONE", "ERROR", "WARNING", "DEBUG"),
  balance = 1.2,
  boundary = 0.25,
  optimizing_cuts = 10L,
  small_component_size = 1000L,
  max_cell_sizes = c(128, 4096, 65536, 2097152),
  quiet = FALSE,
  verbose = FALSE,
  spinner = TRUE,
  echo_cmd = FALSE
) {
  if (!requireNamespace("processx", quietly = TRUE)) {
    stop("'processx' package is required for osrm_partition", call. = FALSE)
  }

  # Extract previous logs if input is an osrm_job object
  input_logs <- if (inherits(input_osrm, "osrm_job")) {
    if (is.list(input_osrm$logs) && length(input_osrm$logs) > 0) {
      input_osrm$logs
    } else {
      list()
    }
  } else {
    list()
  }

  # Extract path from osrm_job object if needed
  input_osrm <- get_osrm_path_from_input(input_osrm)

  if (!is.character(input_osrm) || length(input_osrm) != 1) {
    stop(
      "'input_osrm' must be a single string (path to .osrm base)",
      call. = FALSE
    )
  }

  # Resolve input path (file or directory)
  input_osrm <- resolve_osrm_path(
    input_osrm,
    pattern = "\\.osrm\\.timestamp$",
    file_description = ".osrm.timestamp files",
    error_context = "Please check that you have run `osrm_extract` first."
  )

  if (!file.exists(input_osrm)) {
    stop(
      "File does not exist: ",
      input_osrm,
      "\nPlease check that you have run `osrm_extract` first.",
      call. = FALSE
    )
  }

  osrm_path <- gsub("\\.timestamp$", "", input_osrm)

  # Check for algorithm conflicts (CH files in directory)
  base_name <- sub("\\.osrm$", "", basename(osrm_path))
  dir_path <- dirname(osrm_path)
  check_algorithm_conflict(dir_path, base_name, "mld", "partition")

  verbosity <- match.arg(verbosity)
  arguments <- c(
    osrm_path,
    "-l",
    verbosity,
    "-t",
    as.character(threads),
    "--balance",
    as.character(balance),
    "--boundary",
    as.character(boundary),
    "--optimizing-cuts",
    as.character(optimizing_cuts),
    "--small-component-size",
    as.character(small_component_size),
    "--max-cell-sizes",
    paste(max_cell_sizes, collapse = ",")
  )

  # Determine final processx parameters
  show_echo <- !quiet && verbose
  show_spinner <- !quiet && spinner
  show_echo_cmd <- !quiet && echo_cmd

  logs <- processx::run(
    resolve_osrm_bin("osrm-partition"),
    args = arguments,
    echo = show_echo,
    spinner = show_spinner,
    echo_cmd = show_echo_cmd
  )

  partition_file <- paste0(gsub("\\.osrm$", "", osrm_path), ".osrm.partition")
  if (!file.exists(partition_file)) {
    stop(
      "Partitioning did not produce partition file: ",
      basename(partition_file),
      call. = FALSE
    )
  }

  # Accumulate logs from previous stages
  accumulated_logs <- c(input_logs, list(partition = logs))

  as_osrm_job(
    osrm_job_artifact = normalizePath(partition_file),
    osrm_working_dir = dirname(normalizePath(partition_file)),
    logs = accumulated_logs
  )
}

test_that("osrm_partition runs osrm-partition with expected arguments", {
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  input_osrm <- file.path(
    tmp_dir,
    paste0("osrm-partition-test-", Sys.getpid(), ".osrm")
  )
  on.exit(unlink(input_osrm), add = TRUE)
  file.create(input_osrm)

  captured <- list()
  mock_run <- function(command, args, echo, spinner, echo_cmd, ...) {
    captured <<- list(
      command = command,
      args = args,
      echo = echo,
      spinner = spinner,
      echo_cmd = echo_cmd
    )
    # Create the expected output file
    partition_file <- paste0(input_osrm, ".partition")
    file.create(partition_file)
    list(status = 0, stdout = "", stderr = "")
  }

  result <- with_mocked_bindings(
    osrm_partition(
      input_osrm = input_osrm,
      threads = 6L,
      verbosity = "ERROR",
      balance = 1.5,
      boundary = 0.3,
      optimizing_cuts = 20L,
      small_component_size = 500L,
      max_cell_sizes = c(100, 1000),
      quiet = FALSE,
      verbose = TRUE,
      spinner = TRUE,
      echo_cmd = TRUE
    ),
    run = mock_run,
    .package = "processx"
  )

  expect_equal(tools::file_path_sans_ext(basename(captured$command)), "osrm-partition")
  expect_true("-t" %in% captured$args && "6" %in% captured$args)
  expect_true("-l" %in% captured$args && "ERROR" %in% captured$args)
  expect_true("--balance" %in% captured$args && "1.5" %in% captured$args)
  expect_true("--boundary" %in% captured$args && "0.3" %in% captured$args)
  expect_true("--optimizing-cuts" %in% captured$args && "20" %in% captured$args)
  expect_true(
    "--small-component-size" %in% captured$args && "500" %in% captured$args
  )
  expect_true(
    "--max-cell-sizes" %in% captured$args && "100,1000" %in% captured$args
  )
})

test_that("osrm_partition handles missing input file", {
  expect_error(
    osrm_partition("nonexistent.osrm"),
    "File does not exist"
  )
})

test_that("osrm_partition checks for output file", {
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  input_osrm <- file.path(
    tmp_dir,
    paste0("osrm-fail-", Sys.getpid(), ".osrm")
  )
  on.exit(unlink(input_osrm), add = TRUE)
  file.create(input_osrm)

  mock_run <- function(...) {
    list(status = 0, stdout = "", stderr = "")
  }

  expect_error(
    with_mocked_bindings(
      osrm_partition(input_osrm),
      run = mock_run,
      .package = "processx"
    ),
    "Partitioning did not produce partition file"
  )
})

test_that("osrm_partition accepts a directory with one .osrm.timestamp file", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-partition-dir-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  input_osrm <- file.path(tmp_dir, "test.osrm.timestamp")
  file.create(input_osrm)

  normalized_input <- normalizePath(input_osrm)
  expected_partition <- sub("\\.timestamp$", ".partition", normalized_input)

  mock_run <- function(command, args, echo, spinner, echo_cmd, ...) {
    file.create(expected_partition)
    list(status = 0, stdout = "", stderr = "")
  }

  result <- with_mocked_bindings(
    osrm_partition(
      input_osrm = tmp_dir,
      quiet = TRUE
    ),
    run = mock_run,
    .package = "processx"
  )

  expect_equal(result$osrm_job_artifact, expected_partition)
})

test_that("osrm_partition errors when directory has no .osrm.timestamp files", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-partition-empty-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  expect_error(
    osrm_partition(input_osrm = tmp_dir),
    "No .osrm.timestamp files found"
  )
})

test_that("osrm_partition errors when directory has multiple .osrm.timestamp files", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-partition-multi-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  file.create(file.path(tmp_dir, "test1.osrm.timestamp"))
  file.create(file.path(tmp_dir, "test2.osrm.timestamp"))

  expect_error(
    osrm_partition(input_osrm = tmp_dir),
    "Multiple .osrm.timestamp files found"
  )
})

#' Start an OSRM MLD/CH server with `osrm-routed`
#'
#' @description
#' `r lifecycle::badge("stable")`
#'
#' Launches an `osrm-routed` process pointing at a localized OSRM graph (either
#' `.osrm.mldgr` for MLD or `.osrm.hsgr` for CH/CoreCH).
#'
#' @details
#' The server's standard output and error streams are handled via temporary files
#' by default to prevent deadlocks in R's single-threaded environment. This ensures
#' reliable operation while preserving logs for debugging startup failures.
#'
#' To customize logging behavior, you can use the following approaches:
#' \itemize{
#'   \item **Default (Temp File):** Logs are written to a temporary file. This prevents
#'     deadlocks while keeping logs available for debugging.
#'
#'   \item **Verbose Mode:** Set `verbose = TRUE` to display logs directly in the
#'     R console. Note: This can cause deadlocks in tight loops if R is busy.
#'
#'   \item **Custom Log File:** Set the `osrm.server.log_file` option to redirect
#'     output to a specific file:
#'     `options(osrm.server.log_file = "path/to/osrm.log")`
#'
#'     Note: List specifications (e.g., `list(stdout = "...", stderr = "...")`)
#'     are deprecated and will fall back to the default temporary file behavior.
#' }
#'
#' You can override the `osrm-routed` executable via
#' `options(osrm.routed.exec = "/full/path/to/osrm-routed")`.
#'
#' @param osrm_path Character(1). Path to the `.osrm.mldgr` or `.osrm.hsgr` file
#' @param version Logical; if `TRUE`, prints version and exits
#' @param help Logical; if `TRUE`, prints help and exits
#' @param verbosity Character; one of `"NONE","ERROR","WARNING","INFO","DEBUG"`
#' @param trial Logical or integer; if `TRUE` or >0, quits after initialization (default: `FALSE`)
#' @param ip Character; IP address to bind (default: `"0.0.0.0"`)
#' @param port Integer; TCP port to listen on (default: `5001`). The function checks
#'   if this port is already in use by another running OSRM server (even from
#'   another session) and will stop with an error if a conflict is detected.
#' @param threads Integer; number of worker threads (default: `8`)
#' @param shared_memory Logical; load graph from shared memory (default: `FALSE`)
#' @param memory_file Character or NULL; DEPRECATED (behaves like `mmap`)
#' @param mmap Logical; memory-map data files (default: `FALSE`)
#' @param dataset_name Character or NULL; name of shared memory dataset
#' @param algorithm Character or NULL; one of `"MLD"`, `"CH"`, or `"CoreCH"` (case-insensitive).
#'   If `NULL` (default), auto-selected based on file extension.
#' @param max_viaroute_size Integer (default: `500`)
#' @param max_trip_size Integer (default: `100`)
#' @param max_table_size Integer (default: `100`)
#' @param max_matching_size Integer (default: `100`)
#' @param max_nearest_size Integer (default: `100`)
#' @param max_alternatives Integer (default: `3`)
#' @param max_matching_radius Integer; use `-1` for unlimited (default: `-1`)
#' @param input_osm Character or NULL; path to the original OSM input file (for tracking purposes).
#'   This parameter is typically used internally by [osrm_start()] to record the source data.
#' @param echo_cmd Logical; echo command line to console before launch (default: `FALSE`)
#' @param quiet Logical; when `TRUE`, suppresses package messages.
#' @param verbose Logical; when `TRUE`, routes server stdout and stderr to the R
#'   console for live debugging. Note: This can cause deadlocks in tight loops
#'   if R is busy. Defaults to `FALSE`, which writes logs to a temporary file.
#'
#' @return An OSRM job process (an `osrm_server` object inheriting from
#'   `processx::process`) for the running server (also registered internally).
#' @examples
#' \donttest{
#' if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
#'   install_dir <- osrm_install(
#'     version = "latest",
#'     path_action = "session",
#'     quiet = TRUE
#'   )
#'
#'   # Build a graph then launch an OSRM server on a custom port
#'   pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
#'   osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
#'   dir.create(osrm_dir, recursive = TRUE)
#'   tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
#'   file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
#'
#'   graph <- osrm_prepare_graph(
#'     input_osm = tmp_pbf,
#'     overwrite = TRUE,
#'     threads = 1L,
#'     algorithm = "mld"
#'   )
#'
#'   server <- osrm_start_server(
#'     osrm_path = graph$osrm_job_artifact,
#'     port = 6000,
#'     threads = 1L
#'   )
#'
#'   # Later, stop the server again
#'   osrm_stop(server)
#'
#'   osrm_uninstall(
#'     dest_dir = install_dir,
#'     clear_path = TRUE,
#'     force = TRUE,
#'     quiet = TRUE
#'   )
#'   unlink(osrm_dir, recursive = TRUE)
#' }
#' }
#' @export
osrm_start_server <- function(
  osrm_path,
  version = FALSE,
  help = FALSE,
  verbosity = c("INFO", "ERROR", "WARNING", "NONE", "DEBUG"),
  trial = FALSE,
  ip = "0.0.0.0",
  port = 5001L,
  threads = 8L,
  shared_memory = FALSE,
  memory_file = NULL,
  mmap = FALSE,
  dataset_name = NULL,
  algorithm = NULL,
  max_viaroute_size = 500L,
  max_trip_size = 100L,
  max_table_size = 100L,
  max_matching_size = 100L,
  max_nearest_size = 100L,
  max_alternatives = 3L,
  max_matching_radius = -1L,
  quiet = FALSE,
  verbose = FALSE,
  echo_cmd = FALSE,
  input_osm = NULL
) {
  # Dependencies
  if (!requireNamespace("processx", quietly = TRUE)) {
    stop(
      "'processx' package is required for osrm_start_server()",
      call. = FALSE
    )
  }
  quiet <- isTRUE(quiet)
  verbose <- isTRUE(verbose)

  # Add this at the very beginning of the function body
  osrm_path <- get_osrm_path_from_input(osrm_path)

  # Validate inputs
  if (!is.character(osrm_path) || length(osrm_path) != 1) {
    stop(
      "'osrm_path' must be a single string pointing to .osrm.mldgr or .osrm.hsgr",
      call. = FALSE
    )
  }
  if (!file.exists(osrm_path)) {
    stop("File does not exist: ", osrm_path, call. = FALSE)
  }

  ext <- tolower(sub(".*\\.osrm\\.(.+)$", "\\1", osrm_path))
  if (!ext %in% c("mldgr", "hsgr")) {
    stop("'osrm_path' must end in .osrm.mldgr or .osrm.hsgr", call. = FALSE)
  }

  # Algorithm selection
  if (is.null(algorithm)) {
    algorithm <- if (ext == "mldgr") "MLD" else "CH"
  } else {
    algorithm <- .normalize_algorithm(algorithm)
    if (ext == "mldgr" && algorithm != "MLD") {
      stop(
        "Algorithm must be 'MLD' when using an .osrm.mldgr file",
        call. = FALSE
      )
    }
    if (ext == "hsgr" && !(algorithm %in% c("CH", "CoreCH"))) {
      stop(
        "Algorithm must be 'CH' or 'CoreCH' when using an .osrm.hsgr file",
        call. = FALSE
      )
    }
  }

  # Prefix (graph base path without the .mldgr/.hsgr suffix)
  prefix <- sub(
    "\\.osrm\\.(?:mldgr|hsgr)$",
    "\\.osrm",
    osrm_path,
    ignore.case = TRUE
  )

  # Build CLI arguments (omit defaults where possible; always pass port & algorithm)
  arguments <- character()

  # Standalone flags
  if (isTRUE(version)) {
    arguments <- c(arguments, "-v")
  }
  if (isTRUE(help)) {
    arguments <- c(arguments, "-h")
  }

  # Verbosity
  verbosity <- match.arg(verbosity)
  if (verbosity != "INFO") {
    arguments <- c(arguments, "-l", verbosity)
  }

  # Trial
  if (!identical(trial, FALSE)) {
    val <- if (is.logical(trial) && trial) 1L else as.integer(trial)
    arguments <- c(arguments, "--trial", as.character(val))
  }

  # IP (only if not default)
  if (!identical(ip, "0.0.0.0")) {
    arguments <- c(arguments, "-i", ip)
  }

  # Always pass port
  arguments <- c(arguments, "-p", as.character(as.integer(port)))

  # Threads (only if not default)
  if (!identical(as.integer(threads), 8L)) {
    arguments <- c(arguments, "-t", as.character(as.integer(threads)))
  }

  # Other flags
  if (isTRUE(shared_memory)) {
    arguments <- c(arguments, "--shared-memory")
  }
  if (!is.null(memory_file)) {
    arguments <- c(arguments, "--memory_file", memory_file)
  }
  if (isTRUE(mmap)) {
    arguments <- c(arguments, "-m")
  }
  if (!is.null(dataset_name)) {
    arguments <- c(arguments, "--dataset-name", dataset_name)
  }

  # Always pass algorithm
  arguments <- c(arguments, "-a", algorithm)

  # Size limits (only if not defaults)
  if (!identical(as.integer(max_viaroute_size), 500L)) {
    arguments <- c(
      arguments,
      "--max-viaroute-size",
      as.character(as.integer(max_viaroute_size))
    )
  }
  if (!identical(as.integer(max_trip_size), 100L)) {
    arguments <- c(
      arguments,
      "--max-trip-size",
      as.character(as.integer(max_trip_size))
    )
  }
  if (!identical(as.integer(max_table_size), 100L)) {
    arguments <- c(
      arguments,
      "--max-table-size",
      as.character(as.integer(max_table_size))
    )
  }
  if (!identical(as.integer(max_matching_size), 100L)) {
    arguments <- c(
      arguments,
      "--max-matching-size",
      as.character(as.integer(max_matching_size))
    )
  }
  if (!identical(as.integer(max_nearest_size), 100L)) {
    arguments <- c(
      arguments,
      "--max-nearest-size",
      as.character(as.integer(max_nearest_size))
    )
  }
  if (!identical(as.integer(max_alternatives), 3L)) {
    arguments <- c(
      arguments,
      "--max-alternatives",
      as.character(as.integer(max_alternatives))
    )
  }
  if (!identical(as.integer(max_matching_radius), -1L)) {
    arguments <- c(
      arguments,
      "--max-matching-radius",
      as.character(as.integer(max_matching_radius))
    )
  }

  # Finally, add the graph prefix
  arguments <- c(arguments, prefix)

  # --- PORT CONFLICT CHECK ---
  # Check if the port is already used by a known OSRM server (local or external).
  # This prevents confusing failures where osrm-routed exits immediately.
  known_servers <- osrm_servers(include_all = TRUE)
  if (nrow(known_servers) > 0) {
    conflict <- known_servers[known_servers$port == as.integer(port) & known_servers$alive, ]
    if (nrow(conflict) > 0) {
      stop(sprintf(
        "Port %d is already in use by another OSRM server (pid %d).",
        as.integer(port),
        conflict$pid[1]
      ), call. = FALSE)
    }
  }

  # --- LOGGING CONFIGURATION ---
  # We prefer a temp file to prevent pipe deadlocks while keeping debug info.

  log_opt <- getOption("osrm.server.log_file")

  stdout_dest <- NULL
  stderr_dest <- NULL
  log_file_path <- NULL

  # Helper to validate path and create directory if needed
  prepare_log_path <- function(path) {
    if (
      is.null(path) ||
        !is.character(path) ||
        length(path) != 1 ||
        !nzchar(path) ||
        is.na(path)
    ) {
      return(NULL)
    }
    abs_path <- normalizePath(path, mustWork = FALSE)
    log_dir <- dirname(abs_path)
    if (!dir.exists(log_dir)) {
      dir.create(log_dir, recursive = TRUE, showWarnings = FALSE)
    }
    abs_path
  }

  # Helper to read only the last n lines from a file (memory-efficient)
  read_last_n_lines <- function(file_path, n = 100) {
    tryCatch(
      {
        # Get file size
        file_size <- file.size(file_path)
        if (is.na(file_size) || file_size == 0) {
          return(character())
        }

        # Estimate bytes to read: assume average 100 bytes per line
        bytes_per_line <- 100
        bytes_to_read <- min(file_size, n * bytes_per_line * 2)

        # Open file and seek to position
        con <- suppressWarnings(file(file_path, "rb"))
        on.exit(close(con), add = TRUE)

        # Seek to estimated position (or beginning if file is small)
        seek(con, max(0, file_size - bytes_to_read))

        # Read remaining content
        raw_content <- readBin(con, "raw", n = bytes_to_read)
        content <- rawToChar(raw_content)

        # Split into lines and return last n
        lines <- strsplit(content, "\r?\n")[[1]]

        # Remove potential partial first line (incomplete due to seeking)
        if (file_size > bytes_to_read && length(lines) > 1) {
          lines <- lines[-1]
        }

        # Return last n lines (or all if fewer)
        utils::tail(lines, n)
      },
      error = function(e) {
        # Return empty on any error (permissions, encoding, file deleted, etc.)
        character()
      }
    )
  }

  if (isTRUE(verbose)) {
    # Direct to console (developer mode)
    # Note: This CAN cause deadlocks in tight loops if R is busy!
    stdout_dest <- ""
    stderr_dest <- ""
    # No log file is used in verbose mode; use sentinel for clarity.
    log_file_path <- NA_character_
  } else if (!is.null(log_opt) && is.character(log_opt)) {
    # User override via options - only character path supported
    log_path <- prepare_log_path(log_opt)
    if (!is.null(log_path)) {
      log_file_path <- log_path
      stdout_dest <- log_path
      stderr_dest <- log_path
      if (!quiet) {
        message("Redirecting server stdout and stderr to: ", log_path)
      }
    } else {
      # Invalid character path - fall back to temp file
      log_file_path <- tempfile(pattern = "osrm_", fileext = ".log")
      stdout_dest <- log_file_path
      stderr_dest <- log_file_path
    }
  } else if (is.list(log_opt)) {
    # List option is deprecated - silently fall back to temp file
    # (was causing deadlocks with incomplete configuration)
    log_file_path <- tempfile(pattern = "osrm_", fileext = ".log")
    stdout_dest <- log_file_path
    stderr_dest <- log_file_path
  } else {
    # DEFAULT: Write to a temp file.
    # This prevents deadlocks and allows post-mortem debugging.
    log_file_path <- tempfile(pattern = "osrm_", fileext = ".log")
    stdout_dest <- log_file_path
    stderr_dest <- log_file_path
  }

  # Resolve osrm-routed executable
  osrm_exec <- resolve_osrm_bin("osrm-routed")
  if (!nzchar(Sys.which(osrm_exec)) && !file.exists(osrm_exec)) {
    stop(
      "Cannot find '",
      osrm_exec,
      "' on PATH. Install OSRM or set options(osrm.routed.exec = '/path/to/osrm-routed').",
      call. = FALSE
    )
  }

  # Echo & launch
  show_echo_cmd <- !quiet && isTRUE(echo_cmd)

  if (show_echo_cmd) {
    message(osrm_exec, " ", paste(shQuote(arguments), collapse = " "))
  }

  osrm_server <- processx::process$new(
    osrm_exec,
    args = arguments,
    echo_cmd = show_echo_cmd,
    stdout = stdout_dest,
    stderr = stderr_dest
  )

  # --- HEALTH CHECK & ERROR REPORTING ---

  # Poll for startup failures with multiple checks over a grace period.
  # Most failures (bad port, corrupted graph) happen within milliseconds,
  # but we check multiple times to be robust across different system loads.
  max_checks <- 10
  for (i in seq_len(max_checks)) {
    Sys.sleep(0.1)
    # Early exit if process already failed
    if (!isTRUE(try(osrm_server$is_alive(), silent = TRUE))) {
      break
    }
  }

  alive_now <- FALSE
  err_msg <- NULL

  try(
    {
      alive_now <- osrm_server$is_alive()
      if (!alive_now) {
        status <- try(osrm_server$get_exit_status(), silent = TRUE)
        status <- if (inherits(status, "try-error")) NA_integer_ else status

        # Retrieve logs to explain WHY it failed
        log_content <- character()

        if (!is.null(log_file_path) && file.exists(log_file_path)) {
          # Read only the last ~100 lines to avoid loading large files
          log_content <- read_last_n_lines(log_file_path, n = 100)
        } else if (identical(stdout_dest, "|")) {
          # Fallback for pipes (unlikely with new default)
          log_content <- c(
            tryCatch(osrm_server$read_output_lines(), error = function(e) {
              character()
            }),
            tryCatch(osrm_server$read_error_lines(), error = function(e) {
              character()
            })
          )
        }

        # Format the error message
        err_msg <- paste0(
          "osrm-routed failed to start (exit status: ",
          status,
          ").\n"
        )

        if (length(log_content) > 0) {
          # Show the last 10 lines of the error log
          last_lines <- utils::tail(log_content, 10)
          err_msg <- paste0(
            err_msg,
            "Last 10 log lines:\n",
            paste(last_lines, collapse = "\n")
          )
          if (any(grepl("incompatible with this version of OSRM", last_lines, ignore.case = TRUE))) {
            err_msg <- paste0(
              err_msg,
              "\n\nHint: The OSRM graph files are incompatible with the current OSRM binary version. ",
              "Please rebuild the graph. If using `osrm_start()`, try adding `force_rebuild = TRUE`."
            )
          }
        } else {
          if (isTRUE(verbose) && is.null(log_file_path)) {
            err_msg <- paste0(
              err_msg,
              "Logs were sent to the R console and cannot be recovered. ",
              "Set a log file path to persist logs for debugging."
            )
          } else {
            err_msg <- paste0(
              err_msg,
              "No logs captured (check if 'verbose' or log file options are set)."
            )
          }
        }
      }
    },
    silent = TRUE
  )

  if (!isTRUE(alive_now)) {
    # Ensure we don't leak a defunct process object
    try(osrm_server$kill(), silent = TRUE)
    if (!is.null(err_msg)) {
      stop(err_msg, call. = FALSE)
    }
    stop("osrm-routed failed to start.", call. = FALSE)
  }

  # --- Register the process for later management (stop by id/port/pid across session) ---
  
  # Try to detect profile from metadata
  profile_detected <- NULL
  meta_path <- file.path(dirname(osrm_path), "dataset.meta.json")
  if (file.exists(meta_path)) {
    meta <- tryCatch(jsonlite::read_json(meta_path), error = function(e) NULL)
    if (!is.null(meta$profile)) profile_detected <- meta$profile
  }

  # Calculate center from input OSM PBF to store in registry/metadata
  center_calculated <- NULL
  if (!is.null(input_osm) && file.exists(input_osm)) {
    ext <- tryCatch(.get_pbf_extent(input_osm), error = function(e) NULL)
    if (!is.null(ext)) center_calculated <- ext$center
  }

  # Best-effort: ignore errors if registry is unavailable for any reason.
  try(
    .osrm_register(
      osrm_server,
      port = port,
      prefix = prefix,
      algorithm = algorithm,
      log = log_file_path,
      input_osm = input_osm,
      profile = profile_detected,
      center = center_calculated
    ),
    silent = TRUE
  )

  # Assign custom class and metadata for better UX in this session
  class(osrm_server) <- c("osrm_server", class(osrm_server))
  attr(osrm_server, "osrm_metadata") <- list(
    port = port,
    profile = profile_detected %||% getOption("osrm.profile", "car"),
    algorithm = algorithm %||% (if (grepl("mldgr$", osrm_path)) "MLD" else "CH"),
    path = osrm_path,
    log = log_file_path,
    center = center_calculated
  )

  # Attach log path as attribute for backward compatibility
  if (!is.null(log_file_path)) {
    attr(osrm_server, "log_path") <- log_file_path
  }

  osrm_server
}

test_that("osrm_customize runs osrm-customize with expected arguments", {
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  input_osrm <- file.path(
    tmp_dir,
    paste0("osrm-customize-test-", Sys.getpid(), ".osrm.partition")
  )
  on.exit(unlink(input_osrm), add = TRUE)
  file.create(input_osrm)

  captured <- list()
  mock_run <- function(command, args, echo, spinner, echo_cmd, ...) {
    captured <<- list(
      command = command,
      args = args,
      echo = echo,
      spinner = spinner,
      echo_cmd = echo_cmd
    )
    # Create the expected output file
    mldgr_file <- sub("\\.partition$", ".mldgr", input_osrm)
    file.create(mldgr_file)
    list(status = 0, stdout = "", stderr = "")
  }

  result <- with_mocked_bindings(
    osrm_customize(
      input_osrm = input_osrm,
      threads = 2L,
      verbosity = "WARNING",
      segment_speed_file = "speeds.csv",
      turn_penalty_file = "penalties.csv",
      edge_weight_updates_over_factor = 2.0,
      parse_conditionals_from_now = 987654321,
      time_zone_file = "zones.geojson",
      quiet = TRUE,
      verbose = FALSE,
      spinner = TRUE,
      echo_cmd = FALSE
    ),
    run = mock_run,
    .package = "processx"
  )

  expect_equal(tools::file_path_sans_ext(basename(captured$command)), "osrm-customize")
  expect_true("-t" %in% captured$args && "2" %in% captured$args)
  expect_true("-l" %in% captured$args && "WARNING" %in% captured$args)
  expect_true(
    "--segment-speed-file" %in% captured$args && "speeds.csv" %in% captured$args
  )
  expect_true(
    "--turn-penalty-file" %in%
      captured$args &&
      "penalties.csv" %in% captured$args
  )
  expect_true(
    "--edge-weight-updates-over-factor" %in%
      captured$args &&
      "2" %in% captured$args
  )
  expect_true(
    "--parse-conditionals-from-now" %in%
      captured$args &&
      "987654321" %in% captured$args
  )
  expect_true(
    "--time-zone-file" %in% captured$args && "zones.geojson" %in% captured$args
  )

  # Check processx options (quiet=TRUE overrides verbose/spinner/echo_cmd)
  expect_false(captured$echo)
  expect_false(captured$spinner)
  expect_false(captured$echo_cmd)
})

test_that("osrm_customize handles missing input file", {
  expect_error(
    osrm_customize("nonexistent.osrm.partition"),
    "File does not exist"
  )
})

test_that("osrm_customize checks for output file", {
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  input_osrm <- file.path(
    tmp_dir,
    paste0("osrm-fail-", Sys.getpid(), ".osrm.partition")
  )
  on.exit(unlink(input_osrm), add = TRUE)
  file.create(input_osrm)

  mock_run <- function(...) {
    list(status = 0, stdout = "", stderr = "")
  }

  expect_error(
    with_mocked_bindings(
      osrm_customize(input_osrm),
      run = mock_run,
      .package = "processx"
    ),
    "Customization did not produce MLD graph file"
  )
})

test_that("osrm_customize accepts a directory with one .osrm.partition file", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-customize-dir-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  input_osrm <- file.path(tmp_dir, "test.osrm.partition")
  file.create(input_osrm)

  normalized_input <- normalizePath(input_osrm)
  expected_mldgr <- sub("\\.partition$", ".mldgr", normalized_input)

  mock_run <- function(command, args, echo, spinner, echo_cmd, ...) {
    file.create(expected_mldgr)
    list(status = 0, stdout = "", stderr = "")
  }

  result <- with_mocked_bindings(
    osrm_customize(
      input_osrm = tmp_dir,
      quiet = TRUE
    ),
    run = mock_run,
    .package = "processx"
  )

  expect_equal(result$osrm_job_artifact, expected_mldgr)
})

test_that("osrm_customize errors when directory has no .osrm.partition files", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-customize-empty-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  expect_error(
    osrm_customize(input_osrm = tmp_dir),
    "No .osrm.partition files found"
  )
})

test_that("osrm_customize errors when directory has multiple .osrm.partition files", {
  skip_if_not_installed("processx")

  tmp_dir <- file.path(tempdir(), paste0("osrm-customize-multi-", Sys.getpid()))
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  file.create(file.path(tmp_dir, "test1.osrm.partition"))
  file.create(file.path(tmp_dir, "test2.osrm.partition"))

  expect_error(
    osrm_customize(input_osrm = tmp_dir),
    "Multiple .osrm.partition files found"
  )
})

test_that("osrm_customize gives helpful error when used after osrm_extract without partition", {
  skip_if_not_installed("processx")

  tmp_dir <- tempdir()
  timestamp_file <- file.path(
    tmp_dir,
    paste0("osrm-extract-test-", Sys.getpid(), ".osrm.timestamp")
  )
  on.exit(unlink(timestamp_file), add = TRUE)
  file.create(timestamp_file)

  # Create a mock osrm_job object with a timestamp file (simulating osrm_extract output)
  mock_job <- structure(
    list(
      osrm_job_artifact = timestamp_file,
      osrm_working_dir = tmp_dir,
      logs = list()
    ),
    class = "osrm_job"
  )

  expect_error(
    osrm_customize(input_osrm = mock_job),
    "requires a partitioned graph"
  )
  expect_error(
    osrm_customize(input_osrm = mock_job),
    "MLD pipeline"
  )
  expect_error(
    osrm_customize(input_osrm = mock_job),
    "osrm_partition"
  )
})

---
title: "Quick Start: Install OSRM and calculate routes"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Quick Start: Install OSRM and calculate routes}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
knitr::opts_chunk$set(eval = FALSE)
library(osrm.backend)
library(osrm)
library(sf)
```


# Create a temporary directory and copy example OSM PBF file into it

```{r}
#| label: create-temp-dir
osrm_temp_dir <- file.path(tempdir(), paste0("osrm-data-", Sys.getpid()))
dir.create(osrm_temp_dir, showWarnings = FALSE, recursive = TRUE)
internal_pbf <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
file.copy(from = internal_pbf, to = osrm_temp_dir)
```

# Start local OSRM server

Start the server with `osrm_start()` function pointing to the temporary directory created above. It will automatically check if OSRM Backend binaries are installed, and if not, it will install the latest version, it will also automatically prepare the routing graph from the OSM PBF file if not already done and start the server. If any of the steps are unnecessary, it will skip them, so running this funciton will not lead to re-downloading of OSRM binaries or re-processing of the OSM PBF file if already done.

```{r}
#| label: start-osrm-server
osrm_start(osrm_temp_dir, verbose = TRUE)
```

> **Tip:** You can check if server is running with `osrm_servers()`:
> ```
> # A tibble: 1 × 7
>   id                                   pid  port algorithm started_at          alive has_handle
>   <chr>                              <int> <int> <chr>     <dttm>              <lgl> <lgl>
> 1 osrm-28034-5001-20251103122904.133 28034  5001 MLD       2025-11-03 00:00:00 TRUE  TRUE
> ```
>
> You can kill the server with `osrm_stop()` or `osrm_stop_all()` if you have started several servers.

# Load spatial data

Load some spatial data from the OSM PBF file to sample points for routing and plot the street network:

```{r}
# read the pbf file with sf
osm_points <- read_sf(internal_pbf, layer = "points")
osm_streets <- read_sf(internal_pbf, layer = "multilinestrings")

set.seed(100)
sampled_points <- osm_points[sample(nrow(osm_points), 2), ]
```

# Calculate route between sampled points

To use `{osrm}` package with the local server, you have to set the `osrm.server` option to point to the local server URL. The default port that `osrm_start()` uses is `5001`, so the URL will be `http://localhost:5001/`. Then you can use any `{osrm}` function as usual, here we calculate a route between the two sampled points:

```{r}
#| label: osrm-route
options("osrm.server" = "http://localhost:5001/")
route <- osrm::osrmRoute(loc = sampled_points)
route
```


```
Simple feature collection with 1 feature and 4 fields
Geometry type: LINESTRING
Dimension:     XY
Bounding box:  xmin: -49.27156 ymin: -25.43082 xmax: -49.25396 ymax: -25.42358
Geodetic CRS:  WGS 84
          src dst duration distance                       geometry
3786_503 3786 503     5.55   2.6324 LINESTRING (-49.25534 -25.4...
```

# Plot the route on top of the street network

```{r}
plot(st_geometry(osm_streets), lwd = 0.5, col = "grey70")
plot(st_geometry(route), col = "red", lwd = 5, add = TRUE)
```

```{r}
#| echo: false
# in this block create same plot as above but save it to a png file
dir.create("./man/figures", showWarnings = FALSE, recursive = TRUE)
png("./man/figures/example-route.png", width = 800, height = 600)
plot(st_geometry(osm_streets), lwd = 0.5, col = "grey70")
plot(st_geometry(route), col = "red", lwd = 5, add = TRUE)
dev.off()
```

![](../man/figures/example-route.png)

# Stop local OSRM server and cleanup

You can now stop the server:

```{r}
#| label: stop-osrm-server
osrm_stop()
```

```
Stopped OSRM server id=osrm-29557-5001-20251103124028.405 (pid 29557, port 5001).
```

And remove the temporary directory:

```{r}
#| label: remove-temp-dir
unlink(osrm_temp_dir, recursive = TRUE)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm.backend-package.R
\docType{package}
\name{osrm.backend-package}
\alias{osrm.backend}
\alias{osrm.backend-package}
\title{osrm.backend: Bindings for 'Open Source Routing Machine'}
\description{
Install and control 'Open Source Routing Machine' ('OSRM') backend executables to prepare routing data and run/stop a local 'OSRM' server. For computations with the running server use the 'osrm' R package (\url{https://cran.r-project.org/package=osrm}).
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/e-kotov/osrm.backend}
  \item \url{https://www.ekotov.pro/osrm.backend/}
  \item Report bugs at \url{https://github.com/e-kotov/osrm.backend/issues}
}

}
\author{
\strong{Maintainer}: Egor Kotov \email{kotov.egor@gmail.com} (\href{https://orcid.org/0000-0001-6690-5345}{ORCID}) [copyright holder]

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_which.R
\name{osrm_which}
\alias{osrm_which}
\title{Locate the OSRM Installation Used by \code{osrm.backend}}
\usage{
osrm_which(quiet = FALSE)
}
\arguments{
\item{quiet}{Logical; if \code{FALSE} (default), prints information about the located installation. If \code{TRUE}, suppresses printed output and only returns the information as a list.}
}
\value{
A list with components \code{executable} (full path to
\code{osrm-routed}), \code{directory} (its parent folder), \code{osrm_version} (character
vector of non-empty lines emitted by \code{osrm-routed --version}), and the raw
\code{processx::run} result in \code{logs}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Resolves the \code{osrm-routed} executable available on the current \code{PATH}
(or the override provided via \code{options(osrm.routed.exec)}). Runs
\code{osrm-routed --version} to verify availability, then prints the directory
containing the executable together with the backend version reported by
\code{osrm-routed} so you know what will be used in the current session.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )

  # check which OSRM installation will be used
  osrm_which()

  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_gui.R
\name{osrm_gui}
\alias{osrm_gui}
\title{Launch a GUI to View and Debug OSRM Routing}
\usage{
osrm_gui(
  input_osrm = NULL,
  port = "auto",
  style = "https://basemaps.cartocdn.com/gl/voyager-gl-style/style.json",
  center = NULL,
  zoom = NULL,
  autozoom = TRUE,
  update_while_drag = FALSE,
  debug = FALSE
)
}
\arguments{
\item{input_osrm}{Optional. Can be:
\itemize{
\item An OSRM job process (an \code{osrm_server} object inheriting from
\code{processx::process}) returned by \code{osrm_start()} or \code{osrm_start_server()}.
When providing a process, you must also specify \code{port} explicitly.
\item A path string to an \code{.osrm.hsgr} or \code{.osrm.mldgr} file.
\item A path string to an \code{.osm.pbf} file (will be prepared and started).
\item \code{NULL} (default): Auto-detects a running OSRM server using \code{\link[=osrm_servers]{osrm_servers()}}.
Errors if no servers are running.
}}

\item{port}{Integer or \code{"auto"}. The port the server is running on (or should run on).
Defaults to \code{"auto"}, which attempts to auto-detect a running OSRM server using
\code{\link[=osrm_servers]{osrm_servers()}}. If multiple servers are running, the most recent one is selected
with a warning. If no servers are running, an error is raised.}

\item{style}{Character. Map style for \code{mapgl}. Defaults to "https://basemaps.cartocdn.com/gl/voyager-gl-style/style.json".}

\item{center}{Numeric vector of length 2 (\code{c(lng, lat)}), or named list
(\code{list(lng = ..., lat = ...)}), or \code{NULL} (default). Initial map center.
If \code{NULL} and \code{input_osrm} is a \code{.osm.pbf} file, attempts to auto-center
on the PBF extent. Priority is given to a fast pure R header parser and
\verb{osmium fileinfo} (fast); otherwise estimates the extent by sampling a
small number of features via \code{sf::st_read()} (for example, reading with
a \verb{LIMIT 10} query).}

\item{zoom}{Numeric. Initial zoom level. If \code{NULL} (default) and center is
auto-detected from PBF, defaults to 9. Otherwise uses map default.}

\item{autozoom}{Logical. Whether to enable auto-zoom by default. Defaults to \code{TRUE}.}

\item{update_while_drag}{Logical. Whether to enable live tracking mode by default (updates route while dragging). Defaults to \code{FALSE}.}

\item{debug}{Logical. Whether to enable debug mode (prints OSRM requests to console). Defaults to \code{FALSE}.}
}
\value{
No return value; launches a Shiny Gadget.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Launches a lightweight Shiny application to interactively visualize routing
on a local OSRM server. This interface mimics the \code{r5rgui} experience,
supporting left-click for start, right-click for end, and draggable markers.
}
\details{
The function checks for optional dependencies \code{shiny}, \code{mapgl}, \code{osrm}, \code{sf}, and \code{DT}.
If missing, it prompts the user to install them.

It attempts to detect an active OSRM server. If an OSRM job process (from
\code{osrm_start()}) is passed, it uses that configuration. If a path is passed,
it will start a temporary server for the session.
}
\examples{
if (interactive()) {
  # 1. Auto-detect running server (errors if none running):
  osrm_gui()

  # 2. Connect to specific port:
  # osrm_gui(port = 5001)

  # 3. Start from a graph file (auto-center on PBF):
  # osrm_gui("berlin.osrm.mldgr")

  # 4. Start from PBF with auto-center:
  # osrm_gui("berlin.osm.pbf")

  # 5. Explicit center and zoom:
  # osrm_gui(port = 5001, center = c(13.4, 52.5), zoom = 12)

  # 6. Use an existing process (must specify port):
  # srv <- osrm_start("graph.osrm.mldgr", port = 6000)
  # osrm_gui(srv, port = 6000)

  # 7. Enable debug mode:
  # osrm_gui(debug = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_start_server.R
\name{osrm_start_server}
\alias{osrm_start_server}
\title{Start an OSRM MLD/CH server with \code{osrm-routed}}
\usage{
osrm_start_server(
  osrm_path,
  version = FALSE,
  help = FALSE,
  verbosity = c("INFO", "ERROR", "WARNING", "NONE", "DEBUG"),
  trial = FALSE,
  ip = "0.0.0.0",
  port = 5001L,
  threads = 8L,
  shared_memory = FALSE,
  memory_file = NULL,
  mmap = FALSE,
  dataset_name = NULL,
  algorithm = NULL,
  max_viaroute_size = 500L,
  max_trip_size = 100L,
  max_table_size = 100L,
  max_matching_size = 100L,
  max_nearest_size = 100L,
  max_alternatives = 3L,
  max_matching_radius = -1L,
  quiet = FALSE,
  verbose = FALSE,
  echo_cmd = FALSE,
  input_osm = NULL
)
}
\arguments{
\item{osrm_path}{Character(1). Path to the \code{.osrm.mldgr} or \code{.osrm.hsgr} file}

\item{version}{Logical; if \code{TRUE}, prints version and exits}

\item{help}{Logical; if \code{TRUE}, prints help and exits}

\item{verbosity}{Character; one of \verb{"NONE","ERROR","WARNING","INFO","DEBUG"}}

\item{trial}{Logical or integer; if \code{TRUE} or >0, quits after initialization (default: \code{FALSE})}

\item{ip}{Character; IP address to bind (default: \code{"0.0.0.0"})}

\item{port}{Integer; TCP port to listen on (default: \code{5001}). The function checks
if this port is already in use by another running OSRM server (even from
another session) and will stop with an error if a conflict is detected.}

\item{threads}{Integer; number of worker threads (default: \code{8})}

\item{shared_memory}{Logical; load graph from shared memory (default: \code{FALSE})}

\item{memory_file}{Character or NULL; DEPRECATED (behaves like \code{mmap})}

\item{mmap}{Logical; memory-map data files (default: \code{FALSE})}

\item{dataset_name}{Character or NULL; name of shared memory dataset}

\item{algorithm}{Character or NULL; one of \code{"MLD"}, \code{"CH"}, or \code{"CoreCH"} (case-insensitive).
If \code{NULL} (default), auto-selected based on file extension.}

\item{max_viaroute_size}{Integer (default: \code{500})}

\item{max_trip_size}{Integer (default: \code{100})}

\item{max_table_size}{Integer (default: \code{100})}

\item{max_matching_size}{Integer (default: \code{100})}

\item{max_nearest_size}{Integer (default: \code{100})}

\item{max_alternatives}{Integer (default: \code{3})}

\item{max_matching_radius}{Integer; use \code{-1} for unlimited (default: \code{-1})}

\item{quiet}{Logical; when \code{TRUE}, suppresses package messages.}

\item{verbose}{Logical; when \code{TRUE}, routes server stdout and stderr to the R
console for live debugging. Note: This can cause deadlocks in tight loops
if R is busy. Defaults to \code{FALSE}, which writes logs to a temporary file.}

\item{echo_cmd}{Logical; echo command line to console before launch (default: \code{FALSE})}

\item{input_osm}{Character or NULL; path to the original OSM input file (for tracking purposes).
This parameter is typically used internally by \code{\link[=osrm_start]{osrm_start()}} to record the source data.}
}
\value{
An OSRM job process (an \code{osrm_server} object inheriting from
\code{processx::process}) for the running server (also registered internally).
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Launches an \code{osrm-routed} process pointing at a localized OSRM graph (either
\code{.osrm.mldgr} for MLD or \code{.osrm.hsgr} for CH/CoreCH).
}
\details{
The server's standard output and error streams are handled via temporary files
by default to prevent deadlocks in R's single-threaded environment. This ensures
reliable operation while preserving logs for debugging startup failures.

To customize logging behavior, you can use the following approaches:
\itemize{
\item \strong{Default (Temp File):} Logs are written to a temporary file. This prevents
deadlocks while keeping logs available for debugging.

\item \strong{Verbose Mode:} Set \code{verbose = TRUE} to display logs directly in the
R console. Note: This can cause deadlocks in tight loops if R is busy.

\item \strong{Custom Log File:} Set the \code{osrm.server.log_file} option to redirect
output to a specific file:
\code{options(osrm.server.log_file = "path/to/osrm.log")}

\if{html}{\out{<div class="sourceCode">}}\preformatted{Note: List specifications (e.g., `list(stdout = "...", stderr = "...")`)
are deprecated and will fall back to the default temporary file behavior.
}\if{html}{\out{</div>}}

}

You can override the \code{osrm-routed} executable via
\code{options(osrm.routed.exec = "/full/path/to/osrm-routed")}.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )

  # Build a graph then launch an OSRM server on a custom port
  pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
  dir.create(osrm_dir, recursive = TRUE)
  tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
  file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)

  graph <- osrm_prepare_graph(
    input_osm = tmp_pbf,
    overwrite = TRUE,
    threads = 1L,
    algorithm = "mld"
  )

  server <- osrm_start_server(
    osrm_path = graph$osrm_job_artifact,
    port = 6000,
    threads = 1L
  )

  # Later, stop the server again
  osrm_stop(server)

  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
  unlink(osrm_dir, recursive = TRUE)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_stop_server.R
\name{osrm_servers}
\alias{osrm_servers}
\title{List OSRM servers}
\usage{
osrm_servers(include_all = FALSE, output = c("data.frame", "list"))
}
\arguments{
\item{include_all}{Logical; if \code{TRUE}, scans the system process table for
all \code{osrm-routed} processes, including those not started by this package
in the current session. Default is \code{FALSE}.}

\item{output}{Character string specifying the return format. Either \code{"data.frame"}
(the default) which returns a tabular summary with a custom print method,
or \code{"list"} which returns a detailed list of server metadata objects.}
}
\value{
If \code{output = "data.frame"}, returns a \code{data.frame} (class \code{osrm_server_list})
of OSRM job processes with columns:
\code{id}, \code{pid}, \code{port}, \code{algorithm}, \code{started_at}, \code{alive}, \code{has_handle}, \code{log}, \code{input_osm},
\code{center_lon}, \code{center_lat}. External servers will have \code{id} prefixed with \verb{sys-} and
\code{log} set to \verb{<external>}. If \code{output = "list"}, returns a named list of server metadata.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Lists \code{osrm-routed} processes. By default, it returns a snapshot of servers
started by the current R session (registered via \code{\link[=osrm_start_server]{osrm_start_server()}} or \code{\link[=osrm_start]{osrm_start()}}).
You can optionally list all \code{osrm-routed} processes running on the system,
including those started by other sessions or manually.

You can stop a server by passing its \code{id}, \code{port}, or \code{pid} to \code{\link[=osrm_stop]{osrm_stop()}}.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )

  # copy example OSM PBF into a temporary workspace to avoid polluting pkg data
  pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
  dir.create(osrm_dir, recursive = TRUE)
  tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
  file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
  graph <- osrm_prepare_graph(tmp_pbf, overwrite = TRUE, threads = 1L)

  srv <- osrm_start_server(graph$osrm_job_artifact, port = 6000, threads = 1L)
  osrm_servers()
  osrm_stop(srv)

  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
  unlink(osrm_dir, recursive = TRUE)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_extract.R
\name{osrm_extract}
\alias{osrm_extract}
\title{Extract OSM into OSRM Graph Files}
\usage{
osrm_extract(
  input_osm,
  profile = osrm_find_profile("car.lua"),
  threads = 8L,
  overwrite = FALSE,
  verbosity = c("INFO", "NONE", "ERROR", "WARNING", "DEBUG"),
  data_version = NULL,
  small_component_size = 1000L,
  with_osm_metadata = FALSE,
  parse_conditional_restrictions = FALSE,
  location_dependent_data = NULL,
  disable_location_cache = FALSE,
  dump_nbg_graph = FALSE,
  quiet = FALSE,
  verbose = FALSE,
  spinner = TRUE,
  echo_cmd = FALSE
)
}
\arguments{
\item{input_osm}{A string. Path to the input OSM file
(\code{.osm}, \code{.osm.bz2}, or \code{.osm.pbf}) or a directory containing exactly one
OSM file with a supported extension.}

\item{profile}{A string. Path to the OSRM Lua profile
(e.g. returned by \code{osrm_find_profile("car.lua")}).}

\item{threads}{An integer. Number of threads for
\code{-t/--threads}; default \code{8} (OSRM's default).}

\item{overwrite}{A logical. If \code{FALSE} (default), stops when any
existing \verb{.osrm*} files matching the base name are found alongside
\code{input_osm}. Set to \code{TRUE} to proceed regardless.}

\item{verbosity}{A string. Log verbosity level passed to
\code{-l/--verbosity} (one of \code{"NONE","ERROR","WARNING","INFO","DEBUG"}),
default \code{"INFO"}.}

\item{data_version}{A string or \code{NULL}. Passed to
\code{-d/--data_version}; default \code{NULL}, in which case the option
is omitted.}

\item{small_component_size}{An integer. For
\code{--small-component-size}; default \code{1000} (OSRM's default).}

\item{with_osm_metadata}{A logical. If \code{TRUE}, adds
\code{--with-osm-metadata}; default \code{FALSE}.}

\item{parse_conditional_restrictions}{A logical. If \code{TRUE}, adds
\code{--parse-conditional-restrictions}; default \code{FALSE}.}

\item{location_dependent_data}{A string or \code{NULL}. Path to GeoJSON,
passed to \code{--location-dependent-data}; default \code{NULL}, in which
case the option is omitted.}

\item{disable_location_cache}{A logical. If \code{TRUE}, adds
\code{--disable-location-cache}; default \code{FALSE}.}

\item{dump_nbg_graph}{A logical. If \code{TRUE}, adds
\code{--dump-nbg-graph}; default \code{FALSE}.}

\item{quiet}{A logical. Master switch that suppresses package messages and
process output when \code{TRUE}; default \code{FALSE}.}

\item{verbose}{A logical. When \code{TRUE} and \code{quiet = FALSE}, streams stdout and
stderr from the underlying \code{processx::run} calls.}

\item{spinner}{A logical. When \code{TRUE} and \code{quiet = FALSE}, shows a spinner
instead of live logs; default \code{TRUE}.}

\item{echo_cmd}{A logical. When \code{TRUE} and \code{quiet = FALSE}, prints each
command before running; default \code{FALSE}.}
}
\value{
An object of class \code{osrm_job} with the following elements:
\describe{
\item{osrm_job_artifact}{The path to the generated \code{.osrm.timestamp} file.}
\item{osrm_working_dir}{The directory containing all OSRM files.}
\item{logs}{The \code{processx::run} result object.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Run the \code{osrm-extract} tool to preprocess an OSM file
(\code{.osm}, \code{.osm.bz2}, or \code{.osm.pbf}) into the base \code{.osrm} graph files
using a specified Lua profile.  After running, a companion
\verb{<base>.osrm.timestamp} file must exist to confirm success.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  # Install OSRM (temporary, session PATH)
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )

  # copy example OSM PBF into a temporary workspace to avoid polluting pkg data
  pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
  dir.create(osrm_dir, recursive = TRUE)
  tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
  file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)

  # Find the path to the profile first
  car_profile <- osrm_find_profile("car.lua")

  # extract OSRM graph files
  result <- osrm_extract(
    input_osm                  = tmp_pbf,
    profile                    = car_profile,
    overwrite                  = TRUE,
    threads                    = 1L
  )
  # path to generated .osrm files (specifically, the .osrm.timestamp file)
  result$osrm_job_artifact

  # Clean up binaries and workspace
  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
  unlink(osrm_dir, recursive = TRUE)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pbf_header_reader.R
\name{read_pbf_header_bbox}
\alias{read_pbf_header_bbox}
\title{Read Bounding Box from OSM PBF Header}
\usage{
read_pbf_header_bbox(pbf_file)
}
\arguments{
\item{pbf_file}{Path to an .osm.pbf file}
}
\value{
Named list with \code{center} (c(lng, lat)) and \code{bbox} (c(lng_min, lat_min, lng_max, lat_max))
or NULL if bbox not found or file invalid
}
\description{
Extracts the bbox from the HeaderBlock of an OSM PBF file using pure R.
This reads only the first header blob (OSMHeader), not the data blocks.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_install.R
\name{osrm_clear_path}
\alias{osrm_clear_path}
\title{Clear OSRM Path from Project's .Rprofile}
\usage{
osrm_clear_path(quiet = FALSE)
}
\arguments{
\item{quiet}{A logical value. If \code{TRUE}, suppresses messages. Defaults to \code{FALSE}.}
}
\value{
\code{TRUE} if the file was modified, \code{FALSE} otherwise.
}
\description{
Scans the \code{.Rprofile} file in the current project's root directory and
removes any lines that were added by \code{osrm_install()} to modify the \code{PATH}.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  # Clean up a temporary project's .Rprofile
  old <- setwd(tempdir())
  on.exit(setwd(old), add = TRUE)
  writeLines(
    c(
      "#added-by-r-pkg-osrm.backend",
      'Sys.setenv(PATH = paste("dummy", Sys.getenv("PATH"), sep = .Platform$path.sep))'
    ),
    ".Rprofile"
  )
  osrm_clear_path(quiet = TRUE)
  unlink(".Rprofile")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_prepare_graph.R
\name{osrm_prepare_graph}
\alias{osrm_prepare_graph}
\title{Prepare OSRM Graph for Routing (Extract + Partition/Contract)}
\usage{
osrm_prepare_graph(
  input_osm,
  profile = osrm_find_profile("car.lua"),
  threads = 8L,
  overwrite = FALSE,
  verbosity = c("INFO", "NONE", "ERROR", "WARNING", "DEBUG"),
  data_version = NULL,
  small_component_size = 1000L,
  with_osm_metadata = FALSE,
  parse_conditional_restrictions = FALSE,
  location_dependent_data = NULL,
  disable_location_cache = FALSE,
  dump_nbg_graph = FALSE,
  algorithm = c("mld", "ch"),
  balance = 1.2,
  boundary = 0.25,
  optimizing_cuts = 10L,
  max_cell_sizes = c(128, 4096, 65536, 2097152),
  quiet = FALSE,
  verbose = FALSE,
  spinner = TRUE,
  echo_cmd = FALSE
)
}
\arguments{
\item{input_osm}{A string. Path to the input OSM file (\code{.osm}, \code{.osm.bz2}, or \code{.osm.pbf}) or a directory containing exactly one OSM file with a supported extension.}

\item{profile}{A string. Path to the OSRM Lua profile (e.g. returned by \code{osrm_find_profile("car.lua")}).}

\item{threads}{An integer. Number of threads for extract and partition/contract; default \code{8}.}

\item{overwrite}{A logical. If \code{FALSE}, stops if any existing \verb{.osrm*} files matching
the base name are found alongside \code{input_osm}. Set to \code{TRUE} to overwrite them.}

\item{verbosity}{A string. Log verbosity for extract/partition/contract
(one of \verb{"NONE","ERROR","WARNING","INFO","DEBUG"}); default \code{"INFO"}.}

\item{data_version}{A string or \code{NULL}. Passed to \code{osrm-extract} via \code{-d}; default \code{NULL}.}

\item{small_component_size}{An integer. For extract & partition; default \code{1000}.}

\item{with_osm_metadata}{A logical. Adds \code{--with-osm-metadata} during extract; default \code{FALSE}.}

\item{parse_conditional_restrictions}{A logical. Adds \code{--parse-conditional-restrictions}; default \code{FALSE}.}

\item{location_dependent_data}{A string or \code{NULL}. Path to GeoJSON for extract; default \code{NULL}.}

\item{disable_location_cache}{A logical. Adds \code{--disable-location-cache}; default \code{FALSE}.}

\item{dump_nbg_graph}{A logical. Adds \code{--dump-nbg-graph}; default \code{FALSE}.}

\item{algorithm}{A string. One of \code{"MLD"} (default) or \code{"CH"} (case-insensitive).}

\item{balance}{A numeric. Balance for \code{osrm-partition}; default \code{1.2}.}

\item{boundary}{A numeric. Boundary percentage for \code{osrm-partition}; default \code{0.25}.}

\item{optimizing_cuts}{An integer. Optimizing cuts for \code{osrm-partition}; default \code{10}.}

\item{max_cell_sizes}{A numeric vector. Max cell sizes for \code{osrm-partition}; default \code{c(128,4096,65536,2097152)}.}

\item{quiet}{A logical. Master switch that suppresses package messages and
process output when \code{TRUE}; default \code{FALSE}.}

\item{verbose}{A logical. When \code{TRUE} and \code{quiet = FALSE}, streams stdout and
stderr from the underlying \code{processx::run} calls.}

\item{spinner}{A logical. When \code{TRUE} and \code{quiet = FALSE}, shows a spinner
instead of live logs; default \code{TRUE}.}

\item{echo_cmd}{A logical. When \code{TRUE} and \code{quiet = FALSE}, prints each
command before running; default \code{FALSE}.}
}
\value{
An object of class \code{osrm_job} with the following elements:
\describe{
\item{osrm_job_artifact}{The path to the final routing-ready graph file (\code{.osrm.hsgr} for CH or \code{.osrm.mldgr} for MLD).}
\item{osrm_working_dir}{The directory containing all OSRM files.}
\item{logs}{A list of \code{processx::run} results for each stage:
\code{extract}, \code{partition}/\code{contract}, and \code{customize} (if MLD).}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

High-level wrapper that first runs \code{osrm-extract} on an OSM file
to produce the base \code{.osrm} graph, then prepares it for routing via
either the MLD pipeline (\code{osrm-partition} + \code{osrm-customize}) or the CH pipeline (\code{osrm-contract}).
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )

  # Prepare a routing-ready graph with the default MLD pipeline
  pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
  dir.create(osrm_dir, recursive = TRUE)
  tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
  file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)

  graph <- osrm_prepare_graph(
    input_osm = tmp_pbf,
    overwrite = TRUE,
    threads = 1L,
    algorithm = "mld"
  )
  graph$osrm_job_artifact

  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
  unlink(osrm_dir, recursive = TRUE)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_stop_server.R
\name{osrm_stop_all}
\alias{osrm_stop_all}
\title{Stop all running OSRM servers started via this package}
\usage{
osrm_stop_all(include_all = FALSE)
}
\arguments{
\item{include_all}{A logical value. If \code{TRUE}, stop all OSRM servers found
running on the system, including those not started by the current R session.
Requires \code{ps} and \code{jsonlite} packages. Defaults to \code{FALSE}.}
}
\value{
The number of servers attempted.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )

  pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
  dir.create(osrm_dir, recursive = TRUE)
  tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
  file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
  graph <- osrm_prepare_graph(tmp_pbf, overwrite = TRUE, threads = 1L)

  srv <- osrm_start_server(graph$osrm_job_artifact, port = 6000, threads = 1L)
  stopped <- osrm_stop_all()
  stopped

  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
  unlink(osrm_dir, recursive = TRUE)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_install.R
\name{osrm_check_available_versions}
\alias{osrm_check_available_versions}
\title{Check for Available OSRM Versions}
\usage{
osrm_check_available_versions(prereleases = FALSE)
}
\arguments{
\item{prereleases}{A logical value. If \code{TRUE}, include pre-release versions
in the returned list. Defaults to \code{FALSE}.}
}
\value{
A character vector of available version tags.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Queries the GitHub API to get a list of all available version tags for the
OSRM backend that have binaries for the current platform.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  # Get all stable versions with binaries for this platform
  osrm_check_available_versions()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_contract.R
\name{osrm_contract}
\alias{osrm_contract}
\title{Contract OSRM Graph for Contraction Hierarchies (CH)}
\usage{
osrm_contract(
  input_osrm,
  threads = 8L,
  verbosity = c("INFO", "NONE", "ERROR", "WARNING", "DEBUG"),
  segment_speed_file = NULL,
  turn_penalty_file = NULL,
  edge_weight_updates_over_factor = 0,
  parse_conditionals_from_now = 0,
  time_zone_file = NULL,
  quiet = FALSE,
  verbose = FALSE,
  spinner = TRUE,
  echo_cmd = FALSE
)
}
\arguments{
\item{input_osrm}{A string. Path to a \code{.osrm.timestamp} file, the base path to the \code{.osrm} files (without extension),
or a directory containing exactly one \code{.osrm.timestamp} file.}

\item{threads}{An integer. Number of threads to use; default \code{8}.}

\item{verbosity}{A string. Log verbosity level passed to \code{-l/--verbosity}
(one of \verb{"NONE","ERROR","WARNING","INFO","DEBUG"}); default \code{"INFO"}.}

\item{segment_speed_file}{A string or \code{NULL}. Path to nodeA,nodeB,speed CSV; default \code{NULL}.}

\item{turn_penalty_file}{A string or \code{NULL}. Path to from_,to_,via_nodes,penalties CSV; default \code{NULL}.}

\item{edge_weight_updates_over_factor}{A numeric. Threshold for logging large weight updates; default \code{0}.}

\item{parse_conditionals_from_now}{A numeric. UTC timestamp for conditional restrictions; default \code{0}.}

\item{time_zone_file}{A string or \code{NULL}. GeoJSON file for time zone boundaries; default \code{NULL}.}

\item{quiet}{A logical. Master switch that suppresses package messages and
process output when \code{TRUE}; default \code{FALSE}.}

\item{verbose}{A logical. When \code{TRUE} and \code{quiet = FALSE}, streams stdout and
stderr from the underlying \code{processx::run} calls.}

\item{spinner}{A logical. When \code{TRUE} and \code{quiet = FALSE}, shows a spinner
instead of live logs; default \code{TRUE}.}

\item{echo_cmd}{A logical. When \code{TRUE} and \code{quiet = FALSE}, prints each
command before running; default \code{FALSE}.}
}
\value{
An object of class \code{osrm_job} with the following elements:
\describe{
\item{osrm_job_artifact}{The path to the contracted \code{.osrm.hsgr} file.}
\item{osrm_working_dir}{The directory containing all OSRM files.}
\item{logs}{The \code{processx::run} result object.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Run the \code{osrm-contract} tool to contract an OSRM graph for the CH pipeline.
After running, a companion \verb{<base>.osrm.hsgr} file must exist to confirm success.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )

  # Prepare a small graph then contract it for the CH pipeline
  pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
  dir.create(osrm_dir, recursive = TRUE)
  tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
  file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
  profile <- osrm_find_profile("car.lua")

  extract_job <- osrm_extract(
    input_osm = tmp_pbf,
    profile = profile,
    overwrite = TRUE,
    threads = 1L
  )

  ch_graph <- osrm_contract(extract_job, threads = 1L, verbose = TRUE)
  ch_graph$osrm_job_artifact

  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
  unlink(osrm_dir, recursive = TRUE)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pbf_header_reader.R
\name{.get_pbf_extent}
\alias{.get_pbf_extent}
\title{Get PBF Bounding Box with Fallbacks}
\usage{
.get_pbf_extent(pbf_file)
}
\arguments{
\item{pbf_file}{Path to an .osm.pbf file}
}
\value{
Named list with \code{center} (c(lng, lat)) and \code{bbox} (c(lng_min, lat_min, lng_max, lat_max))
or NULL if bbox could not be determined
}
\description{
Tries multiple methods to extract the bounding box from an OSM PBF file:
\enumerate{
\item Pure R header parsing (fastest, no dependencies)
\item osmium CLI (if available in PATH)
\item sf sampling (if sf is installed, reads first 10 features)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_stop_server.R
\name{osrm_stop}
\alias{osrm_stop}
\title{Stop an OSRM Server}
\usage{
osrm_stop(
  server = NULL,
  id = NULL,
  port = NULL,
  pid = NULL,
  wait = 1000L,
  quiet = FALSE
)
}
\arguments{
\item{server}{Optional OSRM job process (an \code{osrm_server} object
inheriting from \code{processx::process}) returned by \code{osrm_start_server()}.}

\item{id}{Optional character id from \code{osrm_servers()}.}

\item{port}{Optional integer TCP port.}

\item{pid}{Optional integer process id.}

\item{wait}{Integer milliseconds to wait for clean shutdown (default \code{1000}).}

\item{quiet}{Logical; suppress messages (default \code{FALSE}).}
}
\value{
A list with fields \code{id}, \code{pid}, \code{port}, \code{stopped} (logical).
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Terminates an \code{osrm-routed} process launched by \code{osrm_start()} or
\code{osrm_start_server()}. Can also stop external servers by PID or ID.
}
\details{
This function provides a flexible way to stop a running OSRM process. If no
arguments are specified, it defaults to stopping the most recently started
server that is still alive in the current session.

You can also stop a specific server by providing:
\itemize{
\item The OSRM job process (a \code{processx::process} object) returned by
\code{osrm_start()} or \code{osrm_start_server()}.
\item The server's \code{id}, \code{port}, or \code{pid} (use \code{osrm_servers()} to find these).
}

\strong{Advanced Use:}
You can stop an external \code{osrm-routed} process (one not started by the current
R session) by passing its PID, or by finding it via \code{osrm_servers(include_all = TRUE)}
and passing its \code{id} or \code{port}. This requires permission to signal the process.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )

  # copy example OSM PBF into a temporary workspace to avoid polluting pkg data
  pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
  dir.create(osrm_dir, recursive = TRUE)
  tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
  file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
  graph <- osrm_prepare_graph(tmp_pbf, overwrite = TRUE, threads = 1L)

  srv <- osrm_start_server(graph$osrm_job_artifact, port = 6000, threads = 1L)

  # Stop by passing the process object
  osrm_stop(srv)

  # Or stop by port after the process is registered
  osrm_stop(port = 6000)

  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
  unlink(osrm_dir, recursive = TRUE)
}

\dontrun{
  # Advanced: Stop an external server by PID
  # 1. Find the PID of an external server
  srvs <- osrm_servers(include_all = TRUE)
  # 2. Stop it by PID
  if (nrow(srvs) > 0) {
    osrm_stop(pid = srvs$pid[1])
  }
  
  # Or stop by its external ID (e.g., "sys-12345")
  osrm_stop(id = "sys-12345")
}
}
}
\seealso{
\code{\link[=osrm_start]{osrm_start()}}, \code{\link[=osrm_servers]{osrm_servers()}}, \code{\link[=osrm_stop_all]{osrm_stop_all()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_start.R
\name{osrm_start}
\alias{osrm_start}
\title{Start an OSRM Server with Automatic Setup}
\usage{
osrm_start(
  path,
  algorithm = c("mld", "ch"),
  force_rebuild = FALSE,
  show_progress = TRUE,
  quiet = FALSE,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{path}{A string. Path to the input data. Can be one of:
\itemize{
\item A path to an OSM file (e.g., \verb{/path/to/data.osm.pbf}).
\item A path to a directory containing OSRM graph files or an OSM file.
\item A direct path to a final graph file (\code{.osrm.mldgr} or \code{.osrm.hsgr}).
}}

\item{algorithm}{A string specifying the routing algorithm to use for graph
preparation, either \code{"mld"} (Multi-Level Dijkstra, default) or \code{"ch"}
(Contraction Hierarchies). This is only used when \code{osrm_prepare_graph} is
automatically called.}

\item{force_rebuild}{A logical value. If \code{TRUE}, force the rebuilding of the
OSRM graph even if it already exists.}

\item{show_progress}{A logical value. If \code{TRUE} (default), show a progress
spinner during graph preparation.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses installer messages and
warnings. Defaults to \code{FALSE}.}

\item{verbose}{A logical. If \code{FALSE} (default), suppresses detailed console
output from backend commands. If \code{TRUE}, shows all output, which is useful
for debugging.}

\item{...}{Additional arguments passed on to \code{osrm_prepare_graph()} (e.g.,
\code{overwrite = TRUE}) and \code{osrm_start_server()} (e.g., \code{port = 5001}).}
}
\value{
An OSRM job process (an \code{osrm_server} object inheriting from
\code{processx::process}) for the running server.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

A high-level, "one-shot" function to start an OSRM server that automatically
handles OSRM installation and graph preparation. This is the recommended
function for most users to get a server running quickly with minimal steps.
}
\details{
This function is designed for convenience and automates the entire setup
process. By default, it is not verbose and only prints high-level status
messages.

\enumerate{
\item \strong{Check for OSRM Installation:} It first verifies if the \code{osrm-routed}
executable is available in the system's \code{PATH}. If not, it automatically
calls \code{osrm_install()} to download and install the latest stable version.
\item \strong{Process Input Path and Prepare Graph:} The function intelligently
handles the \code{path} argument to find or create the necessary graph files.
If the graph files do not exist, it automatically runs \code{osrm_prepare_graph()}
to build them, which may take some time.
\item \strong{Start Server:} Once the graph files are located or prepared, it
launches the \code{osrm-routed} server and prints a confirmation message
with the server's PID and port.
}
For advanced users or for debugging, set \code{verbose = TRUE} to see the detailed
console output from the installation and graph preparation steps. For full
manual control, use the lower-level functions like \code{osrm_prepare_graph()} and
\code{osrm_start_server()} directly.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )

  # copy example OSM PBF into a temporary workspace to avoid polluting pkg data
  pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
  dir.create(osrm_dir, recursive = TRUE)
  local_pbf <- file.path(osrm_dir, "cur.osm.pbf")
  file.copy(from = pbf_path, to = local_pbf, overwrite = TRUE)

  # Start the server with one command.
  # It will quietly install OSRM and prepare the graph if needed.
  osrm_process <- osrm_start(local_pbf)

  # Stop the server when done.
  osrm_stop()

  # To see all backend logs during setup, use verbose = TRUE
  osrm_process_verbose <- osrm_start(local_pbf, verbose = TRUE)
  osrm_stop()

  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
  unlink(osrm_dir, recursive = TRUE)
}
}
}
\seealso{
\code{\link[=osrm_stop]{osrm_stop()}}, \code{\link[=osrm_start_server]{osrm_start_server()}} for manual server startup.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_uninstall.R
\name{osrm_uninstall}
\alias{osrm_uninstall}
\title{Uninstall OSRM Backend Binaries}
\usage{
osrm_uninstall(
  dest_dir = NULL,
  clear_path = TRUE,
  quiet = FALSE,
  all = FALSE,
  force = FALSE
)
}
\arguments{
\item{dest_dir}{A string specifying the directory from which to remove OSRM binaries.
If \code{NULL} (the default), the function looks for an installation in the
per-version subdirectories inside \code{tools::R_user_dir("osrm.backend", which = "cache")}
and removes it. When multiple versions are installed, interactive sessions
that are not \code{quiet} will be prompted (with a numbered menu and \code{0} to cancel)
to choose a directory; otherwise, \code{dest_dir} must be supplied. Ignored if \code{all = TRUE}.}

\item{clear_path}{A logical value. If \code{TRUE} (default), also removes the
\code{PATH} configuration from the project's \code{.Rprofile} by calling \code{osrm_clear_path()}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses informational messages
and confirmation prompts. Defaults to \code{FALSE}.}

\item{all}{A logical value. If \code{TRUE}, removes all OSRM installations found
in the default cache directory. Will prompt for confirmation unless \code{force = TRUE}.
Defaults to \code{FALSE}. When \code{TRUE}, the \code{dest_dir} parameter is ignored.}

\item{force}{A logical value. If \code{TRUE}, skips all confirmation prompts,
enabling non-interactive usage. Defaults to \code{FALSE}.}
}
\value{
\code{TRUE} if one or more directories were successfully removed,
and \code{FALSE} otherwise.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Removes the OSRM backend binaries and optionally clears the \code{PATH}
configuration from the project's \code{.Rprofile}.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  # Install OSRM temporarily
  install_dir <- osrm_install(path_action = "session", quiet = TRUE)

  # Uninstall that specific version and clear PATH changes
  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )

  # If multiple installs exist, remove them all
  osrm_uninstall(all = TRUE, force = TRUE, quiet = TRUE)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_server_registry.R
\name{osrm_get_server_profile}
\alias{osrm_get_server_profile}
\title{Retrieve the OSRM Profile for a Running Server}
\usage{
osrm_get_server_profile(input_osrm = NULL, port = NULL)
}
\arguments{
\item{input_osrm}{Optional. Can be an OSRM job process (an \code{osrm_server}
object inheriting from \code{processx::process}), a path string, or NULL.}

\item{port}{Optional integer. The port of the server.}
}
\value{
A character string representing the profile name (default "car").
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Attempts to determine the profile (e.g., "car", "bike", "foot") used by an OSRM server.
It follows a priority list:
\enumerate{
\item Checks the active server registry for the given port or ID.
\item Checks for a \code{dataset.meta.json} file in the directory of the graph file.
\item Checks the graph filename for hints (e.g. \code{berlin-car.osrm}).
\item Falls back to \code{getOption("osrm.profile")}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_find_profile.R
\name{osrm_find_profile}
\alias{osrm_find_profile}
\title{Locate an OSRM Lua profile (e.g. car.lua) in a host installation}
\usage{
osrm_find_profile(profile = "car.lua")
}
\arguments{
\item{profile}{A single string, the name of the Lua profile file
(e.g. \code{"car.lua"}). Defaults to \code{"car.lua"}.}
}
\value{
The normalized filesystem path to the profile.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

By default OSRM ships profiles for "car", "bike" and "foot" in a
\verb{profiles/} directory alongside the binaries.  This function will try to
locate \code{osrm-routed} on the \code{PATH}, resolve symlinks, and look first for a
\verb{profiles/} directory next to the binary (as placed there by
\code{osrm_install()}). If that fails, it looks for sibling directories
\code{share/osrm/profiles} and \code{share/osrm-backend/profiles}. IF that fails, it
will try to fall back on \verb{/usr/local/share/osrm/profiles},
\verb{/usr/local/share/osrm-backend/profiles}, \verb{/usr/share/osrm/profiles}, and
\verb{/usr/share/osrm-backend/profiles}.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )
  osrm_find_profile("car.lua")
  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_install.R
\name{osrm_check_latest_version}
\alias{osrm_check_latest_version}
\title{Check for the Latest Stable OSRM Version}
\usage{
osrm_check_latest_version()
}
\value{
A string containing the latest version tag (e.g., \code{"v26.4.1"}).
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Queries the GitHub API to find the most recent stable (non-pre-release)
version tag for the OSRM backend that has binaries available for the current platform.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  # Get the latest stable version number of OSRM backend
  osrm_check_latest_version()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_install.R
\name{osrm_install}
\alias{osrm_install}
\title{Install OSRM Backend Binaries}
\usage{
osrm_install(
  version = "latest",
  dest_dir = NULL,
  force = FALSE,
  path_action = c("session", "project", "none"),
  quiet = FALSE
)
}
\arguments{
\item{version}{A string specifying the OSRM version tag to install.
Defaults to \code{"latest"}. Use \code{"latest"} to automatically find the most
recent stable version (internally calls \code{\link[=osrm_check_latest_version]{osrm_check_latest_version()}}). Versions
other than \code{v5.27.1}, \code{v6.0.0}, \code{v26.4.0} and \code{v26.4.1} will trigger a
warning but are still attempted if binaries are available.}

\item{dest_dir}{A string specifying the directory where OSRM binaries should be
installed. If \code{NULL} (the default), a user-friendly, persistent location is
chosen via \code{tools::R_user_dir("osrm.backend", which = "cache")}, and the
binaries are installed into a subdirectory named after the OSRM version
(e.g. \code{.../cache/v26.4.1}).}

\item{force}{A logical value. If \code{TRUE}, reinstall OSRM even if it's already
found in \code{dest_dir}. If \code{FALSE} (default), the function will stop if an
existing installation is detected.}

\item{path_action}{A string specifying how to handle the system \code{PATH}. One of:
\itemize{
\item \code{"session"} (default): Adds the OSRM bin directory to the \code{PATH}
for the current R session only.
\item \code{"project"}: Modifies the \code{.Rprofile} in the current project to set
the \code{PATH} for all future sessions in that project.
\item \code{"none"}: Does not modify the \code{PATH}.
}}

\item{quiet}{A logical value. If \code{TRUE}, suppresses installer messages and
warnings. Defaults to \code{FALSE}.}
}
\value{
The path to the installation directory.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Downloads and installs pre-compiled binaries for the OSRM backend from the
official GitHub releases. The function automatically detects the user's
operating system and architecture to download the appropriate files. Only the
latest v5 release (\code{v5.27.1}), \code{v6.0.0}, \code{v26.4.0} and \code{v26.4.1} were manually
tested and are known to work well; other releases available on GitHub can be
installed but are not guranteed to function correctly.
}
\details{
The function performs the following steps:
\enumerate{
\item Queries the GitHub API to find the specified release of
\code{Project-OSRM/osrm-backend}.
\item Identifies the correct binary (\code{.tar.gz} archive) for the user's OS
(Linux, macOS, or Windows) and architecture (x64, arm64).
\item Downloads the archive to a temporary location.
\item Extracts the archive and locates the OSRM executables (e.g.,
\code{osrm-routed}, \code{osrm-extract}).
\item Copies these executables to a local directory (defaults to
\verb{file.path(tools::R_user_dir("osrm.backend", which = "cache"), <version>)}).
\item Downloads the matching Lua profiles from the release tarball and installs
them alongside the binaries.
\item Optionally modifies the \code{PATH} environment variable for the current
session or project.
}

macOS users should note that upstream OSRM v6.x (and newer) binaries are
built for macOS 15.0 (Sequoia, Darwin 24.0.0) or newer. \code{osrm_install()}
automatically blocks v6+ installs on older macOS releases and, when
\code{version = "latest"}, selects the most recent v5 build instead while warning
about the requirement. Warnings include both the marketing version and Darwin
kernel so you'll see messages like \verb{macOS 13 Ventura [Darwin 22.6.0]}.

When installing OSRM v6.x or newer for Windows, the upstream release omits
the Intel Threading Building Blocks (TBB) runtime and a compatible \code{bz2} DLL.
To keep the executables runnable out of the box, \code{osrm_install()} fetches TBB
from \href{https://github.com/uxlfoundation/oneTBB/releases/tag/v2022.3.0}{oneTBB
v2022.3.0} and the BZip2 runtime from
\href{https://github.com/philr/bzip2-windows/releases/tag/v1.0.8.0}{bzip2-windows
v1.0.8.0}, verifying their SHA-256 checksums before extraction. Without these
extra libraries, the OSRM v6+ binaries shipped for Windows cannot start.

On macOS, OSRM v6.x and newer binaries also miss the bundled TBB runtime.
The installer reuses the libraries from release \code{v5.27.1} to keep the
binaries functional and patches their \code{libbz2} linkage using
\code{install_name_tool} so that they load the system-provided BZip2 runtime.

Power users (including package authors running cross-platform tests) can
override the auto-detected platform by setting the R options
\code{osrm.backend.override_os} and \code{osrm.backend.override_arch} (e.g.,
\code{options(osrm.backend.override_os = "linux", osrm.backend.override_arch = "arm64")})
before calling \code{osrm_install()}. Overrides allow requesting binaries for any
OS and CPU combination that exists on the GitHub releases.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  old <- setwd(tempdir())
  on.exit(setwd(old), add = TRUE)

  # Install the default stable version and set PATH for this session
  install_dir <- osrm_install(path_action = "session", quiet = TRUE)

  # Install for a project non-interactively (e.g., in a script)
  osrm_install(path_action = "project", quiet = TRUE, force = TRUE)

  # Clean up the project's .Rprofile and uninstall binaries
  osrm_clear_path(quiet = TRUE)
  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/osrm_cleanup.R
\name{osrm_cleanup}
\alias{osrm_cleanup}
\title{Clean Up OSRM Files in a Directory}
\usage{
osrm_cleanup(path, keep_osm = TRUE, dry_run = FALSE, quiet = FALSE)
}
\arguments{
\item{path}{A string. Path to an OSRM file or directory containing OSRM files.
If a file path is provided (e.g., \code{data.osm.pbf} or \code{data.osrm.hsgr}),
the base name will be extracted and all related \verb{.osrm.*} files will be removed.}

\item{keep_osm}{Logical. If \code{TRUE} (default), keeps the original \code{.osm.pbf}
(or \code{.osm}, \code{.osm.bz2}) file. If \code{FALSE}, removes it as well.}

\item{dry_run}{Logical. If \code{TRUE}, shows what would be deleted without
actually deleting. Default is \code{FALSE}.}

\item{quiet}{Logical. If \code{TRUE}, suppresses messages. Default is \code{FALSE}.}
}
\value{
Invisibly returns a character vector of removed file paths.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Remove OSRM-generated files from a directory. This is useful when switching
between algorithms (CH and MLD) or when you want to start fresh.
}
\details{
OSRM creates many \verb{.osrm.*} files during the extract, contract, partition,
and customize stages. This function helps clean up these files.

\strong{Important:} The CH and MLD algorithms cannot safely coexist in the same
directory because the MLD partition stage modifies some extract-stage files.
Use this function to clean up before switching algorithms.
}
\examples{
\donttest{
if (identical(Sys.getenv("OSRM_EXAMPLES"), "true")) {
  install_dir <- osrm_install(
    version = "latest",
    path_action = "session",
    quiet = TRUE
  )

  # Stage a temporary workspace with placeholder OSRM files
  pbf_path <- system.file("extdata/cur.osm.pbf", package = "osrm.backend")
  osrm_dir <- file.path(tempdir(), paste0("osrm-", Sys.getpid()))
  dir.create(osrm_dir, recursive = TRUE)
  tmp_pbf <- file.path(osrm_dir, "cur.osm.pbf")
  file.copy(from = pbf_path, to = tmp_pbf, overwrite = TRUE)
  file.create(
    file.path(osrm_dir, "cur.osrm.timestamp"),
    file.path(osrm_dir, "cur.osrm.hsgr"),
    file.path(osrm_dir, "cur.osrm.mldgr"),
    file.path(osrm_dir, "cur.osrm.partition")
  )

  # Preview what would be deleted
  osrm_cleanup(osrm_dir, dry_run = TRUE)

  # Clean up OSRM artifacts (keep the OSM file)
  osrm_cleanup(osrm_dir)

  # Remove everything including the OSM source
  osrm_cleanup(osrm_dir, keep_osm = FALSE)

  osrm_uninstall(
    dest_dir = install_dir,
    clear_path = TRUE,
    force = TRUE,
    quiet = TRUE
  )
  unlink(osrm_dir, recursive = TRUE)
}
}
}
